/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <gdam/common/gdamautoconf.h>
#include <unistd.h>
#include "cli.h"
#include "dump.h"
#include "../gdam123/getline.h"



/* Map from strings to other strings. */
static GHashTable* variables = NULL;


/* A function to process a single type of command, and a hashtable
 * mapping names to handlers. */
typedef gboolean (*CommandHandler)(LineTokenizer* tokenizer,
                                   CliInfo*       info,
				   GdamRequest**  request);

/* Map from string (command name) => CommandHandler. */
static GHashTable* command_handlers = NULL;

void cli_help(const char* topic);

static void initial_empty_hashtables()
{
  variables = g_hash_table_new (g_str_hash, g_str_equal);
  command_handlers = g_hash_table_new (g_str_hash, g_str_equal);
}

static void cli_info_error (CliInfo *cli_info)
{
  if (cli_info->exit_on_error)
    exit (1);
}

/*
 * 
 *  ___  ___  _   _ _ __ ___ ___
 * / __|/ _ \| | | | '__/ __/ _ \
 * \__ \ (_) | |_| | | | (_|  __/
 * |___/\___/ \__,_|_|  \___\___|
 * 
 * Create a sound source on the server side and assign its
 * id to a variable.
 */
typedef struct _SourceCommandInfo SourceCommandInfo;
struct _SourceCommandInfo {
	char* variable_name;
	char* class_name;
	CliInfo* cli_info;
};

static void add_or_update_variable(const char* var_name, char* value)
{
  char *old_ptr;
  old_ptr = g_hash_table_lookup (variables, var_name);

  if (old_ptr == NULL)
    var_name = g_strdup (var_name);

  g_hash_table_insert (variables, (char *) var_name, value);

  if (old_ptr != NULL)
    g_free (old_ptr);
}

static void handle_source_command_success(int source_id, gpointer ptr_info)
{
  SourceCommandInfo *info = (SourceCommandInfo *) ptr_info;

  if (info->cli_info->interactive)
    fprintf (stderr, "variable `%s' assigned source id %d.\n",
	     info->variable_name, source_id);

  add_or_update_variable (info->variable_name,
			  g_strdup_printf ("%d", source_id));
  cli_info_set_id_type (info->cli_info, source_id, info->class_name);

  g_free (info->variable_name);
  g_free (info->class_name);
  g_free (info);
}

static void handle_source_command_failure(gpointer ptr_info)
{
  SourceCommandInfo *info = (SourceCommandInfo *) ptr_info;
  if (info->cli_info->interactive)
    fprintf (stderr, "failed to create source named `%s' of type %s\n",
	     info->variable_name, info->class_name);
  cli_info_error (info->cli_info);
  g_free (info->variable_name);
  g_free (info->class_name);
  g_free (info);
}


static gboolean handle_source_command(LineTokenizer* tokenizer,
                                      CliInfo*       info,
				      GdamRequest**  request)
{
  char *name = line_tokenizer_get_word (tokenizer);
  char *type = line_tokenizer_get_word (tokenizer);
  ArgHints *hints;
  int num_args = 0, max_args = 16;
  GdamArg *args = g_new (GdamArg, max_args);
  SourceCommandInfo *command_info = g_new (SourceCommandInfo, 1);
  command_info->variable_name = name;
  command_info->cli_info = info;
  command_info->class_name = g_strdup (type);
  hints = arg_hints_table_lookup_arg (tokenizer->hint_table, type);

  while (1)
    {
      if (num_args == max_args)
	{
	  max_args *= 2;
	  args = g_renew (GdamArg, args, max_args);
	}
      if (!line_tokenizer_tokenize (tokenizer, hints, args + num_args))
	break;
      num_args++;
    }
  *request = gdam_request_create_source_v (type,
					   0,
					   num_args,
					   args,
					   handle_source_command_success,
					   handle_source_command_failure,
					   command_info);
  gdam_arg_free_array (num_args, args);
  g_free (type);
  return *request ? TRUE : FALSE;
}

/*
 *           _
 *  ___  ___| |_
 * / __|/ _ \ __|
 * \__ \  __/ |_
 * |___/\___|\__|
 * Set arguments for a source.
 */
static void handle_set_command_success(gpointer unused)
{
  gdam_debug ("set command succeeded");
}
static void handle_set_command_failure(gpointer cli_info)
{
  gdam_debug ("set command failed");
  cli_info_error (cli_info);
}
static gboolean handle_set_command(LineTokenizer* tokenizer,
                                   CliInfo*       info,
				   GdamRequest**  request)
{
  char *id = line_tokenizer_get_word (tokenizer);
  const char *type;
  int source_id;
  int num_args = 0, max_args = 16;
  ArgHints *hints;
  GdamArg *args = g_new (GdamArg, max_args);
  source_id = (int) strtol (id, NULL, 10);
  type = cli_info_get_type_by_id (info, source_id);
  if (type != NULL)
    hints = arg_hints_table_lookup_arg (tokenizer->hint_table, type);
  else
    hints = NULL;
  while (1)
    {
      if (num_args == max_args)
	{
	  max_args *= 2;
	  args = g_renew (GdamArg, args, max_args);
	}
      if (!line_tokenizer_tokenize (tokenizer, hints, args + num_args))
	{
	  break;
	}
      num_args++;
    }
  *request = gdam_request_set_args (source_id,
				    num_args,
				    args,
				    handle_set_command_success,
				    handle_set_command_failure, NULL);
  gdam_arg_free_array (num_args, args);
  g_free (id);
  return *request ? TRUE : FALSE;
}

/*
 *            _     _
 *   __ _  __| | __| |
 *  / _` |/ _` |/ _` |
 * | (_| | (_| | (_| |
 *  \__,_|\__,_|\__,_|
 * Add a source to a mixer.
 */

static void handle_add_success(gpointer unused)
{
}
static void handle_add_failure(gpointer cli_info)
{
  cli_info_error (cli_info);
}
static gboolean handle_add_command(LineTokenizer* tokenizer,
                                   CliInfo*       info,
				   GdamRequest**  request)
{
  guint mix_id = info->mixer_id;
  guint sub_id;
  char *num;
  num = line_tokenizer_get_word (tokenizer);
  if (num == NULL)
    {
      fprintf (stderr, "no source id found...\n");
      return FALSE;
    }
  sub_id = (guint) strtol (num, NULL, 10);
  g_free (num);
  num = line_tokenizer_get_word (tokenizer);
  if (num != NULL)
    {
      mix_id = (guint) strtol (num, NULL, 10);
      g_free (num);
    }
  gdam_debug ("request_add_mix: mix=%d, sub=%d", mix_id, sub_id);
  *request = gdam_request_add_mix (mix_id,
				   sub_id,
				   (GdamTime) - 1,
				   handle_add_success,
				   handle_add_failure, info);
  return *request ? TRUE : FALSE;
}



/*                _    
 *  ___  ___  ___| | __
 * / __|/ _ \/ _ \ |/ /
 * \__ \  __/  __/   < 
 * |___/\___|\___|_|\_\
 */                   
static void handle_seek_success(gpointer unused)
{
}
static void handle_seek_failure(gpointer cli_info)
{
  cli_info_error (cli_info);
}
static gboolean handle_seek_command (LineTokenizer* tokenizer,
                                     CliInfo*       info,
				     GdamRequest**  request)
{
  guint sub_id;
  int samples;
  char *num;
  num = line_tokenizer_get_word (tokenizer);
  if (num == NULL)
    {
      fprintf (stderr, "no source id found...\n");
      return FALSE;
    }
  sub_id = (guint) strtol (num, NULL, 10);
  g_free (num);
  num = line_tokenizer_get_word (tokenizer);
  if (num == NULL)
    {
      fprintf (stderr, "missing size of seek...\n");
      return FALSE;
    }
  samples = (gint) strtol (num, NULL, 10);
  g_free (num);
  gdam_debug ("request_seek_relative: id=%d, size=%d", sub_id, samples);
  *request = gdam_request_seek_relative (sub_id,
					 (GdamTime)samples,
					 handle_seek_success,
					 handle_seek_failure, info);
  return *request ? TRUE : FALSE;
}


/*
 * Cache
 */

typedef struct _CacheData {
	CliInfo* info;
	char* variable_name;
} CacheData;

static void finished_caching(gpointer user_data)
{
  CacheData *cd = user_data;
  CliInfo *info = cd->info;
  *info->is_done_processing = TRUE;
  g_free (cd->variable_name);
  g_free (cd);
}

static void handle_cache_success(int buffer_id, int connection_id,
				 gpointer user_data)
{
  CacheData *cd = user_data;
  CliInfo *info = cd->info;
  add_or_update_variable (cd->variable_name,
			  g_strdup_printf ("%d", buffer_id));
  gdam_debug ("caching: buffer_id=%d event_id=%d", buffer_id, connection_id);
  gdam_channel_register_handler (info->channel,
				 gdam_channel_simple_handler_new
				 (connection_id, TRUE, finished_caching, cd));
}

static void handle_cache_failure(gpointer user_data)
{
  CacheData *cd = user_data;
  CliInfo *info = cd->info;
  cli_info_error (info);
  g_free (cd->variable_name);
  g_free (cd);
  *info->is_done_processing = TRUE;
}
static gboolean handle_cache_command(LineTokenizer* tokenizer,
                                     CliInfo*       info,
				     GdamRequest**  request)
{
  CacheData *cd;
  guint source_id;
  char *num;
  char *var;
  var = line_tokenizer_get_word (tokenizer);
  if (var == NULL)
    {
      fprintf (stderr, "no variable name found...\n");
      return FALSE;
    }
  num = line_tokenizer_get_word (tokenizer);
  if (num == NULL)
    {
      fprintf (stderr, "no source id found...\n");
      return FALSE;
    }
  source_id = (guint) strtol (num, NULL, 10);
  g_free (num);
  cd = g_new (CacheData, 1);
  cd->info = info;
  cd->variable_name = var;
  *request = gdam_request_start_caching (source_id,
					 handle_cache_success,
					 handle_cache_failure, cd);
  *info->is_done_processing = FALSE;
  return *request ? TRUE : FALSE;
}


/*
 *      _ _
 *  ___(_) | ___ _ __   ___ ___
 * / __| | |/ _ \ '_ \ / __/ _ \
 * \__ \ | |  __/ | | | (_|  __/
 * |___/_|_|\___|_| |_|\___\___|
 * 
 */

static void handle_silence_success(gpointer unused)
{
  gdam_debug ("silenced mixer successfully");
}
static void handle_silence_failure(gpointer cli_info)
{
  g_message ("Failed to silence the mixer");
  cli_info_error (cli_info);
}
static gboolean handle_silence_command(LineTokenizer* tokenizer,
                                   CliInfo*       info,
				   GdamRequest**  request)
{
  guint mix_id = info->mixer_id;
  char *num;
  GdamArg arg[1];
  num = line_tokenizer_get_word (tokenizer);
  if (num != NULL)
    {
      fprintf (stderr, "got word mix id word `%s'", num);
      mix_id = (guint) strtol (num, NULL, 10);
      g_free (num);
    }
  gdam_debug ("request_silence_mix: mix=%d", mix_id);
  arg[0].name = "mixed";
  arg[0].type = GDAM_ARG_TYPE_INT64_ARRAY;
  GDAM_ARG_INT64_ARRAY (arg).length = 0;
  GDAM_ARG_INT64_ARRAY (arg).array = NULL;
  *request = gdam_request_set_args (mix_id,
				    1,
				    arg,
				    handle_silence_success,
				    handle_silence_failure, info);
  return *request ? TRUE : FALSE;
}

/*
 *            _
 *  _ __ ___ (_)_  _____ _ __
 * | '_ ` _ \| \ \/ / _ \ '__|
 * | | | | | | |>  <  __/ |
 * |_| |_| |_|_/_/\_\___|_|
 * Print the mixer id.
 */
static gboolean handle_mixer_command(LineTokenizer* tokenizer,
                                     CliInfo*       info,
				     GdamRequest**  request)
{
  fprintf (stderr, "mixer id is %d.\n", info->mixer_id);
  return TRUE;
}

static gboolean handle_device_command(LineTokenizer* tokenizer,
                                      CliInfo*       info,
				      GdamRequest**  request)
{
  fprintf (stderr, "device id is %d.\n", info->device_id);
  return TRUE;
}

/*
 *      _      _             _
 *   __| | ___| |_ __ _  ___| |__
 *  / _` |/ _ \ __/ _` |/ __| '_ \
 * | (_| |  __/ || (_| | (__| | | |
 *  \__,_|\___|\__\__,_|\___|_| |_|
 * Remove a source from its parent.
 */
static void handle_detach_success(gpointer unused)
{
}
static void handle_detach_failure(gpointer cli_info)
{
  cli_info_error (cli_info);
}
static gboolean handle_detach_command(LineTokenizer* tokenizer,
                                      CliInfo*       info,
				      GdamRequest**  request)
{
  guint source_id;
  char *num;
  num = line_tokenizer_get_word (tokenizer);
  if (num == NULL)
    {
      fprintf (stderr, "what do you want to detach?\n");
      return FALSE;
    }
  source_id = (guint) strtoul (num, NULL, 10);
  g_free (num);

  *request = gdam_request_detach_source (source_id,
					 handle_detach_success,
					 handle_detach_failure, info);
  if (!*request)
    {
      fprintf (stderr, "unable to detach source %d.\n", source_id);
      return FALSE;
    }
  return TRUE;
}

/*
 * unref buffer...
 */
static void handle_unref_success(gpointer unused)
{
}
static void handle_unref_failure(gpointer cli_info)
{
  g_message ("unref failed");
  cli_info_error (cli_info);
}

static gboolean handle_unref_generic_command(LineTokenizer* tokenizer,
                                             CliInfo*       info,
				             GdamRequest**  request,
					     int 	    arena_id)
{
  guint source_id;
  char *num;
  num = line_tokenizer_get_word (tokenizer);
  if (num == NULL)
    {
      fprintf (stderr, "what do you want to unref?\n");
      return FALSE;
    }
  source_id = (guint) strtoul (num, NULL, 10);
  g_free (num);

  *request = gdam_request_unref (arena_id,
				 source_id,
				 handle_unref_success,
				 handle_unref_failure, info);
  if (!*request)
    {
      fprintf (stderr, "unable to unref source %d.\n", source_id);
      return FALSE;
    }
  return TRUE;
}

static gboolean handle_unref_buffer_command(LineTokenizer* tokenizer,
                                             CliInfo*       info,
				             GdamRequest**  request)
{
  return handle_unref_generic_command (tokenizer, info, request,
				       GDAM_PROTOCOL_BUFFER_ARG);
}

static gboolean handle_unref_source_command(LineTokenizer* tokenizer,
                                             CliInfo*       info,
				             GdamRequest**  request)
{
  return handle_unref_generic_command (tokenizer, info, request,
				       GDAM_PROTOCOL_SOURCE_ARG);
}


static gboolean handle_feed_command(LineTokenizer* tokenizer,
                                    CliInfo*       info,
				    GdamRequest**  request)
{
  guint device_id, mixer_id;
  char *num;
  num = line_tokenizer_get_word (tokenizer);
  if (num == NULL)
    {
      device_id = info->device_id;
      mixer_id = info->mixer_id;
      g_message ("reopening default device %d", device_id);
    }
  else
    {
      device_id = atoi (num);
      g_free (num);

      num = line_tokenizer_get_word (tokenizer);
      if (num == NULL)
	{
	  mixer_id = info->mixer_id;
	  g_message ("reopening default device %d", device_id);
	}
      else
	{
	  mixer_id = atoi (num);
	  g_free (num);
	}
    }

  *request = gdam_request_feed_sink (device_id, mixer_id, NULL, NULL, NULL);
  if (!*request)
    {
      g_warning ("unable to feed sink %d", device_id);
      return FALSE;
    }
  return TRUE;
}

static gboolean handle_reopen_command(LineTokenizer* tokenizer,
                                      CliInfo*       info,
				      GdamRequest**  request)
{
  guint device_id;
  char *num;
  num = line_tokenizer_get_word (tokenizer);
  if (num == NULL)
    {
      device_id = info->device_id;
      g_message ("reopening default device %d", device_id);
    }
  else
    {
      device_id = atoi (num);
      g_free (num);
    }

  *request = gdam_request_reopen_sink (device_id, NULL, NULL, NULL);
  if (!*request)
    {
      g_warning ("unable to detach source %d", device_id);
      return FALSE;
    }
  return TRUE;
}

typedef struct {
  CliInfo *cli_info;
  int      source_id;
} InfoInt;

/*
 *  _
 * | |___
 * | / __|
 * | \__ \
 * |_|___/
 * 
 */
static void
process_ls_success(int num_args, GdamArg* args, gpointer info_int_struct)
{
  int i;
  InfoInt *info_int = info_int_struct;
  fprintf (stderr, "source %d.\n",  info_int->source_id);
  for (i = 0; i < num_args; i++)
    {
      char *arg_value;
      arg_value = gdam_arg_val_as_string (args + i);
      fprintf (stderr, "  arg %d: %s => %s\n", i, args[i].name, arg_value);
      g_free (arg_value);
    }
  g_free (info_int);
}
static void process_ls_failure(gpointer info_int_struct)
{
  InfoInt *info_int = info_int_struct;
  fprintf (stderr, "ls on source %d failed.\n", info_int->source_id);
  cli_info_error (info_int->cli_info);
  g_free (info_int);
}
static gboolean handle_ls_command(LineTokenizer* tokenizer,
                                  CliInfo*       info,
				  GdamRequest**  request)
{
  guint source_id;
  char *num = NULL;
  gboolean only_vital = TRUE;
  gboolean recursive = FALSE;
  InfoInt *info_int;
  while (num == NULL)
    {
      char *arg = line_tokenizer_get_word (tokenizer);
      if (arg == NULL)
	{
	  fprintf (stderr, "no source id found...\n");
	  return FALSE;
	}
      if (arg[0] == '-')
	{
	  switch (arg[1])
	    {
	    case 'a':
	      only_vital = FALSE;
	      break;
	    case 'R':
	      recursive = TRUE;
	      break;
	    }
	  g_free (arg);
	  continue;
	}
      num = arg;
    }
  source_id = (int) strtol (num, NULL, 10);
  if (recursive)
    {
      cli_do_ls_R (info, source_id, only_vital);
      return TRUE;
    }
  info_int = g_new (InfoInt, 1);
  info_int->cli_info = info;
  info_int->source_id = source_id;
  *request = gdam_request_get_all_args (source_id,
					only_vital,
					process_ls_success,
					process_ls_failure,
					info_int);
  return *request ? TRUE : FALSE;
}

/*
 *            _ _
 *   ___ __ _| | |
 *  / __/ _` | | |
 * | (_| (_| | | |
 *  \___\__,_|_|_|
 * 
 */

static void
process_call_success(int num_args, GdamArg* args, gpointer info_int_struct)
{
  int i;
  InfoInt *info_int = info_int_struct;
  fprintf (stderr, "call: source %d.\n", info_int->source_id);
  for (i = 0; i < num_args; i++)
    {
      char *arg_value;
      arg_value = gdam_arg_val_as_string (args + i);
      fprintf (stderr, "  arg %d: %s => %s\n", i, args[i].name, arg_value);
      g_free (arg_value);
    }
  g_free (info_int);
}
static void process_call_failure(gpointer info_int_struct)
{
  InfoInt *info_int = info_int_struct;
  gdam_debug ("call command failed");
  cli_info_error (info_int->cli_info);
  g_free (info_int);
}

static gboolean handle_general_call_command(LineTokenizer* tokenizer,
                                    CliInfo*       info,
				    GdamRequest**  request,
				    int            call_type)

{
  char *num = line_tokenizer_get_word (tokenizer);
  char *func_name = line_tokenizer_get_word (tokenizer);
  int num_args = 0, max_args = 16;
  int source_id;
  int i;
  const char *type;
  ArgHints *hints;
  InfoInt *info_int;
  GdamArg *args = g_new (GdamArg, max_args);
  g_return_val_if_fail (func_name != NULL, FALSE);
  source_id = (int) strtol (num, NULL, 10);
  if (call_type == GDAM_PROTOCOL_SOURCE_ARG)
    type = cli_info_get_type_by_id (info, source_id);
  else
    type = "GdasCacheBuffer";
  if (type == NULL)
    {
      hints = NULL;
    }
  else
    {
      hints = arg_hints_table_lookup_func (tokenizer->hint_table,
					   type, func_name);
    }
  while (1)
    {
      if (num_args == max_args)
	{
	  max_args *= 2;
	  args = g_renew (GdamArg, args, max_args);
	}
      if (!line_tokenizer_tokenize (tokenizer, hints, args + num_args))
	{
	  break;
	}
      num_args++;
    }
  info_int = g_new (InfoInt, 1);
  info_int->cli_info = info;
  info_int->source_id = source_id;
  *request = gdam_request_call_function (call_type,
					 source_id,
					 func_name,
					 num_args,
					 args,
					 process_call_success,
					 process_call_failure,
					 info_int);
  return *request ? TRUE : FALSE;
}

static gboolean handle_buffer_call_command(LineTokenizer* tokenizer,
                                    CliInfo*       info,
				    GdamRequest**  request)
{
  return handle_general_call_command (tokenizer, info, request,
				      GDAM_PROTOCOL_BUFFER_ARG);
}
static gboolean handle_call_command(LineTokenizer* tokenizer,
                                    CliInfo*       info,
				    GdamRequest**  request)
{
  return handle_general_call_command (tokenizer, info, request,
				      GDAM_PROTOCOL_SOURCE_ARG);
}
static void
process_call_class_success(int num_args, GdamArg* args, gpointer cli_info)
{
  int i;
  fprintf (stderr, "call_class.\n");
  for (i = 0; i < num_args; i++)
    {
      char *arg_value;
      arg_value = gdam_arg_val_as_string (args + i);
      fprintf (stderr, "  arg %d: %s => %s\n", i, args[i].name, arg_value);
      g_free (arg_value);
    }
}
static void process_call_class_failure(gpointer cli_info)
{
  gdam_debug ("call_class command failed");
  cli_info_error (cli_info);
}


static gboolean handle_call_class_command(LineTokenizer* tokenizer,
                                          CliInfo*       info,
				          GdamRequest**  request)
{
  char *class_name = line_tokenizer_get_word (tokenizer);
  char *func_name = line_tokenizer_get_word (tokenizer);
  int num_args = 0, max_args = 16;
  int i;
  const char *type;
  GdamArg *args = g_new (GdamArg, max_args);
  g_return_val_if_fail (func_name != NULL, FALSE);
  while (1)
    {
      if (num_args == max_args)
	{
	  max_args *= 2;
	  args = g_renew (GdamArg, args, max_args);
	}
      if (!line_tokenizer_tokenize (tokenizer, NULL, args + num_args))
	{
	  break;
	}
      num_args++;
    }
  *request = gdam_request_call_class_function (class_name,
					       func_name,
					       num_args,
					       args,
					       process_call_class_success,
					       process_call_class_failure,
					       info);
  return *request ? TRUE : FALSE;
}


/*
 *      _
 *  ___| | ___  ___ _ __
 * / __| |/ _ \/ _ \ '_ \
 * \__ \ |  __/  __/ |_) |
 * |___/_|\___|\___| .__/
 *                 |_|
 * sleep for a number of seconds (possibly fractional).
 */
static gboolean handle_sleep_command(LineTokenizer* tokenizer,
                                     CliInfo*       info,
				     GdamRequest**  request)
{
  char *num_str;
  char *endp;
  double secs;
  num_str = line_tokenizer_get_word (tokenizer);
  if (num_str == NULL)
    {
      g_warning ("sleep needs an argument");
      return FALSE;
    }
  secs = strtod (num_str, &endp);
  if (endp == num_str || *endp != '\0')
    {
      g_warning ("error parsing number");
      return FALSE;
    }
  usleep (secs * 1000000);
  return TRUE;
}

typedef struct {
  CliInfo *cli_info;
  char    *string;
} InfoString;

/*
 *       _           _
 *   ___| |__   __ _| |_
 *  / __| '_ \ / _` | __|
 * | (__| | | | (_| | |_
 *  \___|_| |_|\__,_|\__|
 * Chat with another user.
 */
static void
chat_success (gpointer info_string_struct)
{
  InfoString *info_string = info_string_struct;
  g_message ("sent chat message to %s", info_string->string);
  g_free (info_string->string);
  g_free (info_string);
}
static void
chat_failure (gpointer info_string_struct)
{
  InfoString *info_string = info_string_struct;
  g_message ("no user named %s", info_string->string);
  cli_info_error (info_string->cli_info);
  g_free (info_string->string);
  g_free (info_string);
}
static gboolean handle_chat_command(LineTokenizer* tokenizer,
                                     CliInfo*       info,
				     GdamRequest**  request)
{
  char *username = line_tokenizer_get_word (tokenizer);
  InfoString *info_string;
  if (username == NULL)
    return FALSE;
  while (*tokenizer->at && isspace (*tokenizer->at))
    ++(tokenizer->at);
  info_string = g_new (InfoString, 1);
  info_string->cli_info = info;
  info_string->string = username;
  *request = gdam_request_chat_message (username, tokenizer->at,
					chat_success,
					chat_failure,
					info_string);
  return TRUE;
}

/*
 *           _
 *  ___  ___| |_      _   _ ___  ___ _ __ _ __   __ _ _ __ ___   ___
 * / __|/ _ \ __|____| | | / __|/ _ \ '__| '_ \ / _` | '_ ` _ \ / _ \
 * \__ \  __/ ||_____| |_| \__ \  __/ |  | | | | (_| | | | | | |  __/
 * |___/\___|\__|     \__,_|___/\___|_|  |_| |_|\__,_|_| |_| |_|\___|
 * Set your username for chat purposes.
 */
static void
set_username_success (gpointer info_string_struct)
{
  InfoString *info_string = info_string_struct;
  g_message ("set-username: success: now known as `%s'", info_string->string);
  g_free (info_string->string);
  g_free (info_string);
}

static void
set_username_failure (gpointer info_string_struct)
{
  InfoString *info_string = info_string_struct;
  g_message ("set-username: failure: `%s' already exists", info_string->string);
  cli_info_error (info_string->cli_info);
  g_free (info_string->string);
  g_free (info_string);
}

static gboolean
handle_set_username_command(LineTokenizer* tokenizer,
			    CliInfo*       info,
			    GdamRequest**  request)
{
  char *username = line_tokenizer_get_word (tokenizer);
  InfoString *info_string;
  if (username == NULL)
    return FALSE;
  info_string = g_new (InfoString, 1);
  info_string->cli_info = info;
  info_string->string = username;
  *request = gdam_request_set_username (username,
					set_username_success,
					set_username_failure,
					info_string);
  return TRUE;
}

/*
 *            _ _
 *   _____  _(_) |_
 *  / _ \ \/ / | __|
 * |  __/>  <| | |_
 *  \___/_/\_\_|\__|
 * Exit (or quit) command.
 */
static gboolean handle_exit_command(LineTokenizer* tokenizer,
                                    CliInfo*       info,
				    GdamRequest**  request)
{
  int exit_code = 0;
  char *num;
  num = line_tokenizer_get_word (tokenizer);
  if (num != NULL)
    exit_code = strtol (num, NULL, 10);
  exit (exit_code);
}

static gboolean handle_exit_on_error_command(LineTokenizer* tokenizer,
                                             CliInfo*       info,
				             GdamRequest**  request)
{
  gboolean do_exit_on_error = TRUE;
  char *tkn;
  tkn = line_tokenizer_get_word (tokenizer);
  if (tkn != NULL)
    do_exit_on_error = strtol (tkn, NULL, 10);
  info->exit_on_error = do_exit_on_error;
  return TRUE;
}

static gboolean handle_echo_command(LineTokenizer* tokenizer,
				    CliInfo*       info,
				    GdamRequest**  request)
{
  char *tkn;
  while ((tkn = line_tokenizer_get_word (tokenizer)))
    fprintf (stderr, "%s ", tkn);
  fprintf (stderr, "\n");
  return TRUE;
}



/*
 *  _          _
 * | |__   ___| |_ __
 * | '_ \ / _ \ | '_ \
 * | | | |  __/ | |_) |
 * |_| |_|\___|_| .__/
 *              |_|
 * Help command.
 */
static gboolean handle_help_command(LineTokenizer* tokenizer,
                                    CliInfo*       info,
				    GdamRequest**  request)
{
  char *topic = line_tokenizer_get_word (tokenizer);
  if (topic)
    {
      cli_help (topic);
      g_free (topic);
    }
  else
    {
      fprintf (stderr,
	       "Commands:\n\n"
	       "\tls             - list a source's information.\n"
	       "\tmixer          - get the default mixer id.\n"
	       "\tdevice         - get the default device id.\n"
	       "\tsource         - make a new sound source.\n"
	       "\tcall_class     - call a class function.\n"
	       "\tcall           - make a server-side function call.\n"
	       "\tbuffer_call    - make a server-side function call."
	       "[to a buffer]\n"
	       "\tset            - set a sound source's properties.\n"
	       "\treopen         - close/open a sound device.\n"
	       "\tdetach         - parent a source from its parent.\n"
	       "\tadd            - add a source to a mix.\n"
	       "\tsilence        - remove all sources from a mixer.\n"
	       "\thelp           - online help.\n"
	       "\tsleep #        - sleep a number of seconds.\n"
	       "\tcache name #   - cache a source to a buffer.\n"
	       "\tset_username name - set the username for chatting\n"
	       "\tchat name type msg - talk to another user\n"
	       "\n"
	       "Run `help COMMAND' for more help on one of " "the above.\n");
    }
  return TRUE;
}

/*=====================================================================*/
static void register_command_handler(const char*    command,
                                     CommandHandler handler)
{
  g_assert (g_hash_table_lookup (command_handlers, command) == NULL);
  g_hash_table_insert (command_handlers,
		       g_strdup (command), (gpointer) handler);
}
static CommandHandler lookup_command_handler(const char* command)
{
  if (!command)
    return NULL;
  return (CommandHandler) g_hash_table_lookup (command_handlers, command);
}

static gboolean make_request_from_command(const char*    command,
                                          LineTokenizer* tokenizer,
                                          CliInfo*       info,
				          GdamRequest**  request)
{
  CommandHandler handler = lookup_command_handler (command);
  if (!handler)
    {
      fprintf (stderr, "unknown command type %s.\n", command);
      return FALSE;
    }
  *request = NULL;
  return (*handler) (tokenizer, info, request);
}

/* Build default commands. */
static void initialize_default_comamnds()
{
  register_command_handler ("source", handle_source_command);
  register_command_handler ("detach", handle_detach_command);
  register_command_handler ("unref_buffer", handle_unref_buffer_command);
  register_command_handler ("unref_source", handle_unref_source_command);
  register_command_handler ("feed", handle_feed_command);
  register_command_handler ("add", handle_add_command);
  register_command_handler ("seek", handle_seek_command);
  register_command_handler ("set", handle_set_command);
  register_command_handler ("ls", handle_ls_command);
  register_command_handler ("call", handle_call_command);
  register_command_handler ("call_class", handle_call_class_command);
  register_command_handler ("buffer_call", handle_buffer_call_command);
  register_command_handler ("cache", handle_cache_command);
  register_command_handler ("silence", handle_silence_command);
  register_command_handler ("mixer", handle_mixer_command);
  register_command_handler ("device", handle_device_command);
  register_command_handler ("reopen", handle_reopen_command);
  register_command_handler ("quit", handle_exit_command);
  register_command_handler ("exit", handle_exit_command);
  register_command_handler ("help", handle_help_command);
  register_command_handler ("sleep", handle_sleep_command);
  register_command_handler ("dump", handle_dump_command);
  register_command_handler ("set_username", handle_set_username_command);
  register_command_handler ("chat", handle_chat_command);
  register_command_handler ("exit_on_error", handle_exit_on_error_command);
  register_command_handler ("echo", handle_echo_command);
}
	
static void usage(const char* prog_name)
{
  fprintf (stderr, "usage: %s [gdam-options]\n\n"
	   "Client-Line Interface to gdam.\n\n"
	   "Gdam-Options:\n"
	   "%s"
	   "\t--no-device                        Don't open a device.\n",
	   prog_name, gdam_channel_usage_message);
  exit (1);
}

static gboolean is_blank(const char* str)
{
  while (*str && isspace (*str))
    str++;
  return *str == 0 ? TRUE : FALSE;
}

static char* dollar_expand(const char* orig, GHashTable* vars)
{
  GString *string;
  const char *at;
  char *rv;
  string = g_string_new ("");
  for (at = orig; *at != '\0'; at++)
    {
      if (*at == '$' && at[1] == '{')
	{
	  char *close = strchr (at, '}');
	  char *name;
	  char *value;
	  g_assert (close != NULL);
	  name = g_new (char, close - at);
	  memcpy (name, at + 2, close - (at + 2));
	  name[close - (at + 2)] = '\0';
	  value = g_hash_table_lookup (vars, name);
	  if (value == NULL)
	    {
	      g_warning ("no variable %s", name);
	      g_string_append_c (string, '$');
	      continue;
	    }
	  g_string_append (string, value);
	  at = close;
	  continue;
	}
      if (*at == '$')
	{
	  const char *end_name = at + 1;
	  char *name;
	  char *value;
	  while (*end_name && (isalnum (*end_name) || *end_name == '_'))
	    end_name++;
	  if (end_name == at + 1)
	    {
	      g_string_append_c (string, '$');
	      continue;
	    }
	  name = (char *) alloca (end_name + 1 - at);
	  memcpy (name, at + 1, end_name - (at + 1));
	  name[end_name - (at + 1)] = '\0';
	  value = g_hash_table_lookup (vars, name);
	  if (value == NULL)
	    {
	      g_warning ("no variable %s", name);
	      g_string_append_c (string, '$');
	      continue;
	    }
	  g_string_append (string, value);
	  at = end_name - 1;
	}
      else
	{
	  g_string_append_c (string, *at);
	}
    }
  rv = string->str;
  g_string_free (string, FALSE);
  return rv;
}

void process_file(CliInfo* cli_info, const char* fname, ArgHintsTable* hints);

void process_line(CliInfo* cli_info, const char* line, ArgHintsTable* hints)
{
  LineTokenizer *tokenizer;
  char *command;
  char *expanded_line;
  GdamRequest *request;
  gboolean done_processing = TRUE;
  //tokenizer->variables = variables;
  expanded_line = dollar_expand (line, variables);
  tokenizer = line_tokenizer_new (expanded_line);
  tokenizer->hint_table = hints;

  cli_info->is_done_processing = &done_processing;

  command = line_tokenizer_get_word (tokenizer);
  if (command)
    {
      if (strcmp (command, ".") == 0)
	{
	  char *fname = line_tokenizer_get_word (tokenizer);
	  gboolean old_interactive = cli_info->interactive;
	  g_return_if_fail (fname != NULL);
	  cli_info->interactive = FALSE;
	  process_file (cli_info, fname, hints);
	  cli_info->interactive = old_interactive;
	  g_free (expanded_line);
	  return;
	}
      if (!make_request_from_command (command, tokenizer, cli_info, &request))
	{
	  g_warning ("error parsing %s", line);
	  g_free (command);
	  g_free (expanded_line);
	  line_tokenizer_destroy (tokenizer);
	  return;
	}
      if (request)
	{
	  if (!gdam_channel_process_request (cli_info->channel, request))
	    g_error ("error processing request");
	}
      g_free (command);
    }
  line_tokenizer_destroy (tokenizer);
  g_free (expanded_line);
  if (done_processing)
    g_main_iteration (FALSE);
  while (!done_processing)
    {
      if (!g_main_iteration (TRUE))
	g_error ("g_main_iteration returned FALSE");
    }
}
void process_file(CliInfo* cli_info, const char* fname, ArgHintsTable* hints)
{
  char buf[4096 * 8];
  FILE *fp;
  fp = fopen (fname, "r");
  if (!fp)
    {
      fprintf (stderr, "error opening %s\n", fname);
      return;
    }
  while (fgets (buf, sizeof (buf), fp) != NULL)
    {
      g_strchomp (buf);
      process_line (cli_info, buf, hints);
    }
  fclose (fp);
}

static void
open_device(GdamChannel*   channel,
            const char*    device_name,
	    int*           device_id,
	    int*           mixer_id)
{
  GdamRequestGetIntStruct int_val;
  g_return_if_fail (device_name != NULL);
  gdam_channel_process_request (channel,
				gdam_request_get_server_side_int
				("devices",
				 device_name,
				 gdam_request_get_int,
				 gdam_request_get_failure, &int_val));
  if (!int_val.was_success)
    {
      gdam_channel_process_request (channel,
				    gdam_request_create_sink
				    (device_name,
				     gdam_request_get_int,
				     gdam_request_get_failure, &int_val));
      if (!int_val.was_success)
	g_error ("error opening device %s", device_name);
    }
  g_assert (int_val.was_success);
  *device_id = int_val.rv;


  gdam_channel_process_request (channel,
				gdam_request_get_server_side_int
				("device-feeds",
				 device_name,
				 gdam_request_get_int,
				 gdam_request_get_failure, &int_val));
  if (!int_val.was_success)
    {
      GdamRequest **package = g_new (GdamRequest *, 3);
      int mix_id = gdam_channel_new_local_id (channel);
      int recorder_id = gdam_channel_new_local_id (channel);
      package[0] = gdam_request_create_source ("GdasMix",
					       mix_id,
					       gdam_request_get_int,
					       gdam_request_get_failure,
					       &int_val,
					       1,
					       "mode",
					       GDAM_ARG_TYPE_STRING,
					       "eternal");
      package[1] = gdam_request_create_source ("GdasRecorderExternal",
					       recorder_id,
					       NULL, NULL, NULL,
					       1,
					       "filtered",
					       GDAM_ARG_TYPE_INT, mix_id);
      package[2] = gdam_request_feed_sink (*device_id,
					   recorder_id,
					   NULL, NULL, "eternal");
      gdam_channel_process_request (channel,
				    gdam_request_package_v
				    (3, package, NULL, NULL, NULL));
      if (!int_val.was_success)
	g_error ("error opening device %s", device_name);
    }
  g_assert (int_val.was_success);
  *mixer_id = int_val.rv;
}

static void
print_chat_messages (const char *username,
		     const char *command,
		     const char *text,
		     gpointer    unused)
{
  g_message ("chat from `%s': command=%s: %s", username, command, text);
}

int main(int argc, char** argv)
{
  GdamChannel *channel;
  int device_id = 0, mixer_id = 0;
  char *line;
  char *prompt = "gdam-cli> ";
  int i;
  gboolean no_devices = FALSE;
  GSList *expressions = NULL;
  CliInfo cli_info;
  gboolean hints_inited = FALSE;
  ArgHintsTable *hint_table = NULL;
  for (i = 1; i < argc; i++)
    {
      const char *arg = argv[i];
      if (strcmp (arg, "--help") == 0
	  || strcmp (arg, "-help") == 0 || strcmp (arg, "-h") == 0)
	usage (argv[0]);
      if (strcmp (arg, "-v") == 0 || strcmp (arg, "--version") == 0)
	{
	  printf ("gdamcli version %s.\n"
		  "(C) Copyright Dave Benson, Geoff Matters. 2000.\n"
		  "\n"
		  "THIS PROGRAM COMES WITH NO WARRANTY! "
		  "USE AT YOUR OWN RISK.\n", VERSION);
	  exit (0);
	}
      if (strcmp (arg, "--no-hints") == 0)
	{
	  if (hints_inited)
	    {
	      g_error ("you may only specify --no-hints "
		       "or --hint-file once");
	    }
	  hints_inited = TRUE;
	  continue;
	}
      if (strcmp (arg, "--hint-file") == 0)
	{
	  const char *hint_file = argv[++i];
	  if (hint_file == NULL)
	    g_error ("--hint-file needs argument");
	  if (hints_inited)
	    {
	      g_error ("you may only specify --no-hints "
		       "or --hint-file once");
	    }
	  hints_inited = TRUE;
	  hint_table = arg_hints_table_load (hint_file);
	  continue;
	}
    }
  if (!hints_inited)
    hint_table = arg_hints_table_builtin ();
  initial_empty_hashtables ();
  if (!gdam_channel_init (&argc, &argv, &channel))
    {
      g_error ("error creating connection to client");
      return 1;
    }
  for (i = 1; i < argc; i++)
    {
      /*
      if (strcmp (argv[i], "--device") == 0
	  || strcmp (argv[i], "-device") == 0)
	{
	  open_device (channel, argv[++i], &device_id, &mixer_id);
	  device_opened = TRUE;
	  continue;
	}
      */
      if (strcmp (argv[i], "--no-device") == 0)
	{
	  no_devices = TRUE;
	  continue;
	}

      if (strcmp (argv[i], "-e") == 0)
	{
	  if (argv[++i] == NULL)
	    {
	      g_warning ("no command after `-e'");
	      return 1;
	    }
	  expressions = g_slist_prepend (expressions, argv[i]);
	  continue;
	}
      if (argv[i][0] == '-')
	{
	  g_warning ("unrecognized option: %s", argv[i]);
	  continue;
	}
      expressions = g_slist_prepend (expressions,
				     g_strdup_printf (". %s", argv[i]));
    }
  if (!no_devices)
    {
      char *device;
      if (channel->device_list != NULL)
        {
	  device = channel->device_list->data;
	}
      else
        {
	  device = getenv ("GDAM_DEVICE");
	}
      if (device != NULL)
        {
	  open_device (channel, device, &device_id, &mixer_id);
	  //device_opened = TRUE;
	}
      else
	{
	  g_message ("Using /dev/dsp since no --device was given "
		     "but --no-device was not specified");
	  open_device (channel, "/dev/dsp", &device_id, &mixer_id);
	  //device_opened = TRUE;
	}
    }

  cli_info.interactive = 1;
  cli_info.exit_on_error = 0;
  cli_info.device_id = device_id;
  cli_info.mixer_id = mixer_id;
  cli_info.channel = channel;
  cli_info.id_to_type = g_hash_table_new (NULL, NULL);
  cli_info.const_type_names = g_string_chunk_new (2048);
  g_hash_table_insert (variables, "mixer", g_strdup_printf ("%d", mixer_id));
  g_hash_table_insert (variables,
		       "device", g_strdup_printf ("%d", device_id));

  gdam_channel_trap_chat_messages (channel, NULL, NULL, print_chat_messages,
				   NULL, NULL);

  initialize_default_comamnds ();

  if (expressions)
    {
      GSList *list;
      for (list = expressions; list; list = list->next)
	{
	  char *line = list->data;
	  process_line (&cli_info, line, hint_table);
	}
      return 0;
    }


  while ((line = getline (prompt)) != NULL)
    {
      g_strchomp (line);
      if (!is_blank (line))
	gl_histadd ((char *) line);
      process_line (&cli_info, line, hint_table);
    }

  return 0;
}
