/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamconnect.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <netdb.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#ifndef MAX_UNIX_PATH
#define MAX_UNIX_PATH		108
#endif

/*
 * Lookup a computer's ip address from its host name.
 */
static gboolean gdam_host_name_lookup(const char* host, void* ip_addr_out)
{
  struct hostent *host_ent;
  host_ent = gethostbyname (host);
  if (!host_ent)
    {
      g_warning ("error doing dns lookup of %s: %s",
		 host, hstrerror (h_errno));
      return FALSE;
    }
  /* IPv4 requirement... */
  g_assert (host_ent->h_length == 4);
  memcpy (ip_addr_out, host_ent->h_addr_list[0], 4);
  return TRUE;
}
/* Returns 0 on error. */
static int gdam_parse_port_number(const char* port_spec)
{
  if (isdigit (*port_spec))
    {
      char *ends;
      int port_no = (int) strtol (port_spec, &ends, 10);
      if ((*ends != '\0' && !isspace (*ends)) || ends == port_spec)
	{
	  g_warning ("error parsing port number from `%s'", port_spec);
	  return 0;
	}
      return port_no;
    }

  /* TODO: navigate /etc/services. */
  g_warning ("ports must be specified by number for now...");
  return 0;
}

/*
 * Pack a `struct sockaddr_in' from the host and port.
 */
static gboolean gdam_assemble_tcp_socket_addr(const char*         host,
                                              int                 port,
					      struct sockaddr_in* addr)
{
  memset (addr, 0, sizeof (struct sockaddr_in));
  if (!gdam_host_name_lookup (host, &addr->sin_addr))
    return FALSE;
  addr->sin_family = AF_INET;
  addr->sin_port = htons ((unsigned short) port);
  return TRUE;
}

/*
 * Parse a host:port style address.
 */
static gboolean gdam_parse_tcp_socket_address(const char* addr_spec,
                                              struct sockaddr_in* addr)
{
  const char *colon;
  char *host;
  int port;
  colon = strchr (addr_spec, ':');
  if (!colon)
    {
      g_warning ("address spec `%s' lacked a colon (for port number)",
		 addr_spec);
      return FALSE;
    }
  host = (char *) memcpy (alloca (colon - addr_spec + 1),
			  addr_spec, colon - addr_spec);
  host[colon - addr_spec] = '\0';
  port = gdam_parse_port_number (colon + 1);
  if (!port)
    {
      g_warning ("couldn't parse port number from %s", colon + 1);
      return FALSE;
    }
  return gdam_assemble_tcp_socket_addr (host, port, addr);
}

/*
 * Perform a connection to an arbitrary sockaddr,
 * which may be unix or tcp style.
 *
 * If *pending is set, the connect must be polled and "gdam_connect_finished".
 */
static int
gdam_connect_sockaddr(struct sockaddr* inaddr, gboolean* pending)
{
  int fd;
  int flags;
  int protocol_family;
  int size;
  if (inaddr->sa_family == AF_INET)
    {
      protocol_family = PF_INET;
      size = sizeof (struct sockaddr_in);
    }
  else if (inaddr->sa_family == AF_UNIX)
    {
      protocol_family = PF_UNIX;
      size = sizeof (struct sockaddr_un);
    }
  else
    {
      protocol_family = inaddr->sa_family;
      size = sizeof (struct sockaddr);
    }
  fd = socket (protocol_family, SOCK_STREAM, 0);
  if (fd < 0)
    {
      g_warning ("error creating socket: %s", strerror (errno));
      return -1;
    }
  flags = fcntl (fd, F_GETFL);
  flags |= O_NONBLOCK;
  if (fcntl (fd, F_SETFL, flags) < 0)
    {
      g_warning ("error setting backend fd non-blocking: %s",
		 strerror (errno));
      close (fd);
      return -1;
    }
retry_connect:
  if (connect (fd, inaddr, size) < 0)
    {
      if (errno == EINTR)
	goto retry_connect;
      if (errno != EINPROGRESS && errno != EWOULDBLOCK)
	{
	  g_warning ("error connecting:  %s", strerror (errno));
	  close (fd);
	  return -1;
	}
      *pending = TRUE;
    }
  else
    {
      *pending = FALSE;
    }
  flags &= (~O_NONBLOCK);
  fcntl (fd, F_SETFL, flags);
  return fd;
}

/*
 * Connect to an arbitrary "address spec".
 *
 * TODO:  allow connections to "unix:/path".
 */
GdamConnectResult gdam_connect_generic(const char* addr_spec, int* returned_fd)
{
  struct sockaddr_in inaddr;
  gboolean pending;
  int fd;

  if (!gdam_parse_tcp_socket_address (addr_spec, &inaddr))
    return GDAM_CONNECT_FAILED;
  fd = gdam_connect_sockaddr ((struct sockaddr *) &inaddr, &pending);
  if (fd < 0)
    return GDAM_CONNECT_FAILED;
  *returned_fd = fd;
  return pending ? GDAM_CONNECT_CONNECTING : GDAM_CONNECT_CONNECTED;
}

/*
 * Connect to a host/port pair.
 */
GdamConnectResult gdam_connect_tcp(const char* host, int port, int* returned_fd)
{
  struct sockaddr_in inaddr;
  gboolean pending;
  int fd;

  if (!gdam_assemble_tcp_socket_addr (host, port, &inaddr))
    return GDAM_CONNECT_FAILED;
  fd = gdam_connect_sockaddr ((struct sockaddr *) &inaddr, &pending);
  if (fd < 0)
    return GDAM_CONNECT_FAILED;
  *returned_fd = fd;
  return pending ? GDAM_CONNECT_CONNECTING : GDAM_CONNECT_CONNECTED;
}


/*
 * Connect to a host/port pair.
 */
GdamConnectResult gdam_connect_unix(const char* path, int* returned_fd)
{
  struct sockaddr_un un_addr;
  int fd;
  gboolean pending;
  un_addr.sun_family = AF_UNIX;
  un_addr.sun_path[MAX_UNIX_PATH - 1] = 0;
  strncpy (un_addr.sun_path, path, MAX_UNIX_PATH);
  fd = gdam_connect_sockaddr ((struct sockaddr *) &un_addr, &pending);
  if (fd < 0)
    return GDAM_CONNECT_FAILED;
  *returned_fd = fd;
  if (pending) {
    return GDAM_CONNECT_CONNECTING;
  }
  return GDAM_CONNECT_CONNECTED;
}

gboolean gdam_connect_finish(int fd, gboolean verbose)
{
  int err = 0;
  int len = sizeof (int);

  if (getsockopt (fd, SOL_SOCKET, SO_ERROR, (void *) &err, &len) < 0 || err)
    {
      if (verbose)
	g_warning ("error connecting: %s", strerror (err));
      return FALSE;
    }
  return TRUE;
}
