/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdambindingfile.h"
#include "gdaminputdevice.h"
#include "gdammodel.h"
#include <stdlib.h>

/*                     The binding interface documentation :)

<Gdam-Binding-Interface>
  <constant type="GdamGain">
    <argument>muted</argument>
    <value>1</value>
    <GdamMidiEventSysex>
      <sysex>0x01 0x02 0x03 0x04</sysex>
      <device>/dev/midi</device>
    </GdamMidiEventSysex>
  </constant>

  <range type="GdamGain">
    <argument>volume</argument>
    <GdamMidiEventPitchWheel>
      <input_id>3</input_id>
      <device>/dev/midi</device>
    </GdamMidiEventPitchWheel>
    <min_value>0.0</min_value>
    <max_value>2.0</max_value>
  </range>

  <signal type="GdamPacker">
    <name>whatever</name>
    <handler>gdam_packer_trap_key_press:e:tt1:gdam_skin_arg_int:play_or_synch:-1</handler>
  </signal>

</Gdam-Binding-Interface>

*/

static const char* get_string_content(xmlNode* node)
{
  if (node->content)
    return node->content;
  if (node->childs == NULL)
    return NULL;
  return node->childs->content;
}

GdamInputEvent* parse_input_event(xmlNode* node)
{
  GdamInputEventTypeInfo *type_info;
  int type_id;
  const char *name = node->name;
  GHashTable *table;
  xmlNode *child;
  GdamInputEvent *input_event;
  g_return_val_if_fail (name != NULL, NULL);
  type_id = gdam_input_event_get_type_id_by_name (name);
  if (type_id == 0)
    {
      for (child = node->childs; child != NULL; child = child->next)
	if (strcmp (child->name, "device") == 0)
	  break;
      if (child != NULL)
	{
	  const char *content = child->content;
	  if (content == NULL && child->childs)
	    content = child->childs->content;
	  if (content == NULL)
	    {
	      g_warning ("<device> didn't have text");
	      return NULL;
	    }
	  gdam_input_device_get_by_name (content);
	  type_id = gdam_input_event_get_type_id_by_name (name);
	}
    }

  if (type_id == 0)
    {
      g_warning ("no input event type named `%s' was found", name);
      return NULL;
    }

  g_return_val_if_fail (type_id > 0, NULL);

  table = g_hash_table_new (g_str_hash, g_str_equal);
  for (child = node->childs; child != NULL; child = child->next)
    {
      const char *content;
      if (child->name == NULL)
	{
	  g_warning ("child of <%s> had no xml-name", name);
	  return NULL;
	}
      content = get_string_content (child);
      if (content == NULL)
	{
	  g_warning ("child %s of <%s> had no string content",
		     child->name, name);
	  return NULL;
	}
      g_hash_table_insert (table, (char *) child->name, (char *) content);
    }
  type_info = gdam_input_event_get_type_info_by_name (name);
  input_event = (*type_info->from_keyvalues) (type_id, table);
  g_hash_table_destroy (table);
  return input_event;
}

static gboolean parse_arg(GdamArg*    to_parse,
                          GtkObject*  object,
                          const char* arg_name,
			  xmlNode*    value_node)
{
  int type;
  const char *content;
  gboolean rv;

  type = gdam_arg_query_type (object->klass->type, arg_name);
  if (type == 0)
    {
      g_warning ("no argument %s was found in %s",
		 arg_name, gtk_type_name (object->klass->type));
      return FALSE;
    }
  to_parse->name = g_strdup (arg_name);
  to_parse->type = type;
  if (value_node == NULL)
    {
      GDAM_ARG_DOUBLE (to_parse) = 0.0;
      return TRUE;
    }
  content = get_string_content (value_node);
  if (content == NULL)
    {
      g_warning ("<value> for <%s> had no string content", arg_name);
      return FALSE;
    }
  rv = gdam_arg_parse_val_from_string (to_parse, content);
  if (!rv)
    g_warning ("couldn't parse arg '%s' from content '%s'", arg_name, content);
  return rv;
}

static GList *global_binding_file_node_list = NULL;

GList *
gdam_binding_file_node_type_list ()
{
  return global_binding_file_node_list;
}

/* Register a new type of input binding. */
gboolean 
gdam_binding_file_node_type_register (GdamBindingNodeTypeInfo *info)
{
  global_binding_file_node_list = g_list_append (global_binding_file_node_list,
      						 info);
  g_message ("BINDING NODE TYPE '%s' REGISTERED (node_list %p)", 
             info->node_name, global_binding_file_node_list);
  return TRUE;
}

/* Binds an input event which produces a value (ie Range, Int) and drives an
 * argument.  The event value is multiplied by <factor>, if <is_delta> then
 * result is added to arg's value, else arg's value set to result.
 */
gboolean gdam_binding_parse_variable_xml_node (xmlNode*       const_node,
                                               GtkObject*     object)
{
  xmlNode *node;
  xmlNode *children = const_node->childs;
  GdamArg arg;
  const char *arg_name = NULL;
  const char *is_delta = NULL;
  const char *factor = NULL;
  GdamInputEvent *input_event = NULL;
  guint input_hook;

  g_return_val_if_fail (object != NULL, FALSE);

  for (node = children; node != NULL; node = node->next)
    {
      if (strcmp (node->name, "argument") == 0)
	{
	  arg_name = get_string_content (node);
	  continue;
	}
      if (strcmp (node->name, "is_delta") == 0)
	{
	  is_delta = get_string_content (node);
	  g_return_val_if_fail (is_delta != NULL, FALSE);
	  continue;
	}
      if (strcmp (node->name, "factor") == 0)
	{
	  factor = get_string_content (node);
	  g_return_val_if_fail (factor != NULL, FALSE);
	  continue;
	}
      input_event = parse_input_event (node);
      if (input_event == NULL)
	{
	  g_warning ("couldn't parse input event from "
		     "node named `%s'", node->name);
	  return FALSE;
	}
    }
  if (input_event == NULL)
    {
      g_warning ("<variable> must have GdamInputIntEvent or kin");
      return FALSE;
    }
  if (arg_name == NULL)
    {
      g_warning ("<variable> was missing <argument>");
      return FALSE;
    }
  if (!parse_arg (&arg, object, arg_name, NULL))
    {
      g_warning ("<variable> had error parsing value (or name)");
      return FALSE;
    }
  input_hook = gdam_input_hook_connect_variable
    (gdam_input_trap_find_or_create (input_event), object, &arg, 
     factor ? strtod (factor, NULL) : 1.0, 
     is_delta ? strtod (is_delta, NULL) : 0);
  return TRUE;
}



gboolean gdam_binding_parse_constant_xml_node (xmlNode*       const_node,
                                               GtkObject*     object)
{
  xmlNode *node;
  xmlNode *children = const_node->childs;
  GdamArg arg;
  const char *arg_name = NULL;
  const char *is_delta = NULL;
  const char *is_factor = NULL;
  xmlNode *value_node = NULL;
  GdamInputEvent *input_event = NULL;
  guint input_hook;

  if (object == NULL)
    return FALSE;

  for (node = children; node != NULL; node = node->next)
    {
      if (strcmp (node->name, "argument") == 0)
	{
	  arg_name = get_string_content (node);
	  continue;
	}
      if (strcmp (node->name, "value") == 0)
	{
	  value_node = node->childs;
	  continue;
	}
      if (strcmp (node->name, "is_delta") == 0)
	{
	  is_delta = get_string_content (node);
	  g_return_val_if_fail (is_delta != NULL, FALSE);
	  continue;
	}
      if (strcmp (node->name, "is_factor") == 0)
	{
	  is_factor = get_string_content (node);
	  g_return_val_if_fail (is_factor != NULL, FALSE);
	  continue;
	}
      input_event = parse_input_event (node);
      if (input_event == NULL)
	{
	  g_warning ("couldn't parse input event from "
		     "node named `%s'", node->name);
	  return FALSE;
	}
    }
  if (input_event == NULL)
    {
      g_warning ("constant must have GdamInputEvent or kin");
      return FALSE;
    }
  if (arg_name == NULL || value_node == NULL)
    {
      g_warning ("<constant> was missing <argument> and/or <value>");
      return FALSE;
    }
  if (!parse_arg (&arg, object, arg_name, value_node))
    {
      g_warning ("<constant> had error parsing value (or name)");
      return FALSE;
    }
  input_hook = gdam_input_hook_connect_const_full
    (gdam_input_trap_find_or_create (input_event), object, &arg,
     is_delta ? strtod (is_delta, NULL) : 0,
     is_factor ? strtod (is_factor, NULL) : 0);
  return TRUE;
}

gboolean gdam_binding_parse_range_xml_node    (xmlNode*       range_node,
					       GtkObject*     object)
{
  xmlNode *node;
  GdamArg arg;
  xmlNode *children = range_node->childs;
  const char *arg_name = NULL;
  const char *min_value = NULL;
  const char *max_value = NULL;
  const char *is_logarithmic = "0";
  const char *is_exponent = "0";
  GdamInputEvent *input_event = NULL;
  guint input_hook;
  gboolean is_double;

  if (object == NULL)
    return FALSE;

  for (node = children; node != NULL; node = node->next)
    {
      if (strcmp (node->name, "argument") == 0)
	{
	  arg_name = get_string_content (node);
	  continue;
	}
      if (strcmp (node->name, "min_value") == 0)
	{
	  min_value = get_string_content (node);
	  g_return_val_if_fail (min_value != NULL, FALSE);
	  continue;
	}
      if (strcmp (node->name, "max_value") == 0)
	{
	  max_value = get_string_content (node);
	  g_return_val_if_fail (max_value != NULL, FALSE);
	  continue;
	}
      if (strcmp (node->name, "is_logarithmic") == 0)
	{
	  is_logarithmic = get_string_content (node);
	  g_return_val_if_fail (is_logarithmic != NULL, FALSE);
	  continue;
	}
      if (strcmp (node->name, "is_exponent") == 0)
	{
	  is_exponent = get_string_content (node);
	  g_return_val_if_fail (is_exponent != NULL, FALSE);
	  continue;
	}
      input_event = parse_input_event (node);
      if (input_event == NULL)
	{
	  g_warning ("couldn't parse input event from "
		     "node named `%s'", node->name);
	  return FALSE;
	}
    }
  if (input_event == NULL)
    {
      g_warning ("range must have GdamInputEvent or kin");
      return FALSE;
    }
  if (arg_name == NULL)
    {
      g_warning ("<range> was missing <argument>");
      return FALSE;
    }
  if (min_value == NULL && max_value == NULL)
    {
      gdam_debug ("assuming %s ranges from 0.0 to 1.0", arg_name);
      min_value = "0.0";
      max_value = "1.0";
    }
  if (min_value == NULL || max_value == NULL || is_logarithmic == NULL)
    {
      g_warning ("min_value or max_value was not specified to %s", arg_name);
      return FALSE;
    }

  if (!parse_arg (&arg, object, arg_name, NULL))
    {
      g_warning ("<range> had error parsing arg name");
      return FALSE;
    }
  if (arg.type != GDAM_ARG_TYPE_DOUBLE 
   && arg.type != GDAM_ARG_TYPE_INT)
    {
      g_warning ("argument %s of %s was not an int or double",
		 arg_name, gtk_type_name (object->klass->type));
      return FALSE;
    }

  input_hook = gdam_input_hook_connect_range
			  (gdam_input_trap_find_or_create (input_event),
			   object,
			   arg_name, 
			   strtod (min_value, NULL), 
			   strtod (max_value, NULL),
			   strtod (is_logarithmic, NULL),
			   strtod (is_exponent, NULL)
			   );
  return TRUE;
}



gboolean gdam_binding_file_process      (xmlNode*       node_list, 
                                         GdamModel*     model)
{
  xmlNode *node;
  GtkObject *object;
  int processed = 0;
  object = GTK_OBJECT (model);

  for (node = node_list; node != NULL; node = node->next)
    {
      const char *name;
      const char *property;
      name = node->name;

      property = xmlGetProp (node, "type");
      if (property != NULL)
	{
	  GtkType allow_type;	/* type this tag allows */
	  GtkType model_type = GTK_OBJECT (model)->klass->type;
	  allow_type = gtk_type_from_name (property);
	  if (allow_type == 0 || !gtk_type_is_a (model_type, allow_type))
	    {
	      gdam_debug ("binding: skipping `%s': not a `%s'",
			  gtk_type_name (model_type), property);
	      return FALSE;
	    }
	}


      if (strcmp (name, "constant") == 0)
	{
	  if (gdam_binding_parse_constant_xml_node (node, object))
	    processed++;
	}
      else if (strcmp (name, "range") == 0)
	{
	  if (gdam_binding_parse_range_xml_node (node, object))
	    processed++;
	}
      else if (strcmp (name, "variable") == 0)
	{
	  if (gdam_binding_parse_variable_xml_node (node, object))
	    processed++;
	}
      else
	{
	  g_warning ("input binding tag %s is invalid", name);
	  return FALSE;
	}
    }
  if (processed > 0)
    gdam_debug ("Got %d bindings", processed);
  return TRUE;
}


static gboolean parse_description_xml_node (xmlNode*       children)
{
  xmlNode *node;
  const char *description = NULL;
  GdamInputEvent *input_event = NULL;
  GdamInputTrap *trap;
  for (node = children; node != NULL; node = node->next)
    {
      if (strcmp (node->name, "description") == 0)
	{
	  description = get_string_content (node);
	  continue;
	}
      if (input_event != NULL)
	{
	  g_warning ("only one input event may be specified -- "
		     "skipping %s", node->name);
	  continue;
	}
      input_event = parse_input_event (node);
      if (input_event == NULL)
	{
	  g_warning ("couldn't parse input event from "
		     "node named `%s'", node->name);
	  return FALSE;
	}
    }
  if (input_event == NULL)
    {
      g_warning ("trap must have GdamInputEvent or kin");
      return FALSE;
    }
  if (description == NULL)
    {
      g_warning ("<trap> was missing <description>");
      return FALSE;
    }
  trap = gdam_input_trap_find_or_create (input_event);
  gdam_input_trap_set_name (trap, description);
  return TRUE;
}

gboolean gdam_description_file_process(xmlNode*      node_list)
{
  xmlNode *node;
  int processed = 0;

  for (node = node_list; node != NULL; node = node->next)
    {
      const char *name;
      name = node->name;
      if (strcmp (name, "trap") == 0)
	{
	  if (!parse_description_xml_node (node->childs))
	    g_warning ("parsing of `trap' tag failed");
	  else
	    processed++;
	}
      else
	{
	  g_warning ("not `trap': %s", name);
	  return FALSE;
	}
    }
  if (processed > 0)
    gdam_debug ("Got %d descriptions", processed);
  return TRUE;
}
