/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamechobase.h"
#include <math.h>
#include <gdam/common/gdamsample.h>

/* 
 *    ____ _                 ___       _ _   
 *   / ___| | __ _ ___ ___  |_ _|_ __ (_) |_ 
 *  | |   | |/ _` / __/ __|  | || '_ \| | __|
 *  | |___| | (_| \__ \__ \  | || | | | | |_ 
 *   \____|_|\__,_|___/___/ |___|_| |_|_|\__|
 */

/* Functions for echo base class */
static gboolean gdam_echo_base_real_refresh_echo_array(GdamEchoBase* echo_base);

/* Functions for filter class */
static void gdam_echo_base_on_child_set_beat_info(GdamFilter*           filter,
					     const GdamBeatInfo*   beat_info);
static int gdam_echo_base_append_start_request(GdamFilter* filter,
				               int new_source_id,
				               GdamRequestBuilder* builder);


/* These functions are used to implement the 'reverb' arg.  Although it is
 * an int type, we interpret it as a boolean.
 */
static void gdam_echo_base_arg_set_reverb(GtkObject* object, GdamArg* arg)
{
  GdamEchoBase *echo_base = GDAM_ECHO_BASE (object);
  echo_base->do_reverb = GDAM_ARG_INT (arg);
  gdam_echo_base_refresh_echo_array (echo_base);
}

static void gdam_echo_base_arg_get_reverb(GtkObject* object, GdamArg* arg)
{
  GdamEchoBase *echo_base = GDAM_ECHO_BASE (object);
  GDAM_ARG_INT (arg) = echo_base->do_reverb;
}

static void 
gdam_echo_base_arg_set_passthrough_volume(GtkObject* object, GdamArg* arg)
{
  GdamEchoBase *echo_base = GDAM_ECHO_BASE (object);
  echo_base->passthrough_volume = GDAM_ARG_DOUBLE (arg);
  gdam_echo_base_refresh_echo_array (echo_base);
}

static void 
gdam_echo_base_arg_get_passthrough_volume(GtkObject* object, GdamArg* arg)
{
  GdamEchoBase *echo_base = GDAM_ECHO_BASE (object);
  GDAM_ARG_DOUBLE (arg) = echo_base->passthrough_volume;
}

static void 
gdam_echo_base_arg_set_wet_volume(GtkObject* object, GdamArg* arg)
{
  GdamEchoBase *echo_base = GDAM_ECHO_BASE (object);
  echo_base->wet_volume = GDAM_ARG_DOUBLE (arg);
  gdam_echo_base_refresh_echo_array (echo_base);
}

static void 
gdam_echo_base_arg_set_depth_crossfade (GtkObject* object, GdamArg* arg_in)
{
  GdamEchoBase *echo_base = GDAM_ECHO_BASE (object);
  gdouble fade_amt = GDAM_ARG_DOUBLE (arg_in);
  GdamArg arg;
  gdouble pi = 3.14159265358979;

  arg.name = "passthrough_volume";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = cos(pi/2 * fade_amt);
  gdam_arg_set_object_arg (GTK_OBJECT (echo_base), &arg);
  arg.name = "wet_volume";
  GDAM_ARG_DOUBLE (&arg) = sin(pi/2 * fade_amt);
  gdam_arg_set_object_arg (GTK_OBJECT (echo_base), &arg);
}


static void 
gdam_echo_base_arg_get_wet_volume(GtkObject* object, GdamArg* arg)
{
  GdamEchoBase *echo_base = GDAM_ECHO_BASE (object);
  GDAM_ARG_DOUBLE (arg) = echo_base->wet_volume;
}


/* Preserve server-side buffer with a zero-volume echo.  Unit is beats. */
static void gdam_echo_base_arg_set_dummy_beat (GtkObject *object, GdamArg *arg)
{
  GdamEchoBase *echo_base = GDAM_ECHO_BASE (object);
  echo_base->dummy_beat = GDAM_ARG_DOUBLE (arg);
  gdam_echo_base_refresh_echo_array (echo_base);
}

static void gdam_echo_base_arg_get_dummy_beat (GtkObject *object, GdamArg *arg)
{
  GdamEchoBase *echo_base = GDAM_ECHO_BASE (object);
  GDAM_ARG_DOUBLE (arg) = echo_base->dummy_beat;
}


/* This function is called once when the GdamEchoBaseClass is created.
 * We register our args and assign class functions.  
 */
static void gdam_echo_base_class_init(GdamEchoBaseClass* klass)
{
  GdamFilterClass *filter_class = GDAM_FILTER_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  klass->refresh_echo_array = gdam_echo_base_real_refresh_echo_array;
  filter_class->on_child_set_beat_info = gdam_echo_base_on_child_set_beat_info;
  filter_class->append_start_request = gdam_echo_base_append_start_request;
  gdam_arg_register (object_class,
		     "reverb",
		     GDAM_ARG_TYPE_INT,
		     gdam_echo_base_arg_set_reverb,
		     gdam_echo_base_arg_get_reverb,
		     NULL, "boolean", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "passthrough_volume",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_echo_base_arg_set_passthrough_volume,
		     gdam_echo_base_arg_get_passthrough_volume,
		     NULL, "0,100", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "wet_volume",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_echo_base_arg_set_wet_volume,
		     gdam_echo_base_arg_get_wet_volume,
		     NULL, "0,100", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "depth_crossfade",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_echo_base_arg_set_depth_crossfade,
		     NULL,
		     NULL, "0.0,1.0", 0);
  gdam_arg_register (object_class,
		     "dummy_beat",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_echo_base_arg_set_dummy_beat,
		     gdam_echo_base_arg_get_dummy_beat,
		     NULL, "0,", GDAM_ARG_VITAL);
}

/* This function is called any time an EchoBase class or deriver class
 * is created.  It is responsible for setting default values.
 */
static void gdam_echo_base_init(GdamEchoBase* echo)
{
  echo->num_echoes = 1;
  echo->echoes = g_renew (gdouble, echo->echoes, echo->num_echoes * 2);
  echo->echoes[0] = (int) GdamFixed_ONE;
  echo->echoes[1] = 0;
  echo->wet_volume = 1.0;
}


/* This function defines the GdamEchoBase class, deriving from GdamFilter. */
GtkType gdam_echo_base_get_type()
{
  static GtkType echo_base_type = 0;
  if (!echo_base_type)
    {
      static const GtkTypeInfo echo_info = {
	"GdamEchoBase",
	sizeof (GdamEchoBase),
	sizeof (GdamEchoBaseClass),
	(GtkClassInitFunc) gdam_echo_base_class_init,
	(GtkObjectInitFunc) gdam_echo_base_init,
	/* reserved 1 */ NULL,
	/* reserved 2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      echo_base_type = gtk_type_unique (GDAM_TYPE_FILTER, &echo_info);
    }
  return echo_base_type;
}

/*
 *     ____      _ _ _                _        
 *    / ___|__ _| | | |__   __ _  ___| | _____ 
 *   | |   / _` | | | '_ \ / _` |/ __| |/ / __|
 *   | |__| (_| | | | |_) | (_| | (__|   <\__ \
 *    \____\__,_|_|_|_.__/ \__,_|\___|_|\_\___/
 */

/* on_child_beat_info is a function of GdamFilter.  This is called when
 * our child's beat info changes.  This usually means that the tempo (bpm)
 * of our sound has changed.  As many derivers base their echo arrays on
 * the tempo, we give them the opportunity to rebuild their echo arrays.
 */
static void gdam_echo_base_on_child_set_beat_info(GdamFilter*           filter,
			                     const GdamBeatInfo*   beat_info)
{
  gdam_model_set_beat_info (GDAM_MODEL (filter), beat_info);
  if (!gdam_echo_base_refresh_echo_array (GDAM_ECHO_BASE (filter)))
    g_warning ("error refreshing echo array");
}

/* When a producer begins playing, the start requests of it and its
 * parents are packaged into a single request for efficiency.
 * This function is called with the source id of this filter's child,
 * and the request builder used to package the start requests.
 */
static int gdam_echo_base_append_start_request(GdamFilter* filter,
				               int new_source_id,
				               GdamRequestBuilder* builder)
{
  GdamEchoBase *echo_base = GDAM_ECHO_BASE (filter);
  GdamModel *model = GDAM_MODEL (filter);
  GdamRequest *request;
  gtk_object_ref (GTK_OBJECT (filter));
  /* Get a new source id for our server source.  This id is used to 
   * refer to the source on the server.  
   */
  model->source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source ("GdasEcho",
					model->source_id,
			/* callbacks for source creation */
					gdam_model_got_subsource,
					gdam_model_didnt_get_subsource,
			/* pointer for creation callbacks */
					echo_base,
			/* arguments to new source */
					2,
					"filtered",
					GDAM_ARG_TYPE_INT,
					new_source_id,
					"echoes",
					GDAM_ARG_TYPE_DOUBLE_ARRAY,
					echo_base->num_echoes * 2,
					echo_base->echoes,
					"wet_volume",
					GDAM_ARG_TYPE_DOUBLE,
					echo_base->wet_volume);
			/* XXX: reverb and passthrough volume? */
  /* append our create request to the builder */
  return gdam_request_builder_append_create (builder, request);
}

/* A helper function to check the echo array for sanity.  This was 
 * necessary when the server was less accepting of irregular echo arrays.
 * It's useful to catch bugs in models which derive from GdamEchoBase
 */
static gboolean check_echo_array(GdamEchoBase* echo_base)
{
  int i;

  for (i = 1; i < echo_base->num_echoes * 2; i += 2)
    {
      if (echo_base->echoes[i] < 0)
	{
	  g_warning ("negative time (%f) in echo array!! correcting..",
	             echo_base->echoes[i]);
          echo_base->echoes[i] = 0;
	}
      if (echo_base->echoes[i-1] > 4.0)
	{
	  g_warning ("large volume (%f) in echo array!! muting..",
	             echo_base->echoes[i-1]);
          echo_base->echoes[i-1] = 0;
	}
    }
  return TRUE;
}


/* Called when args have been set on the server side.  We check if there is
 * another request pending, and if so post it.
 */
static void set_arg_complete(gpointer data)
{
  GdamEchoBase *echo_base = GDAM_ECHO_BASE (data);
  if (echo_base->pending_request != NULL)
    {
      gdam_debug ("handling pending_request..");
      gdam_channel_append_request (GDAM_MODEL (echo_base)->channel,
				   echo_base->pending_request);
      echo_base->pending_request = NULL;
    }
  else
    {
      echo_base->making_request = FALSE;
    }
}


/* This function is called to update the echo array on the server.
 * It assumes that the echo_base->echoes field has been filled with
 * the appropriate echo data.
 */
void gdam_echo_base_update(GdamEchoBase*      echo_base)
{
  GdamModel *model = GDAM_MODEL (echo_base);
  GdamRequest *request;
  GdamArg args[4];
  int num_args;

  if (model->source_id == 0)
    return;

  if (!check_echo_array (echo_base))
    {
      g_warning ("Error, bad echo array.");
      return;
    }

  args[0].name = "echoes";
  args[0].type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
  GDAM_ARG_DOUBLE_ARRAY (args + 0).length = echo_base->num_echoes * 2;
  GDAM_ARG_DOUBLE_ARRAY (args + 0).array = echo_base->echoes;
  num_args = 1;
  if (!echo_base->ignore_reverb) 
    {
      args[num_args].name = "reverb";
      args[num_args].type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (args + num_args) = echo_base->do_reverb;
      num_args++;
    }
  if (!echo_base->ignore_passthrough) 
    {
      args[num_args].name = "passthrough_volume";
      args[num_args].type = GDAM_ARG_TYPE_DOUBLE;
      GDAM_ARG_DOUBLE (args + num_args) = echo_base->passthrough_volume;
      num_args++;
    }
  {
      args[num_args].name = "wet_volume";
      args[num_args].type = GDAM_ARG_TYPE_DOUBLE;
      GDAM_ARG_DOUBLE (args + num_args) = echo_base->wet_volume;
      num_args++;
  }
  request = gdam_request_set_args (model->source_id,
				   num_args,
				   args,
				   set_arg_complete, set_arg_complete, model);

  /*  This class implements its own decongestive arg setting.  If a request 
   *  has been passed to the server, but we haven't recieved confirmation that
   *  the request has been processed, then we leave the reuest pending until
   *  our original request has been processed.  If multiple requests are 
   *  pending, only the most recent is kept and the others are discarded.
   */
  if (echo_base->making_request)
    {
      if (echo_base->pending_request != NULL)
	gdam_request_destroy (echo_base->pending_request);
      echo_base->pending_request = request;
    }
  else
    {
      echo_base->making_request = TRUE;
      gdam_channel_append_request (model->channel, request);
    }
}


/* This is a convenience function used to update the server source.  It
 * calls the refresh_echo_array function which has been defined by the
 * deriver class to ensure that echo_base->echoes is correct, and
 * then calls echo_base_update to set those echoes in the source.
 */
gboolean gdam_echo_base_refresh_echo_array(GdamEchoBase* echo_base)
{
  GdamEchoBaseClass *klass;
  klass = GDAM_ECHO_BASE_CLASS (GTK_OBJECT (echo_base)->klass);
  if (!(*klass->refresh_echo_array) (echo_base))
    return FALSE;
  gdam_echo_base_update (echo_base);
  return TRUE;
}


/* The refresh_echo_array function is a virtual which should be overwritten
 * by derivers.  The purpose of this function is to allow each echo class
 * to build the echo_base->echoes array as is appropriate.  If this function 
 * is called, then the deriver has neglected to define a refresh_echo_array
 * function.
 */
static gboolean gdam_echo_base_real_refresh_echo_array(GdamEchoBase* echo_base)
{
  g_warning ("echo_base refresh_echo_array should not be called.");
  return FALSE;
}

