/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdammodel.h"
#include "gdamview.h"
#include <gdam/common/gdamplugin.h>
#include <gdam/client/gdamrequest.h>
#include <gtk/gtksignal.h>
#include <gdk/gdk.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

static gboolean gdam_model_real_set_beat_info  (GdamModel*   model,
	                                            const GdamBeatInfo* match);
static gboolean gdam_model_real_match_beat_info
                                               (GdamModel*   model,
	                                        const GdamBeatInfo* match);
static gboolean gdam_model_real_nudge_by_beat
                                               (GdamModel*   model,
	                                        gdouble          num_beats);
static void gdam_model_real_new_source_id  (GdamModel*   model,
	                                        int              source_id);
static void gdam_model_destroy(GtkObject* object);

static GtkObjectClass* parent_class = NULL;

typedef enum {
	SET_SOURCE_ID,
	SET_BEAT_INFO,
	DETACH_FROM_BUFFER,
	LAST_SIGNAL
} GdamModelSignals;
static guint model_signals[LAST_SIGNAL];


/*  These functions implement the 'anchored' arg.  If a model is
 *  anchored, then other models will look to it for synchronization
 *  info.  The infomanager keeps track of anchored models, so we
 *  notify it that we're the lead model when 'anchored' is set to true.
 */
static void gdam_model_arg_set_anchored(GtkObject* object, 
					GdamArg*   arg)
{
  GdamModel *model = GDAM_MODEL (object);
  model->anchored = GDAM_ARG_INT (arg);
  if (model->anchored)
    gdam_infomanager_set_lead_model (model);
}

static void gdam_model_arg_get_anchored(GtkObject* object, 
					GdamArg*   arg)
{
  GdamModel *model = GDAM_MODEL (object);
  GDAM_ARG_INT (arg) = model->anchored;
}


/*  Writes model info to xml, for debugging. */
static void gdam_model_arg_set_debug_dump(GtkObject* object, GdamArg* arg)
{
  xmlNode *xml_node;
  xmlNode *top_node;
  xmlDoc *doc;
  GdamModel *model;
  GdamXmlInfo *info;
  model = GDAM_MODEL (object);
  info = gdam_xml_info_new ();
  xml_node = gdam_model_to_xml (model, info);
  gdam_xml_info_destroy (info);
  if (xml_node == NULL)
    {
      g_warning ("couldn't xmlify");
      return;
    }
  doc = xmlNewDoc ("1.0");
  top_node = xmlNewNode (NULL, "GDAM-Model-Interface");
  xmlAddChild (top_node, xml_node);
  doc->root = top_node;
  xmlDocDump (stderr, doc);
  xmlSaveFile ("tmp.xml", doc);
  xmlFreeDoc (doc);

}

/* Called once when the GdamModelClass is defined.  We assign the
 * default member functions, register class args, and register
 * class signals.
 */
static void gdam_model_class_init(GdamModelClass* model_class)
{
  GtkObjectClass *object_class = (GtkObjectClass *) model_class;
  object_class->destroy = gdam_model_destroy;
  model_class->set_beat_info = gdam_model_real_set_beat_info;
  model_class->match_beat_info = gdam_model_real_match_beat_info;
  model_class->nudge_by_beat = gdam_model_real_nudge_by_beat;
  model_class->new_source_id = gdam_model_real_new_source_id;
  gdam_arg_register (object_class,
		     "debug_dump",
		     GDAM_ARG_TYPE_INT,
		     gdam_model_arg_set_debug_dump, NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "anchored",
		     GDAM_ARG_TYPE_INT,
		     gdam_model_arg_set_anchored,
		     gdam_model_arg_get_anchored, NULL, "boolean", 0);

  /* This is GTK_RUN_LAST so you have chance to look
   * at the old source_id... */
  model_signals[SET_SOURCE_ID] =
    gtk_signal_new ("set_source",
		    GTK_RUN_LAST | GTK_RUN_NO_HOOKS,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamModelClass,
				       new_source_id),
		    gtk_marshal_NONE__INT, GTK_TYPE_NONE, 1, GTK_TYPE_INT);
  model_signals[SET_BEAT_INFO] =
    gtk_signal_new ("set_beat_info",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamModelClass,
				       set_beat_info),
		    gtk_marshal_NONE__POINTER,
		    GTK_TYPE_NONE, 1, GTK_TYPE_POINTER);
  model_signals[DETACH_FROM_BUFFER] =
    gtk_signal_new ("detach_from_buffer",
		    GTK_RUN_LAST | GTK_RUN_NO_HOOKS,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamModelClass,
				       detach_from_buffer),
		    gtk_marshal_NONE__UINT, GTK_TYPE_NONE, 1, GTK_TYPE_UINT);
  gtk_object_class_add_signals (object_class, model_signals, LAST_SIGNAL);
}

/* Called any time a GdamModel (or deriver) is created.  Any members
 * which should be initialized to values other than zero should be
 * set here.
 */
static void gdam_model_init(GdamModel* model)
{
  /*              -- these are fine being filled with zeroes. */
  /*
     GdamModel*          parent;
     GdamBeatInfo         beat_info;
     int                     source_id;
   */
}

/* Defines the GdamModel type, as deriving from GtkObject. */
GtkType gdam_model_get_type()
{
  static GtkType model_type = 0;
  if (!model_type)
    {
      static const GtkTypeInfo model_info = {
	"GdamModel",
	sizeof (GdamModel),
	sizeof (GdamModelClass),
	(GtkClassInitFunc) gdam_model_class_init,
	(GtkObjectInitFunc) gdam_model_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GTK_TYPE_OBJECT);
      model_type = gtk_type_unique (GTK_TYPE_OBJECT, &model_info);

    }
  return model_type;
}


#if 0
gboolean       gdam_model_nudge_by_beat
                                       (GdamModel*   model,
	                                gdouble          num_beats)
{
...}
#endif


/* Function used to set a model's source id.  It sets the source_id
 * variable and emits the "set_source_id" signal.  Before emitting the
 * signal, we check the number of emissions to see if this funciton is
 * being called in response to the "set_source_id" signal.  If so, we
 * avoid an infinite loop by declining to emit.
 */
void           gdam_model_set_source_id (GdamModel*       model,
				         int              source_id)
{
  model->source_id = source_id;
  if (gtk_signal_n_emissions (GTK_OBJECT (model),
			      model_signals[SET_SOURCE_ID]))
    return;
  gtk_signal_emit (GTK_OBJECT (model),
		   model_signals[SET_SOURCE_ID], source_id);
}


/* Function used to set a model's beat information.  Similar to
 * gdam_model_set_source_id.
 */
void           gdam_model_set_beat_info (GdamModel*   model,
				         const GdamBeatInfo* info)
{
  if (gtk_signal_n_emissions (GTK_OBJECT (model),
			      model_signals[SET_BEAT_INFO]))
    return;

  gtk_signal_emit (GTK_OBJECT (model), model_signals[SET_BEAT_INFO], info);
}



typedef struct _WaitingForSource WaitingForSource;
struct _WaitingForSource {
	GdamModel*          model;
	GdamModelNotify     notify_func;
	GDestroyNotify      destroy;
	gpointer            data;
	guint               connect_id;
};

static void handle_new_source_id(WaitingForSource* waiting)
{
  g_return_if_fail (waiting->model->source_id != 0);
  (*waiting->notify_func) (waiting->model, waiting->data);
  gtk_signal_disconnect (GTK_OBJECT (waiting->model), waiting->connect_id);
}

static void free_waiting_for_source(gpointer waiting_data)
{
  WaitingForSource *waiting = waiting_data;
  if (waiting->destroy)
    (*waiting->destroy) (waiting->data);
  g_free (waiting);
}

void           gdam_model_call_when_source_ready
                                     (GdamModel*      model,
				      GdamModelNotify notify_func,
				      GDestroyNotify      destroy,
				      gpointer            data)
{

  WaitingForSource *waiting;
  g_return_if_fail (model->source_id != 0);
  if (model->source_id)
    {
      (*notify_func) (model, data);
      if (destroy != NULL)
	(*destroy) (data);
    }
  waiting = g_new (WaitingForSource, 1);
  waiting->model = model;
  waiting->notify_func = notify_func;
  waiting->destroy = destroy;
  waiting->data = data;
  waiting->connect_id
    = gtk_signal_connect_full
    (GTK_OBJECT (model), "set_source", GTK_SIGNAL_FUNC (handle_new_source_id), NULL, waiting, free_waiting_for_source, TRUE,	/* is an object signal */
     FALSE);
}

/* Private member functions. */
static gboolean gdam_model_real_set_beat_info  (GdamModel*   model,
	                                        const GdamBeatInfo* new)
{
  g_return_val_if_fail ((new != NULL), FALSE);

  /* TODO: if we are copying beat infos here, should we free them as well? */
  model->beat_info = *gdam_beat_info_copy (new);
  return TRUE;
}

/* if match_beat_info is not defined by a GdamModel's deriver, then
 * the model doesn't support it.
 */
static gboolean gdam_model_real_match_beat_info(GdamModel*   model,
	                                        const GdamBeatInfo* match)
{
  g_warning ("match beat info called on sound source "
	     "which doesn't support it");
  return FALSE;
}

/* if nudge_by_beat is not defined by a GdamModel's deriver, then
 * the model cannot be nudge by beats.
 */
static gboolean
gdam_model_real_nudge_by_beat(GdamModel*   model,
	                      gdouble          num_beats)
{
  g_warning ("nudge by beat called on sound source "
	     "which doesn't support it");
  return FALSE;
}

static void
gdam_model_real_new_source_id(GdamModel*   model,
	                      int          source_id)
{
  gdam_debug ("gdam_model_real_new_source_id, type %s id %d",
      	      gtk_type_name (GTK_OBJECT (model)->klass->type), source_id);
  model->source_id = source_id;
}


/* gdam_model_new_child_request is used to pass a request builder
 * from child to parent all the way up the model tree.  This is
 * most commonly used to start a chain of models playing, or stop 
 * them.  'child' may be NULL if this model has no child.  'play_info'
 * may be NULL if we are not starting to play.
 */
void gdam_model_new_child_request(GdamModel*          parent,
                                  GdamModel*          child,
				  GdamPlayInfo*       play_info,
				  int                 local_src_id,
				  GdamRequestBuilder* builder)
{
  GdamModelClass *klass;
  if (parent == NULL)
    {
      GdamRequest *package;
      package = gdam_request_builder_package (builder);
      if (package)
	gdam_channel_append_request (child->channel, package);
      gdam_request_builder_destroy (builder);
      return;
    }
  klass = GDAM_MODEL_CLASS (GTK_OBJECT (parent)->klass);
  g_return_if_fail (klass != NULL);
#if 0
  if (local_src_id == 0)
    {
      g_warning ("model_new_request, local_id=0, parent=%s child=%s",
	  	 gtk_type_name (GTK_OBJECT (parent)->klass->type),
	  	 gtk_type_name (GTK_OBJECT (child)->klass->type));
    }
#endif
  (*klass->new_request) (parent, child, play_info, local_src_id, builder);
}


/* Called whenever a GdamModel or deriver is destroyed.  We attempt to 
 * silence any source on the server, and free any memory we need to.
 * This function is called after our deriver's destroy func, and we in
 * turn call the destroy func of our parent class.
 */
static void gdam_model_destroy(GtkObject* object)
{
  GdamModel *model;
  model = GDAM_MODEL (object);
  gdam_debug ("gdam_model_destroy: %s", gtk_type_name (object->klass->type));

  /* Last ditch destruction... sufficient the shut the sources
   * up when their tree cannot be followed... figure out
   * how to not do this later... */
  if (model->source_id != 0)
    {
      GdamRequest *request;
      gdam_debug ("gdam_model_destroy DETACHING  source id %d",
	  	  model->source_id);
      request = gdam_request_detach_source (model->source_id,
					    NULL, NULL, NULL);
      gdam_channel_append_request (model->channel, request);
      model->source_id = 0;
    }
  (*parent_class->destroy) (object);
}

static int sort_arg_by_name(const void* arg1, const void* arg2)
{
  const GdamArg *arg_1 = arg1;
  const GdamArg *arg_2 = arg2;
  return strcmp (arg_1->name, arg_2->name);
}

void gdam_model_got_subsource(int source_id, gpointer model_ptr)
{
  GdamModel *model = GDAM_MODEL (model_ptr);
  g_return_if_fail (source_id != 0);
  model->source_id = source_id;
  gtk_object_unref (GTK_OBJECT (model));
}
	
void gdam_model_didnt_get_subsource(gpointer model_ptr)
{
  GdamModel *model = GDAM_MODEL (model_ptr);
  gtk_object_unref (GTK_OBJECT (model));
}


/*
 *                      gdam_model_set_args
 *
 * Helper function to avoid request congestion.  If one model sets
 * several args this way, use arg_index to distinguish between them.
 * Each arg request is sent to the server only if no other arg request
 * of that type is currently being processed by the server.  If we are
 * already waiting on a server request of this type, the arg request
 * becomes pending.  Only the most recent pending arg request is kept.
 */

typedef struct _GdamArgCompleteInfo GdamArgCompleteInfo;
struct _GdamArgCompleteInfo {
	int arg_index;
	GdamModel* model;
	GdamRequest* request;
};

static void set_arg_complete(gpointer data)
{
  GdamArgCompleteInfo *info = data;
  int arg_index = info->arg_index;
  GdamModel *model = info->model;
  if (info->model->pending_arg_requests[arg_index] != NULL)
    {
      gdam_debug ("handling pending_request..");
      gdam_channel_append_request
	(model->channel, model->pending_arg_requests[arg_index]);
      model->pending_arg_requests[arg_index] = NULL;
    }
  else
    {
      model->making_arg_requests[arg_index] = 0;
    }
  g_free (data);
}

void gdam_model_set_args_full(GdamModel*       model, 
			      int              source_id,
			      unsigned         arg_set,
                              int              num_args,
			      GdamArg*         args)
{
  GdamModelClass *model_class;
  GdamRequest *request;
  GdamArgCompleteInfo *complete_info;
  if (source_id == 0)
    {
      source_id = model->source_id;
      if (source_id == 0)
	return;
    }

  model_class = GDAM_MODEL_CLASS (GTK_OBJECT (model)->klass);
  g_return_if_fail (arg_set < model_class->num_arg_sets);

  model->making_arg_requests = g_new0 (guint8, model_class->num_arg_sets);
  model->pending_arg_requests = g_new0 (gpointer, model_class->num_arg_sets);


  complete_info = g_new (GdamArgCompleteInfo, 1);
  complete_info->arg_index = arg_set;
  complete_info->model = model;

  request = gdam_request_set_args (source_id,
				   num_args,
				   args,
				   set_arg_complete, set_arg_complete,
				   complete_info);

  complete_info->request = request;

  if (model->making_arg_requests[arg_set])
    {
      if (model->pending_arg_requests[arg_set] != NULL)
	{
	  GdamArgCompleteInfo *info;
	  info = model->pending_arg_requests[arg_set];
	  gdam_request_destroy (info->request);
	  g_free (info);
	}
      model->pending_arg_requests[arg_set] = complete_info;
    }
  else
    {
      model->making_arg_requests[arg_set] = 1;
      gdam_channel_append_request (model->channel, request);
    }
}

void gdam_model_set_args(GdamModel*       model, 
			 unsigned         arg_set,
                         int              num_args,
			 GdamArg*         args)
{
  gdam_model_set_args_full (model, 0, arg_set, num_args, args);
}

static gboolean arg_to_xml_node(xmlNode* parent, GdamArg* arg)
{
  char *str;
  xmlNode *node;

  str = gdam_arg_val_as_string (arg);
  node = xmlNewChild (parent, NULL, arg->name, str);
  g_free (str);

  return TRUE;
}

/*
 *                          XML descriptions of the models
 *                          and maybe the skins.
 */
static void add_xml_for_skins(GdamModel*         model,
                              xmlNode*           top_node,
			      GdamXmlInfo*       xml_info)
{
  GSList *children;
  for (children = model->skin_list;
       children != NULL; children = children->next)
    {
      GdamView *skin = GDAM_VIEW (children->data);
      xmlNode *node;
      if (gdam_xml_info_lookup_by_object (xml_info, skin) != 0)
	continue;
      node = gdam_view_to_xml (skin, xml_info);
      if (node != NULL)
	{
	  xmlNode *skin_node;
	  skin_node = xmlNewNode (NULL, "skin");
	  xmlAddChild (skin_node, node);
	  xmlAddChild (top_node, skin_node);
	}
    }
}

xmlNode* gdam_model_to_xml(GdamModel*         model,
                           GdamXmlInfo*       xml_info)
{
  GtkObject *object = GTK_OBJECT (model);
  GdamModelClass *model_class = GDAM_MODEL_CLASS (object->klass);
  xmlNode *rv;
  GdamArg *args;
  int num_args, i, id;
  const char *type_name;

  g_return_val_if_fail (gdam_xml_info_lookup_by_object (xml_info, model) == 0,
			NULL);

  id = gdam_xml_info_alloc_id (xml_info, model);

  type_name = gtk_type_name (object->klass->type);
  rv = xmlNewNode (NULL, type_name);
  {
    char buf[128];
    snprintf (buf, sizeof (buf), "%d", id);
    xmlSetProp (rv, "id", buf);
  }

  gdam_arg_get_all (object, TRUE, &num_args, &args);
  for (i = 0; i < num_args; i++)
    arg_to_xml_node (rv, args + i);
  gdam_arg_free_array (num_args, args);

  if (model_class->add_xml != NULL)
    {
      (*model_class->add_xml) (model, rv, xml_info);
    }

  add_xml_for_skins (model, rv, xml_info);
  return rv;
}

static void parse_and_add_skin(GdamModel*   model,
                               xmlNode*     skin_node,
			       GdamXmlInfo* xml_info)
{
  GtkType skin_type;
  GdamViewClass *view_class;
  GdamView *view;
  char *id_str;
  xmlNode *skin_child;

  /* skin_node->name is a GdamSkin type... */
  skin_type = gdam_plugin_get_type_by_name ("skin", skin_node->name);
  if (skin_type == 0)
    {
      g_warning ("cannot open skin %s", skin_node->name);
      return;
    }


  view_class = GDAM_VIEW_CLASS (gtk_type_class (skin_type));
  g_return_if_fail (view_class != NULL);
  view = GDAM_VIEW (gtk_type_new (skin_type));
  g_return_if_fail (view != NULL);
  if (view_class->attach_to_model != NULL)
    (*view_class->attach_to_model) (view, model);

  id_str = xmlGetProp (skin_node, "id");
  if (id_str != NULL)
    {
      gdam_xml_info_assoc (xml_info, atoi (id_str), view);
    }

  if (view_class->parse_xml == NULL)
    {
      g_warning ("we were asked to make a skin that "
		 "couldn't build itself from xml... [%s]", skin_node->name);
      return;
    }
  for (skin_child = skin_node->childs;
       skin_child != NULL; skin_child = skin_child->next)
    {
      if (!(*view_class->parse_xml) (view, skin_child, xml_info))
	{
	  g_warning ("error parsing the xml tags [%s] to "
		     "build this skin", skin_child->name);
	  gtk_object_sink (GTK_OBJECT (view));
	  return;
	}
    }
  gdam_view_add (view, model);
}

static void synchronize_skins(GdamModel* model)
{
  g_slist_foreach (model->skin_list, (GFunc) gdam_view_synchronize, NULL);
}

static gboolean gdam_model_parse_xml(GdamModel*       model,
                                     xmlNode*         children,
				     GdamXmlInfo*     xml_info)
{
  GdamArg *args;
  GdamArg *to_set;
  GtkObject *object;
  xmlNode *child;
  int num_args, num_to_set;
  GdamModelClass *comp_class;

  gboolean destroy_xml_info = FALSE;

  object = GTK_OBJECT (model);
  comp_class = GDAM_MODEL_CLASS (object->klass);
  gdam_arg_get_all (object, FALSE, &num_args, &args);
  qsort (args, num_args, sizeof (GdamArg), sort_arg_by_name);
  to_set = alloca (sizeof (GdamArg) * num_args);
  num_to_set = 0;

  if (xml_info == NULL)
    {
      xml_info = gdam_xml_info_new ();
      destroy_xml_info = TRUE;
    }


  for (child = children; child != NULL; child = child->next)
    {
      GdamArg to_set;
      if (strcmp (child->name, "skin") == 0)
	{
	  parse_and_add_skin (model, child->childs, xml_info);
	  continue;
	}

      to_set.name = (char *) child->name;
      to_set.type = gdam_arg_query_type (object->klass->type, child->name);
      if (to_set.type == 0)
	{
	  if (comp_class->parse_xml)
	    (*comp_class->parse_xml) (model, child, xml_info);
	  else
	    gdam_debug ("ignoring xml tag %s", child->name);
	  continue;
	}
      if (child->childs == NULL || child->childs->content == NULL)
	{
	  gdam_debug ("%s didn't have body to speak of", child->name);
	  continue;
	}
      if (gdam_arg_parse_val_from_string
	  (&to_set, child->childs->content) <= 0)
	{
	  g_warning ("error parsing %s to %s",
		     child->name, gtk_type_name (object->klass->type));
	  continue;
	}
      gdam_arg_set_object_arg (object, &to_set);
      num_to_set++;
    }

  synchronize_skins (model);

  if (destroy_xml_info)
    gdam_xml_info_destroy (xml_info);
  return TRUE;
}
GdamModel* gdam_model_from_xml        (xmlNode*     node,
				       GdamChannel* channel,
				       GdamXmlInfo* xml_info)
{
  GdamModel *rv;
  GtkType type;
  gboolean must_destroy_xml = FALSE;

  if (xml_info == NULL)
    {
      must_destroy_xml = TRUE;
      xml_info = gdam_xml_info_new ();
    }

  if (strcmp (node->name, "id") == 0
      && node->childs != NULL && node->childs->next == NULL)
    {
      int id = atoi (node->childs->content);
      return GDAM_MODEL (gdam_xml_info_lookup_by_id (xml_info, id));
    }


  type = gdam_plugin_get_type_by_name ("model", node->name);
  if (type == 0)
    {
      g_warning ("xmlparse: " "no model of type %s was known", node->name);
      return NULL;
    }

  rv = GDAM_MODEL (gtk_type_new (type));
  g_return_val_if_fail (rv != NULL, NULL);
  rv->channel = channel;

  if (!gdam_model_parse_xml (rv, node->childs, xml_info))
    {
      gtk_object_sink (GTK_OBJECT (rv));
      return NULL;
    }

  if (must_destroy_xml)
    {
      gdam_xml_info_destroy (xml_info);
    }

  return rv;
}

static void set_channel(GdamModel* model, GdamChannel* channel)
{
  model->channel = channel;
}

GdamModel* gdam_model_clone(GdamModel* model)
{
  GdamModel *rv;
  rv = (GdamModel *) gdam_arg_clone_object (GTK_OBJECT (model),
					    GDAM_ARG_CLONE_DEEP);
  g_return_val_if_fail (rv != NULL, NULL);
  gdam_arg_child_forall (GTK_OBJECT (rv),
			 GDAM_TYPE_MODEL,
			 (GdamArgChildForeachFunc) set_channel,
			 model->channel);
  return rv;
}

/*
 *                        _                              _      _
 *  ___ _ __   ___  ___  | |_ ___    _ __ ___   ___   __| | ___| |
 * / __| '_ \ / _ \/ __| | __/ _ \  | '_ ` _ \ / _ \ / _` |/ _ \ |
 * \__ \ |_) |  __/ (__  | || (_) | | | | | | | (_) | (_| |  __/ |
 * |___/ .__/ \___|\___|  \__\___/  |_| |_| |_|\___/ \__,_|\___|_|
 *     |_|
 */

/* this can be used to quickly fabricate a model from a string;
 *       GdamBuilder{component=GdamTurntable{producer=GdamProducerMp3}}{filter#0=GdamGain}
 *
 * syntax for arguments:
 *       GdamProducerMp3(filename=/mp3/???)
 * parentheses-grouped and brace-grouped may be mixed,
 * with no intervening punctuation.
 */
typedef struct _SpecNodeParen SpecNodeParen;
typedef struct _SpecNodeBrace SpecNodeBrace;
typedef struct _SpecNode SpecNode;

struct _SpecNodeParen {
	char*		key;
	char*		value;
};
struct _SpecNodeBrace {
	char*		key;
	SpecNode*	value;
};
struct _SpecNode {
	char*		class_name;
	GSList*		paren_nodes;
	GSList*		brace_nodes;
};

#define L_BRACE		'{'
#define R_BRACE		'}'
#define L_PAREN		'('
#define R_PAREN		')'

static SpecNode* parse_spec(const char* spec);
static void spec_node_free(SpecNode* node);
static GtkObject* spec_node_to_object(SpecNode* node, GdamChannel*);


static SpecNodeParen* parse_spec_paren(const char* start, const char* match)
{
  const char *equal;
  char *key;
  char *value;
  SpecNodeParen *paren_node;

  equal = (char *) memchr ((char *) start, '=', match - start);
  if (equal == NULL)
    {
      g_warning ("no = found in () subexpression in spec");
      return NULL;
    }

  start++;
  key = g_new (char, equal - start + 1);
  memcpy (key, start + 1, equal - start);
  key[equal - start] = '\0';

  equal++;
  value = g_new (char, match - equal + 1);
  memcpy (value, equal + 1, match - equal);
  value[match - equal] = '\0';

  paren_node = g_new (SpecNodeParen, 1);
  paren_node->key = key;
  paren_node->value = value;
  return paren_node;
}

static SpecNodeBrace* parse_spec_brace(const char* start, const char* match)
{
  char *tmp;
  const char *equal;
  char *key;
  SpecNode *sub;
  SpecNodeBrace *brace;

  equal = (char *) memchr ((char *) start, '=', match - start);
  if (equal == NULL)
    {
      g_warning ("no = found in {} subexpression in spec");
      return NULL;
    }

  start++;
  key = g_new (char, equal - start + 1);
  memcpy (key, start, equal - start);
  key[equal - start] = '\0';

  start = equal + 1;
  match--;
  tmp = g_new (char, match - start + 1);
  memcpy (tmp, start, match - start);
  tmp[match - start] = '\0';
  sub = parse_spec (tmp);
  g_free (tmp);
  if (sub == NULL)
    return NULL;
  brace = g_new (SpecNodeBrace, 1);
  brace->key = key;
  brace->value = sub;
  return brace;
}

static void spec_node_paren_free(SpecNodeParen* paren)
{
  g_free (paren->key);
  g_free (paren->value);
  g_free (paren);
}
static void spec_node_brace_free(SpecNodeBrace* brace)
{
  g_free (brace->key);
  spec_node_free (brace->value);
  g_free (brace);
}
static const char* find_matching(const char* input)
{
  char upchar = *input;
  char downchar;
  int count;
  if (upchar == L_BRACE)
    {
      downchar = R_BRACE;
    }
  else if (upchar == L_PAREN)
    {
      downchar = R_PAREN;
    }
  else
    {
      g_warning ("cannot match `%c'", upchar);
      return NULL;
    }
  input++;
  count = 1;
  while (*input != '\0' && count > 0)
    {
      if (*input == upchar)
	count++;
      else if (*input == downchar)
	count--;

      input++;
    }
  if (count == 0)
    return input;
  return NULL;
}

static SpecNode* parse_spec(const char* spec)
{
  const char *at = spec;
  GSList *paren_nodes = NULL;
  GSList *brace_nodes = NULL;
  SpecNode *rv;
  char *name = NULL;
  while (*at && isalpha (*at))
    at++;

  name = g_memdup (spec, at + 1 - spec);
  name[at - spec] = '\0';

  while (*at)
    {
      if (*at == L_BRACE || *at == L_PAREN)
	{
	  const char *match;
	  void *subnode;
	  GSList *list;
	  match = find_matching (at);
	  if (match == NULL)
	    {
	      g_warning ("no match for %c", *at);
	      goto error;
	    }
	  if (*at == L_BRACE)
	    {
	      subnode = parse_spec_brace (at, match);
	      list = brace_nodes;
	    }
	  else
	    {
	      subnode = parse_spec_paren (at, match);
	      list = paren_nodes;
	    }
	  if (subnode == NULL)
	    {
	      g_warning ("error parsing subspec");
	      goto error;
	    }
	  list = g_slist_prepend (list, subnode);
	}
      else
	{
	  g_warning ("unexpected character `%c' [0x%02x]", *at, *at);
	  goto error;
	}
    }
  rv = g_new (SpecNode, 1);
  rv->class_name = name;
  rv->paren_nodes = g_slist_reverse (paren_nodes);
  rv->brace_nodes = g_slist_reverse (brace_nodes);
  return rv;

error:
  g_slist_foreach (paren_nodes, (GFunc) spec_node_paren_free, NULL);
  g_slist_foreach (brace_nodes, (GFunc) spec_node_brace_free, NULL);
  g_slist_free (paren_nodes);
  g_slist_free (brace_nodes);
  g_free (name);
  return NULL;
}

static void spec_node_free(SpecNode* node)
{
  GSList *paren_nodes = node->paren_nodes;
  GSList *brace_nodes = node->brace_nodes;
  g_slist_foreach (paren_nodes, (GFunc) spec_node_paren_free, NULL);
  g_slist_foreach (brace_nodes, (GFunc) spec_node_brace_free, NULL);
  g_slist_free (paren_nodes);
  g_slist_free (brace_nodes);
  g_free (node->class_name);
  g_free (node);
}

static void spec_node_brace_set_object(SpecNodeBrace*      node,
                                       GtkObject*          object)
{
  const char *child_name = node->key;
  GtkObject *child;

  child = spec_node_to_object (node->value, GDAM_MODEL (object)->channel);
  if (child == NULL)
    return;

  gdam_arg_set_child (object, child_name, -1, child);
  gtk_object_unref (child);
}

static void spec_node_paren_set_object(SpecNodeParen*      node,
                                       GtkObject*          object)
{
  int arg_type;
  GdamArg arg;
  arg_type = gdam_arg_query_type (object->klass->type, node->key);
  if (arg_type == 0)
    {
      g_warning ("unknown argument %s to %s", node->key,
		 gtk_type_name (object->klass->type));
      return;
    }
  arg.name = node->key;
  arg.type = arg_type;
  if (gdam_arg_parse_val_from_string (&arg, node->value) <= 0)
    {
      g_warning ("error parsing %s from %s", node->key, node->value);
      return;
    }
  gdam_arg_set_object_arg (object, &arg);
}

static GtkObject* spec_node_to_object(SpecNode* node, GdamChannel* channel)
{
  GtkObject *rv;
  GtkType type;

  type = gdam_plugin_get_type_by_name ("model", node->class_name);
  if (type == 0)
    {
      g_warning ("no type %s found", node->class_name);
      return NULL;
    }

  rv = gtk_type_new (type);
  gtk_object_ref (rv);
  gtk_object_sink (rv);
  GDAM_MODEL (rv)->channel = channel;

  g_slist_foreach (node->paren_nodes, (GFunc) spec_node_paren_set_object, rv);
  g_slist_foreach (node->brace_nodes, (GFunc) spec_node_brace_set_object, rv);
  return rv;
}

GdamModel* gdam_model_from_spec(const char* spec, GdamChannel* channel)
{
  SpecNode *spec_node;
  GtkObject *rv;
  spec_node = parse_spec (spec);
  if (spec_node == NULL)
    {
      g_warning ("gdam_model_from_spec: error parsing spec: %s", spec);
      return NULL;
    }

  rv = spec_node_to_object (spec_node, channel);
  spec_node_free (spec_node);
  g_return_val_if_fail (rv != NULL, NULL);
  return GDAM_MODEL (rv);
}

gdouble gdam_model_get_length(GdamModel* model)
{
  GtkObject *object;
  GdamModelClass *klass;
  object = GTK_OBJECT (model);
  klass = GDAM_MODEL_CLASS (object->klass);
  if (klass->get_length != NULL)
    return (*klass->get_length) (model);
  return -1;
}

