/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamproducer.h"
#include "gdamutil.h"
#include <gtk/gtksignal.h>
#include <stdio.h>

static void     gdam_producer_destroy          (GtkObject*    object);

static void     gdam_producer_real_seek        (GdamProducer* 	    producer,
		          		        gdouble             offset,
					        int 	            whence);
static void     gdam_producer_real_stop_request(GdamProducer*       producer,
                                                GdamRequestBuilder* builder);
static void     gdam_producer_real_set_position(GdamProducer*       producer,
	                                        gdouble             position);
static void     gdam_producer_new_request      (GdamModel*          turntable,
			                        GdamModel*          producer,
						GdamPlayInfo*       play_info,
			                        int                 sub_id,
			                        GdamRequestBuilder* builder);

enum {
	/*PLAY,
	STOP,*/
	SEEK,
	SET_POSITION,
	UPDATE_LENGTH,
	SOURCE_ENDED,
	LAST_SIGNAL
};
static guint producer_signals[LAST_SIGNAL];
static GtkObjectClass* parent_class = 0;

static void gdam_producer_init(GdamProducer* producer)
{
  producer->stopped_position = 0.0;
}

void gdam_producer_stop(GdamProducer* producer){
	GdamArg arg;
	arg.name = "playing";
	arg.type = GDAM_ARG_TYPE_INT;
	GDAM_ARG_INT(&arg) = 0;
	gdam_arg_set_object_arg(GTK_OBJECT(producer), &arg);
}

static void internal_gdam_producer_stop(GdamProducer* producer)
{
  GdamModel *model = GDAM_MODEL (producer);
  GdamRequestBuilder *builder;

  if (model->parent == NULL)
    return;
  builder = gdam_request_builder_new ();
  gdam_producer_stop_request (producer, builder);
  gdam_debug ("internal stop, issuing model_new_child_req on %s",
	      gtk_type_name (GTK_OBJECT (model)->klass->type));
  g_return_if_fail (model->parent != NULL);
  gdam_model_new_child_request (model->parent, model, NULL, 0, builder);
}

static void gdam_producer_arg_set_position(GtkObject* object, GdamArg*   arg)
{
}

static void gdam_producer_arg_get_position(GtkObject* object, GdamArg*   arg)
{
  GdamProducer *producer = GDAM_PRODUCER (object);
  GDAM_ARG_DOUBLE (arg) = producer->stopped_position;
}

/*  Valid values for "playing" are:
 *  0 => stop, 
 *  1 => play, 
 *  2 => play_in_sync or fail. 
 *  3 => play_in_sync or play.
 *  4 => play_in_sync or play, stop if playing.
 **/
static void gdam_producer_arg_set_playing(GtkObject* object, GdamArg*   arg)
{
  GdamModel *model = GDAM_MODEL (object);
  GdamPlayInfo *play_info;
  GdamProducer *producer = GDAM_PRODUCER (object);
  int play_method = GDAM_ARG_INT (arg);;

  play_info = g_new0 (GdamPlayInfo, 1);

  if (play_method == 0)
    {
      if (model->source_id == 0)
	return;
      internal_gdam_producer_stop (producer);
    }
  else if (play_method == 1)
    {
      if (model->source_id != 0)
	{
	  gdam_debug ("already playing, arg_play aborting");
	  return;
	}
      play_info->position = producer->stopped_position;
      gdam_producer_play (producer, play_info);
    }
  else if (play_method == 2 || play_method == 3 || play_method == 4)
    {
      guint leader_id;
      GdamBeatInfo beat_info;
      if (model->source_id != 0)
	{
	  if (play_method == 4)
	    internal_gdam_producer_stop (producer);
	  else
	    gdam_debug ("already playing, arg_play aborting");
	  return;
	}
      if (gdam_infomanager_get_sync_info (model, &leader_id, &beat_info))
	{
	  play_info->lead_id = leader_id;
	  play_info->lead_beat_info = &beat_info;
	  play_info->is_synchronized = TRUE;
	  play_info->position = producer->stopped_position;
	}
      else if (play_method == 3 || play_method == 4)
	{
	  play_info->position = producer->stopped_position;
	}
      else
	{
	  gdam_debug ("aborting arg_play, didn't have info");
	  /* TODO: free play_info ? */
	  return;
	}
      gdam_producer_play (producer, play_info);
    }
  else
    {
      g_warning ("%d invalid value for 'playing'", play_method);
      return;
    }
}

/*  0 => stopped, 1 => playing or request pending  */
static void gdam_producer_arg_get_playing(GtkObject* object, 
					  GdamArg*   arg)
{
  GdamModel *model = GDAM_MODEL (object);

  if (model->source_id != 0)
    {
      gdam_debug ("already playing, arg_play aborting");
      GDAM_ARG_INT (arg) = 1;
    }
  else
    {
      GDAM_ARG_INT (arg) = 0;
    }
}


static void gdam_producer_class_init(GdamProducerClass* producer_class)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (producer_class);
  GdamModelClass *model_class = GDAM_MODEL_CLASS (producer_class);
  producer_class->seek = gdam_producer_real_seek;
  object_class->destroy = gdam_producer_destroy;
  producer_class->set_position = gdam_producer_real_set_position;
  producer_class->stop_request = gdam_producer_real_stop_request;
  model_class->new_request = gdam_producer_new_request;
  /*
     producer_signals[PLAY] =
     gtk_signal_new(
     "play",
     GTK_RUN_LAST|GTK_RUN_NO_HOOKS,
     object_class->type,
     GTK_SIGNAL_OFFSET(GdamProducerClass, play),
     gtk_marshal_NONE__POINTER,
     GTK_TYPE_NONE,
     1,
     GTK_TYPE_POINTER);
     producer_signals[STOP] =
     gtk_signal_new(
     "stop",
     GTK_RUN_LAST|GTK_RUN_NO_HOOKS,
     object_class->type,
     GTK_SIGNAL_OFFSET(GdamProducerClass, stop),
     gtk_marshal_DOUBLE__NONE,
     GTK_TYPE_DOUBLE,
     0); */
  producer_signals[SEEK] =
    gtk_signal_new ("seek",
		    GTK_RUN_LAST | GTK_RUN_NO_HOOKS,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamProducerClass, seek),
		    gtk_marshal_NONE__DOUBLE_INT,
		    GTK_TYPE_NONE, 2, GTK_TYPE_DOUBLE, GTK_TYPE_INT);
  producer_signals[SET_POSITION] =
    gtk_signal_new ("set_position",
		    GTK_RUN_LAST | GTK_RUN_NO_HOOKS,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamProducerClass,
				       set_position),
		    gtk_marshal_NONE__DOUBLE,
		    GTK_TYPE_NONE, 1, GTK_TYPE_DOUBLE);
  producer_signals[UPDATE_LENGTH] =
    gtk_signal_new ("update_length",
		    GTK_RUN_LAST | GTK_RUN_NO_HOOKS,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamProducerClass,
				       update_length),
		    gtk_marshal_NONE__DOUBLE,
		    GTK_TYPE_NONE, 1, GTK_TYPE_DOUBLE);
  producer_signals[SOURCE_ENDED] =
    gtk_signal_new ("source_ended",
		    GTK_RUN_LAST | GTK_RUN_NO_HOOKS,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamProducerClass,
				       source_ended),
		    gtk_marshal_NONE__DOUBLE,
		    GTK_TYPE_NONE, 1, GTK_TYPE_DOUBLE);
  gtk_object_class_add_signals (object_class, producer_signals, LAST_SIGNAL);

  gdam_arg_register (object_class,
		     "playing",
		     GDAM_ARG_TYPE_INT,
		     gdam_producer_arg_set_playing,
		     gdam_producer_arg_get_playing,
		     NULL, "0,4", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "position",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_producer_arg_set_position,
		     gdam_producer_arg_get_position, NULL, NULL, 0);
}

GtkType gdam_producer_get_type()
{
  static GtkType producer_type = 0;
  if (!producer_type)
    {
      static const GtkTypeInfo producer_info = {
	"GdamProducer",
	sizeof (GdamProducer),
	sizeof (GdamProducerClass),
	(GtkClassInitFunc) gdam_producer_class_init,
	(GtkObjectInitFunc) gdam_producer_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAM_TYPE_MODEL);
      producer_type = gtk_type_unique (GDAM_TYPE_MODEL, &producer_info);

    }
  return producer_type;
}

static void producer_ended_callback(gpointer user_data)
{
  GdamProducer *producer = GDAM_PRODUCER (user_data);
  gdam_producer_source_ended (producer, 0.0);
}

static void eof_connect_success(int eof_sub_id, gpointer user_data)
{
  GdamProducer *producer = GDAM_PRODUCER (user_data);
  producer->eof_sub_id = eof_sub_id;
  gdam_channel_register_handler (GDAM_MODEL (producer)->channel,
				 gdam_channel_simple_handler_new (eof_sub_id,
								  TRUE,
								  producer_ended_callback,
								  producer));
}

void gdam_producer_play(GdamProducer* producer,
			GdamPlayInfo* play_info)
{
  GdamRequestBuilder *builder;
  GdamModel *model;
  int source_id;
  g_return_if_fail (producer != 0);

  gdam_debug ("gdam_producer_play: %s",
	      gtk_type_name (GTK_OBJECT (producer)->klass->type));
  model = GDAM_MODEL (producer);
  g_return_if_fail (model->parent != NULL);
  builder = gdam_request_builder_new ();
  source_id = gdam_producer_play_request (producer, play_info, 0, builder);

  /* Should we set producer->source_id, or does each play_request handle it? */
  if (source_id == 0) 
    {
      g_warning("source_id zero in gdam_producer_play, aborting");
      return;
    }


  gdam_request_builder_append_generic (builder,
				       gdam_request_subscribe_source_ended
				       (source_id, eof_connect_success, NULL,
					producer));

  gdam_model_new_child_request (model->parent,
				model, play_info, source_id, builder);
}

void gdam_producer_play_now(GdamProducer* producer)
{
  GdamPlayInfo *play_info = g_new0 (GdamPlayInfo, 1);
  if (GDAM_MODEL (producer)->source_id != 0)
    {
      gdam_debug ("producer_play_now: have id, aborting.");
      return;
    }
  play_info->position = producer->stopped_position;
  gdam_debug ("gdam_producer_play_now:  playing %p at %f sec",
	      producer, play_info->position);
  gdam_producer_play (producer, play_info);
}

void gdam_producer_seek(GdamProducer* producer, gdouble offset, int whence)
{
  gtk_signal_emit (GTK_OBJECT (producer),
		   producer_signals[SEEK], offset, whence);
}

static void 
gdam_producer_real_seek(GdamProducer* producer, gdouble offset, int whence)
{
  int source_id = GDAM_MODEL (producer)->source_id;
  GdamChannel *channel = GDAM_MODEL (producer)->channel;
  GdamTime tim = (GdamTime) (44100.0 * offset);
  GdamArg arg;
  GdamRequest *request;
  if (source_id == 0)
    {
      double stopped_position = producer->stopped_position;
      switch (whence)
	{
	case 0:
	  stopped_position = offset;
	  break;
	case 1:
	  stopped_position += offset;
	  break;
	case 2:
	  g_warning ("cannot seek from end of file");
	  break;
	}
      gdam_producer_set_position (producer, stopped_position);
      return;
    }
  switch (whence)
    {
    case 0:
      arg.name = "position";
      arg.type = GDAM_ARG_TYPE_INT64;
      GDAM_ARG_INT64 (&arg) = tim;
      request = gdam_request_set_args (source_id,
				       1,
				       &arg,
				       gdam_do_nothing,
				       gdam_do_nothing, NULL);
      gdam_channel_append_request (channel, request);
      break;
    case 1:
      request = gdam_request_seek_relative (source_id,
					    tim,
					    gdam_do_nothing,
					    gdam_do_nothing, NULL);
      gdam_channel_append_request (channel, request);
      break;
    case 2:
      g_warning ("cannot seek from end of file");
      break;
    }
}

int         gdam_producer_play_request(GdamProducer*       producer,
                                       GdamPlayInfo*       play_info,
				       int                 source_id,
				       GdamRequestBuilder* builder)
{
  GtkObject *object = (GtkObject *) producer;
  GdamProducerClass *klass = (GdamProducerClass *) (object->klass);
  g_return_val_if_fail (klass->play_request != NULL, 0);
  return (*klass->play_request) (producer, play_info, source_id, builder);
}

void        gdam_producer_stop_request(GdamProducer*       producer,
				       GdamRequestBuilder* builder)
{
  GtkObject *object = (GtkObject *) producer;
  GdamProducerClass *klass = (GdamProducerClass *) (object->klass);
  GdamModel *model = GDAM_MODEL (producer);
  g_return_if_fail (klass->stop_request != NULL);

  if (model->source_id == 0)
    {
      gdam_debug ("producer_stop_request: source_id==0, aborting.");
      return;
    }

  (*klass->stop_request) (producer, builder);
  /*  Cannot have this here, because the source id of the model is
   *  needed to post the stop request... right??  */
  //g_return_if_fail(model->source_id == 0);
}

static void     gdam_producer_new_request      (GdamModel*          model,
			                        GdamModel*          sub,
						GdamPlayInfo*       play_info,
			                        int                 sub_id,
			                        GdamRequestBuilder* builder)
{
  int id;
  GdamProducer *producer = GDAM_PRODUCER (model);
  GdamModel *parent = model->parent;
  g_return_if_fail (parent != NULL);
  /* If our sub id is zero, stop and pass on a value of zero to our parent. 
   */
  if (sub_id == 0)
    {
      gdam_producer_stop_request (producer, builder);
      id = 0;
    }
  /* Otherwise, add an appropriate play request to the builder.
   */
  else
    {
      id = gdam_producer_play_request (producer, play_info, sub_id, builder);
      if (id ==0)
	{
	  g_warning ("producer_new_request: producer_play_request id of zero, type=%s",
	      	     gtk_type_name (GTK_OBJECT (producer)->klass->type));
	}
    }
  gdam_model_new_child_request (parent, model, play_info, id, builder);
}

static void     gdam_producer_destroy       (GtkObject*    object)
{
  GdamProducer *producer = (GdamProducer *) object;
  if (((GdamModel *) object)->source_id != 0)
    {
      gdam_debug ("producer_destroy: calling stop on source_id %d",
		  ((GdamModel *) object)->source_id);
      internal_gdam_producer_stop (producer);
    }
  (*parent_class->destroy) (object);
}

typedef struct _StopReturnValues StopReturnValues;
struct _StopReturnValues {
	GdamRequestGetTimeStruct source_time_return;
	GdamRequestGetIntStruct package_return;
	gboolean detach_source_return;
	GdamProducer* producer;
};

static void gdam_producer_get_stopped_position(gint64 pos, GdamProducer* prod)
{
  gdouble stopped_position = (gdouble) pos / 44100.0;
  gdam_debug ("gdam_producer_get_stopped_position: %f seconds",
	      stopped_position);
  gdam_producer_set_position (prod, stopped_position);
  gtk_object_unref (GTK_OBJECT (prod));
}
static void gdam_producer_failed_get_stopped_position(GdamProducer* prod)
{
  gdam_debug ("failed getting stopped position, song had already ended?");
  gtk_object_unref (GTK_OBJECT (prod));
}

static void zero_source_id(gpointer data)
{
  GdamModel *model;
  model = GDAM_MODEL (data);
  gdam_debug ("producer_stop: about to set source_id to 0");
  gdam_model_set_source_id (model, 0);
}

/* TODO: if making_request,
 *       register for that source to be removed on completion
 */
static void  gdam_producer_real_stop_request(GdamProducer*       producer,
                                             GdamRequestBuilder* builder)
{
  GdamModel *model = GDAM_MODEL (producer);
  StopReturnValues *ret_values = g_new (StopReturnValues, 1);
  int pos_source;
  if (model->source_id == 0)
    {
      gdam_debug ("real_stop_request: source id zero, aborting.");
      return;
    }

  /* This reference will be undone be stop_failed_callback
   * or stop_succeeded_callback, */
  gtk_object_ref (GTK_OBJECT (producer));

  pos_source = model->source_id;
  if (producer->position_source_id != 0)
    pos_source = producer->position_source_id;

  ret_values->producer = producer;
  gdam_request_builder_append_generic 
	  (builder,
	   gdam_request_source_time (pos_source, 0,
				     (GdamRequestGotTimeFunc)
				     gdam_producer_get_stopped_position,
				     (GHookFunc)
				     gdam_producer_failed_get_stopped_position,
				     producer));
  gdam_debug ("producer_real_stop_request on %s, DETACHING source %d.", 
      	      gtk_type_name (GTK_OBJECT (model)->klass->type),
      	      model->source_id);
  gdam_request_builder_append_generic 
    		(builder,
		 gdam_request_detach_source (model->source_id,
					     zero_source_id,
					     zero_source_id,
					     model));
  gdam_model_set_source_id (model, 0);

  if (producer->position_source_id)
    producer->position_source_id = 0;

  return;
}


static void gdam_producer_real_set_position (GdamProducer*  producer,
	                                     gdouble        position)
{
  producer->stopped_position = position;
}

void       gdam_producer_set_position (GdamProducer*  producer,
	                               gdouble        position)
{
  GdamArg arg;
  arg.name = "position";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = position;
  gdam_arg_set_object_arg (GTK_OBJECT (producer), &arg);
  if (gtk_signal_n_emissions (GTK_OBJECT (producer),
			      producer_signals[SET_POSITION]))
    return;
  gtk_signal_emit (GTK_OBJECT (producer),
		   producer_signals[SET_POSITION], position);
}

void       gdam_producer_update_length (GdamProducer*  producer,
	                                gdouble        length)
{
  if (gtk_signal_n_emissions (GTK_OBJECT (producer),
			      producer_signals[UPDATE_LENGTH]))
    return;
  gtk_signal_emit (GTK_OBJECT (producer),
		   producer_signals[UPDATE_LENGTH], length);
}

void       gdam_producer_source_ended (GdamProducer*  producer,
	                               gdouble        position)
{
  if (gtk_signal_n_emissions (GTK_OBJECT (producer),
			      producer_signals[SOURCE_ENDED]))
    return;
  gtk_signal_emit (GTK_OBJECT (producer),
		   producer_signals[SOURCE_ENDED], position);
}
