/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamproducerqueue.h"
#include "gdamutil.h"
#include "gdamsongdatabase.h"
#include <gdam/common/gdammp3info.h>
#include <gtk/gtksignal.h>
#include <glib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <dirent.h>

/* XXX: the following are fairly non-portable and should
 *      be ifdef'd appropriately. */
#include <dirent.h>
#if HAVE_FTW
#include <ftw.h>
#else
#include "myftw.h"
#endif

enum {
	SET_FILENAME,
	SET_QUEUE_LIST,
	LAST_SIGNAL
};

static guint producer_queue_signals[LAST_SIGNAL];


static void     gdam_producer_queue_real_add_filename
					     (GdamProducerQueue* queue,
					      const char*    	 filename);
static void gdam_producer_queue_real_set_filename(GdamProducerQueue* queue,
                                                  const char*      filename);
static void     gdam_producer_queue_real_skip_songs
					     (GdamProducerQueue* queue,
					      int    		 num_songs);
static void gdam_producer_queue_real_skip_to (GdamProducerQueue* queue,
                                             int      	         song_num);
static void gdam_producer_queue_real_set_queue_list(GdamProducerQueue* queue,
						    gpointer 	       data);
static void     gdam_producer_queue_destroy  (GtkObject*         object);
static void     mp3_ended_callback(gpointer user_data);

static void current_song_changed(GdamProducerQueue* queue, 
				       gboolean ignore_server);

static GtkObjectClass* parent_class;
static void gdam_producer_queue_class_init(GdamProducerQueueClass* klass);

static void gdam_producer_queue_init(GdamProducerQueue* queue)
{
  queue->fade_envelope = 4096;
}

GtkType gdam_producer_queue_get_type()
{
  static GtkType queue_type = 0;
  if (!queue_type)
    {
      static const GtkTypeInfo queue_info = {
	"GdamProducerQueue",
	sizeof (GdamProducerQueue),
	sizeof (GdamProducerQueueClass),
	(GtkClassInitFunc) gdam_producer_queue_class_init,
	(GtkObjectInitFunc) gdam_producer_queue_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      queue_type = gtk_type_unique (GDAM_TYPE_PRODUCER, &queue_info);
    }
  return queue_type;
}

static void queued_song_eof_connect_success(int eof_sub_id, gpointer user_data)
{
  GdamProducerQueue *queue = GDAM_PRODUCER_QUEUE (user_data);
  queue->queued_eof_sub_id = eof_sub_id;
  gdam_channel_register_handler (GDAM_MODEL (queue)->channel,
				 gdam_channel_simple_handler_new (eof_sub_id,
								  TRUE,
								  mp3_ended_callback,
								  queue));
  gtk_object_unref (GTK_OBJECT (queue));
}

static void eof_connect_failure(gpointer user_data)
{
  g_warning ("eof connect failure in producerqueue");
  gtk_object_unref (GTK_OBJECT (user_data));
}

static void queued_source_success_callback(int new_id, gpointer user_data)
{
	GdamProducerQueue* queue = GDAM_PRODUCER_QUEUE(user_data);
	queue->queued_id = new_id;
	gtk_object_unref(GTK_OBJECT(queue));
}

static void queued_source_failure_callback(gpointer user_data)
{
	GdamProducerQueue* queue = GDAM_PRODUCER_QUEUE(user_data);
	g_warning("queued source create failure in producerqueue");
	queue->queued_id = 0;
	gtk_object_unref(GTK_OBJECT(queue));
}

/*  This tries to figure out which occurence of a song is the 'current' one.
 *  Returns index in total list, counting from 0 */
static int guess_cur_index(GdamProducerQueue* queue)
{
  int cur_idx = 0;
  int old_idx = queue->current_idx;
  gboolean found;		/*  Whether we've found an index...  */

  if (!queue->current_song)
    return 0;

  /*  get index of first occurence  */
  cur_idx = g_list_index (queue->songinfo_list, queue->current_song);
  /*  if we don't seem to be making forward progress, look for
   *  another occurence of song. */
  if (queue->current_idx > cur_idx)
    {
      GList *latterlist;
      /* find link for first occurence of song */
      latterlist = g_list_find (queue->songinfo_list, queue->current_song);
      while (latterlist && latterlist->next && cur_idx < queue->current_idx)
	{
	  latterlist = g_list_find (latterlist->next, queue->current_song);
	  /* if we found another occurence of song, get its
	   * index in the full list. */
	  if (latterlist)
	    {
	      found = TRUE;
	      cur_idx = g_list_position (queue->songinfo_list, latterlist);
	    }
	}
    }
  /*  Note our new idea of current index for future use  */
  queue->current_idx = cur_idx;
  return cur_idx;
}


/*  Figure out which song, if any, would be the new queued_song, and return
 *  the pointer to it, or NULL of none found.  DOESNT SET 'queue->queued_song' 
 */
static GdamSongInfo* guess_new_queued_song(GdamProducerQueue* queue)
{
  int cur_idx;
  GdamSongInfo *new_queued_song = NULL;

  if (!queue->current_song)
    {
      return NULL;
    }

  /*  Figure out which song, if any, should be the new queued_song  */
  cur_idx = guess_cur_index (queue);
  new_queued_song = g_list_nth_data (queue->songinfo_list, cur_idx + 1);

  if (!new_queued_song && queue->do_repeat)
    new_queued_song = g_list_nth_data (queue->songinfo_list, 0);

  return new_queued_song;
}


/*  Creates a request to create the queued song on the server.  queue->queued_song
 *  should have been set appropriately before this function is called.  Refs the
 *  queue for the create callbacks... meaning the request MUST be added, or
 *  the queue unreffed by hand.
 */
static GdamRequest* get_queued_song_request(GdamProducerQueue* queue)
{
  GdamRequest *request;
  GdamChannel *channel = GDAM_MODEL (queue)->channel;
  request = NULL;

  if (queue->queued_song)
    {
      const char *filename;
      filename = queue->queued_song->filename;
      request = gdam_request_create_source ("GdasSeekableMp3",
					    gdam_channel_new_local_id
					    (channel),
					    queued_source_success_callback,
					    queued_source_failure_callback,
					    queue, 3, "filename",
					    GDAM_ARG_TYPE_STRING, filename,
					    "position", GDAM_ARG_TYPE_INT64,
					    (gint64) 0, "buffer_size",
					    GDAM_ARG_TYPE_INT,
					    (gint) queue->buffer_size);
    }
  gtk_object_ref (GTK_OBJECT (queue));
  return request;
}



/*  Determines which song should come after the current one, 
 *  and sets it up on the server side. */
static void queue_refresh_queued_song(GdamProducerQueue* queue)
{
  int queued_source_id;
  int source_id_array[2];
  int num_ids;
  GdamRequest *request, *srequest;
  GdamRequestBuilder *request_builder;
  GdamArg arg;
  GdamSongInfo *queued_song;
  GdamModel *model = GDAM_MODEL (queue);
  GdamChannel *channel = model->channel;

  num_ids = 2;

  if (!queue->current_song)
    return;

  queued_song = guess_new_queued_song (queue);

  /*  If there is a queued song, but we already have it as the
   *  queued song, return.  */
  if (queued_song && (queued_song == queue->queued_song))
    return;

  queue->queued_song = queued_song;


  /* The rest of the function updates the sources if playing */

  if (!model->source_id)
    return;

  request_builder = gdam_request_builder_new ();

  if (queue->queued_song)
    {
      request = get_queued_song_request (queue);
      queued_source_id = gdam_request_builder_append_create (request_builder,
							     request);
      gdam_request_builder_append_generic (request_builder,
					   gdam_request_subscribe_source_ended
					   (queued_source_id,
					    queued_song_eof_connect_success,
					    eof_connect_failure, queue));
      source_id_array[1] = queued_source_id;
      /*  Ref for eof connect callback...  */
      gtk_object_ref (GTK_OBJECT (queue));
    }
  else
    {
      num_ids = 1;
      source_id_array[1] = 0;
    }

  source_id_array[0] = queue->current_id;

  arg.name = "queue";
  arg.type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (&arg).length = num_ids;
  GDAM_ARG_INT_ARRAY (&arg).array = source_id_array;

  srequest = gdam_request_set_args (GDAM_MODEL (queue)->source_id,
				    1, &arg, NULL, NULL, NULL);

  gdam_request_builder_append_generic (request_builder, srequest);

  gdam_channel_append_request (GDAM_MODEL (queue)->channel,
			       gdam_request_builder_package
			       (request_builder));
  gdam_request_builder_destroy (request_builder);
}


/* Delete a range of songs.  counts from 1. */
void gdam_producer_queue_delete_range(GdamProducerQueue* queue,
				      guint lower, guint upper)
{
  guint n;			/*  n varies from lower to upper, on the same scale.  */
  int to_play = 0;		/* also counted from 1, for now. */
  gboolean delete_need_refresh = FALSE;
  GList *songinfo_list = g_list_copy (queue->songinfo_list);
  int cur_idx;

  g_return_if_fail (lower <= upper);

  cur_idx = guess_cur_index (queue);

  //gdam_debug("queue delete_range: %d %d", lower, upper);

  for (n = upper; n >= lower; n--)
    {
      GdamSongInfo *song_info;

      if (n > g_list_length (songinfo_list))
	continue;
      song_info = g_list_nth_data (songinfo_list, n - 1);
      if (song_info != NULL)
	{
	  if (song_info == queue->current_song
	      && cur_idx < (int) upper && cur_idx >= (int) lower - 1)
	    {
	      if (n == g_list_length (songinfo_list))
		{
		  to_play = n - 1;
		}
	      else
		{
		  to_play = n + 1;
		}
	      delete_need_refresh = TRUE;
	    }
	  songinfo_list = g_list_remove (songinfo_list, song_info);
	  gdam_song_info_unref (song_info);
	}
    }
  /* TODO: set_queue_list does (some) refreshing... */
  gdam_producer_queue_set_queue_list (queue, songinfo_list);

  if (delete_need_refresh)
    {
      if (to_play > 0)
	gdam_producer_queue_skip_to (queue, to_play - 1);
      else
	{
	  gdam_producer_stop (GDAM_PRODUCER (queue));
	  queue->current_song = NULL;
	  current_song_changed (queue, TRUE);
	}
    }
}

/* Move a range of songs.  counts from 1. */
void gdam_producer_queue_move_range(GdamProducerQueue* queue,
				      guint lower, guint upper, guint dest)
{
  guint n;			/*  n varies from lower to upper, on the same scale.  */
  GList *songinfo_list = g_list_copy (queue->songinfo_list);
  GList *current_link = NULL;
  /* TODO: what if dest is  zero? */
  GList *target_link = g_list_nth (songinfo_list, dest - 1);
  GList *mobile_list = NULL;
  int cur_idx = 0;

  g_return_if_fail (lower <= upper);
  if (dest >= lower && dest <= upper + 1)
    return;

  cur_idx = guess_cur_index (queue);
  if ((cur_idx < (int) upper) && (cur_idx <= (int) lower - 1))
    {
      current_link = g_list_nth (songinfo_list, cur_idx);
    }

  for (n = upper; n >= lower; n--)
    {
      gpointer data;

      if (n > g_list_length (songinfo_list))
	continue;
      data = g_list_nth_data (songinfo_list, n - 1);
      if (data)
	{
	  songinfo_list = g_list_remove (songinfo_list, data);
	  mobile_list = g_list_prepend (mobile_list, data);
	}
    }
  //gdam_debug("queue: mobile list length: %d", g_list_length(mobile_list));
  /*  what if dest is NULL?  concat lists.  */
  if (target_link)
    {
      int tar = g_list_position (songinfo_list, target_link);
      while (mobile_list)
	{
	  gpointer data = mobile_list->data;
	  mobile_list = g_list_remove (mobile_list, data);
	  songinfo_list = g_list_insert (songinfo_list, data, tar++);
	}
    }
  else
    {
      g_list_concat (songinfo_list, mobile_list);
    }
  if (current_link)
    queue->current_idx = g_list_position (songinfo_list, current_link);
  gdam_producer_queue_set_queue_list (queue, songinfo_list);
}


static void gdam_producer_queue_real_add_filename(GdamProducerQueue* queue,
                                                  const char*      filename)
{
  GdamSongDatabase *database = gdam_song_database_system ();
  GdamSongInfo *info = NULL;
  GList *songinfo_list = g_list_copy (queue->songinfo_list);

  //gdam_debug("queue_real_add_filename");

  if (database)
    info = gdam_song_database_find (database, filename);

  gdam_song_database_freeze (database);
  if (!info)
    {
      /* TODO: shouldn't require that file exists for client */
      if (gdam_file_exists (filename))
	{
	  GdamMp3Info *mp3_info;
	  mp3_info = gdam_mp3_info_new (filename);
	  if (mp3_info == NULL)
	    {
	      //g_warning("%s was not an mp3 file", filename);
	      return;
	    }
	  info = gdam_song_info_new (filename, 0, 0);
	  info->length = mp3_info->length;
	  /*  TODO: mark as temporary? */
	  gdam_song_database_add (database, info);
	}
      else
	{
	  if (strncmp (filename, "http://", 7) == 0)
	    {
	      info = gdam_song_info_new (filename, 0, 0);
	      /*  TODO: mark as temporary? */
	      gdam_song_database_add (database, info);
	    }
	  else
	    return;
	}
    }
  g_return_if_fail (info != NULL);
  gdam_song_info_ref (info);
  gdam_song_database_thaw (database);

  songinfo_list = g_list_append (songinfo_list, info);

  gdam_producer_queue_set_queue_list (queue, songinfo_list);

  if (g_list_length (songinfo_list) == 1)
    {
      queue->current_song = info;
      queue->current_idx = 0;
      current_song_changed (queue, FALSE);
    }

  if (g_list_length (songinfo_list) == 2)
    {
      queue_refresh_queued_song (queue);
    }
}

static void print_entry(gpointer data, gpointer count_addr)
{
  GdamSongInfo *song_info = (GdamSongInfo *) data;
  int *n = count_addr;
  fprintf (stderr, "%d %s\n", (*n)++, song_info->filename);
}

void gdam_producer_queue_print_queue(GdamProducerQueue*	queue)
{
  int count = 0;
  //gdam_debug("printing %d tracks in queue:",
  //g_list_length(queue->songinfo_list));
  g_list_foreach (queue->songinfo_list, print_entry, &count);
}

static void gdam_producer_queue_real_skip_to(GdamProducerQueue* queue,
                                             int      	        song_num)
{
  int new_idx;
  gpointer data;
  new_idx = song_num - 1;
  new_idx =
    CLAMP (new_idx, 0, (int) g_list_length (queue->songinfo_list) - 1);
  data = g_list_nth_data (queue->songinfo_list, new_idx);
  if (!data)
    {
      g_warning ("Missing data pointer in skip to.");
      return;
    }
  //gdam_debug("skip_to song %d of %d", 
  //new_idx,
  //g_list_length(queue->songinfo_list));
  queue->current_song = (GdamSongInfo *) data;
  queue->current_idx = new_idx;
  current_song_changed (queue, FALSE);
}

static void gdam_producer_queue_real_skip_songs(GdamProducerQueue* queue,
                                                int      	   num_songs)
{
  int cur_idx, new_idx;
  gpointer data;
  int length;
  //gdam_debug("producer_queue_real_skip_songs: %d", num_songs);
  length = g_list_length (queue->songinfo_list);
  g_return_if_fail (queue->songinfo_list != NULL);
  g_return_if_fail (queue->songinfo_list != NULL);

  cur_idx = guess_cur_index (queue);

  if (cur_idx < 0)
    {
      //gdam_debug("no cur_idx, aborting skip.");
      return;
    }
  new_idx = cur_idx + num_songs;

  if (queue->do_repeat)
    {
      new_idx = new_idx % length;
      if (new_idx < 0)
	new_idx += length;
    }

  new_idx = CLAMP (new_idx, 0, (int) length - 1);

  data = g_list_nth_data (queue->songinfo_list, new_idx);
  if (!data)
    {
      g_warning ("Missing data pointer in seek.");
      return;
    }
  //gdam_debug("skip_songs to %d of %d", 
  //new_idx,
  //g_list_length(queue->songinfo_list));
  queue->current_song = (GdamSongInfo *) data;
  queue->current_idx = new_idx;

  current_song_changed (queue, FALSE);
}

static void current_source_success_callback(int new_id, gpointer user_data)
{
  GdamProducerQueue *queue = GDAM_PRODUCER_QUEUE (user_data);
  queue->current_id = new_id;
  gtk_object_unref (GTK_OBJECT (queue));
}

static void current_source_failure_callback(gpointer user_data)
{
  GdamProducerQueue *queue = GDAM_PRODUCER_QUEUE (user_data);
  queue->current_id = 0;
  gtk_object_unref (GTK_OBJECT (queue));
}


/* Called when an mp3 ends.  Assumes that mp3 was the current one,
 * as no other source should experience an eof event. */
static void mp3_ended_callback(gpointer user_data)
{
  GdamProducerQueue *queue = GDAM_PRODUCER_QUEUE (user_data);
  queue->current_song = queue->queued_song;
  queue->current_id = queue->queued_id;
  queue->eof_sub_id = queue->queued_eof_sub_id;

  if (queue->current_song)
    {
      queue->current_idx++;
      /* just want filenames to update, without disturbing the
       * queue source. */
      current_song_changed (queue, TRUE);
    }
  else
    {
      gpointer data;
      int new_index = g_list_length (queue->songinfo_list) - 1;
      if (new_index >= 0)
	{
	  data = g_list_nth_data (queue->songinfo_list, new_index);
	  queue->current_song = (GdamSongInfo *) data;
	  queue->current_idx = new_index;
	}
      else
	{
	  queue->current_song = NULL;
	  queue->current_idx = 0;
	}
      /* just want filenames to update, without disturbing the
       * queue source. */
      current_song_changed (queue, TRUE);
    }
}

static void current_song_eof_connect_success(int eof_sub_id, gpointer user_data)
{
  GdamProducerQueue *queue = GDAM_PRODUCER_QUEUE (user_data);
  queue->eof_sub_id = eof_sub_id;
  gdam_channel_register_handler (GDAM_MODEL (queue)->channel,
				 gdam_channel_simple_handler_new (eof_sub_id,
								  TRUE,
								  mp3_ended_callback,
								  queue));
  gtk_object_unref (GTK_OBJECT (queue));
}


static int      gdam_producer_queue_play_request(GdamProducer*    producer,
					         GdamPlayInfo*    play_info,
						 int              source_id,
						 GdamRequestBuilder* builder)
{
  GdamProducerQueue *queue = GDAM_PRODUCER_QUEUE (producer);
  GdamModel *model;
  GdamChannel *channel;
  GdamRequest *request;
  GdamRequest *prequest, *qrequest;
  int mp3_source_id, queue_source_id;
  int source_id_array[2];
  gdouble position = play_info ? play_info->position : 0.0;
  int num_args;
  int num_ints = 2;
  GdamArg args[4];

  g_return_val_if_fail (queue != NULL, 0);
  if (!queue->current_song)
    {
      g_warning ("gdam_producer_queue_play: no current song");
      return 0;
    }
  model = GDAM_MODEL (queue);
  channel = model->channel;
  if (model->source_id != 0)
    {
      g_warning ("gdam_producer_queue_play: source id != 0, ignoring");
    }
  g_return_val_if_fail (queue->filename != NULL, 0);

  /* Create current song.
   */
  args[0].type = GDAM_ARG_TYPE_STRING;
  args[0].name = "filename";
  GDAM_ARG_STRING (args + 0) = queue->filename;
  num_args = 1;

  args[1].type = GDAM_ARG_TYPE_INT64;
  args[1].name = "position";
  GDAM_ARG_INT64 (args + 1) = (gint64) (position * 44100.0);
  num_args++;
  if (queue->buffer_size > 0)
    {
      args[num_args].type = GDAM_ARG_TYPE_INT;
      args[num_args].name = "buffer_size";
      GDAM_ARG_INT (args + num_args) = queue->buffer_size;
      num_args++;
    }
  if (queue->fade_envelope > 0)
    {
      args[num_args].type = GDAM_ARG_TYPE_INT;
      args[num_args].name = "fade_envelope";
      GDAM_ARG_INT (args + num_args) = queue->fade_envelope;
      num_args++;
    }
  request = gdam_request_create_source_v ("GdasSeekableMp3",
					  gdam_channel_new_local_id (channel),
					  num_args,
					  args,
					  current_source_success_callback,
					  current_source_failure_callback,
					  queue);
  /* Ref for current song create callback, */
  gtk_object_ref (GTK_OBJECT (queue));

  mp3_source_id = gdam_request_builder_append_create (builder, request);
  queue->current_id = mp3_source_id;
  source_id_array[0] = mp3_source_id;

  /* Watch for current song eof event.  
   */
  gdam_request_builder_append_generic (builder,
				       gdam_request_subscribe_source_ended
				       (mp3_source_id,
					current_song_eof_connect_success,
					eof_connect_failure, queue));
  /*  Ref for current sonf eof connect callback...  */
  gtk_object_ref (GTK_OBJECT (queue));


  /* If we have a queued_song info pointer, create an mp3 source on
   * the server and save the source id. 
   */
  if (queue->queued_song != NULL)
    {
      const char *filename;
      filename = queue->queued_song->filename;
      qrequest = get_queued_song_request (queue);
      queue->queued_id =
	gdam_request_builder_append_create (builder, qrequest);
      source_id_array[1] = queue->queued_id;
      gdam_request_builder_append_generic (builder,
					   gdam_request_subscribe_source_ended
					   (source_id_array[1],
					    queued_song_eof_connect_success,
					    eof_connect_failure, queue));
      /*  Ref for eof connect callback...  */
      gtk_object_ref (GTK_OBJECT (queue));
    }
  else
    {
      num_ints = 1;
      source_id_array[1] = 0;
    }

  /* Create the queue source on the server, and  set its children.  
   */
  model->source_id = gdam_channel_new_local_id (model->channel);
  prequest = gdam_request_create_source ("GdasQueue",
					 model->source_id,
					 gdam_model_got_subsource,
					 gdam_model_didnt_get_subsource,
					 GDAM_MODEL (queue),
					 1,
					 "queue",
					 GDAM_ARG_TYPE_INT_ARRAY,
					 num_ints, source_id_array);
  /* Ref once for queue create callback */
  gtk_object_ref (GTK_OBJECT (queue));
  queue_source_id = gdam_request_builder_append_create (builder, prequest);
  return queue_source_id;
}

static void gdam_producer_queue_destroy(GtkObject* object)
{
  /* TODO: free song infos? */
  (*parent_class->destroy) (object);
}

void gdam_producer_queue_skip_songs(GdamProducerQueue* queue,
				    int 		      num)
{
  //gdam_debug("producer_queue_skip_songs");
  gdam_producer_queue_real_skip_songs (queue, num);
}

void gdam_producer_queue_skip_to(GdamProducerQueue* queue,
				 int 		    num)
{
  //gdam_debug("producer_queue_skip_to");
  gdam_producer_queue_real_skip_to (queue, num);
}


void gdam_producer_queue_add_filename(GdamProducerQueue* queue,
				      const char*        filename)
{
  //gdam_debug("producer_queue_add_filename");
  gdam_producer_queue_real_add_filename (queue, filename);
}

void gdam_producer_queue_set_queue_list
				(GdamProducerQueue* queue,
				 GList*		    songinfo_list)
{
  GtkObject *this_object = GTK_OBJECT (queue);

  //gdam_debug("producer_queue_set_queue_list");

  if (gtk_signal_n_emissions (this_object,
			      producer_queue_signals[SET_QUEUE_LIST]))
    return;
  gtk_signal_emit (this_object,
		   producer_queue_signals[SET_QUEUE_LIST], songinfo_list);
}

static void gdam_producer_queue_real_set_queue_list(GdamProducerQueue* queue,
						    gpointer 	       data)
{
  GList *songinfo_list = (GList *) data;
  gboolean had_current = queue->current_song ? TRUE : FALSE;
  //gdam_debug("producer_queue_real_set_queue_list");
  /* XXX: shouldnt we be freeing this songlist??? */
  queue->songinfo_list = g_list_copy (songinfo_list);
  if (g_list_length (songinfo_list) <= 0)
    {
      queue->current_song = NULL;
      queue->current_idx = 0;
      queue->queued_song = NULL;
      gdam_producer_stop (GDAM_PRODUCER (queue));
      return;
    }
  if (!had_current)
    {
      gpointer data;
      data = g_list_nth_data (queue->songinfo_list, 0);
      queue->current_song = (GdamSongInfo *) data;
      queue->current_idx = 0;
      current_song_changed (queue, FALSE);
    }
  if (had_current)
    {
      int new_idx;
      new_idx = guess_cur_index (queue);
      /* TODO.... maybe shouldn't skip to first song? */
      if (new_idx < 0)
	{
	  gpointer data;
	  data = g_list_nth_data (queue->songinfo_list, 0);
	  queue->current_song = (GdamSongInfo *) data;
	  queue->current_idx = 0;
	  current_song_changed (queue, FALSE);
	}
      else
	queue_refresh_queued_song (queue);
    }
}

static void gdam_producer_queue_real_set_filename(GdamProducerQueue* queue,
                                                  const char*      filename){
	GdamArg arg;
	arg.name = "filename";
	arg.type = GDAM_ARG_TYPE_STRING;
	GDAM_ARG_STRING(&arg) = (char*) filename;
	gdam_arg_set_object_arg(GTK_OBJECT(queue), &arg);
}



/*  Called after current song changes.  Sets filename, refreshes queued song. 
 *  If the queue is playing, sets its children appropriately.  Hacked to
 *  optionally update model attributes without disturbing sources, for when 
 *  the current song has ended and been replaced by the queued song. 
 */
static void current_song_changed(GdamProducerQueue* queue, gboolean ignore_server)
{
  GdamSongDatabase *database;
  GdamSongInfo *info;
  GdamBeatInfo *beat_info;
  GdamArg newarg;
  GdamArg queue_arg;
  GdamRequestBuilder *request_builder;
  GdamRequest *request, *srequest;
  gdouble length;
  GdamModel *model;
  GdamChannel *channel;
  const char *filename;
  int songnum;
  int old_current_id;
  int old_queued_id;
  int num_ids;
  int source_id_array[2];

  database = gdam_song_database_system ();
  model = GDAM_MODEL (queue);
  channel = model->channel;
  info = NULL;
  old_current_id = 0;
  old_queued_id = 0;
  filename = NULL;

  if (queue->current_song)
    {
      filename = queue->current_song->filename;
      queue->filename = g_strdup (filename);
      gtk_signal_emit (GTK_OBJECT (queue),
		       producer_queue_signals[SET_FILENAME], queue->filename);
      if (database)
	info = gdam_song_database_find (database, filename);
    }
  else
    {
      queue->filename = "";
      gtk_signal_emit (GTK_OBJECT (queue),
		       producer_queue_signals[SET_FILENAME], queue->filename);
    }

  if (info)
    {
      beat_info = gdam_beat_info_copy (info->beat_info);
      length = info->length;
    }
  else
    {
      beat_info = gdam_beat_info_new ();
      gdam_beat_info_set_first_beat (beat_info, 0);
      gdam_beat_info_set_bpm (beat_info, 0);
      length = 10.0;
    }
  gdam_producer_set_position (GDAM_PRODUCER (queue), 0);
  gdam_model_set_beat_info (GDAM_MODEL (queue), beat_info);
  gdam_producer_update_length (GDAM_PRODUCER (queue), length);
  gdam_beat_info_destroy (beat_info);

  songnum = guess_cur_index (queue) + 1;
  newarg.name = "current_song_num";
  newarg.type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (&newarg) = songnum;
  gdam_arg_set_object_arg (GTK_OBJECT (queue), &newarg);

  queue_refresh_queued_song (queue);

  /* HACK to alert listeners of changed filename when switching to
   * queued_source, without causing song to replay. */
  if (ignore_server)
    return;

  /*  If the queue isn't running on the server, return. */
  if (GDAM_MODEL (queue)->source_id == 0)
    {
      return;
    }

  request_builder = gdam_request_builder_new ();
  num_ids = 2;

  /*  Create current song on server.  
   */
  g_return_if_fail (queue->current_song != NULL);
  if (queue->current_song)
    {
      const char *filename;
      old_current_id = queue->current_id;
      filename = queue->current_song->filename;
      request = gdam_request_create_source ("GdasSeekableMp3",
					    gdam_channel_new_local_id
					    (channel),
					    current_source_success_callback,
					    current_source_failure_callback,
					    queue, 3, "filename",
					    GDAM_ARG_TYPE_STRING, filename,
					    "position", GDAM_ARG_TYPE_INT64,
					    (gint64) 0, "buffer_size",
					    GDAM_ARG_TYPE_INT,
					    (gint) queue->buffer_size);

      queue->current_id = gdam_request_builder_append_create (request_builder,
							      request);
      gdam_request_builder_append_generic (request_builder,
					   gdam_request_subscribe_source_ended
					   (queue->current_id,
					    current_song_eof_connect_success,
					    eof_connect_failure, queue));
      /*  Ref for eof connect callback...  */
      gtk_object_ref (GTK_OBJECT (queue));
      source_id_array[0] = queue->current_id;
    }

  /*  Create queued song on server, if appropriate.  
   */
  queue->queued_song = guess_new_queued_song (queue);
  if (queue->queued_song)
    {
      request = get_queued_song_request (queue);
      old_queued_id = queue->queued_id;
      queue->queued_id = gdam_request_builder_append_create (request_builder,
							     request);
      gdam_request_builder_append_generic (request_builder,
					   gdam_request_subscribe_source_ended
					   (queue->queued_id,
					    queued_song_eof_connect_success,
					    eof_connect_failure, queue));
      /*  Ref for eof connect callback...  */
      gtk_object_ref (GTK_OBJECT (queue));
      source_id_array[1] = queue->queued_id;
    }
  else
    {
      num_ids = 1;
      source_id_array[1] = 0;
    }




  /*  Set queue children
   */
  queue_arg.name = "queue";
  queue_arg.type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (&queue_arg).length = num_ids;
  GDAM_ARG_INT_ARRAY (&queue_arg).array = source_id_array;

  srequest = gdam_request_set_args (GDAM_MODEL (queue)->source_id,
				    1, &queue_arg, NULL, NULL, NULL);

  gdam_request_builder_append_generic (request_builder, srequest);

  /* Ref once for current mp3 request
   */
  gtk_object_ref (GTK_OBJECT (queue));

  gdam_channel_append_request (GDAM_MODEL (queue)->channel,
			       gdam_request_builder_package
			       (request_builder));
  gdam_request_builder_destroy (request_builder);
}




GdamProducerQueue* gdam_producer_queue_new(GdamChannel*	channel)
{
  GdamProducerQueue *rv;
  rv = GDAM_PRODUCER_QUEUE (gtk_type_new (GDAM_TYPE_PRODUCER_QUEUE));
  ((GdamModel *) rv)->channel = channel;
  return rv;
}

void gdam_producer_queue_print_range(GdamProducerQueue* queue,
				     guint lower, guint upper, guint dest)
{
}


void gdam_producer_queue_search_by_string
				    (GdamProducerQueue* queue,
				     const char*	search_string,
				     int		direction)
{
  GList *cur;
  GList *orig;
  if (g_list_length (queue->songinfo_list) == 0)
    return;
  if (queue->current_song)
    orig = g_list_nth (queue->songinfo_list,
		       g_list_index (queue->songinfo_list,
				     queue->current_song));
  else
    orig = queue->songinfo_list;
  cur = orig;
  if (direction > 0)
    cur = cur->next ? cur->next : queue->songinfo_list;
  else
    cur = cur->prev ? cur->prev : g_list_last (queue->songinfo_list);

  while (cur)
    {
      if (cur == orig)
	cur = NULL;
      else
	{
	  GdamSongInfo *song = (GdamSongInfo *) (cur->data);
	  if (strstr (song->filename, search_string))
	    {
	      queue->current_song = (GdamSongInfo *) cur->data;
	      queue->current_idx = g_list_position (queue->songinfo_list,
						    cur);
	      current_song_changed (queue, FALSE);
	      cur = NULL;
	    }
	  else if (direction > 0)
	    cur = cur->next ? cur->next : queue->songinfo_list;
	  else
	    cur = cur->prev ? cur->prev : g_list_last (queue->songinfo_list);
	}
    }
}

void gdam_queue_arg_get_filename(GtkObject* object, GdamArg* arg)
{
  GdamProducerQueue *queue = GDAM_PRODUCER_QUEUE (object);
  if (queue->current_song)
    GDAM_ARG_STRING (arg) =
      g_strdup (queue->current_song->filename);
  else
    GDAM_ARG_STRING (arg) = "no file";
}

void gdam_queue_set_repeat(GdamProducerQueue* queue, gboolean val){
	GdamArg arg;
	arg.type = GDAM_ARG_TYPE_INT;
	arg.name = "repeat";
	GDAM_ARG_INT(&arg) = val;
	gdam_arg_set_object_arg(GTK_OBJECT(queue), &arg);
}

static void gdam_queue_arg_set_repeat(GtkObject* object, GdamArg* arg)
{
  GdamProducerQueue *queue = GDAM_PRODUCER_QUEUE (object);
  queue->do_repeat = GDAM_ARG_INT (arg);
  queue_refresh_queued_song (queue);
}

static void gdam_queue_arg_get_repeat(GtkObject* object, GdamArg* arg)
{
  GdamProducerQueue *queue = GDAM_PRODUCER_QUEUE (object);
  GDAM_ARG_INT (arg) = queue->do_repeat;
}


static void gdam_queue_arg_set_shuffle(GtkObject* object, GdamArg* arg)
{
  GdamProducerQueue *queue = GDAM_PRODUCER_QUEUE (object);
  queue->do_shuffle = GDAM_ARG_INT (arg);
  if (queue->do_shuffle)
    gdam_producer_queue_shuffle_list (queue);
}

static void gdam_queue_arg_get_shuffle(GtkObject* object, GdamArg* arg)
{
  GdamProducerQueue *queue = GDAM_PRODUCER_QUEUE (object);
  GDAM_ARG_INT (arg) = queue->do_shuffle;
}

static void gdam_queue_arg_set_filename(GtkObject* object, GdamArg* arg)
{
  //g_warning("use set_queue_list rather than set_filename.... arg_set_filename ignored.");
}

static void gdam_queue_arg_set_queue_list(GtkObject* object, GdamArg* arg)
{
  GdamProducerQueue *queue;
  char *nl;
  char *cur_file;
  char *filenames = g_strdup (GDAM_ARG_STRING (arg));
  gboolean keep_going = TRUE;
  GdamArg newarg;
  int num_songs;
  queue = GDAM_PRODUCER_QUEUE (object);
  cur_file = filenames;
  g_list_free (queue->songinfo_list);
  queue->songinfo_list = NULL;
  while (keep_going && *cur_file != '\0')
    {
      nl = strchr (cur_file, '\n');
      if (nl)
	*nl = '\0';
      else
	keep_going = FALSE;
      //gdam_debug("arg set queue list: adding %s", cur_file);
      gdam_producer_queue_add_filename (queue, cur_file);
      cur_file = nl + 1;
    }
  num_songs = 0;
  if (queue->songinfo_list)
    num_songs = g_list_length (queue->songinfo_list);
  newarg.name = "total_song_num";
  newarg.type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (&newarg) = num_songs;
  gdam_arg_set_object_arg (object, &newarg);
}

/* XXX: Needs to be arbitrarily long!!  */
static void gdam_queue_arg_get_queue_list(GtkObject* object, GdamArg* arg)
{
  GdamProducerQueue *queue;
  char str[16383] = "";
  GList *l;
  queue = GDAM_PRODUCER_QUEUE (object);
  for (l = queue->songinfo_list; l != NULL; l = l->next)
    {
      strcat (str, ((GdamSongInfo *) (l->data))->filename);
      strcat (str, "\n");
    }
  GDAM_ARG_STRING (arg) = strdup (str);
}

static void gdam_queue_arg_set_current_song_num(GtkObject* object, GdamArg* arg)
{
}

static void gdam_queue_arg_get_current_song_num(GtkObject* object, GdamArg* arg)
{
  GdamProducerQueue *queue;
  int songnum;
  queue = GDAM_PRODUCER_QUEUE (object);
  songnum = guess_cur_index (queue);
  songnum += 1;
  GDAM_ARG_INT (arg) = songnum;
}

static void gdam_queue_arg_set_total_song_num(GtkObject* object, GdamArg* arg)
{
}

static void gdam_queue_arg_get_total_song_num(GtkObject* object, GdamArg* arg)
{
  GdamProducerQueue *queue;
  int songnum;
  queue = GDAM_PRODUCER_QUEUE (object);
  songnum = 0;
  if (queue->songinfo_list)
    songnum = g_list_length (queue->songinfo_list);
  GDAM_ARG_INT (arg) = songnum;
}

static void gdam_producer_queue_class_init(GdamProducerQueueClass* klass)
{
  GdamProducerClass *producer_class = GDAM_PRODUCER_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  klass->skip_songs = gdam_producer_queue_real_skip_songs;
  klass->skip_to = gdam_producer_queue_real_skip_to;
  klass->add_filename = gdam_producer_queue_real_add_filename;
  klass->set_filename = gdam_producer_queue_real_set_filename;
  klass->set_queue_list = gdam_producer_queue_real_set_queue_list;
  producer_class->play_request = gdam_producer_queue_play_request;
  object_class->destroy = gdam_producer_queue_destroy;
  parent_class = gtk_type_class (GDAM_TYPE_PRODUCER);
  gdam_arg_register (object_class,
		     "queue_list",
		     GDAM_ARG_TYPE_STRING,
		     gdam_queue_arg_set_queue_list,
		     gdam_queue_arg_get_queue_list,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "current_song_num",
		     GDAM_ARG_TYPE_INT,
		     gdam_queue_arg_set_current_song_num,
		     gdam_queue_arg_get_current_song_num, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "total_song_num",
		     GDAM_ARG_TYPE_INT,
		     gdam_queue_arg_set_total_song_num,
		     gdam_queue_arg_get_total_song_num, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "filename",
		     GDAM_ARG_TYPE_STRING,
		     gdam_queue_arg_set_filename,
		     gdam_queue_arg_get_filename, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "repeat",
		     GDAM_ARG_TYPE_INT,
		     gdam_queue_arg_set_repeat,
		     gdam_queue_arg_get_repeat, NULL, "boolean", 0);
  gdam_arg_register (object_class,
		     "shuffle",
		     GDAM_ARG_TYPE_INT,
		     gdam_queue_arg_set_shuffle,
		     gdam_queue_arg_get_shuffle, NULL, "boolean", 0);
  producer_queue_signals[SET_QUEUE_LIST] =
    gtk_signal_new ("set_queue_list",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamProducerQueueClass,
				       set_queue_list),
		    gtk_marshal_NONE__POINTER,
		    GTK_TYPE_NONE, 1, GTK_TYPE_POINTER);
  producer_queue_signals[SET_FILENAME] =
    gtk_signal_new ("set_filename",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamProducerQueueClass, set_filename),
		    gtk_marshal_NONE__STRING,
		    GTK_TYPE_NONE, 1, GTK_TYPE_STRING);
  gtk_object_class_add_signals (object_class,
				producer_queue_signals, LAST_SIGNAL);
}



/*
 *   _____ _ _        ____                       _             
 *  |  ___(_) | ___  / ___|  ___  _   _ _ __ ___(_)_ __   __ _ 
 *  | |_  | | |/ _ \ \___ \ / _ \| | | | '__/ __| | '_ \ / _` |
 *  |  _| | | |  __/  ___) | (_) | |_| | | | (__| | | | | (_| |
 *  |_|   |_|_|\___| |____/ \___/ \__,_|_|  \___|_|_| |_|\__, |
 *      						 |___/ 
 */

static gboolean is_url(const char* filename)
{
  if (strncmp (filename, "http://", 7) == 0)
    return TRUE;
  return FALSE;
}

static gboolean is_mp3_filename(const char* filename)
{
  gboolean rv;
  const char *extension = strrchr (filename, '.');
  if (extension != NULL && strcasecmp (extension, ".mp3") == 0)
    {
      if (!g_dirname (filename))
	{
	  return FALSE;
	}
      if (g_basename (filename))
	{
	  return TRUE;
	}
    }
  return FALSE;
}


/* TODO: treat special cases like the word 'and' */
static char* title_canon(const char* title)
{
  char rv[2048];
  int i;
  int j = 0;
  //fprintf(stderr, "canon: %s\r\n", title);
  for (i = 0; title[i]; i++)
    {
      if (isalnum (title[i]))
	{
	  if (isupper (title[i]))
	    rv[j] = tolower (title[i]);
	  else
	    rv[j] = title[i];
	  j++;
	}
    }
  rv[j] = '\0';
  //fprintf(stderr, "canon: %s \r\n into: %s \r\n", title, rv);
  return strdup (rv);
}


static void parse_cddb_data(GdamProducerQueue* queue, const char* cddb_filename)
{
  FILE *fp;
  char buf[2048];
  //fprintf(stderr, "parse_cddb: %s\r\n", cddb_filename);
  fp = fopen (cddb_filename, "r");
  if (!fp)
    return;
  while (fgets (buf, sizeof (buf), fp))
    {
      char *title;
      if (buf[strlen (buf) - 1] != '\n')
	{
	  fprintf (stderr, "Error [%s] - "
		   "line too long, binary file?\n", cddb_filename);
	  break;
	}
      buf[strlen (buf) - 1] = '\0';	/* strip newline */
      title = buf;
      /* Should lead with whitespace */
      if (!title || !isspace (*title))
	continue;
      while (title && isspace (*title))
	title++;
      /* Should be followed by 2 digits */
      if (!title || !isdigit (*title) || !isdigit (*(title + 1)))
	continue;
      title += 2;
      /* Should be followed by whitespace */
      if (!title || !isspace (*title))
	continue;
      while (title && isspace (*title))
	title++;
      /* Should be followed by tracklength "nn:nn"  */
      if (!title || !isdigit (*title)
	  || !isdigit (*(title + 1))
	  || *(title + 2) != ':'
	  || !isdigit (*(title + 3)) || !isdigit (*(title + 4)))
	continue;
      title += 5;
      /* Should be followed by whitespace */
      if (!title || !isspace (*title))
	continue;
      while (title && isspace (*title))
	title++;
      /* title should now point at the title */
      {
	char *canon_entry;
	canon_entry = title_canon (title);
	queue->cddb_canon_titles =
	  g_list_append (queue->cddb_canon_titles, canon_entry);
      }
    }
  fclose (fp);
}

static void 
append_playlist_canon(GdamProducerQueue* queue, const char* filename)
{
  const char *fn;
  char *canon_entry;
  char *last_slash;
  char f1[2048] = "\0";

  fn = filename;
  last_slash = strrchr (filename, '/');
  if (last_slash)
    fn = last_slash + 1;
  if (isdigit (fn[0]) && isdigit (fn[1]) && (fn[2] == '-'))
    strncpy (f1, fn + 3, strlen (fn) - 7);
  else
    strncpy (f1, fn, strlen (fn) - 4);
  canon_entry = title_canon (f1);
  queue->playlist_canon_titles =
    g_list_append (queue->playlist_canon_titles, canon_entry);
}

/* TODO: block against mutual recursion among playlists.... */
static gboolean source_songdb_file(GdamProducerQueue* queue, 
	            		   const char* playlist_filename, 
				   gboolean append_to_playlist)
{
  FILE *fp;
  char buf[2048] = "\0";
  gboolean success;
  success = FALSE;

  fp = fopen (playlist_filename, "r");
  if (!fp)
    return FALSE;
  while (fgets (buf, sizeof (buf), fp))
    {
      char full_name[2048] = "\0";
      if (buf[strlen (buf) - 1] != '\n')
	{
	  //fprintf(stderr, "missed newline in playlist");
	  continue;
	}
      buf[strlen (buf) - 1] = '\0';	/* strip newline */
      /* If song pathname is not absolute, append base path */
      if ((strncmp (buf, "file: ", 6) == 0) ||
	  (strncmp (buf, "File: ", 6) == 0))
	{
	  strcat (full_name, buf + 6);
	}
      else
	{
	  continue;
	}
      if (is_mp3_filename (full_name))
	{
	  success = TRUE;
	  if (append_to_playlist)
	    //fprintf(stderr, "sourcing: appending %s\r\n", 
	    //full_name);
	    gdam_producer_queue_add_filename (queue, full_name);
	}

    }
  fclose (fp);
  return success;
}


/* TODO: block against mutual recursion among playlists.... */
/* TODO: special sortings for: Track_n[n].mp3 n[n].mp3 .... */
static gboolean source_playlist_file(GdamProducerQueue* queue, 
	            		     const char* playlist_filename, 
				     gboolean append_to_playlist)
{
  FILE *fp;
  char buf[2048] = "\0";
  char pbp[2048] = "\0";
  char *last_slash = strrchr (playlist_filename, '/');
  char *playlist_base_path = "";
  gboolean success;
  success = FALSE;
  if (last_slash)
    {
      int base_len = strlen (playlist_filename) - strlen (last_slash);
      strncpy (pbp, playlist_filename, base_len);
      playlist_base_path = pbp;
    }

  //fprintf(stderr, "sourcing '%s' '%s'\n", playlist_base_path, 
  //playlist_filename);
  fp = fopen (playlist_filename, "r");
  if (!fp)
    return FALSE;
  while (fgets (buf, sizeof (buf), fp))
    {
      char full_name[2048] = "\0";
      if (buf[strlen (buf) - 1] != '\n')
	{
	  //fprintf(stderr, "missed newline in playlist");
	  continue;
	}
      buf[strlen (buf) - 1] = '\0';	/* strip newline */
      if (is_url (buf))
	{
	  success = TRUE;
	  if (append_to_playlist)
	    gdam_producer_queue_add_filename (queue, buf);
	  append_playlist_canon (queue, buf);
	  continue;
	}
      /* If song pathname is not absolute, append base path */
      if (buf[0] != '/')
	{
	  strcat (full_name, playlist_base_path);
	  strcat (full_name, "/");
	}
      strcat (full_name, buf);
      if (is_mp3_filename (full_name))
	{
	  success = TRUE;
	  if (append_to_playlist)
	    //fprintf(stderr, "sourcing: appending %s\r\n", 
	    //full_name);
	    gdam_producer_queue_add_filename (queue, full_name);
	  append_playlist_canon (queue, buf);
	}

    }
  playlist_base_path = "";
  fclose (fp);
  return success;
}

static int match_title(gconstpointer p1, gconstpointer p2){
	int rv;
	char* song1 = (char*)p1;
	char* song2 = (char*)p2;
	rv = strcmp(song1, song2);
	return rv;
}

static GdamProducerQueue* hackqueue = NULL;

static int compare_smart_filename(gconstpointer p1, gconstpointer p2){
	int pos1p = -1;
	int pos1c = -1;
	int pos2p = -1;
	int pos2c = -1;
	char f1[2048] = "\0";
	char f2[2048] = "\0";
	char* fn;
	char* fn2;
	char* last_slash;
	int rv = 0;
	GList* found;
	fn = g_strdup((char*)p1);
	fn2 = g_strdup((char*)p2);

	//fprintf(stderr, "compare_smart: %s \r\n          %s\r\n", fn, fn2);

	last_slash = strrchr(fn, '/');
	if (last_slash)
		fn = last_slash+1;
	if (isdigit(fn[0]) && isdigit(fn[1]) && (fn[2] == '-'))
		strncpy(f1, fn+3, strlen(fn)-7);
	else
		strncpy(f1, fn, strlen(fn)-4);
	found = g_list_find_custom(hackqueue->playlist_canon_titles, 
				   title_canon(f1), match_title);
	pos1p = g_list_position(hackqueue->playlist_canon_titles, found);
	found = g_list_find_custom(hackqueue->cddb_canon_titles, 
				   title_canon(f1), match_title);
	pos1c = g_list_position(hackqueue->cddb_canon_titles, found);
	//fprintf(stderr, "got index %d for %s\r\n", pos1p, f1);

	last_slash = strrchr(fn2, '/');
	if (last_slash)
		fn2 = last_slash+1;
	if (isdigit(fn2[0]) && isdigit(fn2[1]) && (fn2[2] == '-'))
		strncpy(f2, fn2+3, strlen(fn2+3)-4);
	else
		strncpy(f2, fn2, strlen(fn2)-4);
	found = g_list_find_custom(hackqueue->playlist_canon_titles, 
				   title_canon(f2), match_title);
	pos2p = g_list_position(hackqueue->playlist_canon_titles, found);
	found = g_list_find_custom(hackqueue->cddb_canon_titles, 
				   title_canon(f2), match_title);
	pos2c = g_list_position(hackqueue->cddb_canon_titles, found);
	//fprintf(stderr, "got index %d for %s\r\n", pos2p, f2);

	if (pos1p >= 0 && pos2p >= 0){
		rv =  pos1p - pos2p;
	}
	else if (pos1c >= 0 && pos2c >= 0){
		//fprintf(stderr, "compare_smart returning %d ", pos1c - pos2c);
		rv =  pos1c - pos2c;
	}
	  /* if neither list contained both... */
	else if (pos1p >= 0 || pos1c >= 0)
		rv =  -1;
	else if (pos2p >= 0 || pos2c >= 0)
		rv =  1;
	  /* if neither list contained either... */
	else 
		/* TODO: sort alphabetically by directory? */
		rv = strcmp(fn, fn2);
	//fprintf(stderr, "compare_smart returning 0: %d %d %d %d", 
					//pos1p, pos2p, pos1c, pos2c);
	return rv;
}

/*  idle add to avoid recursion  */
static void handle_directory(GdamProducerQueue* queue, const char* dir_name)
{
  GList *tmp;
  DIR *dir;
  struct dirent *dirent;
  dir = opendir (dir_name);
  //fprintf(stderr, "handling directory: %s\r\n", dir_name);
  if (!dir)
    return;
  if (queue->minilist)
    g_list_free (queue->minilist);
  queue->minilist = NULL;
  dirent = readdir (dir);
  while (dirent)
    {
      char filename[1024] = "";
      strcat (filename, dir_name);
      strcat (filename, "/");
      strcat (filename, dirent->d_name);
      if (is_mp3_filename (filename))
	{
	  queue->minilist = g_list_append (queue->minilist,
					   strdup (filename));
	}
      else if (source_playlist_file (queue, filename, FALSE))
	{
	}
      else
	{
	  parse_cddb_data (queue, filename);
	}
      dirent = readdir (dir);
    }
  g_return_if_fail (hackqueue == NULL);
  hackqueue = queue;
  queue->minilist = g_list_sort (queue->minilist, compare_smart_filename);
  hackqueue = NULL;
  tmp = queue->minilist;
  while (tmp)
    {
      //fprintf(stderr, "handle dir: adding %s\r\n", (char*)tmp->data);
      gdam_producer_queue_add_filename (queue, (char *) tmp->data);
      //free((char*)tmp->data);
      tmp = tmp->next;
    }
}


static void handle_directory_list(GdamProducerQueue* queue, GList* tmp){
	while(tmp){
		handle_directory(queue, (char*)tmp->data);
		//free((char*)tmp->data);
		tmp = tmp->next;
	}
}

/* XXX: hack for ftw callback. */
static GList* directory_list = NULL;

/*  Only collect directories */
static int directory_walk_callback(const char*        file,
                                   const struct stat* sb,
				   int                flag)
{
  if (flag == FTW_D)
    {
      directory_list = g_list_append (directory_list, strdup (file));
    }
  return 0;			/*  keep walking  */
}

/*  This is the outermost function used to add files, dirs, and playlists to
 *  the queue. */
void 
gdam_producer_queue_source_file(GdamProducerQueue* queue, const char* _filename)
{
  struct stat file_stat;
  char filename[2048] = "";
  if ((*_filename != '/') && !is_url (_filename))
    {
      getcwd (filename, sizeof (filename));
      strcat (filename, "/");
    }
  strcat (filename, _filename);
  stat (filename, &file_stat);
  if (S_ISDIR (file_stat.st_mode))
    {
      g_return_if_fail (directory_list == NULL);
      ftw (filename, directory_walk_callback, 3);
      handle_directory_list (queue, directory_list);
      if (directory_list)
	g_list_free (directory_list);
      directory_list = NULL;
    }
  else if (is_mp3_filename (filename))
    {
      gdam_producer_queue_add_filename (queue, filename);
    }
  else if (is_url (filename))
    {
      gdam_producer_queue_add_filename (queue, filename);
    }
  else if (source_songdb_file (queue, filename, TRUE))
    {
    }
  else if (source_playlist_file (queue, filename, TRUE))
    {
    }
  else
    {
      parse_cddb_data (queue, filename);
    }
}

void gdam_producer_queue_shuffle_list(GdamProducerQueue* queue)
{
  GList *new_list = NULL;
  float random_num;
  int length = g_list_length (queue->songinfo_list);
  while (length)
    {
      GdamSongInfo *temp_si;
      random_num = rand () % length;
      temp_si = g_list_nth_data (queue->songinfo_list, random_num);
      queue->songinfo_list = g_list_remove (queue->songinfo_list, temp_si);
      new_list = g_list_prepend (new_list, temp_si);
      length--;
    }
  length = g_list_length (queue->songinfo_list);
  if (length != 0)
    g_warning ("old list has %d entries", length);
  gdam_producer_queue_set_queue_list (queue, new_list);
}

