/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsequencer.h"
#include "gdambuffermanager.h"
#include <gdam/common/gdamtildeexpand.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtksignal.h>

#define MAX_SPLIT 12

typedef struct _GdamSequencerLoadThread GdamSequencerLoadThread;
struct _GdamSequencerLoadThread {
	GdamSequencer*			sequencer;
	GdamChannel*			channel;
	GList*				elements;
	FILE*				fp;
	int				destroy_connect_id;
	gboolean			is_mp3;

	char*				filename;
	GdamSequencerElement*		under_construction;
	GdamLocalSampleInfo*		local_under_construction;
};

static void continue_processing_thread(GdamSequencerLoadThread* load_thread);
static void on_sequencer_destroy(GdamSequencer* sequencer,
                                 GdamSequencerLoadThread* thread)
{
  thread->destroy_connect_id = 0;
  thread->sequencer = NULL;
}

static void free_load_thread(GdamSequencerLoadThread* load_thread)
{
  if (load_thread->fp)
    fclose (load_thread->fp);
  g_list_foreach (load_thread->elements, (GFunc) g_free, NULL);
  g_list_free (load_thread->elements);

  if (load_thread->sequencer && load_thread->destroy_connect_id)
    {
      gtk_signal_disconnect (GTK_OBJECT (load_thread->sequencer),
			     load_thread->destroy_connect_id);
      load_thread->destroy_connect_id = 0;
      load_thread->sequencer = NULL;
    }
  g_free (load_thread);
}

/* Whichever type of element we were building, complete it. */
static void finished_constructing_element(GdamSequencerLoadThread* thread)
{
  if (thread->under_construction)
    {
      thread->elements = g_list_prepend (thread->elements,
					 thread->under_construction);
      thread->under_construction = NULL;
    }
  if (thread->local_under_construction)
    {
      thread->local_under_construction = NULL;
    }

  g_free (thread->filename);
  thread->filename = NULL;
}


static gboolean match_filename_extension (const char* filename,
    					  const char* ext)
{
  gboolean rv;
  const char *extension;
  g_return_val_if_fail (filename != NULL, FALSE);
  extension = strrchr (filename, '.');
  if (extension != NULL && strcasecmp (extension, ext) == 0)
    {
      if (!g_dirname (filename))
	  return FALSE;
      if (g_basename (filename))
	  return TRUE;
    }
  return FALSE;
}

static gboolean is_mp3_filename(const char* filename)
{
  return match_filename_extension (filename, ".mp3");
}

static gboolean is_wav_filename(const char* filename)
{
  return match_filename_extension (filename, ".wav");
}

static void buffer_manager_cached (int buffer_id, gpointer udata)
{
  GdamSequencerLoadThread *load_thread;
  load_thread = (GdamSequencerLoadThread *) udata;
  if (load_thread->under_construction)
    {
      load_thread->under_construction->buffer_id = buffer_id;
    }
  else if (load_thread->local_under_construction)
    {
      load_thread->local_under_construction->buffer_id = buffer_id;
    }
  if (load_thread->sequencer)
    {
      finished_constructing_element (load_thread);
      continue_processing_thread (load_thread);
    }
  else
    free_load_thread (load_thread);
}

static void buffer_manager_failed (gpointer udata)
{
  GdamSequencerLoadThread *load_thread;
  load_thread = (GdamSequencerLoadThread *) udata;

  /* Continue processing??? */
  /*continue_processing_thread(load_thread); */
  /* Or not??? */
  free_load_thread (load_thread);
}

static void append_cache_request(GdamSequencerLoadThread* load_thread,
                                 char*                    filename)
{
  GdamRequestBuilder *builder = gdam_request_builder_new ();
  int local_id;
  const char* source_type;
  char* description;
  GdamChannel *channel = load_thread->channel;

  g_assert (load_thread->filename == NULL);
  gdam_debug ("append_cache_request: loading %s", filename);

  g_return_if_fail (filename != NULL);
  load_thread->filename = filename;

  if (is_mp3_filename (filename))
    {
      source_type = "GdasMp3";
      description = g_strdup_printf ("Mp3 %s", filename);
    }
  else
    {
      source_type = "GdasWav";
      description = g_strdup_printf ("Wav %s", filename);
    }

  local_id = gdam_request_builder_append_create (builder,
						 gdam_request_create_source
						 (source_type,
						  gdam_channel_new_local_id
						  (channel), NULL, NULL, NULL,
						  1, "filename",
						  GDAM_ARG_TYPE_STRING,
						  filename));
  gdam_buffer_manager_start_caching (builder, local_id, channel, TRUE,
      				     description,
                                     buffer_manager_cached,
				     buffer_manager_failed,
				     load_thread);
}


static GdamLocalSampleInfo* 
parse_local_sample (GdamSequencerLoadThread *load_thread,
    		    char **split, int num_split, char **filename_out) 
{
  GdamLocalSampleInfo *linfo;
  GdamBufferManagerInfo *info;
  char *filename;
  char *description;
  int li_idx;

  g_return_val_if_fail (num_split == 7, NULL);

  li_idx = gdam_sequencer_add_local_info (load_thread->sequencer, 0, 0, 0);
  linfo = &g_array_index (load_thread->sequencer->local_infos, 
      		          GdamLocalSampleInfo, li_idx);
  filename = split[1];
  linfo->local_id = atoi (split[2]);
  linfo->buffer_id = atoi (split[3]);
  linfo->rate = atof (split[3]);
  linfo->gain = atof (split[4]);
  linfo->sample_offset = atoi (split[5]);
  linfo->output_length = atoi (split[6]);

  if (is_mp3_filename (filename))
    description = g_strdup_printf ("Mp3 %s", filename);
  else
    description = g_strdup_printf ("Wav %s", filename);
  info = gdam_buffer_manager_info_by_desc (description);
  if (info != NULL)
    linfo->buffer_id = info->buffer_id;
  else
    {
      linfo->buffer_id = 0;
      *filename_out = g_strdup (filename);
    }
  g_free (description);
  return linfo;
}


/* Parses a line of input until an element or local sample info is encountered.
 */
static gboolean
parse_element (GdamSequencerLoadThread* load_thread, char** filename_out)
{
  GdamSequencerElement *rv;
  GdamBufferManagerInfo *info;
  char buf[1024];
  FILE* fp;
  fp = load_thread->fp; 
  while (fgets (buf, sizeof (buf), fp) != NULL)
    {
      char *at = buf;
      char *split[128];
      char *description = "";
      char *filename;
      int num_split = 0;
      while (*at && isspace (*at))
	at++;
      /* Skip comment lines. */
      if (*at == 0 || *at == '#')
	continue;

      g_strchomp (at);

      /* Split the line into whitespace delimited pieces. */
      /* TODO: handle filenames with spaces, by ignoring spaces within quotes */
      while (*at && num_split < MAX_SPLIT)
	{
	  char *end_word = at;
	  while (*end_word && !isspace (*end_word))
	    end_word++;
	  split[num_split++] = at;
	  if (*end_word == '\0')
	    break;
	  *end_word = 0;
	  at = end_word + 1;
	  while (*at && isspace (*at))
	    at++;
	}
      if (num_split == 2 && strcmp(split[0], "bpm:") == 0)
        {
	  GdamBeatInfo* new_beat_info;
	  float bpm = atof (split[1]);
	  new_beat_info = gdam_beat_info_copy 
			  (&GDAM_MODEL(load_thread->sequencer)->beat_info);
	  gdam_beat_info_set_bpm (new_beat_info, bpm);
	  gdam_model_set_beat_info (GDAM_MODEL(load_thread->sequencer), 
				    new_beat_info);
	  gdam_beat_info_destroy(new_beat_info);
	  continue;
        }
      else if (num_split == 2 && strcmp(split[0], "beats:") == 0)
        {
	  float beats = atoi (split[1]);
	  gdam_sequencer_set_num_beats(load_thread->sequencer, beats);
	  continue;
        }
      else if (strcmp(split[0], "GdamLocalSampleInfo") == 0)
        {
	  GdamLocalSampleInfo *linfo;
	  linfo = parse_local_sample (load_thread, split, num_split, 
	      			      filename_out);
	  if (linfo)
	    {
	      load_thread->local_under_construction = linfo;
	      return TRUE;
	    }
	  g_warning ("error parsing local sample.");
	  continue;
        }
      else if (num_split != 6 && num_split != 7 && num_split != 8)
	{
	  g_warning ("bad number of fields (%d) in line %s parsing sequence",
	      	     num_split, buf);
	  continue;
	}

      g_return_val_if_fail (num_split >= 6, FALSE);
      g_return_val_if_fail (split[5] != NULL, FALSE);

      rv =  gdam_sequencer_new_element ();
      filename = split[0];
      rv->beat_offset = atof (split[1]);
      rv->rate = atof (split[2]);
      rv->gain = atof (split[3]);
      rv->sample_offset = atoi (split[4]);
      rv->output_length = atoi (split[5]);
      if (num_split >= 7) 
	rv->pattern_number = atoi (split[6]);
      if (num_split == 8) 
	rv->local_id = atoi (split[7]);

      /* If this sequence element is based on a local id, we needn't worry
       * about the buffer_id.
       */
      if (strcmp (filename, "none") == 0 || strcmp (filename, "local") == 0)
	{
	  if (num_split < 8 || rv->local_id < 0)
	    {
	      g_warning ("missing filename and local id");
	      g_free (rv);
	      continue;
	    }
	  /* TODO: breaks tracking of samples in sequencer. */
	  rv->buffer_id = 0;
	}
      /* If we need a buffer_id, decide whether we need to cache any file. */
      else
	{
	  if (is_mp3_filename (filename))
	    description = g_strdup_printf ("Mp3 %s", filename);
	  else
	    description = g_strdup_printf ("Wav %s", filename);
	  info = gdam_buffer_manager_info_by_desc (description);
	  if (info != NULL)
	    rv->buffer_id = info->buffer_id;
	  else
	    {
	      rv->buffer_id = 0;
	      *filename_out = g_strdup (filename);
	    }
	  g_free (description);
	}

      load_thread->under_construction = rv;
      return TRUE;
    }
  return FALSE;
}

/* Do what needs to be done in the buffer-loading thread.  This is called
 * whenever we are in a position to cache a new buffer... ie we have
 * finished caching a buffer or we are beginning to cache the first. 
 */
static void continue_processing_thread(GdamSequencerLoadThread* load_thread)
{
  char *filename;
  filename = NULL;

  /* We shouldn't be waiting on any sort of sample. */
  g_assert (load_thread->under_construction == NULL);
  g_assert (load_thread->local_under_construction == NULL);

  gdam_debug ("processing thread...");

  /* Check if the sequencer has been destroyed. */
  if (load_thread->sequencer == NULL)
    {
      free_load_thread (load_thread);
      return;
    }

  /* While we can read an element or local info from the file we're loading */
  while (parse_element (load_thread, &filename))
    {
      /* If we have a filename to cache, start it caching */
      if (filename)
	{
	  append_cache_request (load_thread, filename);
	  return;
	}
      finished_constructing_element (load_thread);
    }
  if (load_thread->elements)
    {
      gdam_sequencer_set_elements (load_thread->sequencer,
				   load_thread->elements);
      load_thread->elements = NULL;
    }
  free_load_thread (load_thread);
}

void                    gdam_sequencer_load       (GdamSequencer* sequencer,
                                                   const char*    filename)
{
  GdamSequencerLoadThread *loader;
  FILE *fp;
  char *expanded_filename;
  /* Open the file */
  expanded_filename = gdam_tilde_expand (filename);
  fp = fopen (expanded_filename, "r");
  g_free (expanded_filename);
  if (!fp)
    {
      g_warning ("gdam_sequencer_load: error loading %s", filename);
      return;
    }
  gdam_debug ("loading file %s", expanded_filename);
  /* Create a thread to load buffers. */
  loader = g_new0 (GdamSequencerLoadThread, 1);
  loader->fp = fp;
  loader->sequencer = sequencer;
  loader->channel = GDAM_MODEL (sequencer)->channel;
  loader->destroy_connect_id
    = gtk_signal_connect (GTK_OBJECT (sequencer),
			  "destroy",
			  GTK_SIGNAL_FUNC (on_sequencer_destroy), loader);
  loader->elements = NULL;
  loader->under_construction = NULL;
  loader->local_under_construction = NULL;
  loader->filename = NULL;
  /* Start the buffer-loading thread */
  continue_processing_thread (loader);
}

static void write_local_sample (GdamLocalSampleInfo *linfo, FILE* fp)
{
  GdamBufferManagerInfo *info;
  const char *filename = "local";
  info = gdam_buffer_manager_info_by_id (linfo->buffer_id);
  g_return_if_fail (info != NULL);
  if (strncmp (info->description, "Wav ", 4)    != 0
      && strncmp (info->description, "Mp3 ", 4) != 0)
    {
      g_warning ("only sequence elements of type wav, mp3 can be saved.");
      return;
    }
  filename = info->description + 4;
  fprintf (fp, "GdamLocalSampleInfo %s %d %4.4g %4.4g %d %d\n",
	   filename,
	   linfo->local_id,
	   linfo->rate,
	   linfo->gain,
	   linfo->sample_offset,
	   linfo->output_length); 
}

static void write_element(GdamSequencerElement* element, FILE* fp)
{
  GdamBufferManagerInfo *info;
  const char *filename = "none";
  if (element->local_id < 0)
    {
      info = gdam_buffer_manager_info_by_id (element->buffer_id);
      g_return_if_fail (info != NULL);
      if (strncmp (info->description, "Wav ", 4)    != 0
	  && strncmp (info->description, "Mp3 ", 4) != 0)
	{
	  g_warning ("only sequence elements of type wav, mp3 can be saved.");
	  return;
	}
      filename = info->description + 4;
    }
  fprintf (fp, "%s %4.4g %4.4g %4.4g %d %d %d %d\n",
	   filename,
	   element->beat_offset,
	   element->rate,
	   element->gain,
	   element->sample_offset,
	   element->output_length, 
	   element->pattern_number,
	   element->local_id);
}
	

void                    gdam_sequencer_save       (GdamSequencer* sequencer,
                                                   const char*    filename)
{
  GdamBeatInfo* bi;
  guint i;
  FILE *fp;
  char *expanded_filename = gdam_tilde_expand (filename);
  fp = fopen (expanded_filename, "w");
  g_free (expanded_filename);
  if (!fp)
    {
      g_warning ("gdam_sequencer_save: error creating %s", filename);
      return;
    }
  bi = &(GDAM_MODEL (sequencer)->beat_info);
  fprintf (fp, "beats: %4.4g\n", sequencer->num_beats);
  fprintf (fp, "bpm: %4.4g\n", gdam_beat_info_get_bpm(bi, 0));
  g_list_foreach (sequencer->elements, (GFunc) write_element, fp);
  if (sequencer->local_infos)
    {
      for (i = 0; i < sequencer->local_infos->len; i++)
	{
	  GdamLocalSampleInfo *linfo;
	  linfo = &g_array_index (sequencer->local_infos, 
	      			  GdamLocalSampleInfo,
				  i);
	  write_local_sample (linfo, fp);
	}
    }
  fclose (fp);
}
