/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsequencer.h"
#include <gdam/common/gdamsample.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkmarshal.h>
#include "gdambuffermanager.h"
#include <gdam/component/gdamutil.h>
#include <math.h>

enum {
	SET_NUM_BEATS,
	SEQUENCE_CHANGED,
	LAST_SIGNAL
};
static guint sequencer_signals[LAST_SIGNAL] = {0};

typedef struct _GdamSequencerRescaleInfo GdamSequencerRescaleInfo;

struct _GdamSequencerRescaleInfo {
	GdamSequencer* 	sequencer;
	double		rescale_factor;
};

static gboolean gdam_sequencer_match_beat_info(GdamModel*	    model,
				  	       const GdamBeatInfo*  new_info);
static gboolean gdam_sequencer_set_beat_info  (GdamModel*           model,
	                                       const GdamBeatInfo*  new_info);
static void gdam_sequencer_detach_from_buffer(GdamModel*     model,
				              guint          buffer_id);
static void play_by_local_id 	             (GdamSequencer *sequencer,
                                              int            local_id);
static void stop_by_local_id 	             (GdamSequencer *sequencer,
                                              int            local_id);

static void maybe_extend_patterns (GdamSequencer *sequencer, guint num)
{
  g_return_if_fail (sequencer != NULL);
  g_return_if_fail (sequencer->active_patterns != NULL);
  if (num > sequencer->active_patterns->len)
    sequencer->active_patterns = g_array_set_size (sequencer->active_patterns, 
						   num);
}

static void 
arg_set_play_by_local_id (GtkObject *object,
			  GdamArg   *arg)
{
  GdamSequencer *sequencer = GDAM_SEQUENCER (object);
  play_by_local_id (sequencer, GDAM_ARG_INT (arg));
}

static void 
arg_set_stop_by_local_id (GtkObject *object,
			  GdamArg   *arg)
{
  GdamSequencer *sequencer = GDAM_SEQUENCER (object);
  stop_by_local_id (sequencer, GDAM_ARG_INT (arg));
}

static void 
arg_toggle_pattern 	(GtkObject *object, 
                    	 GdamArg   *arg)
{
  GdamSequencer *sequencer = GDAM_SEQUENCER (object);
  int val = GDAM_ARG_INT (arg);

  if (gdam_sequencer_pattern_get_active (sequencer, val))
    gdam_sequencer_pattern_off (sequencer, val);
  else
    gdam_sequencer_pattern_on (sequencer, val);
}

/* Sets the current pattern, i.e. the pattern which is being edited */
static void gdam_sequencer_arg_set_pattern(GtkObject* object,
				            GdamArg*   arg)
{
  GdamSequencer *sequencer = GDAM_SEQUENCER (object);
  int val = GDAM_ARG_INT (arg);
  sequencer->current_pattern = val;
  maybe_extend_patterns (sequencer, val);
  gdam_sequencer_update_sequence (sequencer);
}

static void gdam_sequencer_arg_load_file(GtkObject* object, 
				         GdamArg*   arg)
{
  GdamSequencer *sequencer = GDAM_SEQUENCER (object);
  const char *val = GDAM_ARG_STRING (arg);
  gdam_sequencer_load (sequencer, val);
}

static void gdam_sequencer_arg_get_pattern(GtkObject* object, 
				            GdamArg*   arg)
{
  GdamSequencer *sequencer = GDAM_SEQUENCER (object);
  GDAM_ARG_INT (arg) = sequencer->current_pattern;
}
	
	

void                    gdam_sequencer_set_pattern(GdamSequencer* sequencer,
						   int		  pattern_no)
{
  GdamArg arg;
  arg.type = GDAM_ARG_TYPE_INT;
  arg.name = "pattern";
  GDAM_ARG_INT (&arg) = pattern_no;
  gdam_arg_set_object_arg (GTK_OBJECT (sequencer), &arg);
}

void gdam_sequencer_pattern_on (GdamSequencer* sequencer,
			        int pattern_no)
{
  maybe_extend_patterns (sequencer, pattern_no);
  g_return_if_fail (sequencer->active_patterns != NULL);
  g_array_index (sequencer->active_patterns, gboolean, pattern_no) = TRUE;
  gdam_sequencer_update_sequence (sequencer);
}

void gdam_sequencer_pattern_off(GdamSequencer* sequencer,
				int pattern_no)
{
  maybe_extend_patterns (sequencer, pattern_no);
  g_return_if_fail (sequencer->active_patterns != NULL);
  g_array_index (sequencer->active_patterns, gboolean, pattern_no) = FALSE;
  gdam_sequencer_update_sequence (sequencer);
}

gboolean gdam_sequencer_pattern_get_active (GdamSequencer* sequencer, 
    					int pattern_no)
{
  maybe_extend_patterns (sequencer, pattern_no);
  g_return_val_if_fail (sequencer->active_patterns != NULL, FALSE);
  return g_array_index (sequencer->active_patterns, gboolean, pattern_no);
}

static inline gdouble beat_to_samples(GdamSequencer* sequencer, gdouble beat)
{
  gdouble bpm =
    gdam_beat_info_get_bpm (&GDAM_MODEL (sequencer)->beat_info, 0);
  return (60.0 / bpm * beat * 44100.0);
}

static inline gdouble samples_to_beats(GdamSequencer* sequencer, gdouble pos)
{
  gdouble bpm =
    gdam_beat_info_get_bpm (&GDAM_MODEL (sequencer)->beat_info, 0);
  return (bpm / 60.0 / 44100.0) * pos;
}

static inline gdouble seconds_to_beats(GdamSequencer* sequencer, gdouble pos)
{
  gdouble bpm =
    gdam_beat_info_get_bpm (&GDAM_MODEL (sequencer)->beat_info, 0);
  return (bpm / 60.0) * pos;
}


typedef struct {
  GdamSequencer *sequencer;
  int		 local_buffer_index;
  int		 source_id;
} LocalPlayInfo;

static void 
local_eof (gpointer user_data)
{
  LocalPlayInfo *pi = (LocalPlayInfo*) user_data;
  GdamSequencer *sequencer = pi->sequencer;
  sequencer->local_buffer_id_list =
    g_slist_remove (sequencer->local_buffer_id_list, pi);
  g_free (pi);
}

static void 
detach_success (gpointer user_data)
{
  LocalPlayInfo *pi = (LocalPlayInfo*) user_data;
  GdamSequencer *sequencer = pi->sequencer;
  sequencer->local_buffer_id_list =
    g_slist_remove (sequencer->local_buffer_id_list, pi);
  /* XXX: free here, or on eof??  What if we get both?? */
  g_free (pi);
}

static void
stop_by_local_id (GdamSequencer *sequencer, int local_id)
{
  GSList *cur;
  for (cur = sequencer->local_buffer_id_list; cur != NULL; cur = cur->next)
    {
      LocalPlayInfo *pi = (LocalPlayInfo*)cur->data;
      if (pi->local_buffer_index == local_id)
	{
	  GdamRequestBuilder *builder;
	  GdamRequest *request;
	  builder = gdam_request_builder_new ();
          request = gdam_request_detach_source (pi->source_id, detach_success,
	                                        NULL, pi);
	  gdam_request_builder_append_generic (builder, request);
	  request = gdam_request_builder_package (builder);
	  gdam_channel_append_request (GDAM_MODEL (sequencer)->channel,
	                               request);
	  gdam_request_builder_destroy (builder);
	}
    }
}

static void 
eof_connect_success (int eof_sub_id, gpointer user_data)
{
  LocalPlayInfo *pi = (LocalPlayInfo*) user_data;
  GdamSequencer *sequencer = pi->sequencer;
  gdam_channel_register_handler (GDAM_MODEL (sequencer)->channel,
				 gdam_channel_simple_handler_new (eof_sub_id,
								  TRUE,
								  local_eof,
								  pi));
}

/* Play one sample, without synch or looping. */
static void 
play_by_local_id 	(GdamSequencer *sequencer,
                         int            local_id)
{
  GdamLocalSampleInfo *linfo= NULL;
  GdamModel *model = GDAM_MODEL (sequencer);
  GdamRequestBuilder *builder;
  GdamRequest *request;
  LocalPlayInfo *pi;
  int num_infos = sequencer->local_infos->len;
  int cache_source_id, rate_source_id, gain_source_id;
  int local_source_id = 0;
  int output_offset, output_length, buffer_id;
  int mix_id;
  gdouble rate;

  g_return_if_fail (local_id >= 0);

  if (GDAM_MODEL (sequencer)->source_id == 0)
    {
      g_warning ("can't play by local id, missing source_id");
      return;
    }

  if (local_id >= num_infos)
    {
      g_warning ("invalid local id %d...", local_id);
      return;
    }

  linfo = &g_array_index (sequencer->local_infos,
			  GdamLocalSampleInfo,
			  local_id);
  g_return_if_fail (linfo != NULL);

  builder = gdam_request_builder_new ();

  /* Create a new source with the local sample's attributes. */
  output_offset = linfo->sample_offset;
  output_length = linfo->output_length;
  buffer_id = linfo->buffer_id;

  cache_source_id = gdam_channel_new_local_id (model->channel);

  request = gdam_request_create_source ("GdasCache",
					cache_source_id,
					NULL, NULL,
					sequencer, 3,
					"buffer_id",
					GDAM_ARG_TYPE_INT,
					buffer_id,
					"position",
					GDAM_ARG_TYPE_INT64,
					(gint64) output_offset,
					"length",
					GDAM_ARG_TYPE_INT,
					output_length);
  g_return_if_fail (request != NULL);
  local_source_id = gdam_request_builder_append_create (builder, request);

  /* XXX: get total rate and gain from local, pattern, and buffer infos */
  gain_source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source ("GdasGain",
					gain_source_id,
					NULL,
					NULL,
					//NULL, NULL,
					sequencer, 2,
					"filtered",
					GDAM_ARG_TYPE_INT,
					local_source_id,
					"gain",
					GDAM_ARG_TYPE_DOUBLE,
					linfo->gain);
  g_return_if_fail (request != NULL);
  local_source_id = gdam_request_builder_append_create (builder, request);

  rate = linfo->rate;
  if (linfo->trigger_rate != 0)
    rate *= linfo->trigger_rate;
  rate_source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source ("GdasResample",
					rate_source_id,
					NULL,
					NULL,
					sequencer, 2,
					"filtered",
					GDAM_ARG_TYPE_INT,
					local_source_id,
					"rate",
					GDAM_ARG_TYPE_DOUBLE,
					rate);
  g_return_if_fail (request != NULL);
  local_source_id = gdam_request_builder_append_create (builder, request);

  /* Add new source to sequencer's GdasMix */
  mix_id = model->source_id;
  gdam_request_builder_append_generic (builder, 
            gdam_request_set_args (mix_id,
                                   1,
				   gdam_arg_new_from_valist
				     (1, "add_with_latency",
				      GDAM_ARG_TYPE_INT,
				      local_source_id),
				   NULL, NULL, NULL));
  pi = g_new (LocalPlayInfo, 1);
  pi->sequencer = sequencer;
  pi->local_buffer_index = local_id;
  pi->source_id = local_source_id;
  sequencer->local_buffer_id_list =
    g_slist_prepend (sequencer->local_buffer_id_list, pi);

  /* Track for eof */
  gdam_request_builder_append_generic (builder,
				       gdam_request_subscribe_source_ended
				         (local_source_id,
					  eof_connect_success,
					  NULL, 
					  pi));

  /* Post requests to server */
  request = gdam_request_builder_package (builder);
  gdam_request_builder_destroy (builder);
  gdam_channel_append_request (model->channel, request);
}

static gint* gdam_sequencer_create_array(GdamSequencer* sequencer,
                                         int*           array_length_out)
{
  int num_elements = g_list_length (sequencer->elements);
  gint *rv = g_new (gint, num_elements * 6);
  int i;
  GList *at = sequencer->elements;
  gdouble beat_length = beat_to_samples (sequencer, 1.0);
  gdouble rate, gain;

  for (i = 0; at; i++, at = at->next)
    {
      GdamSequencerElement *element = at->data;
      GdamLocalSampleInfo *linfo= NULL;

      /* If the element will not be heard, skip it entirely */
      if (element->muted
	  || !gdam_sequencer_pattern_get_active (sequencer, 
	   				         element->pattern_number))
	{
	  i--;
	  continue;
	}

      if (element->beat_offset < 0)
	{
	  g_warning ("Negative time value in sequencer...");
	  i--;
	  continue;
	}
      if (element->beat_offset > sequencer->num_beats)
	{
	  g_warning ("Oversized time value in sequencer...");
	  element->beat_offset -= sequencer->num_beats;
	}

      if (element->local_id >= 0)
	{
	  int num_infos = sequencer->local_infos->len;
	  if (element->local_id >= num_infos)
	    {
	      g_warning ("invalid local id %d...", element->local_id);
	      i--;
	      continue;
	    }
	  linfo = &g_array_index (sequencer->local_infos, 
	      			  GdamLocalSampleInfo,
				  element->local_id);
	}

      if (linfo)
      {
	rv[i * 6 + 0] = linfo->buffer_id;
	  if (!linfo->buffer_id)
	    g_warning ("missing linfo->buffer_id");
      }
      else
      {
	rv[i * 6 + 0] = element->buffer_id;
	if (!element->buffer_id)
	  g_warning ("missing el->buffer_id");
      }

      rv[i * 6 + 1] = (gint) (element->beat_offset * beat_length);

      rate = element->rate;
      if (element->rate_factor != 0)
	rate *= element->rate_factor;
      if (linfo)
	rate *= linfo->rate;
      rv[i * 6 + 2] = (gint) (rate * GdamFixed_ONE);

      gain = element->gain * element->gain_factor;
      if (linfo)
	gain *= linfo->gain;
      rv[i * 6 + 3] = (gint) (gain * GdamFixed_ONE);

      if (linfo)
	rv[i * 6 + 4] = linfo->sample_offset + element->sample_offset;
      else
	rv[i * 6 + 4] = element->sample_offset;

      rv[i * 6 + 5] = element->output_length;

      /* If the local info struct limits our length... */
      if (linfo && linfo->output_length >= 0)
	{
	  int possible_length;
	  possible_length = linfo->output_length - element->sample_offset;
	  if (possible_length < 0)
	    {
	      g_warning ("bad possible length %d from li->len %d el->offset %d",
			 possible_length, linfo->output_length, 
			 element->sample_offset);
	      possible_length = 0;
	    }
	  /* If the element requests a certain length, and it is not possible,
	   * then trim the length */
	  if (element->output_length >= 0
	      && possible_length < element->output_length)
	    rv[i * 6 + 5] = possible_length;
	  /* If the element requests full length, give it */
	  else if (element->output_length < 0)
	    rv[i * 6 + 5] = possible_length;
	  /* If the element requests a possible length, honor it */
	  else
	    rv[i * 6 + 5] = element->output_length;
	}

      /* However the length was determined, it refers to the length within
       * the original, unscaled sample.  The sequencer applies this length to
       * the scaled sample.  If we are changing the pitch, then
       * we need to scale the length here to account for it.
       */
      if (rv[i * 6 + 5] > 0)
	rv[i * 6 + 5] = (gint)((gdouble)rv[i * 6 + 5] / rate);
    }

  rv = g_renew (gint, rv, i * 6);
  *array_length_out = i * 6;
  return rv;
}


void gdam_sequencer_update_sequence(GdamSequencer* sequencer)
{
  gint *array;
  int array_length;
  GdamRequest *request;
  GdamModel *this_model = GDAM_MODEL (sequencer);
  GdamArg *arg;

  if (sequencer->frozen)
    return;

  if (this_model->source_id != 0)
    {
      gdouble scale_factor = 1.0;
#if 0
      /* XXX: scale here? */
      int old_len = sequencer->last_sequence_length_samples;
      gdouble scale_factor = sequencer->last_sequence_length_samples;
      scale_factor = sequencer->last_sequence_length_samples / scale_factor;
      g_message ("scale_factor %f from old_len %d len %d",
	         scale_factor, old_len, sequencer->last_sequence_length_samples);
#endif
      array = gdam_sequencer_create_array (sequencer, &array_length);

      arg = gdam_arg_new_from_valist (2,
				      "elements",
				      GDAM_ARG_TYPE_INT_ARRAY,
				      array_length, array,
				      "scale_position",
				      GDAM_ARG_TYPE_DOUBLE,
				      scale_factor);
      request = gdam_request_set_args (sequencer->sequence_source_id,
				       2, arg, NULL, NULL, NULL);
      gdam_channel_append_request (this_model->channel, request);
    }
  gtk_signal_emit(GTK_OBJECT (sequencer), sequencer_signals[SEQUENCE_CHANGED]);
}

/* When the buffer has been created, we set its beat information
 * in the buffer manager. 
 */
static void buffer_create_success(int buffer_id, gpointer data)
{
  GdamBufferManagerInfo *bufferinfo;
  GdamBeatInfo *beat_info;

  beat_info = gdam_beat_info_copy (&GDAM_MODEL (data)->beat_info);
  bufferinfo = gdam_buffer_manager_info_by_id (buffer_id);
  bufferinfo->beat_info = beat_info;
}

static void buffer_create_failure(gpointer data)
{
  gtk_object_unref (GTK_OBJECT (data));
}

/* XXX: the sequence plays repeatedly, making huge cache */
void  gdam_sequencer_cache (GdamSequencer*  sequencer)
{
  GdamRequest *request;
  GdamRequestBuilder *request_builder;
  GdamModel *model = GDAM_MODEL (sequencer);
  gint *array;
  int array_length, sequence_length, source_id, local_source_id;
  char description[256];
  g_return_if_fail (sequencer != NULL);
  sequence_length = (int) beat_to_samples (sequencer, sequencer->num_beats);
  array = gdam_sequencer_create_array (sequencer, &array_length);
  g_return_if_fail (array != NULL);

  source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source ("GdasSequencer2",
					source_id,
					NULL,
					NULL,
					NULL,
					3,
					"sequence_length",
					GDAM_ARG_TYPE_INT,
					sequence_length,
					"elements",
					GDAM_ARG_TYPE_INT_ARRAY,
					array_length, 
					array,
					"length",    /* play only once */
					GDAM_ARG_TYPE_INT64,
					(gint64)sequence_length);
  g_free (array);
  request_builder = gdam_request_builder_new ();
  local_source_id = gdam_request_builder_append_create (request_builder,
							request);
  sprintf (description, "sequence of %f beats", sequencer->num_beats);
  gtk_object_ref (GTK_OBJECT (sequencer));
  gdam_buffer_manager_start_caching (request_builder,
				     local_source_id,
				     model->channel,
				     TRUE,
				     description,
				     buffer_create_success,
				     buffer_create_failure,
				     model);
}


/* TODO: start within the sequence */
static int  gdam_sequencer_play_request(GdamProducer*       producer, 
					GdamPlayInfo*       play_info,
					int                 source_id,
					GdamRequestBuilder* builder)
{
  GdamSequencer *sequencer = GDAM_SEQUENCER (producer);
  GdamRequest *request;
  GdamModel *model = GDAM_MODEL (producer);
  gint *array;
  int rv;
  int array_length;
  int sequence_length;
  g_return_val_if_fail (sequencer != NULL, 0);
  if (GDAM_MODEL (producer)->source_id != 0)
    {
      gdam_debug ("sequencer already playing");
      return 0;
    }
  sequence_length = (int) beat_to_samples (sequencer, sequencer->num_beats);
  array = gdam_sequencer_create_array (sequencer, &array_length);
  g_return_val_if_fail (array != NULL, 0);
  sequencer->last_sequence_length_samples = sequence_length;
  gtk_object_ref (GTK_OBJECT (sequencer));
  sequencer->sequence_source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source ("GdasSequencer2",
					sequencer->sequence_source_id,
					NULL, NULL,
					sequencer,
					2,
					"sequence_length",
					GDAM_ARG_TYPE_INT,
					sequence_length,
					"elements",
					GDAM_ARG_TYPE_INT_ARRAY,
					array_length, array);
  gdam_request_builder_append_create (builder, request);
  g_free (array);

  /* Wrap the sequence in a mix, so we can later add additional sources */
  {
    gint64 mixed[2];
    mixed[0] = 0;  /* position in samples */
    mixed[1] = sequencer->sequence_source_id;

    model->source_id = gdam_channel_new_local_id (model->channel);
    GDAM_PRODUCER (sequencer)->position_source_id = sequencer->sequence_source_id;
    rv = gdam_request_builder_append_create
			  (builder,
			   gdam_request_create_source
			   ("GdasMix", 
			    model->source_id,
			    gdam_model_got_subsource,
			    gdam_model_didnt_get_subsource,
			    sequencer, 2, 
			    "mixed", GDAM_ARG_TYPE_INT64_ARRAY, 2, mixed,
			    "mode", GDAM_ARG_TYPE_STRING, "eternal"));

  }

  sequencer->making_request = 1;
  return rv;
}

static void gdam_sequencer_init(GdamSequencer* sequencer)
{
  sequencer->num_beats = 4;
  sequencer->num_sub_beats = 4;
  gdam_beat_info_set_bpm (&GDAM_MODEL (sequencer)->beat_info, 60.0);

  sequencer->active_patterns = g_array_new (FALSE, TRUE, sizeof (gboolean));
  sequencer->active_patterns 
    			= g_array_set_size (sequencer->active_patterns, 16);
  sequencer->local_infos = g_array_new (FALSE, 
      					TRUE, 
      					sizeof (GdamLocalSampleInfo));
  gdam_sequencer_pattern_on (sequencer, 0);
}

static void gdam_sequencer_class_init(GdamSequencerClass* sequencer_class)
{
  GdamModelClass *model_class;
  GdamProducerClass *producer_class;
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (sequencer_class);
  producer_class = GDAM_PRODUCER_CLASS (sequencer_class);
  model_class = GDAM_MODEL_CLASS (sequencer_class);
  producer_class->play_request = gdam_sequencer_play_request;
  model_class->match_beat_info = gdam_sequencer_match_beat_info;
  model_class->set_beat_info = gdam_sequencer_set_beat_info;
  model_class->detach_from_buffer = gdam_sequencer_detach_from_buffer;
  sequencer_signals[SET_NUM_BEATS] =
    gtk_signal_new ("set_num_beats",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamSequencerClass, set_num_beats),
		    gtk_marshal_NONE__DOUBLE,
		    GTK_TYPE_NONE, 1, GTK_TYPE_DOUBLE);
  sequencer_signals[SEQUENCE_CHANGED] =
    gtk_signal_new ("sequence_changed",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamSequencerClass, sequence_changed),
		    gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class, sequencer_signals, LAST_SIGNAL);
  gdam_arg_register (object_class,
		     "pattern",
		     GDAM_ARG_TYPE_INT,
		     gdam_sequencer_arg_set_pattern,
		     gdam_sequencer_arg_get_pattern, NULL, "0,", 0);
  gdam_arg_register (object_class,
		     "toggle_pattern",
		     GDAM_ARG_TYPE_INT,
		     arg_toggle_pattern,
		     NULL, NULL,
		     "0,", 0);
  gdam_arg_register (object_class,
		     "load_file",
		     GDAM_ARG_TYPE_STRING,
		     gdam_sequencer_arg_load_file,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "sequencer_play_by_local_id",
		     GDAM_ARG_TYPE_INT,
		     arg_set_play_by_local_id,
		     NULL, NULL, "0,", 0);
  gdam_arg_register (object_class,
		     "sequencer_stop_by_local_id",
		     GDAM_ARG_TYPE_INT,
		     arg_set_stop_by_local_id,
		     NULL, NULL, "0,", 0);
}

GtkType gdam_sequencer_get_type()
{
  static GtkType sequencer_type = 0;
  if (!sequencer_type)
    {
      static const GtkTypeInfo sequencer_info = {
	"GdamSequencer",
	sizeof (GdamSequencer),
	sizeof (GdamSequencerClass),
	(GtkClassInitFunc) gdam_sequencer_class_init,
	(GtkObjectInitFunc) gdam_sequencer_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_PRODUCER;
      sequencer_type = gtk_type_unique (parent, &sequencer_info);
    }
  return sequencer_type;
}

GdamSequencer* gdam_sequencer_new()
{
  GdamSequencer *seq = GDAM_SEQUENCER (gtk_type_new (GDAM_TYPE_SEQUENCER));
  return seq;
}


int gdam_sequencer_add_local_info (GdamSequencer   *sequencer,
	                           int              buffer_id,
	                           int              sample_offset,
	                           int              output_len)
{
  GdamLocalSampleInfo *linfo;

  /* XXX: ref?? */
  //gdam_buffer_manager_ref (buffer_id, GDAM_MODEL (sequencer));

  linfo = g_new (GdamLocalSampleInfo, 1);
  linfo->buffer_id = buffer_id;
  linfo->rate = 1.0;
  linfo->gain = 1.0;
  linfo->trigger_rate = 1.0;
  linfo->sample_offset = sample_offset;
  linfo->output_length = output_len;
  linfo->local_id = sequencer->local_infos->len;
  sequencer->local_infos = g_array_append_val (sequencer->local_infos, *linfo);
  return linfo->local_id;
}

void gdam_sequencer_set_num_local_infos (GdamSequencer *sequencer,
    					 int		num)
{
  int cur_num = sequencer->local_infos->len;
  while (cur_num < num)
    cur_num = gdam_sequencer_add_local_info (sequencer, 0, 0, -1) + 1;
}

/* Allocate a new element and set default values */
GdamSequencerElement *gdam_sequencer_new_element ()
{
  GdamSequencerElement *element;
  element = g_new (GdamSequencerElement, 1);
  element->buffer_id = 0;
  element->beat_offset = 0;
  element->rate = 1.0;
  element->gain = 1.0;
  element->sample_offset = 0;
  element->output_length = -1;
  element->gain_factor = 1;
  element->rate_factor = 1;
  element->muted = 0;
  element->local_id = -1;
  element->pattern_number = 0;
  return element;
}

GdamSequencerElement*	gdam_sequencer_add_element(GdamSequencer* sequencer,
	                                           int            buffer_id,
						   gdouble        beat_offset,
	                                           gdouble        rate,
	                                           gdouble        gain,
	                                           int            sample_offset,
	                                           int            output_len)
{
  GdamSequencerElement *element;
  g_return_val_if_fail (buffer_id != 0, NULL);

  gdam_debug ("gdam_sequencer_add_element: "
	      "id=%d, beat=%g, rate/gain=%g/%g; outputlen=%d",
	      buffer_id, beat_offset, rate, gain, output_len);

  gdam_buffer_manager_ref (buffer_id, GDAM_MODEL (sequencer));

  element = gdam_sequencer_new_element ();
  element->buffer_id = buffer_id;
  element->beat_offset = beat_offset;
  element->rate = rate;
  element->gain = gain;
  element->sample_offset = sample_offset;
  element->output_length = output_len;
  element->pattern_number = sequencer->current_pattern;
  sequencer->elements = g_list_prepend (sequencer->elements, element);
  if (sequencer->frozen == 0)
    gdam_sequencer_update_sequence (sequencer);
  return element;
}

typedef struct _ImmediateAddition ImmediateAddition;
struct _ImmediateAddition {
	GdamSequencerElement* element;
	GdamSequencer* sequencer;
};

static void successfully_got_time(int num_args, GdamArg* args, gpointer data)
{
  ImmediateAddition *ia = (ImmediateAddition *) data;
  gint64 position;
  gint64 sequence_length;
  gdouble beatnum, orig;
  GdamSequencer *sequencer = ia->sequencer;
  position = GDAM_ARG_INT64 (args);
  sequence_length = (int) beat_to_samples (sequencer, sequencer->num_beats);
  if (sequencer->immediate_lag > 0)
    {
      gint64 pos = position;
      pos -= (gint64) (sequencer->immediate_lag * 44100.0);
      if (pos < 0)
	{
	  pos += sequence_length;
	}
      position = pos;
    }

  orig = beatnum = samples_to_beats (sequencer, position);

  /* TODO: make this snapping optional?? */
  if (sequencer->num_sub_beats)
    {
      beatnum *= sequencer->num_sub_beats;
      beatnum += 0.5;
      beatnum = floor (beatnum);
      beatnum /= sequencer->num_sub_beats;
    }

  gdam_debug ("SEQ: ADD_IMM: delta latency: %f", orig - beatnum);

  ia->element->beat_offset = beatnum;
  sequencer->elements = g_list_prepend (sequencer->elements, ia->element);
  if (sequencer->frozen == 0)
    gdam_sequencer_update_sequence (sequencer);
  g_free (ia);
}
static void didnt_successfully_get_time(gpointer data)
{
  ImmediateAddition *ia = (ImmediateAddition *) data;
  g_free (ia->element);
  g_free (ia);
}
GdamSequencerElement*	gdam_sequencer_add_element_immediate
                                                  (GdamSequencer* sequencer,
	                                           int            buffer_id,
	                                           gdouble        rate,
	                                           gdouble        gain,
	                                           int            sample_offset,
	                                           int            output_len)
{
  ImmediateAddition *ia = g_new (ImmediateAddition, 1);
  GdamSequencerElement *element;
  GdamModel *model = GDAM_MODEL (sequencer);
  const char *arg_names[] = { "position" };

  gdam_debug ("gdam_sequencer_add_element: "
	      "id=%d, beat=TBD, rate/gain=%g/%g; outputlen=%d",
	      buffer_id, rate, gain, output_len);

  gdam_buffer_manager_ref (buffer_id, GDAM_MODEL (sequencer));
  element = gdam_sequencer_new_element ();
  element->buffer_id = buffer_id;
  element->beat_offset = -1;
  element->rate = rate;
  element->gain = gain;
  element->sample_offset = sample_offset;
  element->output_length = output_len;
  element->pattern_number = sequencer->current_pattern;
  ia->element = element;
  ia->sequencer = sequencer;
  if (model->source_id == 0)
    {
      gdouble stopped_position;
      stopped_position = GDAM_PRODUCER (sequencer)->stopped_position;
      ia->element->beat_offset = seconds_to_beats (sequencer,
						   stopped_position);
      sequencer->elements = g_list_prepend (sequencer->elements, ia->element);
      return element;
    }

  gdam_channel_append_request (model->channel,
			       gdam_request_get_args (sequencer->sequence_source_id,
						      1,
						      arg_names,
						      successfully_got_time,
						      didnt_successfully_get_time,
						      ia));
  return element;
}

static void remove_model(GdamSequencerElement* element, GdamSequencer* seq)
{
  gdam_sequencer_remove_element (seq, element);
}

static void gdam_sequencer_detach_from_buffer(GdamModel*     model,
				              guint          buffer_id)
{
  GSList *kill_elements = NULL;
  GList *elements;
  GdamSequencer *sequencer = GDAM_SEQUENCER (model);
  for (elements = sequencer->elements;
       elements != NULL; elements = elements->next)
    {
      GdamSequencerElement *element = elements->data;
      if (element->buffer_id == (int) buffer_id)
	kill_elements = g_slist_prepend (kill_elements, element);
    }
  if (kill_elements)
    {
      gdam_sequencer_freeze (sequencer);
      g_slist_foreach (kill_elements, (GFunc) remove_model, sequencer);
      gdam_sequencer_thaw (sequencer);
      gtk_signal_emit (GTK_OBJECT (sequencer),
		       sequencer_signals[SEQUENCE_CHANGED]);
      g_slist_free (kill_elements);
    }
}

void gdam_sequencer_remove_element(GdamSequencer* sequencer,
				   GdamSequencerElement* element)
{
  sequencer->elements = g_list_remove (sequencer->elements, element);
  gdam_buffer_manager_unref (element->buffer_id, GDAM_MODEL (sequencer));
  g_free (element);
  gdam_sequencer_update_sequence (sequencer);
}

void gdam_sequencer_set_num_beats (GdamSequencer* sequencer,
                                   gdouble        num_beats)
{
  gdouble scale_factor = 1.0;
  int old_len = sequencer->last_sequence_length_samples;
  GdamModel *model = GDAM_MODEL (sequencer);
  int length = beat_to_samples (sequencer, num_beats);
  gdam_debug("sequencer_set_num_beats to %f", num_beats);
  sequencer->last_sequence_length_samples = length;

  if (old_len > 0)
    scale_factor = (gdouble)sequencer->last_sequence_length_samples / old_len;

  sequencer->num_beats = num_beats;

  gtk_signal_emit (GTK_OBJECT (sequencer),
		   sequencer_signals[SET_NUM_BEATS], num_beats);
  gtk_signal_emit_by_name (GTK_OBJECT (sequencer),
			   "update_length", (gdouble) length / 44100.0);

  if (model->source_id == 0)
    return;

  gdam_channel_append_request (model->channel,
			       gdam_request_set_args (sequencer->sequence_source_id,
						      2,
						      gdam_arg_new_from_valist
							(2,
							 "scale_position",
							 GDAM_ARG_TYPE_DOUBLE,
							 scale_factor,
							 "sequence_length",
							 GDAM_ARG_TYPE_INT,
							 length),
						      NULL, NULL, NULL));
}

void gdam_sequencer_freeze(GdamSequencer* sequencer)
{
  sequencer->frozen++;
}

void gdam_sequencer_thaw(GdamSequencer* sequencer)
{
  g_return_if_fail (sequencer->frozen > 0);
  if (--sequencer->frozen == 0)
    gdam_sequencer_update_sequence (sequencer);
}

static void clear_one(GdamSequencerElement* element, GdamModel* sequencer)
{
  gdam_buffer_manager_unref (element->buffer_id, sequencer);
  g_free (element);
}
void gdam_sequencer_clear(GdamSequencer* sequencer)
{
  g_list_foreach (sequencer->elements, (GFunc) clear_one, sequencer);
  g_list_free (sequencer->elements);
  sequencer->elements = NULL;
  gdam_sequencer_update_sequence (sequencer);
}

static void reference_element_buffer(GdamSequencerElement* element,
                                     GdamModel*            model)
{
  gdam_buffer_manager_ref (element->buffer_id, model);
}

void gdam_sequencer_set_elements(GdamSequencer* sequencer,
                                 GList*         elements)
{
  g_list_foreach (elements, (GFunc) reference_element_buffer, sequencer);
  gdam_sequencer_freeze (sequencer);
  gdam_sequencer_clear (sequencer);
  sequencer->elements = elements;
  gdam_sequencer_thaw (sequencer);
  gtk_signal_emit (GTK_OBJECT (sequencer),
		   sequencer_signals[SEQUENCE_CHANGED]);
}

static void rescale_one(GdamSequencerElement* element, 
			GdamSequencerRescaleInfo* rescale_info)
{
  element->beat_offset *= rescale_info->rescale_factor;
}

/* Rescale the position of elements within a sequence
   by scale_factor.  TODO: incorporate set_length? */
void gdam_sequencer_rescale_elements(GdamSequencer* sequencer, 
				     gdouble	    scale_factor)
{
  GdamSequencerRescaleInfo rescale_info;
  rescale_info.sequencer = sequencer;
  rescale_info.rescale_factor = scale_factor;
  g_list_foreach (sequencer->elements, (GFunc) rescale_one, &rescale_info);
}

static gboolean gdam_sequencer_set_beat_info(GdamModel*           model,
	                                      const GdamBeatInfo* new_info)
{
  model->beat_info = *new_info;
  gdam_sequencer_update_sequence (GDAM_SEQUENCER (model));
  return TRUE;
}

/* Match beat info - adjusts length of sequence and position of
 * samples so that the tempo of the sequence matches the tempo
 * passed it.  As the sequence can only match beat info by setting
 * its real beat info, this is done.
 */
static gboolean gdam_sequencer_match_beat_info(GdamModel*	      model,
				  	       const GdamBeatInfo*    new_info)
{
  GdamSequencer *sequencer;
  sequencer = GDAM_SEQUENCER (model);
  gdam_sequencer_freeze (sequencer);
  gdam_model_set_beat_info (model, new_info);
  gdam_sequencer_set_num_beats (sequencer, sequencer->num_beats);
  gdam_sequencer_thaw (sequencer);
  return TRUE;
}
