/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdambender.h"
#include <gdam/component/gdampitch.h>
#include <gdam/common/gdamsample.h>
#include <gdam/common/gdamplugin.h>
#include <gtk/gtksignal.h>
#include <gdam/component/gdammodel.h>
#include <gdam/component/gdamfilter.h>
#include <gdam/component/gdamutil.h>

enum {
	SLOW_TO_STALL,
	LAST_SIGNAL
};

static guint bender_signals[LAST_SIGNAL];

/* Bender class function */
static void bender_real_slow_to_stall(GdamBender* bender, gdouble slow_time);
/* Filter class function */
static int gdam_bender_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder);

static void gdam_bender_init(GdamBender* bender)
{
  bender->backbuffer = 8.0;
  bender->beat_factor = 0.5;
  bender->rate_factor = 1.0;
  bender->mode = GDAM_BENDER_KEEP_TIME;
}

static void gdam_bender_class_init(GdamBenderClass* klass);
GtkType gdam_bender_get_type()
{
  static GtkType bender_type = 0;
  if (!bender_type)
    {
      static const GtkTypeInfo bender_info = {
	"GdamBender",
	sizeof (GdamBender),
	sizeof (GdamBenderClass),
	(GtkClassInitFunc) gdam_bender_class_init,
	(GtkObjectInitFunc) gdam_bender_init,
	/* reserved 1 */ NULL,
	/* reserved 2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      bender_type = gtk_type_unique (GDAM_TYPE_FILTER, &bender_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return bender_type;
}

static void made_request(gpointer data)
{
  GdamBender *bender = GDAM_BENDER (data);
  gdam_debug ("made request");
  if (bender->pending_request != NULL)
    {
      gdam_debug ("handling pending request");
      gdam_channel_append_request (GDAM_MODEL (bender)->channel,
				   bender->pending_request);
      bender->pending_request = NULL;
    }
  else
    {
      bender->making_request = FALSE;
    }
}

static void bender_real_slow_to_stall(GdamBender* bender, gdouble slow_time)
{
  g_warning ("deprecated method: slow_to_stall");
}

GdamBender* 	gdam_bender_new		()
{
  GdamBender *bender;
  bender = GDAM_BENDER (gtk_type_new (GDAM_TYPE_BENDER));
  return bender;
}

static gdouble
get_block_size_seconds (GdamBender *bender)
{
  gdouble block_size;
  gdouble bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bender)->beat_info, 0);
  if (bpm > 1)
    block_size = bender->beat_factor * 60.0 / bpm;
  else
    block_size = bender->beat_factor;
  if (block_size < 0.01)
    block_size = 1.0;
  return block_size;
}

static char *
get_mode_string (GdamBender *bender)
{
  if (bender->mode == GDAM_BENDER_KEEP_TIME)
    return g_strdup ("keep_time");
  else if (bender->mode == GDAM_BENDER_SNAP_TIME)
    return g_strdup ("snap_time");
  else if (bender->mode == GDAM_BENDER_CONTINUOUS)
    return g_strdup ("continuous");
  else if (bender->mode == GDAM_BENDER_RETURN)
    return g_strdup ("return");
  else if (bender->mode == GDAM_BENDER_KEEP_IN_BEAT)
    return g_strdup ("keep_in_beat");
  g_warning ("unknown mode %d", bender->mode);
  return NULL;
}

static int gdam_bender_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder)
{
  GdamBender *bender = GDAM_BENDER (filter);
  GdamModel *model = GDAM_MODEL (filter);
  GdamRequest *request;
  char *mode_string = get_mode_string (bender);
  gdouble block_size;

  if (!mode_string)
    {
      g_warning ("unknown mode %d", bender->mode);
      mode_string = g_strdup ("keep time");
    }

  block_size = get_block_size_seconds (bender);

  gtk_object_ref (GTK_OBJECT (bender));
  gdam_debug ("append_start_request: building bender from %d", new_source_id);
  model->source_id = gdam_channel_new_local_id (model->channel);
  /* TODO: set back buffer, mode */
  request = gdam_request_create_source ("GdasBender",
					model->source_id,
					gdam_model_got_subsource,
					gdam_model_didnt_get_subsource,
					bender,
					4,
					"filtered",
					GDAM_ARG_TYPE_INT,
					new_source_id,
					"buffer_length",
					GDAM_ARG_TYPE_INT,
					(int) (bender->backbuffer * 44100.0),
					"beat_length",
					GDAM_ARG_TYPE_INT,
					(int) (block_size * 44100.0),
					"change_mode",
					GDAM_ARG_TYPE_STRING,
					mode_string
					);
  return gdam_request_builder_append_create (builder, request);
}

void gdam_bender_slow_to_stall(GdamBender* bender, float slow_time)
{
  GtkObject *this_object = GTK_OBJECT (bender);

  if (gtk_signal_n_emissions (this_object, bender_signals[SLOW_TO_STALL]))
    return;

  gtk_signal_emit (this_object, bender_signals[SLOW_TO_STALL], slow_time);
}

void		gdam_bender_set_mode	(GdamBender*		bender,
					 GDAM_BENDER_MODE	mode)
{
  GdamArg args[2];
  int num_args = 1;
  char *mode_string;

  bender->mode = mode;
  gdam_debug ("Bender set_mode to %d", mode);

  args[0].name = "change_mode";
  args[0].type = GDAM_ARG_TYPE_STRING;

  mode_string = get_mode_string (bender);
  if (mode_string == NULL)
    return;

  GDAM_ARG_STRING (args) = mode_string;

  if (mode == GDAM_BENDER_KEEP_IN_BEAT)
    {
      gdouble block_size;
      gdouble bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bender)->beat_info, 0);
      if (bpm > 1)
	block_size = bender->beat_factor * 60.0 / bpm;
      else
	block_size = bender->beat_factor;
      num_args = 2;
      args[1].name = "beat_length";
      args[1].type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (args+1) = (int)(44100.0 * block_size);
      GDAM_ARG_STRING (args) = g_strdup ("keep_in_beat");
    }

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       num_args, args, NULL, NULL, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
    }
}


/* Posts a set_args request, avoiding congestion by metering out one
 * request at a time, and keeping only the most recent update. */
static void
bender_post_arg_requests 	(GdamBender 	*bender,
                                 GdamArg        *args,
				 int             num_args)
{
  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       num_args,
				       args,
				       made_request, made_request, bender);
      if (bender->making_request)
	{
	  if (bender->pending_request != NULL)
	    gdam_request_destroy (bender->pending_request);
	  bender->pending_request = request;
	}
      else
	{
	  bender->making_request = TRUE;
	  gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
	}
    }
  else
    {
      g_message ("skipping bender arg requests, no source id");
    }
}

/* Set the bender's changes to a list of [time_in_samples_output,rate] pairs. */
void		gdam_bender_set_changes	(GdamBender*		bender,
					 int			num_points,
					 int*			changes)
{
  int *drag_data = changes;
  GdamArg args[2];

  /* TODO: fix this memory leak */
  /*memcpy(drag_data, changes, sizeof(int)*num_points*2); */

#if 0
  int n;
  gdam_debug ("Bender set_changes, %d points", num_points);
  fprintf (stderr, "Bender array:");
  for (n = 0; n < num_points * 2; n += 2)
    {
      fprintf (stderr, "  %f %f", (float) drag_data[n] / 44100,
	       (float) drag_data[n + 1] / GdamFixed_ONE);
    }
#endif

  /* TODO: copy this data? */
  args[0].name = "changes";
  args[0].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args).length = num_points ? num_points * 2 : 0;
  GDAM_ARG_INT_ARRAY (args).array = drag_data;

  bender_post_arg_requests (bender, args, 1);
}



/*  If we are in KEEP_TIME mode, where the underlying song continues to
 *  play while the output is bent, then snap back to where we should
 *  have been if no bending had occurred.
 */
void		gdam_bender_jump_to_mark(GdamBender*		bender)
{
  GdamArg args[2];

  gdam_debug ("Bender jump_to_mark");

  if (bender->mode != GDAM_BENDER_KEEP_TIME &&
      bender->mode != GDAM_BENDER_KEEP_IN_BEAT)
    {
      g_warning ("can't jump to mark in continuous mode...");
      return;
    }

  args[0].name = "jump_to_mark";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args) = 0;

  args[1].name = "changes";
  args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 1).length = 0;
  GDAM_ARG_INT_ARRAY (args + 1).array = NULL;

  bender_post_arg_requests (bender, args, 2);
}

/* Set the bender to play at 100% speed. */
void		gdam_bender_passthrough (GdamBender*		bender)
{
  GdamArg args[1];

  g_message ("bender passthrough!");

  args[0].name = "changes";
  args[0].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 0).length = 0;
  GDAM_ARG_INT_ARRAY (args + 0).array = NULL;

  bender_post_arg_requests (bender, args, 1);
}



void		gdam_bender_set_back_buffer	
					(GdamBender*		bender,
					 float			length)
{
  GdamArg args[1];

  gdam_debug ("Bender set_back_buffer");

  args[0].name = "back_buffer";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args) = length * 44100.0;

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       1,
				       args,
				       gdam_do_nothing,
				       gdam_do_nothing, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
    }
  bender->backbuffer = length;
}

void		gdam_bender_rewind	(GdamBender*		bender,
					 float			rewind_length)
{
  float r_length = 0;
  int num_points = 1;
  int *drag_data = g_new (int, num_points * 2);
  float r_f = bender->rate_factor ? bender->rate_factor : 1.0;
  GdamArg args[2];

  drag_data[0] = 1000.0 * 44100.0 / r_f;
  drag_data[1] = -1 * r_f * GdamFixed_ONE;

  args[0].name = "change_mode";
  args[0].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args) = g_strdup ("keep_time");
  args[1].name = "changes";
  args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 1).length = num_points * 2;
  GDAM_ARG_INT_ARRAY (args + 1).array = drag_data;

  bender_post_arg_requests (bender, args, 2);
}

static int delta_millis(GTimeVal* older, GTimeVal* younger)
{
  int millis = (older->tv_sec - younger->tv_sec) * 1000
    + (older->tv_usec - younger->tv_usec) / 1000;
  return millis;
}

static void gdam_bender_arg_jump_to_mark(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  if (GDAM_ARG_INT (arg) >= 1)
    gdam_bender_jump_to_mark (bender);
}

static void 
gdam_bender_arg_set_position_marker (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamArg args[1];

  args[0].name = "set_position_marker";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args) = GDAM_ARG_INT (arg);

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       1,
				       args,
				       gdam_do_nothing,
				       gdam_do_nothing, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
    }
}


static void 
gdam_bender_arg_jump_to_position_marker (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamArg args[1];

  args[0].name = "jump_to_position_marker";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args) = GDAM_ARG_INT (arg);

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       1,
				       args,
				       gdam_do_nothing,
				       gdam_do_nothing, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
    }
}


static void 
gdam_bender_arg_jump_to_nearest_mark (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamArg args[3];
  gdouble bpm, block_size;
  int block_samples;

  g_return_if_fail (bender->beat_factor > 0);

  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bender)->beat_info, 0);
  if (bpm > 1)
    block_size = bender->beat_factor * 60.0 / bpm;
  else
    block_size = bender->beat_factor;

  block_samples = (int)(44100.0 * block_size);

  gdam_debug ("Bender jump_to_nearest_mark");

  if (bender->mode != GDAM_BENDER_KEEP_TIME &&
      bender->mode != GDAM_BENDER_KEEP_IN_BEAT)
    {
      g_warning ("can't jump to mark in continuous mode...");
      return;
    }

  args[0].name = "jump_to_nearest_mark";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args) = block_samples;

  args[1].name = "changes";
  args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 1).length = 0;
  GDAM_ARG_INT_ARRAY (args + 1).array = NULL;

  args[2].name = "change_mode";
  args[2].type = GDAM_ARG_TYPE_STRING;
  if (bender->mode == GDAM_BENDER_KEEP_TIME)
    GDAM_ARG_STRING (args+ 2) = g_strdup ("keep_time");
  else
    GDAM_ARG_STRING (args+ 2) = g_strdup ("keep_in_beat");

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       3,
				       args,
				       gdam_do_nothing,
				       gdam_do_nothing, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
    }

}

static void gdam_bender_arg_set_autotune(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->autotune = GDAM_ARG_INT (arg);
}

static void gdam_bender_arg_get_autotune(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->autotune;
}

#define GDAM_BEND_DELTA_CEILING 200  
static void gdam_bender_arg_set_bend_amount(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GTimeVal real_time;
  int num_points = 10;
  float total_time = 0.2;
  int *change_array = g_new (int, num_points * 2);
  int n, delmil;
  float last_rate = bender->last_rate;
  float new_rate;
  float cur_rate = last_rate;
  float step_rate;
  float step_time;
  int pitch = (int) (GDAM_ARG_DOUBLE (arg) * 8192.0 * 2);
  gboolean jump_to_mark = FALSE;

  /* XXX: for now, out-of-range value means not bending. */
  if ((pitch < 0 || pitch > 16383))
    {
      if (!bender->is_bending)
	{
	  return;
	}
      gdam_debug ("HANDS OFF - no longer bending...");
      /*  TODO: re-enable this once it doesn't crash server. */
      jump_to_mark = TRUE;
      /*gdam_bender_jump_to_mark(bender); */
      new_rate = 1.0;
      bender->is_bending = FALSE;
    }
  else
    {
      bender->is_bending = TRUE;
      /* We fake (for now) raw pitch range: 0 <- 8192 -> 16383 */
      /* Pitch Wheel Mode:  0 <- 1 -> 2 */
      if (bender->bend_mode == 0)
	{
	  new_rate = (float) pitch / 8192;
	}
      /* Pitch Wheel Mode:  -2 <- 0 |1| 0-> 2 */
      else if (bender->bend_mode == 1)
	{
	  if (pitch == 8192)
	    new_rate = 1.0;
	  else if (pitch < 8192)
	    new_rate = (float) (8192 - pitch) * -2 / 8192.0;
	  else
	    new_rate = (float) (8192 - pitch) * -2 / 8192.0;
	}
      /* Pitch Wheel Mode:  -2 <- 1 -> 2 */
      else if (bender->bend_mode == 2)
	{
	  if (pitch >= 8192)
	    new_rate = (float) pitch / 8192.0;
	  else if (pitch > 5000)
	    new_rate = (float) (pitch - 5000) * -1 / 3192.0;
	  else
	    new_rate = (float) (pitch - 5000) * 2 / 5000.0;
	}
      /* TODO: Pitch Wheel Mode:  -3 <- -1 |1| 1 -> 0 */
      else
	{
	  g_warning ("Confused in bend.");
	  new_rate = 1.0;
	}
    }

  if (bender->autotune)
    new_rate = gdam_pitch_get_magic_rate (new_rate);

  step_rate = (new_rate - last_rate) / num_points;

  g_get_current_time (&real_time);
  delmil = delta_millis (&real_time, &bender->bend_adj_time);
  g_get_current_time (&bender->bend_adj_time);

  if (delmil < GDAM_BEND_DELTA_CEILING)
    {
      total_time = 0.001 * delmil;
    }
  if (delmil == 0)
    {
      total_time = 0.001;
    }

  step_time = total_time / num_points;
  bender->last_rate = new_rate;

  gdam_debug ("midi bend wheel changed!!!");
  gdam_debug ("delmil: %d", delmil);

  for (n = 0; n < num_points * 2; n += 2)
    {
      cur_rate += step_rate;
      change_array[n] = step_time * 44100;
      if (n == num_points * 2 - 2)
	change_array[n] = 1000.0 * 44100;
      change_array[n + 1] = cur_rate * GdamFixed_ONE;
    }
  gdam_bender_set_changes (bender, num_points, change_array);
  if (jump_to_mark)
    gdam_bender_jump_to_mark (bender);
}


static void 
gdam_bender_stall (GdamBender *bender, gdouble factor)
{
  //float last_rate = bender->last_rate;
  float last_rate = 1.0;
  float cur_rate = last_rate;
  float new_rate, step_rate, step_time;
  float bpm, beat_length;
  float total_time = 0.2;
  int num_points = 300;
  int *change_array = g_new (int, num_points * 2);
  int n;


  if (bender->beat_factor <= 0)
    return;
  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bender)->beat_info, 0);
  if (bpm > 1)
    total_time = factor * bender->beat_factor * 60.0 / bpm;
  else
    total_time = factor * bender->beat_factor;

  new_rate = 0.05;
  step_rate = (new_rate - last_rate) / num_points;
  step_time = total_time / num_points;
  bender->last_rate = new_rate;

  for (n = 0; n < num_points * 2; n += 2)
    {
      cur_rate += step_rate;
      change_array[n] = step_time * 44100;
      if (n == num_points * 2 - 2)
	change_array[n] = 1000.0 * 44100;
      change_array[n + 1] = cur_rate * GdamFixed_ONE;
    }
  gdam_bender_set_changes (bender, num_points, change_array);
}





void		gdam_bender_rub		(GdamBender*		bender,
					 float			rub_width,
					 float			rub_length)
{
  float r_length = 0;
  float r_width = 0;
  float r_f = bender->rate_factor ? bender->rate_factor : 1.0;
  int num_points = 1;
  int *drag_data;
  int n;
  GdamArg args[2];

  if (!rub_width)
    {
      float bpm;
      if (bender->beat_factor <= 0)
	return;
      bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bender)->beat_info, 0);
      if (bpm > 1)
	r_width = bender->beat_factor * 60.0 / bpm;
      else
	r_width = bender->beat_factor;
    }
  else if (rub_width < 0)
    {
      /* TODO: what to do? */
      return;
    }
  else
    r_width = rub_width;
  if (!rub_length)
    {
      r_length = 180.0;
    }
  else if (rub_length < 0)
    {
      return;
    }
  else
    r_length = rub_length;

  g_return_if_fail (r_width > 0);
  g_return_if_fail (r_length > 0);


  gdam_debug ("Bender rub, width: %f length: %f", r_width, r_length);


  /*  Rub rates weighted +/- rate */
  if (bender->rub_mode == 1)
    {
      float rate0 = r_f;
      float rate1 = (1.0 + (1.0 - r_f));
      float time0 = rate1 * 2 * r_width / (rate0 + rate1);
      float time1 = 2 * r_width - time0;

      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = time0 * 44100;
	  drag_data[n + 1] = -1.0 * rate0 * GdamFixed_ONE;
	  drag_data[n + 2] = time1 * 44100;
	  drag_data[n + 3] = rate1 * GdamFixed_ONE;
	}
    }
  /*  Rub rates weighted +rate, normal */
  else if (bender->rub_mode == 2)
    {
      float rate0 = r_f;
      float rate1 = 1.0;
      float time0 = rate1 * 2 * r_width / (rate0 + rate1);
      float time1 = 2 * r_width - time0;

      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = time0 * 44100;
	  drag_data[n + 1] = -1.0 * rate0 * GdamFixed_ONE;
	  drag_data[n + 2] = time1 * 44100;
	  drag_data[n + 3] = rate1 * GdamFixed_ONE;
	}
    }
  else if (bender->rub_mode == 3)
    {
      float cur_rate = r_f * GdamFixed_ONE;
      float num_samples = r_width * 44100 / r_f;
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 3);
      for (n = 0; n < num_points * 3; n += 6)
	{
	  drag_data[n + 0] = num_samples / 2;
	  drag_data[n + 1] = -2.0 * cur_rate;
	  drag_data[n + 2] = num_samples * 5 / 10;
	  drag_data[n + 3] = 0;
	  drag_data[n + 4] = num_samples;
	  drag_data[n + 5] = 1.0 * cur_rate;
	  cur_rate *= 0.9;
	}
    }
  else if (bender->rub_mode == 4)
    {
      float cur_rate = r_f * GdamFixed_ONE;
      float num_samples = r_width * 44100 / r_f;
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = num_samples;
	  drag_data[n + 1] = -1.0 * cur_rate;
	  drag_data[n + 2] = num_samples;
	  drag_data[n + 3] = 1.0 * cur_rate;
	  cur_rate *= 1.12;
	}
    }
  else if (bender->rub_mode == 5)
    {
      float cur_rate = r_f * GdamFixed_ONE;
      float num_samples = r_width * 44100 / r_f;
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = num_samples;
	  drag_data[n + 1] = -1.0 * cur_rate;
	  drag_data[n + 2] = num_samples;
	  drag_data[n + 3] = 1.0 * cur_rate;
	  cur_rate *= 1.12;
	  num_samples *= 0.75;
	}
    }
  else if (bender->rub_mode == 6)
    {
      float cur_rate = r_f * GdamFixed_ONE;
      float num_samples = r_width * 44100 / r_f;
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = num_samples;
	  drag_data[n + 1] = -1.0 * cur_rate;
	  if (n == 12)
	    {
	      cur_rate *= 1.12;
	    }
	  drag_data[n + 2] = num_samples;
	  drag_data[n + 3] = 1.0 * cur_rate;
	  if (n == 4)
	    {
	      num_samples /= 3;
	    }
	  else if (n == 8)
	    {
	      num_samples *= 3;
	      num_samples /= 4;
	      cur_rate *= 1.12;
	    }
	  else if (n == 12)
	    {
	      cur_rate *= 1.12;
	    }
	  else if (n == 16)
	    {
	      cur_rate /= 1.12;
	      cur_rate /= 1.12;
	      cur_rate /= 1.12;
	      cur_rate /= 1.12;
	    }
	  else if (n == 20)
	    {
	      cur_rate *= 1.12;
	    }
	}
    }
  else if (bender->rub_mode == 7)
    {
      float cur_rate = r_f * GdamFixed_ONE;
      float num_samples = r_width * 44100 / r_f;
      float factor = 1.12;
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = num_samples;
	  drag_data[n + 1] = 1.0 * cur_rate;
	  drag_data[n + 2] = num_samples;
	  drag_data[n + 3] = 1.0 * cur_rate * 0;
	  cur_rate *= factor;
	  if ((n == 16) || (n == 32) || (n == 48))
	    {
	      factor = 1 / factor;
	    }
	  if (n == 32) 
	    {
	      num_samples /= 2;
	    }
	  else if (n == 40) 
	    {
	      num_samples *= 2;
	    }
	}
    }
  /* loops the recent past in reverse, jumping forward when we reach the
   * beginning of the sample. */
  else if (bender->rub_mode == 8)
    {
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = r_width * 44100 / r_f;
	  drag_data[n + 1] = -1.0 * r_f * GdamFixed_ONE;
	  drag_data[n + 2] = 0;
	  drag_data[n + 3] = r_width * 44100 / r_f;
	}
    }
  /* spin via backwards jumps */
  else if (bender->rub_mode == 9)
    {
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = r_width * 44100 / r_f;
	  drag_data[n + 1] = 1.0 * r_f * GdamFixed_ONE;
	  drag_data[n + 2] = 0;
	  drag_data[n + 3] = r_width * -44100 / r_f;
	}
    }
  /*  Rub rates even.  */
  else
    {
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = r_width * 44100 / r_f;
	  drag_data[n + 1] = -1.0 * r_f * GdamFixed_ONE;
	  drag_data[n + 2] = r_width * 44100 / r_f;
	  drag_data[n + 3] = r_f * GdamFixed_ONE;
	}
    }

  args[0].name = "change_mode";
  args[0].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args) = g_strdup ("keep_time");
  args[1].name = "changes";
  args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 1).length = num_points * 2;
  GDAM_ARG_INT_ARRAY (args + 1).array = drag_data;

  bender_post_arg_requests (bender, args, 2);
}

/* Helper which determines the segment size of a bend effect */
static gdouble
get_rub_width 		(GdamBender 	*bender)
{
  float bpm;
  if (bender->beat_factor <= 0)
    return 1.0;
  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bender)->beat_info, 0);
  if (bpm > 1)
    return bender->beat_factor * 60.0 / bpm;
  return bender->beat_factor;
}


/* Implement 'spin', on-the-fly looping of the recent past or near future
 * with variable speed and direction.   While spinning, the underlying
 * song does not advance in time.
 *
 * Rub Width:  size of looped segment, in seconds
 * Rub Length: how long the looping should last
 * Mode:   1:  spin recent past forward 
 *         2:  near future forward
 *         3:  recent past in reverse 
 *         4:  near future reverse
 */
void		gdam_bender_spin	(GdamBender*		bender,
					 float			rub_width,
					 float			rub_length,
                                         int                    spin_mode)
{
  float r_length = 0;
  float r_width = 0;
  float r_f = bender->rate_factor ? bender->rate_factor : 1.0;
  int num_points = 1;
  int *drag_data;
  int n;
  GdamArg args[2];

  if (!rub_width)
    {
      r_width = get_rub_width (bender);
    }
  else if (rub_width < 0)
    {
      /* TODO: what to do? */
      g_warning ("negative rub width, aborting bender spin");
      return;
    }
  else
    r_width = rub_width;
  if (!rub_length)
    {
      r_length = 180.0;
    }
  else if (rub_length < 0)
    {
      return;
    }
  else
    r_length = rub_length;

  g_return_if_fail (r_width > 0);
  g_return_if_fail (r_length > 0);

  num_points = r_length / r_width;
  if (num_points % 2)
    num_points++;
  drag_data = g_new (int, num_points * 2);

  if (spin_mode == 1)
    {
      /* spin the recent past by making backwards jumps of r_width seconds
       * every r_width seconds. */
      for (n = 0; n < num_points * 2; n += 4)
	{
	  /* The following pair, with a time value of zero, indicates a 
	   * jump of -r_width seconds. */
	  drag_data[n + 0] = 0;
	  drag_data[n + 1] = r_width * -44100 / r_f;
	  /* This pair says to play at 100% speed for r_width seconds */
	  drag_data[n + 2] = r_width * 44100 / r_f;
	  drag_data[n + 3] = 1.0 * r_f * GdamFixed_ONE;
	}
    }
  else if (spin_mode == 2)
    {
      /* spin the near future.  as above, but the jump comes after the forward
       * play. */
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = r_width * 44100 / r_f;
	  drag_data[n + 1] = 1.0 * r_f * GdamFixed_ONE;
	  drag_data[n + 2] = 0;
	  drag_data[n + 3] = r_width * -44100 / r_f;
	}
    }
  else if (spin_mode == 3)
    {
      /* spin the recent past in reverse by playing at -100% speed and making 
       * forward jumps.  */
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = r_width * 44100 / r_f;
	  drag_data[n + 1] = -1.0 * r_f * GdamFixed_ONE;
	  drag_data[n + 2] = 0;
	  drag_data[n + 3] = r_width * 44100 / r_f;
	}
    }
  else if (spin_mode == 4)
    {
      /* spin the near future in reverse by playing the near future at 100% 
       * speed, then repeatedly at -100% speed with forward jumps.  */
      /* TODO: initial forward jump rather than initial forward play?? */
      drag_data[0] = r_width * 44100 / r_f;
      drag_data[1] = 1.0 * r_f * GdamFixed_ONE;
      for (n = 2; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = r_width * 44100 / r_f;
	  drag_data[n + 1] = -1.0 * r_f * GdamFixed_ONE;
	  drag_data[n + 2] = 0;
	  drag_data[n + 3] = r_width * 44100 / r_f;
	}
    }

  /* XXX: set component's mode arg */
  args[0].name = "change_mode";
  args[0].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args) = g_strdup ("continuous");
  args[1].name = "changes";
  args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 1).length = num_points * 2;
  GDAM_ARG_INT_ARRAY (args + 1).array = drag_data;

  bender_post_arg_requests (bender, args, 2);
}


static void gdam_bender_arg_set_rate_factor(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->rate_factor = GDAM_ARG_DOUBLE (arg);
}
static void gdam_bender_arg_get_rate_factor(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_DOUBLE (arg) = bender->rate_factor;
}

static void gdam_bender_arg_set_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  gdam_bender_set_mode (bender, GDAM_ARG_INT (arg));
}
static void gdam_bender_arg_get_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  if (arg->type == GDAM_ARG_TYPE_INT)
    {
      GDAM_ARG_INT (arg) = bender->mode;
    }
  else
    {
      GDAM_BENDER_MODE mode = bender->mode;
      if (mode == GDAM_BENDER_KEEP_TIME)
	GDAM_ARG_STRING (arg) = g_strdup ("keep_time");
      else if (mode == GDAM_BENDER_KEEP_IN_BEAT)
	GDAM_ARG_STRING (arg) = g_strdup ("keep_in_beat");
      else if (mode == GDAM_BENDER_SNAP_TIME)
	GDAM_ARG_STRING (arg) = g_strdup ("snap_time");
      else if (mode == GDAM_BENDER_CONTINUOUS)
	GDAM_ARG_STRING (arg) = g_strdup ("continuous");
      else if (mode == GDAM_BENDER_RETURN)
	GDAM_ARG_STRING (arg) = g_strdup ("return");
    }
}

static void gdam_bender_arg_set_bend_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->bend_mode = GDAM_ARG_INT (arg);
}

static void gdam_bender_arg_get_bend_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->bend_mode;
}

static void gdam_bender_arg_set_rub_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->rub_mode = GDAM_ARG_INT (arg);
}

static void gdam_bender_arg_set_rub_by_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  arg->name = "rub_mode";
  gdam_arg_set_object_arg (object, arg);
  arg->name = "rub_by_mode";
  gdam_bender_rub (bender, 0, 0);
}


static void gdam_bender_arg_set_passthrough (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  gdam_bender_passthrough (bender);
}

static void gdam_bender_arg_set_spin (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  gdam_bender_spin (bender, 0, 0, GDAM_ARG_INT (arg));
}

static void 
gdam_bender_arg_set_stall_by_factor(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  gdam_bender_stall (bender, GDAM_ARG_DOUBLE (arg));
}


static void
arg_get_position_bending (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->position_bending;
}

static void
arg_set_position_bending (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamArg args[3];
  int num_args = 2;

  if (GDAM_ARG_INT (arg) == 0)
    {
      bender->position_bending = FALSE;
      /* jump to mark */
      args[0].name = "jump_to_mark";
      args[0].type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (args) = 0;
      /* erase changes */
      args[1].name = "changes";
      args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
      GDAM_ARG_INT_ARRAY (args + 1).length = 0;
      GDAM_ARG_INT_ARRAY (args + 1).array = NULL;
    }
  else
    {
      GTimeVal timeval;
      int change_array[2];
      g_get_current_time (&timeval);
      bender->position_bending = TRUE;
      bender->last_bend_timeval = timeval;
      bender->last_bend_position = 0;
      bender->last_bend_rate = 0;

      change_array[0] = 1000.0 * 44100;  /*  play this rate indefinitely */
      change_array[1] = 0.0 * GdamFixed_ONE;

      /* set the mark to the current position */
      args[0].name = "change_mode";
      args[0].type = GDAM_ARG_TYPE_STRING;
      if (bender->mode == GDAM_BENDER_KEEP_TIME)
	GDAM_ARG_STRING (args+ 0) = g_strdup ("keep_time");
      else
	GDAM_ARG_STRING (args+ 0) = g_strdup ("keep_in_beat");
      /* set to 0% speed */
      args[1].name = "changes";
      args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
      GDAM_ARG_INT_ARRAY (args + 1).length = 2;
      GDAM_ARG_INT_ARRAY (args + 1).array = change_array;
    }

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       2, args, NULL, NULL, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
    }
}
static void 
gdam_bender_arg_get_bend_position(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_DOUBLE (arg) = bender->last_bend_position;
}

/* map  -/+1  to  -/+ (beat_factor / 2) 
 *
 *                       mark (bend start)
 *   mark - (bf/2)       |
 *   |-------------------+----------------------| mark + (bf/2)
 *   -1                  0                      1
 *
 * */
#define MIN_MILLIS_PER_UPDATE 100.0
static void 
gdam_bender_arg_set_bend_position(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GTimeVal timeval;
  gdouble millis, beat_factor, bpm, rate, cur_server_position;
  gdouble last_bend_position = bender->last_bend_position;
  gdouble bend_position = GDAM_ARG_DOUBLE (arg);
  GdamArg args[3];
  int cur_server_offset;
  int change_array[2];

  g_get_current_time (&timeval);
  /* millis is negative ?? */
  millis = delta_millis (&bender->last_bend_timeval, &timeval);
  millis *= -1;

  if (!bender->position_bending)
    return;

  if (millis < MIN_MILLIS_PER_UPDATE)
    return;

  /* calculate overall rate since last position event */
  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bender)->beat_info, 0);
  if (bpm < 1)
    bpm = 60.0;
  beat_factor = bender->beat_factor * 1000 * 60.0 / bpm; /* millis per beat */
  rate = (bend_position - last_bend_position) * beat_factor / 2.0;
  rate /= millis;  /* millis change in position over millis elapsed time */
  g_message ("bend by position: rate %f from input %f millis %f", rate, bend_position, millis);

  /* use our knowledge of the server's state to estimate current position
   * and rate */
  cur_server_position = last_bend_position * beat_factor / 2.0; /* last start */
  cur_server_position += millis * bender->last_bend_rate;
  cur_server_offset = cur_server_position * 44.1; /* convert to samples */

  change_array[0] = 1000.0 * 44100;  /*  play this rate indefinitely */
  change_array[1] = rate * GdamFixed_ONE;

  /* force our estimate of current offset to be correct (this prevents
   * position drift over time) */
  args[0].name = "jump_to_mark";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args) = cur_server_offset;
  args[1].name = "changes";
  args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 1).length = 2;
  GDAM_ARG_INT_ARRAY (args + 1).array = change_array;

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       2, args, NULL, NULL, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
    }

  bender->last_bend_timeval = timeval;
  bender->last_bend_position = bend_position;
  bender->last_bend_rate = rate;
}

static void gdam_bender_arg_set_rewind_by_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  gdam_bender_rewind (bender, -1);
}

static void gdam_bender_arg_get_rub_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->rub_mode;
}

static void gdam_bender_arg_set_eighth_beat_factor(GtkObject* object, 
						   GdamArg*   arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamArg new_arg;
  if (bender->beat_factor == GDAM_ARG_DOUBLE (arg) / 8)
    return;
  bender->beat_factor = GDAM_ARG_DOUBLE (arg) / 8;
  new_arg.name = "beat_factor";
  new_arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&new_arg) = bender->beat_factor;
  gdam_arg_set_object_arg (object, &new_arg);
}
static void gdam_bender_arg_get_eighth_beat_factor(GtkObject* object, 
						   GdamArg*   arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_DOUBLE (arg) = 8 * bender->beat_factor;
}

static void gdam_bender_arg_set_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamArg new_arg;
  if (bender->beat_factor == GDAM_ARG_DOUBLE (arg))
    return;
  bender->beat_factor = GDAM_ARG_DOUBLE (arg);
  new_arg.name = "eighth_beat_factor";
  new_arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&new_arg) = bender->beat_factor * 8;
  gdam_arg_set_object_arg (object, &new_arg);

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      GdamArg beat_length_arg;
      gdouble block_size = get_block_size_seconds (bender);
      beat_length_arg.name = "beat_length";
      beat_length_arg.type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (&beat_length_arg) = (int) (block_size * 44100.0);
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       1, &beat_length_arg, NULL, NULL, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
    }
}

static void gdam_bender_arg_get_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_DOUBLE (arg) = bender->beat_factor;
}


static void gdam_bender_class_init(GdamBenderClass* klass)
{
  GdamFilterClass *filter_class;
  GdamModelClass *model_class;
  GtkObjectClass *object_class;
  filter_class = GDAM_FILTER_CLASS (klass);
  model_class = GDAM_MODEL_CLASS (klass);
  object_class = GTK_OBJECT_CLASS (klass);
  klass->slow_to_stall = bender_real_slow_to_stall;
  filter_class->append_start_request = gdam_bender_append_start_request;
  gdam_arg_register (object_class,
		     "rate_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_rate_factor,
		     gdam_bender_arg_get_rate_factor,
		     NULL, ",", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "beat_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_beat_factor,
		     gdam_bender_arg_get_beat_factor,
		     NULL, ",", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "eighth_beat_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_eighth_beat_factor,
		     gdam_bender_arg_get_eighth_beat_factor,
		     NULL, ",", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "mode",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_mode,
		     gdam_bender_arg_get_mode, NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "bend_mode",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_bend_mode,
		     gdam_bender_arg_get_bend_mode,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "bend_amount",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_bend_amount, NULL, NULL, ",", 0);
  gdam_arg_register (object_class,
		     "jump_to_mark",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_jump_to_mark, NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "jump_to_nearest_mark",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_jump_to_nearest_mark, NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "set_position_marker",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_position_marker, 
		     NULL, NULL, "0,3", 0);
  gdam_arg_register (object_class,
		     "jump_to_position_marker",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_jump_to_position_marker, 
		     NULL, NULL, "0,3", 0);
  gdam_arg_register (object_class,
		     "autotune",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_autotune,
		     gdam_bender_arg_get_autotune,
		     NULL, "0,1", 0);
  gdam_arg_register (object_class,
		     "rub_mode",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_rub_mode,
		     gdam_bender_arg_get_rub_mode,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "rub_by_mode",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_rub_by_mode,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "rewind_by_mode",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_rewind_by_mode,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "stall_by_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_stall_by_factor,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "spin",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_spin,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "passthrough",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_passthrough,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "bend_position",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_bend_position,
		     gdam_bender_arg_get_bend_position,
		     NULL, "-1.0,1.0", 0);
  gdam_arg_register (object_class,
		     "position_bending",
		     GDAM_ARG_TYPE_INT,
		     arg_set_position_bending,
		     arg_get_position_bending,
		     NULL, "0,1", 0);

  bender_signals[SLOW_TO_STALL] =
    gtk_signal_new ("slow_to_stall",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamBenderClass, slow_to_stall),
		    gtk_marshal_NONE__DOUBLE,
		    GTK_TYPE_NONE, 1, GTK_TYPE_DOUBLE);
  gtk_object_class_add_signals (object_class, bender_signals, LAST_SIGNAL);
}

void gdam_bender_set_mark(GdamBender* bender){
	if (bender->mode != GDAM_BENDER_KEEP_TIME &&
	    bender->mode != GDAM_BENDER_KEEP_IN_BEAT){
		g_warning("can't set mark in continuous mode...");
		return;
	}
	gdam_bender_set_mode(bender, bender->mode);
}

