/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamgong.h"

typedef struct _GdamGongStyleCoeff GdamGongStyleCoeff;
struct _GdamGongStyleCoeff {
	int		num_coeffs;
	gdouble*	coeffs;
};
static GdamGongStyleCoeff gong_coeffs[GDAM_GONG_NUM_STYLES];

static gdouble free_oscillation_coeffs[] = {
	1.0,
	-1.0
};
static gdouble clamped_circular_plate_coeffs[] = {
	0.4694, 2.08, 3.41, 5.00, 6.82,
	3.89, 5.95, 8.28, 10.87, 13.71,
	8.72, 11.75, 15.06, 18.63, 22.47,
	-1.0
};

static void gdam_gong_init(GdamGong* gong)
{
}

static void make_generic(int index, gdouble* coeffs)
{
  int num;
  int i;
  gdouble norm;
  for (i = 0; coeffs[i] > 0.0; i++)
    ;
  num = i;
  gong_coeffs[index].num_coeffs = num * 3;
  gong_coeffs[index].coeffs = g_new (gdouble, num * 3);
  norm = 1.0 / coeffs[0];
  for (i = 0; i < num; i++)
    {
      gdouble c = coeffs[i] * norm;
      gong_coeffs[index].coeffs[i * 3 + 0] = c;
      gong_coeffs[index].coeffs[i * 3 + 1] = 1.0 / c;
      gong_coeffs[index].coeffs[i * 3 + 2] = 1.0;
    }
}
static int gdam_gong_play_request(GdamProducer*       producer,
			          GdamPlayInfo*       play_info,
		                  int                 source_id,
			          GdamRequestBuilder* builder)
{
  GdamModel *model;
  GdamRequest *request;

  model = GDAM_MODEL (producer);

  gdam_debug ("Creating gong");
  gtk_object_ref (GTK_OBJECT (model));
  model->source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source_v ("GdasGong",
					  model->source_id,
					  0, NULL,
					  gdam_model_got_subsource,
					  gdam_model_didnt_get_subsource,
					  producer);
  return gdam_request_builder_append_create (builder, request);
}

static void gdam_gong_class_init(GdamGongClass* gong_class)
{
  GdamProducerClass *producer_class;
  producer_class = GDAM_PRODUCER_CLASS (gong_class);
  make_generic (GDAM_GONG_FREE_OSCILLATION, free_oscillation_coeffs);
  make_generic (GDAM_GONG_CLAMPED_CIRCULAR_PLATE,
		clamped_circular_plate_coeffs);
  producer_class->play_request = gdam_gong_play_request;
}

GtkType gdam_gong_get_type()
{
  static GtkType gong_type = 0;
  if (!gong_type)
    {
      static const GtkTypeInfo gong_info = {
	"GdamGong",
	sizeof (GdamGong),
	sizeof (GdamGongClass),
	(GtkClassInitFunc) gdam_gong_class_init,
	(GtkObjectInitFunc) gdam_gong_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_PRODUCER;
      gong_type = gtk_type_unique (parent, &gong_info);
    }
  return gong_type;
}

GdamGongStyle* gdam_gong_add   (GdamGong*        gong,
                                GdamGongType     gong_type)
{
  GdamGongStyle *style;
  g_return_val_if_fail (gong_type < GDAM_GONG_NUM_STYLES, NULL);

  style = g_new (GdamGongStyle, 1);
  style->description = NULL;
  style->coefficients = gong_coeffs[gong_type].coeffs;
  style->num_coefficients = gong_coeffs[gong_type].num_coeffs;
  style->overall_volume = 1.0;
  style->hertz = 220.0;
  style->halflife_scale = 1.0;
  style->gong = gong;
  gong->style_list = g_slist_prepend (gong->style_list, style);
  return style;
}

void           gdam_gong_strike(GdamGong*      gong,
                                GdamGongStyle* style)
{
  gdouble *coeffs;
  gdouble volume = style->overall_volume;
  gdouble hertz = style->hertz;
  gdouble decay = style->halflife_scale;
  int num = style->num_coefficients / 3;
  int i;
  GdamArg arg[1];
  int source_id;
  GdamRequest *request;
  source_id = GDAM_MODEL (gong)->source_id;
  g_return_if_fail (source_id != 0);

  coeffs = alloca (sizeof (gdouble) * style->num_coefficients);
  for (i = 0; i < num; i++)
    {
      coeffs[i * 3 + 0] = style->coefficients[i * 3 + 0] * hertz;
      coeffs[i * 3 + 1] = style->coefficients[i * 3 + 1] * volume;
      coeffs[i * 3 + 2] = style->coefficients[i * 3 + 2] * decay;
    }
  arg[0].name = "strike";
  arg[0].type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
  GDAM_ARG_DOUBLE_ARRAY (arg + 0).array = coeffs;
  GDAM_ARG_DOUBLE_ARRAY (arg + 0).length = num * 3;
  request = gdam_request_set_args (source_id, 1, arg, NULL, NULL, NULL);
  gdam_channel_append_request (GDAM_MODEL (gong)->channel, request);
}

