/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamproducermp3cacheable.h"
#include "../gdamsongdatabase.h"
#include <gdam/common/gdamplugin.h>
#include "../gdambuffermanager.h"
#include <gtk/gtksignal.h>
#include <math.h>


static void     gdam_producer_mp3_cacheable_destroy  (GtkObject*       object);

static int      gdam_producer_mp3_cacheable_play_request(
					       GdamProducer*       producer,
					       GdamPlayInfo*       play_info,
					       int                 source_id,
					       GdamRequestBuilder* builder);

static GtkObjectClass* parent_class;

static void gdam_producer_cacheable_arg_set_punch_in(GtkObject* object, 
						     GdamArg*   arg)
{
  GdamProducerMp3Cacheable *mp3;
  gdouble time = GDAM_ARG_DOUBLE (arg);
  mp3 = GDAM_PRODUCER_MP3_CACHEABLE (object);
  mp3->cache_start = time;
}

static void gdam_producer_cacheable_arg_get_punch_in(GtkObject* object, 
						     GdamArg*   arg)
{
  GdamProducerMp3Cacheable *mp3;
  mp3 = GDAM_PRODUCER_MP3_CACHEABLE (object);
  GDAM_ARG_DOUBLE (arg) = mp3->cache_start;
}

static void gdam_producer_cacheable_arg_set_length_in_beats(GtkObject* object, 
						            GdamArg*   arg)
{
  GdamProducerMp3Cacheable *mp3;
  gdouble num_beats = GDAM_ARG_DOUBLE (arg);
  mp3 = GDAM_PRODUCER_MP3_CACHEABLE (object);
  mp3->cache_beats = num_beats;
}

static void gdam_producer_cacheable_arg_get_length_in_beats(GtkObject* object, 
						            GdamArg*   arg)
{
  GdamProducerMp3Cacheable *mp3;
  mp3 = GDAM_PRODUCER_MP3_CACHEABLE (object);
  GDAM_ARG_DOUBLE (arg) = mp3->cache_beats;
}

static void gdam_producer_cacheable_arg_set_sub_length_in_beats
					(GtkObject* object, GdamArg*   arg)
{
  GdamProducerMp3Cacheable *mp3;
  gdouble num_beats = GDAM_ARG_DOUBLE (arg);
  double bf = GDAM_ARG_DOUBLE (arg);
  mp3 = GDAM_PRODUCER_MP3_CACHEABLE (object);
  mp3->cache_beats = num_beats;
}

static void gdam_producer_cacheable_arg_get_sub_length_in_beats
					(GtkObject* object, GdamArg*   arg)
{
  GdamProducerMp3Cacheable *mp3;
  mp3 = GDAM_PRODUCER_MP3_CACHEABLE (object);
  GDAM_ARG_DOUBLE (arg) = mp3->cache_beats;
}

static void gdam_producer_cacheable_arg_set_loop_name(GtkObject* object, 
						      GdamArg*   arg)
{
  GdamProducerMp3Cacheable *mp3;
  mp3 = GDAM_PRODUCER_MP3_CACHEABLE (object);
  mp3->loop_name = g_strdup (GDAM_ARG_STRING (arg));
}

static void gdam_producer_cacheable_arg_get_loop_name(GtkObject* object, 
						      GdamArg*   arg)
{
  GdamProducerMp3Cacheable *mp3;
  mp3 = GDAM_PRODUCER_MP3_CACHEABLE (object);
  GDAM_ARG_STRING (arg) = mp3->loop_name ? g_strdup (mp3->loop_name) : "";
}

/*  When a new song has been loaded, we reset some loop info. */
static void mp3_cacheable_filename_changed(gpointer info_ptr,
    					   GdamArg *arg,
					   GtkObject* obj)
{
  GdamProducer* producer = GDAM_PRODUCER (info_ptr);
  GdamArg newarg;
  newarg.name = "loop_name";
  newarg.type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (&newarg) = "";
  gdam_arg_set_object_arg (GTK_OBJECT (producer), &newarg);
}

static void 
gdam_producer_cacheable_arg_get_play_buffer (GtkObject* object, GdamArg* arg)
{
  GdamProducerMp3Cacheable *mp3 = GDAM_PRODUCER_MP3_CACHEABLE (object);
  GDAM_ARG_INT (arg) = mp3->play_buffer;
}

static void 
gdam_producer_cacheable_arg_set_play_buffer (GtkObject* object, GdamArg* arg)
{
  GdamProducerMp3Cacheable *mp3 = GDAM_PRODUCER_MP3_CACHEABLE (object);
  GdamBufferManagerInfo *info;
  gdouble position;
  mp3->play_buffer = GDAM_ARG_INT (arg);

  /* If we are changing to "true", set our beat info the buffer's beatinfo.  */
  if (mp3->play_buffer)
    {
      GdamBeatInfo* bi;
      info = gdam_buffer_manager_info_by_id (mp3->buffer_id);
      position = 0;
      if (info)
        {
	  if (info->beat_info)
	    gdam_model_set_beat_info (GDAM_MODEL (mp3), info->beat_info);
	}
      else
	{
	  GdamBeatInfo *newbeat;
	  newbeat = gdam_beat_info_new ();
	  gdam_model_set_beat_info (GDAM_MODEL (mp3), newbeat);
	  gdam_beat_info_destroy (newbeat);
	}
      gdam_producer_update_length (GDAM_PRODUCER (mp3),
				   (float) mp3->buffer_length / 44100.0);
    }
  /* Otherwise, set our beat info to that of the mp3 file. */
  else
    {
      GdamSongInfo* si;
      si = gdam_song_database_find (gdam_song_database_system (),
				    GDAM_PRODUCER_MP3 (mp3)->filename);
      g_return_if_fail (si != NULL);
      gdam_model_set_beat_info (GDAM_MODEL (mp3), si->beat_info);
      gdam_producer_update_length (GDAM_PRODUCER (mp3), si->length);
      position = gdam_beat_info_get_first_beat (si->beat_info, 0);
    }
    gdam_producer_set_position (GDAM_PRODUCER (mp3), position);
}


static void 
gdam_producer_cacheable_arg_get_buffer_id (GtkObject* object, GdamArg* arg)
{
  GdamProducerMp3Cacheable *mp3 = GDAM_PRODUCER_MP3_CACHEABLE (object);
  GDAM_ARG_INT (arg) = mp3->buffer_id;
}

static void 
gdam_producer_cacheable_arg_set_buffer_id (GtkObject* object, GdamArg* arg)
{
  /*GdamBeatInfo beat_info; */
  GdamProducerMp3Cacheable *mp3 = GDAM_PRODUCER_MP3_CACHEABLE (object);
  GdamBufferManagerInfo *info;
  GdamModel *model = GDAM_MODEL (object);
  int new_buffer_id = GDAM_ARG_INT (arg);
  gdam_producer_stop (GDAM_PRODUCER (mp3));

  gdam_buffer_manager_ref (new_buffer_id, model);
  if (mp3->buffer_id > 0)
    gdam_buffer_manager_unref (mp3->buffer_id, model);

  mp3->buffer_id = new_buffer_id;
  info = gdam_buffer_manager_info_by_id (new_buffer_id);

  if (info == NULL)
    {
      mp3->buffer_length = 0;
      return;
    }
  mp3->buffer_length = info->length;
  if (info->beat_info && mp3->play_buffer)
    {
      gdam_model_set_beat_info (GDAM_MODEL (mp3), info->beat_info);
    }
  else if (mp3->play_buffer)
    {
      GdamBeatInfo *newbeat;
      newbeat = gdam_beat_info_new ();
      gdam_model_set_beat_info (GDAM_MODEL (mp3), newbeat);
      gdam_beat_info_destroy (newbeat);
    }
}


static void gdam_producer_cacheable_arg_set_nudge_length_in_beats
					(GtkObject* object, GdamArg* arg)
{
  GdamProducerMp3Cacheable *mp3;
  GdamArg new_arg;
  int val = GDAM_ARG_INT (arg);
  double sbf;
  mp3 = GDAM_PRODUCER_MP3_CACHEABLE (object);
#if 0
  if (spin->beat_factor == 0)
    sbf = spin->beat_factor;
  else if (spin->beat_factor < 1)
    sbf = (-1.0 / spin->beat_factor);
  else
    sbf = spin->beat_factor;
  if (sbf <= 1 && sbf >= -1)
    {
      if (val > 0)
	sbf = 2;
      else
	sbf = -2;
    }
  else
    {
      sbf += val;
      if (sbf == -1)
	sbf = 1;
    }
  new_arg.name = "sub_beat_factor";
  new_arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&new_arg) = sbf;
  gdam_arg_set_object_arg (object, &new_arg);
#endif
}

static void gdam_producer_mp3_cacheable_class_init(GdamProducerMp3CacheableClass* klass)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  GdamProducerClass *producer_class = GDAM_PRODUCER_CLASS (klass);
  object_class->destroy = gdam_producer_mp3_cacheable_destroy;
  parent_class = gtk_type_class (GDAM_TYPE_PRODUCER_MP3);
  producer_class->play_request = gdam_producer_mp3_cacheable_play_request;
  gdam_arg_register (object_class,
		     "punch_in",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_producer_cacheable_arg_set_punch_in,
		     gdam_producer_cacheable_arg_get_punch_in,
		     NULL, ",", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "loop_name",
		     GDAM_ARG_TYPE_STRING,
		     gdam_producer_cacheable_arg_set_loop_name,
		     gdam_producer_cacheable_arg_get_loop_name,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "length_in_beats",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_producer_cacheable_arg_set_length_in_beats,
		     gdam_producer_cacheable_arg_get_length_in_beats,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "sub_length_in_beats",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_producer_cacheable_arg_set_sub_length_in_beats,
		     gdam_producer_cacheable_arg_get_sub_length_in_beats,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "nudge_length_in_beats",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_producer_cacheable_arg_set_nudge_length_in_beats,
		     NULL, NULL, "-1,1", 0);
  gdam_arg_register (object_class,
		     "buffer_id",
		     GDAM_ARG_TYPE_INT,
		     gdam_producer_cacheable_arg_set_buffer_id,
		     gdam_producer_cacheable_arg_get_buffer_id,
		     NULL, NULL, GDAM_ARG_VITAL);
  /* Whether to play the buffer rather than the mp3 file */
  gdam_arg_register (object_class,
		     "play_buffer",
		     GDAM_ARG_TYPE_INT,
		     gdam_producer_cacheable_arg_set_play_buffer,
		     gdam_producer_cacheable_arg_get_play_buffer,
		     NULL, "boolean", GDAM_ARG_VITAL);
}

static void gdam_producer_mp3_cacheable_init(GdamProducerMp3Cacheable* mp3)
{
  mp3->fname_arg_connect_id = gdam_arg_connect (GTK_OBJECT (mp3),
						"filename", 
						mp3_cacheable_filename_changed,
						mp3);
}

GtkType gdam_producer_mp3_cacheable_get_type()
{
  static GtkType mp3_cacheable_type = 0;
  if (!mp3_cacheable_type)
    {
      static const GtkTypeInfo mp3_cacheable_info = {
	"GdamProducerMp3Cacheable",
	sizeof (GdamProducerMp3Cacheable),
	sizeof (GdamProducerMp3CacheableClass),
	(GtkClassInitFunc) gdam_producer_mp3_cacheable_class_init,
	(GtkObjectInitFunc) gdam_producer_mp3_cacheable_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      mp3_cacheable_type = gtk_type_unique (GDAM_TYPE_PRODUCER_MP3,
					    &mp3_cacheable_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return mp3_cacheable_type;
}

GdamProducerMp3Cacheable* gdam_producer_mp3_cacheable_new(GdamChannel*	channel)
{
  GdamProducerMp3Cacheable *rv;
  rv = GDAM_PRODUCER_MP3_CACHEABLE (gtk_type_new
				    (GDAM_TYPE_PRODUCER_MP3_CACHEABLE));
  ((GdamModel *) rv)->channel = channel;
  return rv;
}

static void gdam_producer_mp3_cacheable_destroy(GtkObject* object)
{
  (*parent_class->destroy) (object);
}


void gdam_producer_mp3_cacheable_set_start(GdamProducerMp3Cacheable* mp3)
{
  GdamArg arg;
  arg.name = "punch_in";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = GDAM_PRODUCER (mp3)->stopped_position;
  gdam_arg_set_object_arg (GTK_OBJECT (mp3), &arg);
}

void gdam_producer_mp3_cacheable_set_beats(GdamProducerMp3Cacheable* mp3,
					   float 		     num_beats)
{
  GdamArg arg;
  arg.name = "length_in_beats";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = num_beats;
  gdam_arg_set_object_arg (GTK_OBJECT (mp3), &arg);
}

typedef struct _BufferCreateInfo {
  GdamBeatInfo *beat_info;
  GdamProducerMp3Cacheable *mp3;
  int buffer_id;
} BufferCreateInfo;

static gboolean idle_select (gpointer data)
{
  BufferCreateInfo *createinfo;
  GdamArg arg;
  createinfo = (BufferCreateInfo*)data;

  arg.name = "buffer_id";
  arg.type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (&arg) = createinfo->buffer_id;
  gdam_arg_set_object_arg (GTK_OBJECT (createinfo->mp3), &arg);
  arg.name = "play_buffer";
  arg.type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (&arg) = 1;
  gdam_arg_set_object_arg (GTK_OBJECT (createinfo->mp3), &arg);
  g_free (createinfo);
  return FALSE;
}

/* When the buffer has been created, we set its beat information
 * in the buffer manager. 
 */
static void buffer_create_success(int buffer_id, gpointer data)
{
  GdamBufferManagerInfo *bufferinfo;
  BufferCreateInfo *createinfo;

  createinfo = (BufferCreateInfo*)data;
  bufferinfo = gdam_buffer_manager_info_by_id (buffer_id);
  bufferinfo->beat_info = createinfo->beat_info;
  createinfo->buffer_id = buffer_id;
  g_idle_add (idle_select, createinfo);
}


/* Writes a description of the current loop to the song database.  This
 * allows the loop to be quickly recreated in the future.
 */
void 
gdam_producer_mp3_cacheable_write_to_database(GdamProducerMp3Cacheable* mp3)
{
  gdouble position = mp3->cache_start;
  gdouble length;
  GdamBeatInfo *beat_info = &GDAM_MODEL (mp3)->beat_info;
  GdamSongInfo *si;
  double beat_length;
  char *name;

  g_return_if_fail (mp3->cache_beats >= 0);

  si = gdam_song_database_find (gdam_song_database_system (),
				GDAM_PRODUCER_MP3 (mp3)->filename);
  g_return_if_fail (si != NULL);
  if (gdam_beat_info_get_bpm (beat_info, position) < 1)
    beat_length = 1.0;
  else
    beat_length = gdam_beat_info_get_beat_length (beat_info, position);
  length = mp3->cache_beats * beat_length;
  name = mp3->loop_name ? mp3->loop_name : "untitled";
  gdam_beat_info_add_loop (si->beat_info, name, position, length);
}

void 
gdam_producer_mp3_cacheable_read_from_database(GdamProducerMp3Cacheable* mp3)
{
  gdouble position = mp3->cache_start;
  GdamBeatInfo *beat_info = &GDAM_MODEL (mp3)->beat_info;
  GdamSongInfo *si;
  GdamLoopInfo *li;
  GdamArg arg;
  const GList *loops;
  double beat_length;
  int list_length;

  si = gdam_song_database_find (gdam_song_database_system (),
				GDAM_PRODUCER_MP3 (mp3)->filename);
  g_return_if_fail (si != NULL);

  loops = gdam_beat_info_get_loops (beat_info);
  if (loops == NULL)
    return;
  list_length = g_list_length ((GList *) loops);
  if (mp3->loop_index >= list_length - 1)
    mp3->loop_index = 0;
  else
    mp3->loop_index++;
  li = (GdamLoopInfo *) g_list_nth_data ((GList *) loops, mp3->loop_index);
  g_return_if_fail (li != NULL);

  arg.name = "punch_in";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = li->start;
  gdam_arg_set_object_arg (GTK_OBJECT (mp3), &arg);

  if (gdam_beat_info_get_bpm (beat_info, li->start) < 1)
    beat_length = 1.0;
  else
    beat_length = gdam_beat_info_get_beat_length (beat_info, li->start);
  if (beat_length < 0.01)
    {
      g_warning ("beat_length too small in read_loop");
      beat_length = 1.0;
    }

  arg.name = "loop_name";
  arg.type = GDAM_ARG_TYPE_STRING;
  if (li->name != NULL)
    GDAM_ARG_STRING (&arg) = li->name;
  else
    GDAM_ARG_STRING (&arg) = "";
  gdam_arg_set_object_arg (GTK_OBJECT (mp3), &arg);
  gdam_producer_mp3_cacheable_set_beats (mp3, li->length / beat_length);
}


/*  Plays a portion of an mp3 files and caches it into memory.  Calculates
 *  the beat info of the cached sample, gives it a name, and registers the
 *  new cache with the buffer manager.
 */
void gdam_producer_mp3_cacheable_export(GdamProducerMp3Cacheable* mp3)
{
  GdamChannel *channel;
  GdamModel *model;
  GdamRequest *request;
  int local_source_id;
  int num_args;
  GdamArg args[4];
  GdamBufferManagerInfo *bufferinfo;
  GdamRequestBuilder *request_builder;
  gdouble position = mp3->cache_start;
  gdouble length;
  GdamBeatInfo *beat_info = &GDAM_MODEL (mp3)->beat_info;
  GdamBeatInfo *new_beat_info;
  BufferCreateInfo *create_info;
  gboolean use_buffer_manager = TRUE;
  double beat_length;
  double beat_frac;
  double beat_number;
  double first_beat;
  char description[2048] = "";

  g_return_if_fail (mp3 != NULL);
  g_return_if_fail (mp3->cache_beats != 0);

  if (!GDAM_PRODUCER_MP3 (mp3)->filename)
    {
      g_warning ("gdam_producer_mp3_cache: no filename");
      return;
    }

  if (gdam_beat_info_get_bpm (beat_info, position) < 1)
    beat_length = 1.0;
  else
    beat_length = gdam_beat_info_get_beat_length (beat_info, position);

  length = mp3->cache_beats * beat_length;

  sprintf (description, "%s %3.2f  %2.2f  %s",
	   mp3->loop_name ? mp3->loop_name : "", position, length,
	   GDAM_PRODUCER_MP3 (mp3)->filename);

  bufferinfo = gdam_buffer_manager_info_by_desc (description);
  if (bufferinfo)
    {
      GdamArg arg;
      arg.name = "buffer_id";
      arg.type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (&arg) = bufferinfo->buffer_id;
      gdam_arg_set_object_arg (GTK_OBJECT (mp3), &arg);
      arg.name = "play_buffer";
      arg.type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (&arg) = 1;
      gdam_arg_set_object_arg (GTK_OBJECT (mp3), &arg);
      return;
    }

  model = GDAM_MODEL (mp3);
  channel = model->channel;
  g_return_if_fail (channel != NULL);

  args[0].type = GDAM_ARG_TYPE_STRING;
  args[0].name = "filename";
  GDAM_ARG_STRING (args + 0) = GDAM_PRODUCER_MP3 (mp3)->filename;

  args[1].type = GDAM_ARG_TYPE_INT64;
  args[1].name = "position";
  GDAM_ARG_INT64 (args + 1) = (gint64) (position * 44100.0);

  args[2].type = GDAM_ARG_TYPE_INT64;
  args[2].name = "length";
  GDAM_ARG_INT64 (args + 2) = (gint64) (length * 44100.0);

  num_args = 3;

  model->source_id = gdam_channel_new_local_id (channel);
  request = gdam_request_create_source_v ("GdasMp3",
					  model->source_id,
					  num_args, args, NULL, NULL, NULL);

  gdam_debug ("Creating cached mp3 with position %f length %f",
	      position, length);

  request_builder = gdam_request_builder_new ();
  local_source_id = gdam_request_builder_append_create (request_builder,
							request);

  first_beat = gdam_beat_info_get_first_beat (beat_info, position);
  beat_number = (position - first_beat) / beat_length;
  beat_frac = fmod (beat_number, 1.0);
  if (beat_number < 0)
    beat_frac = -beat_frac;
  else if (beat_number > 0)
    beat_frac = 1.0 - beat_frac;
  new_beat_info = gdam_beat_info_copy (beat_info);
  gdam_beat_info_set_first_beat (new_beat_info, (beat_length * (beat_frac)));

  create_info = g_new0 (BufferCreateInfo, 1);
  create_info->beat_info = new_beat_info;
  create_info->mp3 = mp3;

  gdam_buffer_manager_start_caching (request_builder,
				     local_source_id,
				     channel,
				     use_buffer_manager,
				     description,
				     buffer_create_success,
				     NULL, create_info);
}


static int      gdam_producer_mp3_cacheable_play_request(
					       GdamProducer*       producer,
					       GdamPlayInfo*       play_info,
					       int                 source_id,
					       GdamRequestBuilder* builder)
{
  GdamProducerMp3Cacheable *mp3 = GDAM_PRODUCER_MP3_CACHEABLE (producer);
  GdamChannel *channel;
  GdamModel *model;
  GdamRequest *request;
  gdouble position;
  int local_source_id;

  if (!mp3->play_buffer)
    {
      return (*GDAM_PRODUCER_CLASS (parent_class)->play_request) 
					   (producer, 
	  				    play_info, 
	  				    source_id, 
					    builder);
    }

  if (mp3->buffer_id < 0)
    {
      g_warning ("gdam_producer_mp3_cacheable_play: no buffer_id");
      return 0;
    }

  model = GDAM_MODEL (mp3);
  channel = model->channel;
  g_return_val_if_fail (channel != NULL, 0);

  position = play_info->position;

  gdam_debug ("producer mp3 cacheable creating producing wavetable with position %f", position);
  gtk_object_ref (GTK_OBJECT (mp3));
  model->source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source ("GdasWavetable",
					model->source_id,
					gdam_model_got_subsource,
					gdam_model_didnt_get_subsource,
					mp3,
					3,
					"buffer_id",
					GDAM_ARG_TYPE_INT,
					mp3->buffer_id,
					"position",
					GDAM_ARG_TYPE_INT64,
					(gint64) (position * 44100.0),
					"terminates",
					GDAM_ARG_TYPE_INT,
					FALSE);
  local_source_id = gdam_request_builder_append_create (builder, request);
  return local_source_id;
}


#if 0
/* TODO:  add mix,echo to create loop on the fly? */
static int      gdam_producer_mp3_cacheable_play_request(
					       GdamProducer*       producer,
					       GdamPlayInfo*       play_info,
					       int                 source_id,
					       GdamRequestBuilder* builder)
{
  GdamProducerMp3 *mp3 = GDAM_PRODUCER_MP3 (producer);
  GdamChannel *channel;
  GdamModel *model;
  GdamRequest *request;
  int local_source_id;
  int num_args;
  GdamArg args[4];
  gdouble position = play_info ? play_info->position : 0.0;

  g_return_val_if_fail (mp3 != NULL, 0);
  if (!mp3->filename)
    {
      g_warning ("gdam_producer_mp3_play_request: no filename");
      return 0;
    }
  model = GDAM_MODEL (mp3);
  channel = model->channel;
  g_return_val_if_fail (channel != NULL, 0);

  gdam_debug ("producermp3: not in sync.", position);
  args[0].type = GDAM_ARG_TYPE_STRING;
  args[0].name = "filename";
  GDAM_ARG_STRING (args + 0) = mp3->filename;

  args[1].type = GDAM_ARG_TYPE_INT64;
  args[1].name = "position";
  GDAM_ARG_INT64 (args + 1) = (gint64) (position * 44100.0);

  num_args = 2;
  if (mp3->buffer_size > 0)
    {
      args[num_args].type = GDAM_ARG_TYPE_INT;
      args[num_args].name = "buffer_size";
      GDAM_ARG_INT (args + num_args) = mp3->buffer_size;
      num_args++;
    }
  if (mp3->fade_envelope > 0)
    {
      args[num_args].type = GDAM_ARG_TYPE_INT;
      args[num_args].name = "fade_envelope";
      GDAM_ARG_INT (args + num_args) = mp3->fade_envelope;
      num_args++;
    }
  if (play_info->length > 0)
    {
      args[num_args].type = GDAM_ARG_TYPE_INT64;
      args[num_args].name = "length";
      GDAM_ARG_INT64 (args + num_args) =
	(gint64) (play_info->length * 44100.0);
      num_args++;
    }

  request = gdam_request_create_source_v ("GdasSeekableMp3",
					  num_args,
					  args,
					  gdam_model_got_subsource,
					  gdam_model_didnt_get_subsource,
					  mp3);

  gdam_debug ("Creating producing mp3 with position %f", position);
  gtk_object_ref (GTK_OBJECT (mp3));

  local_source_id = gdam_request_builder_append_create (builder, request);
  return local_source_id;
}

#endif
