/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamproducerwavetable.h"
#include <gdam/common/gdamplugin.h>
#include "../gdambuffermanager.h"
#include <gtk/gtksignal.h>
#include <gdam/common/gdamplugin.h>
#include "../gdamutil.h"

enum {
	RATE_CHANGED,
	LAST_SIGNAL
};

static guint 	producer_wavetable_signals[LAST_SIGNAL];
static void 
gdam_producer_wavetable_class_init	(GdamProducerWavetableClass* klass);
void gdam_producer_wavetable_arg_changed(GdamProducerWavetable* wavetable);

static int gdam_producer_wavetable_play_request(GdamProducer*       producer,
					        GdamPlayInfo*       play_info,
					        int                 source_id,
					        GdamRequestBuilder* builder);

static void
gdam_producer_wavetable_detach_from_buffer(GdamModel*        model,
                                           guint             buffer_id);
static void 
gdam_producer_wavetable_init	(GdamProducerWavetable* producer_wavetable)
{
  producer_wavetable->hertz = 440.0;
  producer_wavetable->buffer_id = -1;
}

GtkType gdam_producer_wavetable_get_type()
{
  static GtkType producer_wavetable_type = 0;
  if (!producer_wavetable_type)
    {
      static const GtkTypeInfo producer_wavetable_info = {
	"GdamProducerWavetable",
	sizeof (GdamProducerWavetable),
	sizeof (GdamProducerWavetableClass),
	(GtkClassInitFunc) gdam_producer_wavetable_class_init,
	(GtkObjectInitFunc) gdam_producer_wavetable_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = gdam_plugin_get_type_by_name ("model",
						     "GdamProducer");
      producer_wavetable_type = gtk_type_unique (parent,
						 &producer_wavetable_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return producer_wavetable_type;
}

static int gdam_producer_wavetable_play_request(GdamProducer*       producer,
					        GdamPlayInfo*       play_info,
					        int                 source_id,
					        GdamRequestBuilder* builder)
{
  GdamProducerWavetable *wavetable = GDAM_PRODUCER_WAVETABLE (producer);
  GdamChannel *channel;
  GdamModel *model;
  GdamRequest *request;
  gdouble position;
  int local_source_id;

  g_return_val_if_fail (wavetable != NULL, 0);

  if (wavetable->buffer_id < 0)
    {
      g_warning ("gdam_producer_wavetable_play: no buffer_id");
      return 0;
    }
  model = GDAM_MODEL (wavetable);
  channel = model->channel;
  g_return_val_if_fail (channel != NULL, 0);

  position = play_info->position;

  gdam_debug ("Creating producing wavetable with position %f", position);
  gtk_object_ref (GTK_OBJECT (wavetable));
  model->source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source ("GdasWavetable",
					model->source_id,
					gdam_model_got_subsource,
					gdam_model_didnt_get_subsource,
					wavetable,
					3,
					"buffer_id",
					GDAM_ARG_TYPE_INT,
					wavetable->buffer_id,
					"position",
					GDAM_ARG_TYPE_INT64,
					(gint64) (position * 44100.0),
					"terminates",
					GDAM_ARG_TYPE_INT,
					wavetable->terminates);

  local_source_id = gdam_request_builder_append_create (builder, request);
  return local_source_id;
}

static void
gdam_producer_wavetable_detach_from_buffer(GdamModel*        model,
                                           guint             buffer_id)
{
  GdamProducerWavetable *wavetable;
  wavetable = GDAM_PRODUCER_WAVETABLE (model);
  if (wavetable->buffer_id != 0)
    gdam_producer_wavetable_set_buffer_id (wavetable, 0);
}

static void 
gdam_producer_wavetable_arg_get_buffer_id (GtkObject* object, GdamArg* arg)
{
  GdamProducerWavetable *wavetable = GDAM_PRODUCER_WAVETABLE (object);
  GDAM_ARG_INT (arg) = wavetable->buffer_id;
}

static void 
gdam_producer_wavetable_arg_set_buffer_id (GtkObject* object, GdamArg* arg)
{
  /*GdamBeatInfo beat_info; */
  GdamProducerWavetable *wavetable = GDAM_PRODUCER_WAVETABLE (object);
  GdamBufferManagerInfo *info;
  GdamModel *model = GDAM_MODEL (object);
  int new_buffer_id = GDAM_ARG_INT (arg);
  gdam_producer_stop (GDAM_PRODUCER (wavetable));

  gdam_buffer_manager_ref (new_buffer_id, model);
  if (wavetable->buffer_id > 0)
    gdam_buffer_manager_unref (wavetable->buffer_id, model);

  wavetable->buffer_id = new_buffer_id;
  info = gdam_buffer_manager_info_by_id (new_buffer_id);

  if (info == NULL)
    {
      wavetable->buffer_length = 0;
      return;
    }
  wavetable->buffer_length = info->length;
  if (info->beat_info)
    {
      gdam_model_set_beat_info (GDAM_MODEL (wavetable), info->beat_info);
    }
  else
    {
      GdamBeatInfo *newbeat;
      newbeat = gdam_beat_info_new ();
      gdam_model_set_beat_info (GDAM_MODEL (wavetable), newbeat);
      gdam_beat_info_destroy (newbeat);
    }
  gdam_producer_update_length (GDAM_PRODUCER (wavetable),
			       (float) info->length / 44100.0);

  /*
     gtk_signal_emit_by_name(GTK_OBJECT(wavetable),
     "update_length",
     info->length);
     beat_info.first_beat = info->first_beat;
     beat_info.bpm = info->bpm;
     gdam_model_set_beat_info(GDAM_MODEL(wavetable), &beat_info);
   */
}

void gdam_producer_wavetable_term_changed(GdamProducerWavetable* wavetable)
{
  GtkObject *object = GTK_OBJECT (wavetable);
  GdamArg arg[1];
  GdamRequest *request;
  GdamModel *model = GDAM_MODEL (wavetable);
  if (!model->source_id)
    return;

  arg[0].name = "terminates";
  arg[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (arg + 0) = wavetable->terminates;

  request = gdam_request_set_args (model->source_id,
				   1,
				   arg,
				   gdam_do_nothing, gdam_do_nothing, NULL);
  gdam_channel_append_request (model->channel, request);
}

static void 
gdam_producer_wavetable_arg_set_terminates (GtkObject* object, GdamArg* arg)
{
  GdamProducerWavetable *wavetable = GDAM_PRODUCER_WAVETABLE (object);
  wavetable->terminates = GDAM_ARG_INT (arg) ? TRUE : FALSE;
  gdam_producer_wavetable_term_changed (wavetable);
}

static void 
gdam_producer_wavetable_arg_get_terminates (GtkObject* object, GdamArg* arg)
{
  GdamProducerWavetable *wavetable = GDAM_PRODUCER_WAVETABLE (object);
  GDAM_ARG_INT (arg) = wavetable->terminates ? 1 : 0;
}

void gdam_producer_wavetable_set_buffer_id(GdamProducerWavetable* wavetable,
                                           int                    buffer_id)
{
  gdam_arg_set (GTK_OBJECT (wavetable), "buffer_id", buffer_id);
}

void gdam_producer_wavetable_set_frequency(GdamProducerWavetable* wavetable,
                                           gdouble                hertz)
{
  wavetable->hertz = hertz;
  wavetable->rate = (gdouble) wavetable->buffer_length
    / (44100.0 / wavetable->hertz);
  gdam_producer_wavetable_arg_changed (wavetable);
}
void gdam_producer_wavetable_set_rate     (GdamProducerWavetable* wavetable,
                                           gdouble                rate)
{
  wavetable->rate = rate;
  wavetable->hertz = (gdouble) 44100.0 / (wavetable->buffer_length / rate);
  gdam_producer_wavetable_arg_changed (wavetable);
}

void gdam_producer_wavetable_frequency_changed(GdamProducerWavetable* wavetable)
{
  gdam_producer_wavetable_arg_changed (wavetable);
}

void gdam_producer_wavetable_arg_changed(GdamProducerWavetable* wavetable)
{
  GtkObject *object = GTK_OBJECT (wavetable);
  GdamArg arg[2];
  GdamRequest *request;
  GdamModel *model = GDAM_MODEL (wavetable);
  guint signal = producer_wavetable_signals[RATE_CHANGED];
  int num_args;
  if (gtk_signal_n_emissions (object, signal))
    return;
  if (!model->source_id)
    return;

  num_args = 1;
  arg[0].name = "terminates";
  arg[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (arg + 0) = wavetable->terminates;

  if (wavetable->hertz > 0)
    {
      num_args = 2;
      arg[1].name = "hertz";
      arg[1].type = GDAM_ARG_TYPE_DOUBLE;
      GDAM_ARG_DOUBLE (arg + 1) = wavetable->hertz;
    }

  request = gdam_request_set_args (model->source_id,
				   num_args,
				   arg,
				   gdam_do_nothing, gdam_do_nothing, NULL);

  gdam_channel_append_request (model->channel, request);
  gtk_signal_emit (object, signal);
}

static gdouble wavetable_get_length(GdamModel* model)
{
  GdamProducerWavetable *wavetable = GDAM_PRODUCER_WAVETABLE (model);
  return wavetable->buffer_length / 44100.0;
}

static void 
gdam_producer_wavetable_class_init	(GdamProducerWavetableClass* klass)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  GdamModelClass *model_class = GDAM_MODEL_CLASS (klass);
  GdamProducerClass *producer_class = GDAM_PRODUCER_CLASS (klass);
  producer_class->play_request = gdam_producer_wavetable_play_request;
  model_class->detach_from_buffer
    = gdam_producer_wavetable_detach_from_buffer;
  model_class->get_length = wavetable_get_length;
  gdam_arg_register (object_class,
		     "buffer_id",
		     GDAM_ARG_TYPE_INT,
		     gdam_producer_wavetable_arg_set_buffer_id,
		     gdam_producer_wavetable_arg_get_buffer_id,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "terminates",
		     GDAM_ARG_TYPE_INT,
		     gdam_producer_wavetable_arg_set_terminates,
		     gdam_producer_wavetable_arg_get_terminates,
		     NULL, "boolean", GDAM_ARG_VITAL);

  producer_wavetable_signals[RATE_CHANGED] =
    gtk_signal_new ("rate_changed",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamProducerWavetableClass,
				       rate_changed),
		    gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class,
				producer_wavetable_signals, LAST_SIGNAL);
}

