/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamconnection.h"
#include "gdamevent.h"
#include "gdamserverdebug.h"
#include "gdamdispatch.h"
#include "gdamcachebuffer.h"
#include <gdam/common/gdam-protocol.h>
#include <gdam/common/gdamsampleconverter.h>
#include <gtk/gtksignal.h>
#include <stdio.h>
#include <stdlib.h>

static GtkObjectClass* parent_class = NULL;

static void      gdam_cache_buffer_finished(GdamCacheBuffer*);
static void      gdam_cache_buffer_destroy(GtkObject*);
static gboolean  gdam_cache_buffer_prepare        (gpointer         src_data,
                                                   GTimeVal        *curtime,
                                                   gint            *timeout,
                                                   gpointer         user_data);
static gboolean  gdam_cache_buffer_check          (gpointer         src_data,
                                                   GTimeVal        *curtime,
                                                   gpointer         user_data);
static gboolean  gdam_cache_buffer_dispatch       (gpointer         src_data,
                                                   GTimeVal        *curtime,
                                                   gpointer         user_data);

static GSourceFuncs source_funcs = { 
        gdam_cache_buffer_prepare,
        gdam_cache_buffer_check,
	gdam_cache_buffer_dispatch,
	NULL
};

enum {
	FINISHED,
	LAST_SIGNAL
};

static guint cache_buffer_signals[LAST_SIGNAL];


static void gdam_cache_buffer_arg_get_length(GtkObject *object, GdamArg *arg)
{
  GdamCacheBuffer *cache_buffer = GDAM_CACHE_BUFFER (object);
  GDAM_ARG_INT64 (arg) = cache_buffer->buffer_length;
}

#if 0
static void gdam_cache_buffer_taper_region  (GtkObject *object,
                                             int        num_in,
                                             GdamArg   *args_in,
                                             int       *num_out,
                                             GdamArg  **args_out)
{
  GdamCacheBuffer *cache_buffer = GDAM_CACHE_BUFFER (object);
  int i;
  gint64 start = 0, length = -1;
  gboolean fade_in = TRUE;
  GdamFixed start_level, end_level;
  GdamSample *at;
  for (i = 0; i < num_in; i++)
    {
      GdamArg *arg = args_in + i;
      if (strcmp (arg->name, "fade_in") == 0
	  && arg->type == GDAM_ARG_TYPE_INT)
	{
	  fade_in = (GDAM_ARG_INT (arg));
	  continue;
	}
      if (strcmp (arg->name, "start") == 0
	  && arg->type == GDAM_ARG_TYPE_INT64)
	{
	  start = GDAM_ARG_INT64 (arg);
	  continue;
	}
      if (strcmp (arg->name, "length") == 0
	  && arg->type == GDAM_ARG_TYPE_INT64)
	{
	  length = GDAM_ARG_INT64 (arg);
	  continue;
	}
      g_warning ("unknown type/arg %s", arg->name);
      return;
    }

  if (fade_in)
    {
      start_level = 0;
      delta_level = GdamFixed_ONE;
    }
  else
    {
      start_level = GdamFixed_ONE;
      delta_level = -GdamFixed_ONE;
    }

  at = cache_buffer->buffer + start;
  for (i = 0; i < length; i++)
    {
    volume = (GdamSampleProduct) start_level *i +}
}
#endif

static void gdam_cache_buffer_cut           (GdamCacheBuffer     *buffer,
                                             int                  start,
					     int                  length)
{
  char *buf;
  int sample_size;
  if (buffer->_buffer == NULL)
    {
      g_warning ("cannot `cut' from virtual buffer");
      return;
    }
  if (buffer->read_only)
    {
      g_warning ("cannot `cut' from read-only buffer");
      return;
    }
  g_return_if_fail (length >= 0);
  g_return_if_fail (start >= 0);
  g_return_if_fail (start + length <= buffer->buffer_length);

  sample_size = buffer->is_16bit ? sizeof (GdamSample16) : sizeof (GdamSample);
  buf = buffer->_buffer;

  g_memmove (buf + start * sample_size,
             buf + (start + length) * sample_size,
	     (buffer->buffer_length - start - length) * sample_size);
  buffer->buffer_length -= length;
}
static GdamCacheBuffer *gdam_cache_buffer_make_subbuffer
                                            (GdamCacheBuffer     *buffer,
                                             int                  start,
					     int                  length,
					     gboolean             is_16bit)
{
  GdamCacheBuffer *rv_buffer;
  void *rv;
  int sample_size;
  sample_size = is_16bit ? sizeof (GdamSample16) : sizeof (GdamSample);
  rv = g_new (char, sample_size * length);
  if (buffer->_buffer != NULL && is_16bit)
    {
      if (buffer->is_16bit)
        gdam_sample16_copy ((GdamSample16 *) rv,
	                    (GdamSample16 *) buffer->_buffer + start,
			    length);
      else
        gdam_sample16_from_normal ((GdamSample16 *) rv,
	                           (GdamSample *) buffer->_buffer + start,
			           length);
    }
  else if (is_16bit)
    {
      GdamSample *tmp;
      tmp = g_new (GdamSample, length);
      /* XXX: we should have a ``read_defaults'' which is like this... */
      gdam_cache_buffer_read (buffer, tmp, start, 0, NULL, NULL, length,
                              GdamFixed_ONE, GdamFixed_ONE, FALSE);
      
      gdam_sample16_from_normal ((GdamSample16 *) rv, tmp, length);
    }
  else
    {
      /* XXX: we should have a ``read_defaults'' which is like this... */
      gdam_cache_buffer_read (buffer, (GdamSample *) rv,
                              start, 0, NULL, NULL, length,
                              GdamFixed_ONE, GdamFixed_ONE, FALSE);
    }
    
  rv_buffer = gdam_cache_buffer_memory (length, rv, g_free, is_16bit);
  g_return_val_if_fail (rv_buffer != NULL, NULL);
  return rv_buffer;
}

static void gdam_cache_buffer_process_region(GtkObject *object,
                                             int        num_in,
                                             GdamArg   *args_in,
                                             int       *num_out,
                                             GdamArg  **args_out)
{
  int length;
  int start;
  int i;
  gboolean cut;
  gboolean make_buffer;
  gboolean hi_quality;
  GdamCacheBuffer *orig = GDAM_CACHE_BUFFER (object);
  if (orig->read_only)
    {
      *num_out = 0;
      return;
    }
  length = gdam_arg_scan_int (num_in, args_in, "length", -1);
  start = gdam_arg_scan_int (num_in, args_in, "start", 0);
  cut = gdam_arg_scan_int (num_in, args_in, "cut", FALSE);
  make_buffer = gdam_arg_scan_int (num_in, args_in, "make_buffer", FALSE);
  hi_quality = gdam_arg_scan_int (num_in, args_in, "hi_quality", FALSE);
  if (start < 0)
    {
      length += start;
      start = 0;
    }
  if (start + length > orig->buffer_length)
    length = orig->buffer_length - start;
  if (length <= 0)
    return;

  if (make_buffer)
    {
      GdamCacheBuffer *new_buffer = NULL;
      new_buffer = gdam_cache_buffer_make_subbuffer (orig, start, length,
                                                     ! hi_quality);
      g_return_if_fail (new_buffer != NULL);

      *num_out = 1;
      *args_out = g_new (GdamArg, 1);
      (*args_out)[0].name = g_strdup ("buffer_id");
      (*args_out)[0].type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT ((*args_out) + 0) = gdam_cache_buffer_get_id (new_buffer);
    }
  if (cut)
    gdam_cache_buffer_cut (orig, start, length);
}

static void gdam_cache_buffer_replace   (GdamCacheBuffer     *buffer,
                                         GdamCacheBuffer     *insert_buffer,
					 int                  buf_start,
					 int                  buf_length,
					 int                  insert_start,
					 int                  insert_length,
					 int                  insert_repeat)
{
  int post_length;
  int new_length;
  if (insert_length == -1)
    insert_length = insert_buffer->buffer_length;
  if (insert_start < 0)
    {
      insert_length += insert_start;
      insert_start = 0;
    }
  if (insert_start + insert_length > insert_buffer->buffer_length)
    insert_length = insert_buffer->buffer_length - insert_start;
  if (insert_length < 0)
    return;

  post_length = buffer->buffer_length - buf_start - buf_length;
  new_length = buffer->buffer_length
             - buf_length
	     + insert_length * insert_repeat;

  if (buffer->is_16bit)
    {
      GdamSample16 *buf;
      buf = g_renew (GdamSample16, buffer->_buffer, new_length);
      buffer->_buffer = buf;
      memmove (buf + buf_start + insert_length,
	       buf + buf_start + buf_length,
	       post_length * sizeof (GdamSample16));
      if (insert_buffer->is_16bit)
	{
	  GdamSample16 *insert_buf = insert_buffer->_buffer;
	  while (insert_repeat-- > 0)
	    {
	      gdam_sample16_copy (buf + buf_start,
				  insert_buf + insert_start,
				  insert_length);
	      buf_start += insert_length;
	    }
        }
      else
	{
	  GdamSample *insert_buf = insert_buffer->_buffer;
	  insert_repeat--;
	  buf += buf_start;
	  gdam_sample16_from_normal (buf,
				     insert_buf + insert_start,
				     insert_length);
	  while (insert_repeat-- > 0)
	    {
	      GdamSample16 *new_buf = buf + insert_length;
	      gdam_sample16_copy (new_buf, buf, insert_length);
	      buf = new_buf;
	    }
        }
    }
  else
    {
      GdamSample *buf;
      buf = g_renew (GdamSample, buffer->_buffer, new_length);
      buffer->_buffer = buf;
      memmove (buf + buf_start + insert_length,
	       buf + buf_start + buf_length,
	       post_length * sizeof (GdamSample));
      if (insert_buffer->is_16bit)
	{
	  GdamSample16 *insert_buf = insert_buffer->_buffer;
	  while (insert_repeat-- > 0)
	    {
	      gdam_sample_from_16 (buf + buf_start,
				   insert_buf + insert_start,
				   insert_length);
	      buf_start += insert_length;
	    }
        }
      else
	{
	  GdamSample *insert_buf = insert_buffer->_buffer;
	  buf += buf_start;
	  while (insert_repeat-- > 0)
	    {
	      gdam_sample_copy (buf, insert_buf + insert_start, insert_length);
	      buf += insert_length;
	    }
        }
    }
  buffer->buffer_length = new_length;
}
static void gdam_cache_buffer_reverse   (GdamCacheBuffer     *buffer,
					 int                  buf_start,
					 int                  buf_length)
{
  if (buf_start < 0)
    {
      buf_length += buf_start;
      buf_start = 0;
    }
  if (buf_start + buf_length > buffer->buffer_length)
    buf_length = buffer->buffer_length - buf_start;
  if (buf_length <= 0)
    return;
  if (buffer->is_16bit)
    {
      GdamSample16 *buf = (GdamSample16 *) buffer->_buffer;
      GdamSample16 *start_ptr;
      GdamSample16 *end_ptr;
      start_ptr = buf + buf_start;
      end_ptr = buf + buf_start + buf_length - 1;
      while (start_ptr < end_ptr)
	{
	  GdamSample16 tmp = *start_ptr;
	  *start_ptr = *end_ptr;
	  *end_ptr = tmp;
	  start_ptr++;
	  end_ptr--;
	}
    }
  else
    {
      GdamSample *buf = (GdamSample *) buffer->_buffer;
      GdamSample *start_ptr;
      GdamSample *end_ptr;
      start_ptr = buf + buf_start;
      end_ptr = buf + buf_start + buf_length - 1;
      while (start_ptr < end_ptr)
	{
	  GdamSample tmp = *start_ptr;
	  *start_ptr = *end_ptr;
	  *end_ptr = tmp;
	  start_ptr++;
	  end_ptr--;
	}
    }
}

static void gdam_cache_buffer_replace_region(GtkObject *object,
                                             int        num_in,
                                             GdamArg   *args_in,
                                             int       *num_out,
                                             GdamArg  **args_out)
{
  int length = 0, start = 0, i;
  int insert_length = -1, insert_start = 0;
  int insert_buffer_id = 0;
  GdamCacheBuffer *insert_buffer;
  GdamCacheBuffer *orig = GDAM_CACHE_BUFFER (object);
  if (orig->read_only)
    {
      *num_out = 0;
      return;
    }
  for (i = 0; i < num_in; i++)
    {
      if (strcmp (args_in[i].name, "length") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  length = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "start") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  start = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "insert_buffer_id") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  insert_buffer_id = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "insert_start") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  insert_start = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "insert_length") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  insert_length = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      g_warning ("bad argument/type-mismatch: buffer_cut: %s",
		 args_in[i].name);
    }
  if (start < 0)
    {
      length += start;
      start = 0;
    }
  if (start + length > orig->buffer_length)
    length = orig->buffer_length - start;
  if (length <= 0)
    return;
  if (insert_buffer_id == 0)
    return;
  insert_buffer = gdam_cache_buffer_get_by_id (insert_buffer_id);
  if (insert_buffer == NULL)
    {
      gdam_debug ("buffer to insert %d not found", insert_buffer_id);
      return;
    }

  gdam_cache_buffer_replace (orig, insert_buffer, start, length,
                             insert_start, insert_length, 1);
}

static void gdam_cache_buffer_beat_times    (GtkObject *object,
                                             int        num_in,
                                             GdamArg   *args_in,
                                             int       *num_out,
                                             GdamArg  **args_out)
{
  gdouble *rv;
  gdouble window_size = 0.05;
  gdouble amplitude_ratio = 2.0;
  gdouble min_sample_length = 0.05;
  int amplitude_threshold = (int)(0.05 * GDAM_SAMPLE_ONE);
  int num_beat_times;
  int i;
  GdamCacheBuffer *buffer = GDAM_CACHE_BUFFER (object);

  for (i = 0; i < num_in; i++)
    {
      if (strcmp (args_in[i].name, "window_size") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_DOUBLE)
	{
	  window_size = GDAM_ARG_DOUBLE (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "amplitude_threshold") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  amplitude_threshold = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "amplitude_ratio") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_DOUBLE)
	{
	  amplitude_ratio = GDAM_ARG_DOUBLE (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "min_sample_length") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_DOUBLE)
	{
	  min_sample_length = GDAM_ARG_DOUBLE (args_in + i);
	  continue;
	}
      g_warning ("bad argument/type-mismatch: buffer_beat_times: %s",
		 args_in[i].name);
    }
  rv = gdam_cache_buffer_guess_beat_times (buffer,
					  window_size,
					  amplitude_threshold,
					  amplitude_ratio,
					  min_sample_length,
					  &num_beat_times);
  *num_out = 1;
  *args_out = g_new (GdamArg, 1);
  (*args_out)[0].name = g_strdup ("beat_times");
  (*args_out)[0].type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
  GDAM_ARG_DOUBLE_ARRAY ((*args_out) + 0).length = num_beat_times;
  GDAM_ARG_DOUBLE_ARRAY ((*args_out) + 0).array = rv;
}

static void gdam_cache_buffer_insert        (GtkObject *object,
                                             int        num_in,
                                             GdamArg   *args_in,
                                             int       *num_out,
                                             GdamArg  **args_out)
{
  GdamCacheBuffer *cache_buffer = GDAM_CACHE_BUFFER (object);
  GdamCacheBuffer *insert;
  GdamSample *at;
  const GdamSample *start_copy;
  int insertion_point = 0, multiplicity = 1, buffer_id = 0;
  int insert_offset = 0, insert_length = -1;
  int i;
  int new_length;
  int second_half_length;	/* length of original buffer after
				 * the insertion point. */

  if (cache_buffer->read_only)
    {
      *num_out = 0;
      return;
    }
  for (i = 0; i < num_in; i++)
    {
      if (strcmp (args_in[i].name, "multiplicity") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  multiplicity = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "insert_buffer_id") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  buffer_id = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "start") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  insertion_point = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "insert_offset") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  insert_offset = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "insert_length") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  insert_length = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      gdam_debug ("buffer_insert: unknown argument: %s", args_in[i].name);
    }

  if (buffer_id == 0
      || (insert = gdam_cache_buffer_get_by_id (buffer_id)) == NULL)
    {
      gdam_debug ("insert: no buffer id %d found", buffer_id);
      return;
    }

  if (insert_length == -1)
    insert_length = insert->buffer_length;
  if (insert_offset < 0)
    {
      insert_length += insert_offset;
      insert_offset = 0;
    }
  if (insert_offset + insert_length > insert->buffer_length)
    insert_length = insert->buffer_length - insert_offset;
  if (insert_length <= 0)
    return;

  if (insertion_point > cache_buffer->buffer_length)
    insertion_point = cache_buffer->buffer_length;
  if (insertion_point < 0)
    insertion_point = 0;

  gdam_cache_buffer_replace (cache_buffer, insert,
                             insertion_point, 0,
			     insert_offset, insert_length,
			     multiplicity);
}
static void gdam_cache_buffer_reverse_region(GtkObject* object,
				             int        num_in,
                                             GdamArg*   args_in,
				             int*       num_out,
				             GdamArg**  args_out)
{
  int length = -1, start = 0, i;
  GdamCacheBuffer *cache = GDAM_CACHE_BUFFER (object);
  GdamSample *start_ptr, *end_ptr;
  if (cache->read_only)
    {
      *num_out = 0;
      return;
    }
  for (i = 0; i < num_in; i++)
    {
      if (strcmp (args_in[i].name, "length") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  length = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "start") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  start = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      g_warning ("bad argument/type-mismatch: buffer_reverse: %s",
		 args_in[i].name);
    }
  gdam_cache_buffer_reverse (cache, start, length);
}

typedef struct _SavingThread SavingThread;

struct _SavingThread {
	char* filename;
	FILE* fp;
	GdamCacheBuffer* buffer;
	int at;
	int remaining;
};

static void free_saving_thread(SavingThread* saving_thread)
{
  if (saving_thread->fp)
    {
      FILE *fp = saving_thread->fp;
      saving_thread->fp = NULL;
      fclose (fp);
    }
  gtk_object_unref (GTK_OBJECT (saving_thread->buffer));
  g_free (saving_thread);
}
#define MAX_ATOMIC_SAVE		8192


/* TODO:  could use some cleanup.... */
static gboolean save_more(gpointer user_data)
{
  SavingThread *saving_thread = user_data;
  int to_write = MAX_ATOMIC_SAVE;
  int amt;
  char *data;
  GdamCacheBuffer *buffer = saving_thread->buffer;
  char *raw_buffer = buffer->_buffer;
  int sample_size;
  int sample_bits = buffer->is_16bit ? 16 : GDAM_SAMPLE_BITS;
  GdamSampleConverter wav_converter = NULL;
  sample_size = buffer->is_16bit ? sizeof (GdamSample16) : sizeof (GdamSample);
  if (saving_thread->at >= buffer->buffer_length)
    {
      free_saving_thread (saving_thread);
      return FALSE;
    }
  if (saving_thread->remaining >= 0 && to_write > saving_thread->remaining)
    to_write = saving_thread->remaining;
  if (to_write + saving_thread->at > buffer->buffer_length)
    to_write = buffer->buffer_length - saving_thread->at;
  if (to_write <= 0)
    {
      free_saving_thread (saving_thread);
      return FALSE;
    }
  data = alloca (4 * to_write);
  wav_converter = gdam_sample_converter_get
      (sample_bits,
       TRUE,
       (G_BYTE_ORDER == G_BIG_ENDIAN),
       16, TRUE, (G_BYTE_ORDER == G_BIG_ENDIAN));

  (*wav_converter) (data, raw_buffer + saving_thread->at * sample_size,
                    to_write * 2);
  amt = fwrite (data, 4, to_write, saving_thread->fp);
  if (amt <= 0)
    {
      g_warning ("fwrite failed writing %s", saving_thread->filename);
      free_saving_thread (saving_thread);
      return FALSE;
    }
  saving_thread->at += amt;
  if (saving_thread->at == buffer->buffer_length)
    {
      free_saving_thread (saving_thread);
      return FALSE;
    }
  if (saving_thread->remaining >= 0)
    {
      if ((saving_thread->remaining -= amt) <= 0)
	{
	  free_saving_thread (saving_thread);
	  return FALSE;
	}
    }
  return TRUE;
}

#define WAV_RIFF   0x46464952	/* "RIFF" */
#define WAV_WAVE   0x45564157	/* "WAVE" */
#define WAV_LIST   0x5453494c	/* "LIST" */
#define WAV_INFO   0x4f464e49	/* "INFO" */
#define WAV_ISFT   0x54465349	/* "ISFT" */
#define WAV_ICRD   0x44524349	/* "ICRD" */
#define WAV_FMT    0x20746d66	/* "fmt " */
#define WAV_DATA   0x61746164	/* "data" */
#define WAV_MONO   0x00010001	/* 0x0001 0x0001 */
#define WAV_STEREO 0x00020001	/* 0x0001 0x0002 */

static gboolean write_wave_file_header(FILE* fp, int length)
{
  guint32 byte_length;
  guint32 h[11];
  guint32 fmt = (16 << 16) | 4;

  byte_length = length * 4;
  memcpy (h + 0, "RIFF", 4);
  h[1] = byte_length + 36;
  memcpy (h + 2, "WAVE", 4);
  memcpy (h + 3, "fmt ", 4);
  h[4] = GINT32_TO_LE (16);
  h[5] = GINT32_TO_LE (WAV_STEREO);
  h[6] = 44100;
  h[7] = 44100 * 4;
  h[8] = fmt;			/* TODO: see what this field should be */
  memcpy (h + 9, "data", 4);
  h[10] = byte_length;
  if (fwrite (h, 4, 11, fp) != 11 || ferror (fp))
    return FALSE;
  return TRUE;
}

static void gdam_cache_buffer_save_region   (GtkObject* object,
				             int        num_in,
                                             GdamArg*   args_in,
				             int*       num_out,
				             GdamArg**  args_out)
{
  int length = -1, start = 0, i;
  GdamCacheBuffer *cache_buffer = GDAM_CACHE_BUFFER (object);
  char *filename = NULL;
  SavingThread *thread_data;
  char *extension;
  FILE *fp;
  for (i = 0; i < num_in; i++)
    {
      if (strcmp (args_in[i].name, "length") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  length = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "start") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  start = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "filename") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_STRING)
	{
	  filename = g_strdup (GDAM_ARG_STRING (args_in + i));
	  continue;
	}
      g_warning ("bad argument/type-mismatch: buffer_save: %s",
		 args_in[i].name);
    }
  if (filename == NULL)
    {
      g_warning ("no filename was given to save_region");
      return;
    }
  extension = strrchr (filename, '.');
  if (extension == NULL || g_strcasecmp (extension, ".wav") != 0)
    {
      g_warning ("files must end in .wav (for security)");
      return;
    }
  fp = fopen (filename, "w");
  if (fp == NULL)
    {
      g_warning ("opening create %s", filename);
      return;
    }
  thread_data = g_new (SavingThread, 1);
  thread_data->filename = filename;
  thread_data->fp = fp;
  thread_data->buffer = cache_buffer;
  thread_data->at = start;
  thread_data->remaining = length;
  gtk_object_ref (GTK_OBJECT (cache_buffer));
  if (length < 0)
    length = cache_buffer->buffer_length - start;
  if (!write_wave_file_header (fp, length))
    {
      g_warning ("error writing wave file header");
      free_saving_thread (thread_data);
      return;
    }
  g_idle_add (save_more, thread_data);
}

static void gdam_cache_buffer_class_init(GdamCacheBufferClass* klass)
{
  GtkObjectClass *object_class = (GtkObjectClass *) klass;

  klass->finished = gdam_cache_buffer_finished;
  object_class->destroy = gdam_cache_buffer_destroy;

  cache_buffer_signals[FINISHED] =
    gtk_signal_new ("finished",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamCacheBufferClass,
				       finished),
		    gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class, cache_buffer_signals,
				LAST_SIGNAL);

  gdam_arg_register (object_class,
		     "length",
		     GDAM_ARG_TYPE_INT64,
		     NULL, gdam_cache_buffer_arg_get_length, NULL, NULL, 0);
  gdam_arg_function_register (object_class,
			      "guess_beat_times",
			      gdam_cache_buffer_beat_times);
  gdam_arg_function_register (object_class,
			      "process_region",
			      gdam_cache_buffer_process_region);
  gdam_arg_function_register (object_class,
			      "replace_region",
			      gdam_cache_buffer_replace_region);
  gdam_arg_function_register (object_class,
			      "insert_region", gdam_cache_buffer_insert);
  gdam_arg_function_register (object_class,
			      "reverse_region",
			      gdam_cache_buffer_reverse_region);
  gdam_arg_function_register (object_class,
			      "save", gdam_cache_buffer_save_region);
}

static void gdam_cache_buffer_init(GdamCacheBuffer* cache)
{
  gdam_buffer_construct (&cache->loading_buffer);
  cache->buffer_length = -1;
}

GtkType gdam_cache_buffer_get_type ()
{
  static GtkType cache_buffer_type = 0;
  if (!cache_buffer_type)
    {
      static const GtkTypeInfo cache_buffer_info = {
	"GdamCacheBuffer",
	sizeof (GdamCacheBuffer),
	sizeof (GdamCacheBufferClass),
	(GtkClassInitFunc) gdam_cache_buffer_class_init,
	(GtkObjectInitFunc) gdam_cache_buffer_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAM_TYPE_SERVER_OBJECT);
      cache_buffer_type = gtk_type_unique (GDAM_TYPE_SERVER_OBJECT,
					   &cache_buffer_info);
    }
  return cache_buffer_type;
}

GdamCacheBuffer* gdam_cache_buffer_new(GdasSource* to_drain)
{
  GdamCacheBuffer *buffer;
  GtkObject *to_drain_object = GTK_OBJECT (to_drain);
  g_return_val_if_fail (to_drain_object != NULL, NULL);

  buffer = (GdamCacheBuffer *) (gtk_type_new (GDAM_TYPE_CACHE_BUFFER));
  buffer->to_drain = to_drain;
  gtk_object_ref (to_drain_object);
  gtk_object_sink (to_drain_object);
  buffer->source_id = g_source_add (G_PRIORITY_LOW, FALSE, &source_funcs, NULL,	/* no source_data */
				    buffer, NULL);
  g_return_val_if_fail (buffer->source_id != 0, NULL);
  gdam_debug ("new cache buffer (source_id=%d) ref_count %d",
	      buffer->source_id, GTK_OBJECT (buffer)->ref_count);
  return buffer;
}

GdamCacheBuffer  *gdam_cache_buffer_memory    (int              buffer_length,
                                               void            *buffer,
                                               GDestroyNotify   free_func,
					       gboolean         is_16bit)
{
  GdamCacheBuffer *cache_buffer;
  cache_buffer = (GdamCacheBuffer *) (gtk_type_new (GDAM_TYPE_CACHE_BUFFER));
  cache_buffer->buffer_length = buffer_length;
  cache_buffer->_buffer = buffer;
  cache_buffer->free_func = free_func;
  cache_buffer->is_16bit = is_16bit;
  return cache_buffer;
}
GdamCacheBuffer  *gdam_cache_buffer_memory_16 (int              buffer_length,
                                               GdamSample16    *buffer,
                                               GDestroyNotify   free_func)
{
  GdamCacheBuffer *cache_buffer;
  cache_buffer = (GdamCacheBuffer *) (gtk_type_new (GDAM_TYPE_CACHE_BUFFER));
  cache_buffer->buffer_length = buffer_length;
  cache_buffer->_buffer = buffer;
  cache_buffer->free_func = free_func;
  cache_buffer->is_16bit = TRUE;
  return cache_buffer;
}

static void my_free_wrapper(gpointer ptr)
{
  if (ptr)
    free (ptr);
}


/* We use malloc and free here rather than g_malloc/g_free
 * b/c this is a rather likely place to run out of memory,
 * and b/c handling the out-of-memory condition in this case
 * is quite easy...
 */
static void gdam_cache_buffer_finished(GdamCacheBuffer* cache)
{
  int n;
  int sample_size;
  g_return_if_fail (cache->buffer_length == -1);
  sample_size = cache->is_16bit ? sizeof (GdamSample16) : sizeof (GdamSample);
  cache->buffer_length = cache->loading_buffer.size / sizeof (GdamSample);
  cache->_buffer = malloc (sample_size * cache->buffer_length);
  cache->free_func = my_free_wrapper;
  if (cache->_buffer == NULL)
    {
      cache->buffer_length = 0;
      g_message ("couldn't create buffer: out of memory");
      gdam_buffer_destruct (&cache->loading_buffer);
      return;
    }
  n = gdam_buffer_read (&cache->loading_buffer,
			cache->_buffer,
			cache->buffer_length * sample_size);
  gdam_debug ("done caching: buffer id = %d, length = %d, ref_count = %d",
	      gdam_cache_buffer_get_id (cache),
	      cache->buffer_length, GTK_OBJECT (cache)->ref_count);
  g_return_if_fail (n == cache->buffer_length * (int) sample_size);
  gdam_buffer_destruct (&cache->loading_buffer);
}

static gboolean  gdam_cache_buffer_prepare        (gpointer         src_data,
                                                   GTimeVal*        curtime,
                                                   gint*            timeout,
		                                   gpointer         user_data)
{
  GdamCacheBuffer *cache_buffer = (GdamCacheBuffer *) user_data;
  gboolean rv = gdas_source_is_ready_to_read (cache_buffer->to_drain);
  return rv;
}
static gboolean  gdam_cache_buffer_check          (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data)
{
  GdamCacheBuffer *cache_buffer = (GdamCacheBuffer *) user_data;
  gboolean rv = gdas_source_is_ready_to_read (cache_buffer->to_drain);
  if (GDAM_SERVER_DEBUG_POLL)
    gdam_debug ("gdam_cache_buffer_check: is_ready=%d", rv);
  return rv;
}

static gboolean  gdam_cache_buffer_dispatch       (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data)
{
  /* XXX: had to decrease this because macosx seg faults on all reads >= 4096*/
  //GdamSample tmp_buf[4096];
  GdamSample tmp_buf[2048];
  int n;
  GdamCacheBuffer *cache = GDAM_CACHE_BUFFER (user_data);
  g_return_val_if_fail (cache->to_drain != NULL, TRUE);
  g_return_val_if_fail (cache->buffer_length == -1, TRUE);

  n = gdas_source_read (cache->to_drain, tmp_buf,
			sizeof (tmp_buf) / sizeof (GdamSample));
  if (n <= 0)
    {
      GtkObject *sub;

      gtk_signal_emit (GTK_OBJECT (cache), cache_buffer_signals[FINISHED]);
      cache->source_id = 0;
      sub = GTK_OBJECT (cache->to_drain);
      cache->to_drain = NULL;
      gtk_object_unref (sub);
      return FALSE;
    }
  gdam_buffer_append (&cache->loading_buffer,
		      tmp_buf, n * sizeof (GdamSample));
  return TRUE;
}

static void  gdam_cache_buffer_destroy(GtkObject* cache)
{
  GdamCacheBuffer *cache_buffer = GDAM_CACHE_BUFFER (cache);
  if (cache_buffer->name != NULL)
    gdam_cache_buffer_set_name (cache_buffer, NULL);
  if (cache_buffer->source_id)
    {
      int source_id = cache_buffer->source_id;
      cache_buffer->source_id = 0;
      g_source_remove (source_id);
    }
  if (cache_buffer->to_drain != NULL)
    {
      GtkObject *sub;
      sub = GTK_OBJECT (cache_buffer->to_drain);
      cache_buffer->to_drain = NULL;
      gtk_object_unref (sub);
    }
  gdam_debug ("destroying buffer %d, size %d",
	      gdam_cache_buffer_get_id (cache_buffer),
	      cache_buffer->buffer_length);
  gdam_buffer_destruct (&cache_buffer->loading_buffer);
  if (cache_buffer->free_func)
    (*cache_buffer->free_func) (cache_buffer->_buffer);
  (*parent_class->destroy) (cache);
}

GdamCacheBuffer*  gdam_cache_buffer_get_by_id (guint            buffer_id)
{
  GdamServerObject *object;
  object = gdam_dispatch_get_server_object (buffer_id);
  if (object == NULL)
    return NULL;
  gtk_object_unref ((GtkObject *) object);
  if (!GDAM_IS_CACHE_BUFFER (object))
    return NULL;
  return (GdamCacheBuffer *) object;
}
int               gdam_cache_buffer_get_id    (GdamCacheBuffer* buffer)
{
  GdamServerObject *object = GDAM_SERVER_OBJECT (buffer);
  return object->global_id;
}


static void marshal_cache_buffer_finish(GdamConnection*    registered,
                                        GtkObject*         actor,
			                int                connection_id,
			                gpointer           event_type_data,
			                gpointer           connection_data)
{
  GdamCacheBuffer *external = GDAM_CACHE_BUFFER (actor);
  GdamBuffer event;
  gdam_buffer_construct (&event);
  gdam_buffer_append_int (&event, GDAM_EVENT_CACHE_BUFFER_FINISH);
  gdam_buffer_append_int (&event, connection_id);
  gdam_buffer_append_int (&event, gdam_cache_buffer_get_id (external));
  gdam_connection_emit_event (registered, &event);
}

guint gdam_cache_buffer_get_done_event()
{
  static guint id = 0;
  if (id == 0)
    {
      id = gdam_event_register (GDAM_TYPE_CACHE_BUFFER,
				"finished",
				marshal_cache_buffer_finish, NULL, NULL);
    }
  return id;
}

GdamSample        gdam_cache_buffer_index     (GdamCacheBuffer *buffer,
                                               int              index)
{
  if (buffer->_buffer == NULL)
    {
      GtkObject *object = (GtkObject*) buffer;
      GdamCacheBufferClass *klass = (GdamCacheBufferClass*) object->klass;
      return (*klass->index)(buffer, index);
    }
  else if (buffer->is_16bit)
    {
      GdamSample rv;
      const gint16 *in = (const gint16 *)(buffer->_buffer);
      in += 2 * index;
      rv.s[0] = (int)in[0] << (GDAM_SAMPLE_BITS - 16);
      rv.s[1] = (int)in[1] << (GDAM_SAMPLE_BITS - 16);
      return rv;
    }
  else
    {
      GdamSample *buf = buffer->_buffer;
      return buf[index];
    }
}

static GHashTable *named_buffers = NULL;

void
gdam_cache_buffer_set_name  (GdamCacheBuffer *buffer,
			     const char      *name)
{
  if (buffer->name != NULL)
    {
      g_hash_table_remove (named_buffers, buffer->name);
      buffer->name = NULL;
    }
  if (name != NULL)
    {
      if (named_buffers == NULL)
	named_buffers = g_hash_table_new (g_str_hash, g_str_equal);
      g_hash_table_remove (named_buffers, name);
      buffer->name = g_strdup (name);
      g_hash_table_insert (named_buffers, buffer->name, buffer);
    }
}

GdamCacheBuffer *
gdam_cache_buffer_peek_by_name(const char      *name)
{
  if (named_buffers == NULL)
    return NULL;
  return g_hash_table_lookup (named_buffers, name);
}

static int
gdam_cache_buffer_add_1to1 (GdamCacheBuffer    *buffer,
			    GdamSample         *output,
			    int                 index,
			    int                *end_index,
			    int                 max_output,
			    GdamFixed           gain)
{
  int rv = 0;
  g_assert (buffer->_buffer != NULL);
  if (index < 0)
    {
      rv = -index;
      if (rv > max_output)
        rv = max_output;
    }
  if (max_output > buffer->buffer_length - index)
    max_output = buffer->buffer_length - index;
  if (rv >= max_output)
    {
      if (end_index != NULL)
        *end_index = max_output + index;
      return rv;
    }
  if (buffer->is_16bit)
    gdam_sample_addto_from_16_scaled
			      (output + rv,
			       ((GdamSample16*)buffer->_buffer) + index + rv,
			       max_output - rv,
			       gain);
      
  else
    gdam_sample_add_scaled (output + rv,
			    ((GdamSample*)buffer->_buffer) + index + rv,
			    max_output - rv,
			    gain);
  if (end_index != NULL)
    *end_index = max_output + index;
  return max_output;
}
static int gdam_cache_buffer_read_1to1 (GdamCacheBuffer    *buffer,
                                        GdamSample         *output,
					int                 index,
					int                *end_index,
					int                 max_output,
					GdamFixed           gain)
{
  int rv = 0;
  g_assert (buffer->_buffer != NULL);
  if (index < 0)
    {
      rv = -index;
      if (rv > max_output)
        rv = max_output;
      gdam_sample_zero (output, rv);
    }
  if (max_output > buffer->buffer_length - index)
    max_output = buffer->buffer_length - index;
  if (rv >= max_output)
    {
      if (end_index != NULL)
        *end_index = max_output + index;
      return rv;
    }
  if (buffer->is_16bit)
    gdam_sample_from_16 (output + rv,
			 ((GdamSample16*)buffer->_buffer) + index + rv,
			 max_output - rv);
  else
    gdam_sample_copy (output + rv,
                      ((GdamSample*)buffer->_buffer) + index + rv,
		      max_output - rv);
  if (gain != GdamFixed_ONE)
    gdam_sample_scale (output + rv, max_output - rv, gain);
  if (end_index != NULL)
    *end_index = max_output + index;
  return max_output;
}

int               gdam_cache_buffer_read      (GdamCacheBuffer *buffer,
                                               GdamSample      *output,
                                               int              start_index,
					       guint32          index_frac,
					       int             *end_index,
					       guint32         *end_index_frac,
                                               int              max_output,
                                               GdamFixed        rate,
                                               GdamFixed        gain,
                                               gboolean         add)
{
  /* number of samples to step at each iteration of the inner loop. */
  int step_whole;

  /* number of fractional samples to step at each iteration. */
  guint32 step_frac;

  if (buffer->_buffer == NULL)
    {
      GtkObject *object = (GtkObject*) buffer;
      GdamCacheBufferClass *klass = (GdamCacheBufferClass*) object->klass;
      return (*klass->read)(buffer, output, start_index, index_frac,
                            end_index, end_index_frac, 
			    max_output, rate,
                            gain, add);
    }
  else
    {
      int rv = 0;
      if (rate == 0)
        {
	  gdam_sample_zero (output, max_output);
	  return max_output;
	}
      if (rate < 0)
        {
	  gint64 end_index_fixed;
	  gint64 max_samples;
	  /* clip the samples. (we are traversing the sample backward) */
	  /* clip off stuff before the end of the sample. */
	  if (start_index > buffer->buffer_length)
	    {
	      int input_offset = start_index - buffer->buffer_length;
	      int output_offset;
	      gint64 index_delta;
	      output_offset = ((gint64) input_offset << GdamFixed_LOG_ONE)
	                    / (-rate);
	      if (output_offset > max_output)
	        output_offset = max_output;
	      gdam_sample_zero (output, output_offset);
	      rv = output_offset;

	      /* update start_index and index_frac */
	      index_delta = output_offset;
	      index_delta *= rate;
	      start_index += (index_delta >> GdamFixed_LOG_ONE);
	      index_frac = (index_delta & (GdamFixed_ONE - 1))
	                 << (32 - GdamFixed_LOG_ONE);
	    }

	  /* clip off stuff after the beginning of the buffer */
	  max_samples = ((gint64) start_index << GdamFixed_LOG_ONE)
	              / (-rate);
	  if (max_samples < (gint64) (max_output + rv))
	    {
	      max_output = max_samples - rv;
	      if (max_output < 0)
	        max_output = 0;
	    }
	}
      else
        {
	  gint64 end_index_fixed;
	  gint64 max_to_end;
	  if (rate == GdamFixed_ONE)
	    {
	      if (end_index_frac != NULL)
	        *end_index_frac = index_frac;
	      if (add)
		return gdam_cache_buffer_add_1to1 (buffer, output, start_index,
						   end_index, max_output, gain);
	      else
		return gdam_cache_buffer_read_1to1 (buffer, output, start_index,
						   end_index, max_output, gain);
	    }
	  /* clip off stuff before the beginning of the buffer. */
	  if (start_index < 0)
	    {
	      int output_offset;
	      gint64 input_offset;
	      output_offset = ((gint64) (-start_index) << GdamFixed_LOG_ONE)
	                    / rate;
	      if (output_offset > max_output)
	        output_offset = max_output;
	      rv = output_offset;

	      input_offset = (gint64) rate * output_offset;
	      start_index += (input_offset >> GdamFixed_LOG_ONE);
	      index_frac += ((input_offset & (GdamFixed_ONE-1))
	                      << (32 - GdamFixed_LOG_ONE));
	      gdam_sample_zero (output, rv);
	    }

	  /* clip off stuff after the end of the buffer. */
	  end_index_fixed = buffer->buffer_length - start_index;
	  end_index_fixed <<= GdamFixed_LOG_ONE;
	  end_index_fixed -= (index_frac >> (32 - GdamFixed_LOG_ONE));
	  max_to_end = (gint64) (end_index_fixed / rate);
	  if (max_to_end < max_output - rv)
	    {
	      max_output = max_to_end + rv;
	    }
	}

      /* ok, we're all clipped; just do the output. */
      step_whole = rate >> GdamFixed_LOG_ONE;
      step_frac = ((unsigned)rate & (GdamFixed_ONE - 1));
      step_frac <<= (32 - GdamFixed_LOG_ONE);
      if (buffer->is_16bit)
	{
	  GdamSample16 *at = buffer->_buffer;
	  rv += max_output;
	  at += start_index;
#define WRITE_LOOP_INCREMENT()					\
	G_STMT_START{ guint32 tmp_incr;                         \
	              at += step_whole;				\
		      tmp_incr = step_frac + index_frac;	\
		      if (tmp_incr < index_frac)		\
	  	        at++;                              	\
		      output++;					\
		      index_frac = tmp_incr; }G_STMT_END
	    
	  if (gain == GdamFixed_ONE)
	    if (add)
	      while (max_output-- > 0)
		{
		  /* 16-bit, unit gain, adding */
		  output->s[0] += ((int)at->s[0]) << (GDAM_SAMPLE_BITS - 16);
		  output->s[1] += ((int)at->s[1]) << (GDAM_SAMPLE_BITS - 16);
		  WRITE_LOOP_INCREMENT();
		}
	    else
	      while (max_output-- > 0)
		{
		  /* 16-bit, unit gain, replacing */
		  output->s[0] = ((int)at->s[0]) << (GDAM_SAMPLE_BITS - 16);
		  output->s[1] = ((int)at->s[1]) << (GDAM_SAMPLE_BITS - 16);
		  WRITE_LOOP_INCREMENT();
		}
	  else  /* non unit gains ... */
	    if (add)
	      while (max_output-- > 0)
	        {
		  /* 16-bit, non-unit gain, adding */
		  int tmp;
		  tmp = ((int)at->s[0]) << (GDAM_SAMPLE_BITS - 16);
		  tmp = GDAM_FIXED_MUL (tmp, gain);
		  output->s[0] += tmp;
		  tmp = ((int)at->s[1]) << (GDAM_SAMPLE_BITS - 16);
		  tmp = GDAM_FIXED_MUL (tmp, gain);
		  output->s[1] += tmp;
		  WRITE_LOOP_INCREMENT();
		}
	    else
	      while (max_output-- > 0)
	        {
		  /* 16-bit, non-unit gain, not adding */
		  int tmp;
		  tmp = ((int)at->s[0]) << (GDAM_SAMPLE_BITS - 16);
		  tmp = GDAM_FIXED_MUL (tmp, gain);
		  output->s[0] = tmp;
		  tmp = ((int)at->s[1]) << (GDAM_SAMPLE_BITS - 16);
		  tmp = GDAM_FIXED_MUL (tmp, gain);
		  output->s[1] = tmp;
		  WRITE_LOOP_INCREMENT();
		}
        }
      else
        {
	  GdamSample *at = buffer->_buffer;
	  rv += max_output;
	  at += start_index;
	  if (gain == GdamFixed_ONE)
	    if (add)
	      while (max_output-- > 0)
		{
		  /* native-bit, unit gain, adding */
		  output->s[0] += at->s[0];
		  output->s[1] += at->s[1];
		  WRITE_LOOP_INCREMENT();
		}
	    else
	      while (max_output-- > 0)
		{
		  /* native-bit, unit gain, replacing */
		  output->s[0] = at->s[0];
		  output->s[1] = at->s[1];
		  WRITE_LOOP_INCREMENT();
		}
	  else  /* non unit gains ... */
	    if (add)
	      while (max_output-- > 0)
	        {
		  /* native-bit, non-unit gain, adding */
		  output->s[0] += GDAM_FIXED_MUL (at->s[0], gain);
		  output->s[1] += GDAM_FIXED_MUL (at->s[1], gain);
		  WRITE_LOOP_INCREMENT();
		}
	    else
	      while (max_output-- > 0)
	        {
		  /* native-bit, non-unit gain, replacing */
		  output->s[0] = GDAM_FIXED_MUL (at->s[0], gain);
		  output->s[1] = GDAM_FIXED_MUL (at->s[1], gain);
		  WRITE_LOOP_INCREMENT();
		}
        }
      if (end_index != NULL)
        *end_index = start_index;
      if (end_index_frac != NULL)
        *end_index_frac = index_frac;

      return rv;
    }
}
