/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamconnection.h"
#include "gdamdispatch.h"
#include "gdamnamespace.h"
#include "gdamevent.h"
#include "gdamcachebuffer.h"
#include "gdamserverdebug.h"
#include <unistd.h>
#include <errno.h>

int             gdam_connection_total_log_level = 0;


GdamConnection* gdam_connection_new(int client_fd,
                                    GdamAuthenticator* authenticator)
{
  GdamConnection *connection;

  if (!authenticator)
    {
      g_warning ("no authenticator for connection:  cannot accept");
      return NULL;
    }
  connection = g_new0 (GdamConnection, 1);
  gdam_buffer_construct (&connection->client_commands);
  connection->authenticator = authenticator;
  connection->fd = client_fd;
  connection->fd_poll = NULL;
  connection->local_id_table = g_hash_table_new (NULL, NULL);
  gdam_buffer_construct (&connection->outgoing);
  connection->is_readable = 1;
  connection->ref_table = g_hash_table_new (NULL, NULL);
  gdam_permissions_none (&connection->permissions);

  return connection;
}

static void do_unreferences(gpointer key, gpointer value, gpointer con)
{
  GtkObject *object = key;
  int count = GPOINTER_TO_INT (value);
  while (count-- > 0)
    gtk_object_unref (object);
}
static void prune_connections_from_server_objects(gpointer k, gpointer value,
                                                  gpointer unused)
{
  GdamServerObject *server_object;
  server_object = GDAM_SERVER_OBJECT (value);
  server_object->connection = NULL;
  server_object->local_id = 0;
  gtk_object_sink (GTK_OBJECT (server_object));
}

static void gdam_connection_destroy(GdamConnection* connection)
{
  if (connection->username != NULL)
    gdam_connection_set_username (connection, NULL);
  gdam_event_disconnect_all_from_connection (connection);

  gdam_debug ("<l>gdam_connection_destroy: %d references to clear",
	      connection->
	      ref_table ? g_hash_table_size (connection->ref_table) : 0);

  gdam_connection_set_log_level (connection, 0);

  if (connection->ref_table)
    {
      g_hash_table_foreach (connection->ref_table,
			    do_unreferences, connection);
      g_hash_table_destroy (connection->ref_table);
      connection->ref_table = NULL;
    }

  g_hash_table_foreach (connection->local_id_table,
			prune_connections_from_server_objects, NULL);
  g_hash_table_destroy (connection->local_id_table);
  connection->local_id_table = NULL;

  if (connection->fd_poll != NULL)
    {
      gdam_fd_poll_destroy (connection->fd_poll);
      connection->fd_poll = NULL;
    }
  if (connection->fd >= 0)
    {
      close (connection->fd);
      connection->fd = -1;
    }

  gdam_buffer_destruct (&connection->client_commands);
  gdam_buffer_destruct (&connection->outgoing);

  /* XXX: should we call g_main_quit instead? */
  if (connection->exit_upon_close)
    exit (0);

  g_free (connection);
}

#if 0
static void dispatch_disconnected(gpointer conn)
{
  GdamConnection *connection = (GdamConnection *) conn;
  if (connection->client_poll.fd >= 0)
    {
      g_main_remove_poll (&connection->client_poll);
      close (connection->client_poll.fd);
      connection->client_poll.fd = -1;
    }
  connection->source_id = 0;
  gdam_connection_destroy (connection);
}
#endif
static inline GIOCondition
compute_events (GdamConnection *connection)
{
  GIOCondition rv = connection->is_readable ? (G_IO_IN | G_IO_HUP) : 0;
  if (connection->outgoing.size > 0)
    rv |= G_IO_OUT;
  rv |= G_IO_ERR;
  return rv;
}

static inline GdamBuffer *
do_dispatch(GdamConnection* connection)
{
  GdamBuffer command;
  GdamBuffer *rv;
  int length;
  gdam_buffer_peek_int (&connection->client_commands, &length);
  gdam_buffer_construct (&command);
  gdam_debug ("<c>do_dispatch:  starting size=%d, length=%d",
	      connection->client_commands.size, length);
  gdam_buffer_send_buffer (&command,
			   &connection->client_commands, length, TRUE);
  g_assert (command.size == length);
  gdam_debug ("<c>do_dispatch:  remaining client input's length=%d",
	      connection->client_commands.size);

  gdam_buffer_parse_int (&command, &length);

  if (connection->authenticator)
    {
      GdamAuthenticator *auth = connection->authenticator;
      gboolean is_done;
      gboolean failed = TRUE;
      gdam_debug ("<c>doing authentification");
      rv = (*auth->authenticate) (auth,
				  &command,
				  &connection->permissions,
				  &is_done, &failed);
      if (is_done)
	{
	  if (auth->destroy)
	    (*auth->destroy) (auth);
	  connection->authenticator = NULL;
	  if (failed)
	    {
	      gdam_debug ("failed authentication");
	      return NULL;
	    }
	}
    }
  else
    {
      rv = gdam_dispatch_buffer (connection,
				 &connection->permissions, &command);
    }
  gdam_buffer_destruct (&command);
  gdam_buffer_prepend_length (rv);
  return rv;
}

static gboolean 
gdam_connection_handle_events  (GdamFdPoll     *fd_poll,
				int             fd,
				GIOCondition    events,
				gpointer        data)
{
  GdamConnection *connection = (GdamConnection *) data;

  int client_fd = connection->fd;
  g_return_val_if_fail (client_fd == fd, FALSE);
  g_return_val_if_fail (client_fd >= 0, FALSE);

  if (events & G_IO_IN)
    {
      char buf[16 * 1024];
      /* The buffer for client input. */
      GdamBuffer *client_buf = &connection->client_commands;
      int num_read;

      /* Do the reading. (and error handling) */
      num_read = read (client_fd, buf, sizeof (buf));
      if (num_read < 0)
	{
	  if (errno == EINTR || errno == EWOULDBLOCK || errno == EINPROGRESS)
	    {
	      return TRUE;
	    }
	  g_warning ("dispatch command:  reading from client: %s",
		     strerror (errno));
	  connection->fd_poll = NULL;
	  return FALSE;
	}

      /* Add newly-read data into "incoming" buffer. */
      gdam_buffer_append (client_buf, buf, num_read);
      /*gdam_debug("appending %d bytes from client [new size=%d]",
         num_read, client_buf->size); */

      if (num_read == 0)
	{
	  /* end-of-file condition... */
	  connection->is_readable = 0;
	}
      else
	{
	  GdamBuffer *response;
	  int cmd_len = 0;
	  while (gdam_buffer_peek_int (client_buf, &cmd_len)
		 && cmd_len <= client_buf->size)
	    {
#if DEBUG_DISPATCHING
	      gdam_debug ("<d>dispatching message... len=%d", cmd_len);
#endif
	      response = do_dispatch (connection);
	      if (!response)
		{
		  g_warning ("error processing command");
		  connection->fd_poll = NULL;
		  return FALSE;
		}
	      gdam_buffer_append_buffer (&connection->outgoing,
					 response, TRUE);
	      gdam_buffer_destroy (response);
	    }
#if DEBUG_DISPATCHING
	  if (gdam_buffer_peek_int (client_buf, &cmd_len))
	    gdam_debug ("waiting to dispatch message: "
			"cur_len=%d, cmd_len=%d", client_buf->size, cmd_len);
#endif
	}
    }

  /* Handle outgoing data. */
  if (events & G_IO_OUT && connection->outgoing.size > 0)
    {
      if (!gdam_buffer_write_one (&connection->outgoing, client_fd))
	{
	  connection->fd_poll = NULL;
	  return FALSE;
	}
    }

  if (connection->outgoing.size == 0 && !connection->is_readable)
    {
      connection->fd_poll = NULL;
      return FALSE;
    }

  if (events & (G_IO_HUP | G_IO_ERR))
    {
      connection->fd_poll = NULL;
      return FALSE;
    }

  gdam_fd_poll_set_events (fd_poll, compute_events (connection));
  return TRUE;
}

void gdam_connection_attach_to_main_loop(GdamConnection* connection)
{
  g_return_if_fail (connection->fd_poll == NULL);
  g_return_if_fail (connection->fd >= 0);

  connection->fd_poll = gdam_fd_poll_new (connection->fd,
				          G_PRIORITY_DEFAULT,
				          compute_events (connection),
				          gdam_connection_handle_events,
				          connection,
				          NULL);
}


void  gdam_connection_emit_event(GdamConnection* connection,
                                 GdamBuffer*     event)
{
  gdam_buffer_prepend_length (event);
  gdam_buffer_append_buffer (&connection->outgoing, event, TRUE);
  if (connection->fd_poll != NULL)
    gdam_fd_poll_set_events (connection->fd_poll, compute_events (connection));
}

gboolean gdam_connection_ref(GdamConnection* connection, GtkObject* object)
{
  int old_count;
  g_return_val_if_fail (object != NULL, FALSE);
  old_count = (int) (g_hash_table_lookup (connection->ref_table, object));
  gtk_object_ref (object);
  if (old_count == 0)
    gtk_object_sink (object);
  g_hash_table_insert (connection->ref_table,
		       object, GINT_TO_POINTER (old_count + 1));
  gdam_debug
    ("<r>"
     "gdam_connection_ref: %s: new_ref_count=%d; con_ref_count=%d; floating=%d",
     gtk_type_name (object->klass->type), object->ref_count, old_count + 1,
     GTK_OBJECT_FLOATING (object));
  return TRUE;
}

gboolean gdam_connection_unref(GdamConnection* connection, GtkObject* object)
{
  int old_count;
  old_count = (int) (g_hash_table_lookup (connection->ref_table, object));
  if (old_count <= 0)
    {
      gdam_debug ("tried to unref too many times");
      return FALSE;
    }
  gdam_debug ("<r>connection_unref old count = %d object ref = %d",
	      old_count, object->ref_count);
  gtk_object_unref (object);
  g_hash_table_insert (connection->ref_table,
		       object, GINT_TO_POINTER (old_count - 1));
  return TRUE;
}

/*
 *  _                    _   _     _
 * | |    ___   ___ __ _| | (_) __| |
 * | |   / _ \ / __/ _` | | | |/ _` |
 * | |__| (_) | (_| (_| | | | | (_| |
 * |_____\___/ \___\__,_|_| |_|\__,_|
 * 
 */

GdamServerObject*    gdam_connection_get_local          (GdamConnection* conn,
                                                         int        local_id)
{
  return g_hash_table_lookup (conn->local_id_table,
			      GINT_TO_POINTER (local_id));
}
void                 gdam_connection_set_local          (GdamConnection* conn,
                                                         int        local_id,
							 GdamServerObject* obj)
{
  g_return_if_fail (obj->connection == NULL);
  g_return_if_fail (obj->local_id == 0);
  obj->local_id = local_id;
  obj->connection = conn;
  return g_hash_table_insert (conn->local_id_table,
			      GINT_TO_POINTER (local_id), obj);
}

GdasSource*          gdam_connection_get_source         (GdamConnection* conn,
                                                         int        local_id)
{
  GdamServerObject *object;
  g_return_val_if_fail (local_id != 0, NULL);
  if (conn == NULL)
    conn = gdam_connection_get_top ();
  if (local_id < 0)
    {
      if (conn == NULL)
	return NULL;
      object = gdam_connection_get_local (conn, local_id);
      if (object == NULL)
	return NULL;
      gtk_object_ref (GTK_OBJECT (object));
      if (GDAS_IS_SOURCE (object))
	return (GdasSource *) object;
    }
  else
    {
      return gdam_dispatch_get_source (local_id);
    }
  return NULL;
}

/*
 *  ____  _             _
 * / ___|| |_ __ _  ___| | __
 * \___ \| __/ _` |/ __| |/ /
 *  ___) | || (_| | (__|   <
 * |____/ \__\__,_|\___|_|\_\
 * 
 */
static GSList* connection_stack = NULL;
void gdam_connection_push(GdamConnection* connection)
{
  connection_stack = g_slist_prepend (connection_stack, connection);
}
void gdam_connection_pop()
{
  GdamConnection *connection;
  g_return_if_fail (connection_stack != NULL);
  connection = connection_stack->data;
  connection_stack = g_slist_remove (connection_stack, connection);
}
GdamConnection* gdam_connection_get_top()
{
  g_return_val_if_fail (connection_stack != NULL, NULL);
  return connection_stack->data;
}

/*
 *  _   _
 * | | | |___  ___ _ __ _ __   __ _ _ __ ___   ___  ___
 * | | | / __|/ _ \ '__| '_ \ / _` | '_ ` _ \ / _ \/ __|
 * | |_| \__ \  __/ |  | | | | (_| | | | | | |  __/\__ \
 *  \___/|___/\___|_|  |_| |_|\__,_|_| |_| |_|\___||___/
 * 
 */
static GHashTable *by_username = NULL;
gboolean
gdam_connection_set_username (GdamConnection    *connection,
			      const char        *username)
{
  if (by_username == NULL)
    by_username = g_hash_table_new (g_str_hash, g_str_equal);
  if (username != NULL && g_hash_table_lookup (by_username, username) != NULL)
    return FALSE;
  if (connection->username != NULL)
    {
      g_hash_table_remove (by_username, connection->username);
      g_free (connection->username);
    }
  connection->username = g_strdup (username);
  if (username != NULL)
    g_hash_table_insert (by_username, connection->username, connection);
}
GdamConnection *
gdam_connection_get_by_username (const char *username)
{
  if (by_username == NULL)
    return NULL;
  return g_hash_table_lookup (by_username, username);
}

/*
 *  _                  _                   _
 * | |    ___   __ _  | |    _____   _____| |
 * | |   / _ \ / _` | | |   / _ \ \ / / _ \ |
 * | |__| (_) | (_| | | |__|  __/\ V /  __/ |
 * |_____\___/ \__, | |_____\___| \_/ \___|_|
 *             |___/
 * implementing the lists for each flag.
 */
#define MAX_FLAGS		31

static GSList* log_level_lists[MAX_FLAGS] = { 0 };

void  gdam_connection_set_log_level     (GdamConnection* connection,
                                         int         new_level)
{
  int i;
  int old_level = connection->log_level;
  for (i = 0; i < MAX_FLAGS; i++)
    {
      int flag = (1 << i);
      if ((flag & old_level) == (flag & new_level))
	continue;
      if ((flag & new_level) == flag)
	{
	  /* new log level */
	  log_level_lists[i]
	    = g_slist_prepend (log_level_lists[i], connection);
	}
      else
	{
	  log_level_lists[i]
	    = g_slist_remove (log_level_lists[i], connection);
	}
    }
  connection->log_level = new_level;

  /* rebuild the log level flag */
  gdam_connection_total_log_level = 0;
  for (i = 0; i < MAX_FLAGS; i++)
    {
      if (log_level_lists[i] != NULL)
	gdam_connection_total_log_level |= (1 << i);
    }
}

static inline int fast_bit_index(int flag)
{
  int rv = 0;
  g_assert (MAX_FLAGS <= 31);
  if (flag >= (1 << 16))
    {
      flag >>= 16;
      rv += 16;
    }
  if (flag >= (1 << 8))
    {
      flag >>= 8;
      rv += 8;
    }
  if (flag >= (1 << 4))
    {
      flag >>= 4;
      rv += 4;
    }
  if (flag >= (1 << 2))
    {
      flag >>= 2;
      rv += 2;
    }
  if (flag >= (1 << 1))
    {
      flag >>= 1;
      rv += 1;
    }
  return rv;
}

GSList* gdam_connections_for_level_flag   (int         log_flag)
{
  int index = fast_bit_index (log_flag);
  g_return_val_if_fail (index >= 0, NULL);
  g_return_val_if_fail (index < MAX_FLAGS, NULL);
  return log_level_lists[index];
}
