/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

/*
 * Unconfuse the autopluginlister: NOT A PLUGIN.
 */

#define DO_FORCED_READ 0

#include "gdamdevicecoreaudio.h"

#if HAVE_CORE_AUDIO

#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <gdam/server/gdamconfigfile.h>
#include <gdam/common/gdamfdpoll.h>

static void gdam_device_core_audio_init(GdamDeviceCoreAudio* device_core_audio)
{
}

/* we provide this function to service the audio device. 
   output_data should be filled with data before returning. */
static OSStatus 
audio_io_proc (AudioDeviceID device,
               const AudioTimeStamp  *now,
               const AudioBufferList *input_data,
               const AudioTimeStamp  *in_time,
               AudioBufferList       *output_data,
               const AudioTimeStamp  *out_time,
               void                  *client_data)
{
  GdamDeviceCoreAudio *device_core_audio;
  GdamCallbackDevice *callback_device;
  GdamOutputDevice *output_device;
  GdamSink *sink;
  int num_buffers, num_channels;
  int i;

  device_core_audio = (GdamDeviceCoreAudio *) client_data;
  callback_device = (GdamCallbackDevice *) device_core_audio;
  output_device = (GdamOutputDevice *) device_core_audio;
  sink = (GdamSink *) output_device;
  num_buffers = output_data->mNumberBuffers;

  if (!sink->feed) {
    bzero (output_data->mBuffers[0].mData, 
           output_data->mBuffers[0].mDataByteSize);
    return;
  }


  for (i=0; i < num_buffers; i++)
    {
      long samples_out;
      gboolean rv;
      samples_out = output_data->mBuffers[i].mDataByteSize / sizeof(float);
      num_channels = output_data->mBuffers[i].mNumberChannels;

      //fprintf (stderr, "audio_io_proc on %d buffers of %d bytes %d samples %d channels\n", num_buffers, output_data->mBuffers[i].mDataByteSize, samples_out, num_channels);

      //rv = one_frag (output_device, 
      rv = gdam_callback_device_fill_one_frag
                    (GDAM_CALLBACK_DEVICE (output_device),
                     output_data->mBuffers[i].mData, 
		     callback_device->frag_buffer_size,
		     samples_out / num_channels);
      if (!rv) {
	g_warning ("couldn't read one frag");
      }
    }
}
               

static int    
gdam_device_core_audio_write_pcm     (GdamOutputDevice*  device, 
				      gconstpointer      buffer, 
				      int                length)
{
  g_message ("CoreAudio: write_pcm");
  return 0;
}

static gboolean 
gdam_device_core_audio_open_fd (GdamOutputDevice *output_device)
{
  GdamDeviceCoreAudio *device_core_audio;
  GdamCallbackDevice *callback_device;
  UInt32 size, bufsize, latency;
  AudioStreamBasicDescription format;
  AudioTimeStamp cur_when;
  AudioDeviceID our_id;
  AudioDeviceID *device_list;
  AudioDeviceIOProc io_proc;
  int num_devices;
  OSStatus status;
  int pipe_fds[2];
  char s[10];
  gboolean is_writeable;

  device_core_audio = GDAM_DEVICE_CORE_AUDIO (output_device);
  callback_device = GDAM_CALLBACK_DEVICE (output_device);

  g_return_val_if_fail (!callback_device->is_open, TRUE);
  callback_device->is_open = TRUE;

/* Select default device */
  size = sizeof (AudioDeviceID);
  status =  AudioHardwareGetProperty (kAudioHardwarePropertyDefaultOutputDevice, &size, &our_id);

  size = sizeof (format);
  status = AudioDeviceGetProperty (our_id, 0, 0, kAudioDevicePropertyStreamFormat, &size, &format);

  if (status) {
    g_warning ("return code %d from AddDeviceGetProperty", status);
  }


/* Check format */
  if(format.mFormatID != kAudioFormatLinearPCM) {
    g_warning ("formatID %d rather than %d", format.mFormatID, kAudioFormatLinearPCM);
  } else {
    g_message ("CoreAudio: selected linear pcm format rate %f %d channels per frame",
	       format.mSampleRate, format.mChannelsPerFrame);
  }


/* Set block size */
  output_device->options.block_size = 
			(1 << output_device->options.log_frag_size);
  bufsize = output_device->options.block_size
            * output_device->options.bits_per_sample / 8
            * output_device->options.num_channels;

  g_message ("CoreAudio: buffer size of %d bytes from block %d log_frag %d bits/samp %d channels %d",
             bufsize, output_device->options.block_size, 
	     output_device->options.log_frag_size,
             output_device->options.bits_per_sample,
             output_device->options.num_channels);

  size = sizeof (bufsize);
  status = AudioDeviceGetCurrentTime (our_id, &cur_when);
  status = AudioDeviceSetProperty (our_id, &cur_when, 0, 0, kAudioDevicePropertyBufferSize, size, &bufsize);
  if (status) {
    g_warning ("return code %d from SetPropertyBufferSize", status);
  }
  else {
    g_message ("CoreAudio: set buffer size of %d bytes", bufsize);
  }


/* Check block size */
  status = AudioDeviceGetProperty (our_id, 0, 0, kAudioDevicePropertyBufferSize, &size, &bufsize);
  if (status) {
    g_warning ("return code %d from GetPropertyBufferSize", status);
  }
  else {
    g_message ("CoreAudio: got buffer size of %d bytes", bufsize);
  }

#if 0
/* Check / set latency */
  status = AudioDeviceGetPropertyInfo (our_id, 0, FALSE, 
           kAudioDevicePropertyStreamLatency, NULL, &is_writeable);
  if (status == kAudioHardwareNoError) 
    {
      if (is_writeable)
	{
	}
    }

  status = AudioDeviceGetProperty (our_id, 0, 0, kAudioDevicePropertyStreamLatency, &size, &latency);
  if (status) {
    g_warning ("return code %d from GetPropertyLatency", status);
  }
  else {
    g_message ("CoreAudio: got latency %d", latency);
  }
#endif

  gdam_callback_device_add_ipc_poll (callback_device);

/* Attach a process to serve our device */
  io_proc = audio_io_proc;
  status = AudioDeviceAddIOProc (our_id, io_proc, device_core_audio);
  if (status) {
    g_warning ("return code %d from AddIOProc", status);
  }
  else {
    g_message ("CoreAudio: added ioproc to audio device");
  }

/* And start it running */
  status = AudioDeviceStart (our_id, io_proc);
  if (status) {
    g_warning ("return code %d from AudioDeviceStart", status);
  }
  else {
    g_message ("CoreAudio: started audio device");
  }

/* XXX: free these buffer (and the semaphores) on destroy */
  callback_device->frag_buffer_size = bufsize;
  callback_device->temp_buffer = (GdamSample *)malloc (bufsize);
  callback_device->fragment_buffer_out = (GdamSample*)malloc (bufsize);
  return TRUE;
}

static void 
gdam_device_core_audio_class_init (GdamDeviceCoreAudioClass *core_audio_class)
{
  GdamOutputDeviceClass* klass;
  klass = GDAM_OUTPUT_DEVICE_CLASS(core_audio_class);
  klass->open_fd = gdam_device_core_audio_open_fd;
  //klass->write_pcm = gdam_device_core_audio_write_pcm;
}


GtkType gdam_device_core_audio_get_type()
{
	static GtkType device_core_audio_type = 0;
	if (!device_core_audio_type) {
		static const GtkTypeInfo device_core_audio_info = {
			"GdamDeviceCoreAudio",
			sizeof(GdamDeviceCoreAudio),
			sizeof(GdamDeviceCoreAudioClass),
			(GtkClassInitFunc) gdam_device_core_audio_class_init,
			(GtkObjectInitFunc) gdam_device_core_audio_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_CALLBACK_DEVICE;
		device_core_audio_type = gtk_type_unique(parent,
		                                 &device_core_audio_info);
	}
	return device_core_audio_type;
}

/* 
 * This function is mandatory for plugin devices.
 */
GdamSink* gdam_device_core_audio_new(gpointer      device_data,
                              const char*   device_name,
			      const char*   device_options)
{
	GtkType type = gdam_device_core_audio_get_type();
	GdamDeviceCoreAudio* dev_core_audio = (GdamDeviceCoreAudio*)gtk_type_new(type);
	if (!gdam_sink_open(GDAM_SINK(dev_core_audio), device_name, device_options)) {
		gtk_object_unref(GTK_OBJECT(dev_core_audio));
		return NULL;
	}
	return GDAM_SINK(dev_core_audio);
}


#endif /* HAVE_CORE_AUDIO */
