#include "gdamrealtime.h"
#include "gdamconfigfile.h"
#include <gdam/common/gdamautoconf.h>
#include <sched.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <errno.h>
#include <glib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>


#if HAVE_MLOCKALL
static void
do_mlock (int kilobytes)
{
  int i;
  char** reserved_space;
  reserved_space = g_malloc (sizeof (char*) * kilobytes);

  /* For each kilobyte reserved, alloc a new block and touch it. */
  for (i = 0; i < kilobytes; i++)
    {
      char* new_block;
      reserved_space[i] = g_malloc (sizeof (char) * 1024);
      new_block = reserved_space[i];
      new_block[0] = 0;
      new_block[512] = 0;
    }

  mlockall(MCL_CURRENT);
  g_message ("locked current ram pages plus %dk extra", kilobytes);

  /* Free each kilobyte reserved. */
  for (i = 0; i < kilobytes; i++)
    {
      char* new_block;
      g_free (reserved_space[i]);
      reserved_space[i] = 0;
    }
  g_free (reserved_space);
}
#else
#define do_mlock(extra_kilobytes)
#endif

#ifdef  HAVE_SYS_PRIOCNTL_H

/* XXX: untested */
#include <sys/priocntl.h>
#include <sys/rtpriocntl.h>


static void
switch_to_realtime_priority()
{
  pcinfo_t info;
  pcparms_t param;
  rtinfo_t rtinfo;
  rtparms_t rtparam;
  int pid;

  pid = getpid ();

  /* get info */
  strcpy (info.pc_clname, "RT");
  if (priocntl (P_PID, pid, PC_GETCID, (void *) &info) < 0)
    {
      g_warning ("Cannot get priority class id priocntl(PC_GETCID): %s",
		 g_strerror (errno));
    }

  movebytes (info.pc_clinfo, &rtinfo, sizeof (rtinfo_t));

  /* set priority not to the max */
  rtparam.rt_pri = rtinfo.rt_maxpri - 2;
  rtparam.rt_tqsecs = 0;
  rtparam.rt_tqnsecs = RT_TQDEF;
  param.pc_cid = info.pc_cid;
  movebytes (&rtparam, param.pc_clparms, sizeof (rtparms_t));

  if (priocntl (P_PID, pid, PC_SETPARMS, (void *) &param) < 0)
    {
      if (errno == EPERM && getuid () != 0 && geteuid () != 0)
	{
	  g_message ("Since you were not root, "
		     "could not priocntl to hi-priority");
	}
      else
	{
	  g_warning ("priocntl failed: %s", g_strerror (errno));
	}
    }
}
#endif

#if      defined _POSIX_PRIORITY_SCHEDULING

#include <sched.h>

static void
switch_to_realtime_priority()
{
  int sched_fifo_min, sched_fifo_max;
  int rv;
  int pid;
  struct sched_param sched_parms;

#ifdef  _SC_PRIORITY_SCHEDULING
  if (sysconf (_SC_PRIORITY_SCHEDULING) == -1)
    {
      g_warning ("WARNING: RR-scheduler not available, disabling.\n");
      return;
    }
#endif

#if HAVE_SCHED_SETSCHEDULER
  pid = getpid();
  memset (&sched_parms, 0, sizeof (sched_parms));
  sched_fifo_min = sched_get_priority_min (SCHED_FIFO);
  sched_fifo_max = sched_get_priority_max (SCHED_FIFO);
  sched_parms.sched_priority = sched_fifo_max;
  rv = sched_setscheduler (pid, SCHED_FIFO, &sched_parms);
  if (rv < 0)
    {
      if (errno == EPERM && getuid () != 0 && geteuid () != 0)
	g_message ("Since you were not root, "
		   "we could not enter posix realtime priority");
      else
	g_warning ("sched_setscheduler error: %s", strerror (errno));
      return;
    }
  g_message ("set posix \"realtime\" mode for pid %d", pid);
#endif /* HAVE_SCHED_SETSCHEDULER */

  do_mlock (gdam_mlock_kilos);
}

#elif HAVE_PTHREAD_SETSCHEDPARAM

static void
switch_to_realtime_priority()
{
  struct sched_param info;
  int policytype;

  /* use 'nice' to get preference over competing processes */
  setpriority (0, 0, -20);

  if (pthread_getschedparam (pthread_self(), &policytype, &info) == 0)
    {
      info.sched_priority = sched_get_priority_max (policytype);
      pthread_setschedparam (pthread_self (), policytype, &info);
      policytype = SCHED_FIFO;
      info.sched_priority = sched_get_priority_max (SCHED_FIFO);
      pthread_setschedparam (pthread_self (), SCHED_FIFO, &info);
    }
  fprintf (stderr, "NI: %d\n", getpriority(0,0));
  fprintf (stderr, "PRI: %d\n", info.sched_priority);
  fprintf (stderr, "POLI: %d\n", policytype);
}

#else

static void
switch_to_realtime_priority()
{
  g_warning ("Cannot switch to realtime priority, not compiled in.");
}

#endif



void gdam_realtime_set()
{
  switch_to_realtime_priority ();
  do_mlock (gdam_mlock_kilos);
}

void gdam_realtime_unset()
{
#if HAVE_SCHED_SETSCHEDULER
  struct sched_param param;
  int pid = getpid ();
  param.sched_priority = 0;
  sched_setscheduler (pid, SCHED_OTHER, &param);
#endif
/* XXX: non-setsched implementation */
}

