#include "gdassequencer2.h"
#include "gdamcachebuffer.h"

struct _GdasSequencer2Element
{
  /* Index of the first sample in the loop to fill with this
   * the first sample of this element. */
  int                     start_offset;

  /* Speedup/slowdown */
  GdamFixed               rate;

  /* Revolumize */
  GdamFixed               gain;


  /* First sample to output. */
  int                     in_buffer_offset;

  /* Number of samples to output with this element. */
  GdamFixed               sample_output_length;

  /* The buffer this sample comes from. */
  GdamCacheBuffer        *buffer;

  /* Internal linked-list of elements in this sequence */
  GdasSequencer2Element  *next;
};

static GMemChunk *element_chunk = NULL;
static GdasSequencer2Element *gdas_sequencer2_element_alloc ()
{
  if (element_chunk == NULL)
    element_chunk = g_mem_chunk_create (GdasSequencer2Element, 16,
				        G_ALLOC_AND_FREE);
  return g_chunk_new (GdasSequencer2Element, element_chunk);
}
static void gdas_sequencer2_element_free (GdasSequencer2Element *element)
{
  g_chunk_free (element, element_chunk);
}

static int         gdas_sequencer2_read         (GdasSource*    input,
	                                         GdamSample*    samples,
						 int            max_read);
static void        gdas_sequencer2_miss_data    (GdasSource*    input,
	                                         int            samples_lost);
static gboolean    gdas_sequencer2_seek         (GdasSource*    source,
                                                 GdamTime       offset,
						 int            whence);
static GdamTime    gdas_sequencer2_get_position (GdasSource*    source);
static void        gdas_sequencer2_destroy      (GtkObject*     object);

static GtkObjectClass *parent_class = NULL;

static void gdas_sequencer2_init (GdasSequencer2* sequencer2)
{
  sequencer2->map = gdam_interval_map_new ();
  sequencer2->sequence_length = 44100;
}

/*
 *            _     _
 *   __ _  __| | __| |
 *  / _` |/ _` |/ _` |
 * | (_| | (_| | (_| |
 *  \__,_|\__,_|\__,_|
 * (Part of the public interface).
 */
gboolean gdas_sequencer2_add       (GdasSequencer2      *sequencer,
                                    int                  buffer_id,
				    int                  start_offset,
				    GdamFixed            rate,
				    GdamFixed            gain,
				    int                  offset,
				    int                  output_length)
{
  GdasSequencer2Element *element;
  GdamCacheBuffer *buffer;
  int max_sample_output_length;
  int last_sample;
  buffer = gdam_cache_buffer_get_by_id (buffer_id);

  if (buffer == NULL)
    return FALSE;
  if (rate == 0)
    return FALSE;

  element = gdas_sequencer2_element_alloc ();
  element->start_offset = start_offset;
  element->buffer = buffer;
  element->rate = rate;
  element->gain = gain;
  element->in_buffer_offset = offset;
  element->sample_output_length = output_length;
  if (rate > 0)
    {
      int max_sample_input_length = buffer->buffer_length - offset;
      max_sample_output_length = (gint64) max_sample_input_length 
                               * GdamFixed_ONE
			       / rate;
    }
  else
    {
      int max_sample_input_length = offset;
      max_sample_output_length = (gint64) max_sample_input_length 
                               * GdamFixed_ONE
			       / (-rate);
   }

  if (element->sample_output_length < 0
   || element->sample_output_length > max_sample_output_length)
    {
      element->sample_output_length = max_sample_output_length;
    }
  if (element->sample_output_length < 0)
    {
      gdas_sequencer2_element_free (element);
      return FALSE;
    }
  if (element->sample_output_length > sequencer->sequence_length)
    element->sample_output_length = sequencer->sequence_length;
  
  gtk_object_ref (GTK_OBJECT (buffer));

  /* add to interval map */
  last_sample = element->start_offset + element->sample_output_length - 1;
  if (last_sample >= sequencer->sequence_length)
    {
      gdam_interval_map_add (sequencer->map, element->start_offset,
			     sequencer->sequence_length - 1, element);
      gdam_interval_map_add (sequencer->map, 0,
			     last_sample - sequencer->sequence_length,
			     element);
    }
  else
    {
      gdam_interval_map_add (sequencer->map, element->start_offset,
			     last_sample, element);
    }

  /* add to the element list */
  element->next = sequencer->elements;
  sequencer->elements = element;

  return TRUE;
}

/*
 *       _
 *   ___| | ___  __ _ _ __
 *  / __| |/ _ \/ _` | '__|
 * | (__| |  __/ (_| | |
 *  \___|_|\___|\__,_|_|
 * (Part of the public interface).
 */
void     gdas_sequencer2_clear     (GdasSequencer2      *sequencer)
{
  GdasSequencer2Element *element;

  gdam_interval_map_clear (sequencer->map);
  element = sequencer->elements;
  while (element != NULL)
    {
      GdasSequencer2Element *next = element->next;
      gdas_sequencer2_element_free (element);
      element = next;
    }
  sequencer->elements = NULL;
  sequencer->num_elements = 0;
}

/*
 *                     _
 *  _ __ ___  __ _  __| |
 * | '__/ _ \/ _` |/ _` |
 * | | |  __/ (_| | (_| |
 * |_|  \___|\__,_|\__,_|
 * (GdasSource virtual)
 */

static void       gdas_sequencer2_element_fill (GdasSequencer2Element *element,
                                                int                    position,
						int                    length,
						int                    seq_len,
						GdamSample            *output,
						gboolean               add)
{
  int start_index;
  guint32 start_frac;
  gint64 start;

  if (position < element->start_offset)
    position += seq_len;

  start = ((gint64) element->in_buffer_offset << 32)
        + (((gint64) element->rate * (position - element->start_offset))
	                         << (32 - GdamFixed_LOG_ONE));
  start_frac = (guint32) start;
  start_index = start >> 32;

  gdam_cache_buffer_read (element->buffer,
                          output,
			  start_index,
			  start_frac,
			  NULL, NULL,
			  length,
			  element->rate,
			  element->gain,
			  add);
}

static void       element_list_fill      (GSList     *elements,
                                          int         position,
					  int         length,
					  int         sequence_length,
					  GdamSample *output)
{
  gdam_sample_zero (output, length);
  if (elements == NULL)
    {
      return;
    }

  while (elements != NULL)
    {
      GdasSequencer2Element *element = elements->data;
      gdas_sequencer2_element_fill (element, position, length, sequence_length,
                                    output, TRUE);
      elements = elements->next;
    }
}

static int gdas_sequencer2_read (GdasSource     *source,
                                 GdamSample     *buffer,
				 int             max_read)
{
  GdasSequencer2 *sequencer = GDAS_SEQUENCER2 (source);
  GdamIntervalMapPost *post = NULL;
  int position = sequencer->position;
  int rv = max_read;
  int seq_length = sequencer->sequence_length;
  while (max_read > 0)
    {
      int chunk_size;
      GdamIntervalMapPost *next;
      GSList *objects;
      if (post == NULL)
        post = gdam_interval_map_lookup (sequencer->map, position);
      if (post == NULL)
        next = gdam_interval_map_first (sequencer->map);
      else
        next = post->next;

      chunk_size = max_read;
      if (chunk_size > sequencer->sequence_length - position)
	chunk_size = sequencer->sequence_length - position;
      if (next != NULL && chunk_size > next->value - position)
	chunk_size = next->value - position;

      g_return_val_if_fail (chunk_size >= 0, -1);
      if (post != NULL)
        g_assert (post->value <= position);

      objects = (post != NULL) ? post->objects : NULL;
      element_list_fill (objects, position, chunk_size, seq_length, buffer);
      buffer += chunk_size;
      max_read -= chunk_size;
      position += chunk_size;
      if (position >= seq_length)
	{
          position = 0;
	  next = NULL;
	}
      post = next;
    }

  sequencer->position = position;
  return rv;
}

/*
 *                _
 *  ___  ___  ___| | __
 * / __|/ _ \/ _ \ |/ /
 * \__ \  __/  __/   <
 * |___/\___|\___|_|\_\
 * 
 */


static GdamTime gdas_sequencer2_get_position(GdasSource* source)
{
  return GDAS_SEQUENCER2 (source)->position;
}

static gboolean    gdas_sequencer2_seek         (GdasSource*    source,
                                                 GdamTime       offset,
						 int            whence)
{
  GdasSequencer2 *sequencer = GDAS_SEQUENCER2 (source);
  int sequence_length = sequencer->sequence_length;
  if (sequence_length <= 0)
    return FALSE;
  if (whence == 1)
    offset += sequencer->position;
  else if (whence == 2)
    offset += sequence_length;
  offset %= sequence_length;
  while (offset < 0)
    offset += sequence_length;
  sequencer->position = offset;
  return TRUE;
}

static void      gdas_sequencer2_miss_data     (GdasSource*      input,
	                                        int              samples_lost)
{
  GdasSequencer2 *sequencer = GDAS_SEQUENCER2 (input);
  sequencer->position += samples_lost;
  sequencer->position %= sequencer->sequence_length;
}
static void        gdas_sequencer2_destroy      (GtkObject*     object)
{
  gdas_sequencer2_clear (GDAS_SEQUENCER2 (object));
  (*parent_class->destroy) (object);
}

/*
 *           _      __         _          _                           _
 *  ___  ___| |_   / /_ _  ___| |_    ___| | ___ _ __ ___   ___ _ __ | |_ ___
 * / __|/ _ \ __| / / _` |/ _ \ __|  / _ \ |/ _ \ '_ ` _ \ / _ \ '_ \| __/ __|
 * \__ \  __/ |_ / / (_| |  __/ |_  |  __/ |  __/ | | | | |  __/ | | | |_\__ \
 * |___/\___|\__/_/ \__, |\___|\__|  \___|_|\___|_| |_| |_|\___|_| |_|\__|___/
 *                  |___/
 */
static void gdas_sequencer2_set_elements (GtkObject       *object,
                                          GdamArg         *arg)
{
  int num_el = GDAM_ARG_INT_ARRAY (arg).length;
  int *elements = GDAM_ARG_INT_ARRAY (arg).array;
  int i;
  GdasSequencer2 *sequencer = GDAS_SEQUENCER2 (object);
  if (num_el % 6 != 0)
    {
      gdam_debug ("gdas_sequencer_set_elements, num ints was not "
		  "a multiple of 6");
    }
  num_el /= 6;

  gdas_sequencer2_clear (sequencer);
  for (i = 0; i < num_el; i++)
    {
      if (!gdas_sequencer2_add (sequencer,
				elements[i * 6 + 0],
				elements[i * 6 + 1],
				elements[i * 6 + 2],
				elements[i * 6 + 3],
				elements[i * 6 + 4],
				elements[i * 6 + 5]))
	{
	  gdam_debug ("error creating sequencer subelement %d", i);
	}
    }
}

static void gdas_sequencer2_get_elements(GtkObject* object, GdamArg* arg)
{
  GdasSequencer2 *sequencer = GDAS_SEQUENCER2 (object);
  int *elements = g_new (int, sequencer->num_elements * 6);
  int *rv = elements;
  GdasSequencer2Element *element;
  for (element = sequencer->elements; element != NULL; element = element->next)
    {
      *rv++ = gdam_cache_buffer_get_id (element->buffer);
      *rv++ = element->start_offset;
      *rv++ = element->rate;
      *rv++ = element->gain;
      *rv++ = element->in_buffer_offset;
      *rv++ = element->sample_output_length;
    }
  GDAM_ARG_INT_ARRAY (arg).length = rv - elements;
  GDAM_ARG_INT_ARRAY (arg).array = elements;
}

/*
 *           _      __         _     _                  _   _
 *  ___  ___| |_   / /_ _  ___| |_  | | ___ _ __   __ _| |_| |__
 * / __|/ _ \ __| / / _` |/ _ \ __| | |/ _ \ '_ \ / _` | __| '_ \
 * \__ \  __/ |_ / / (_| |  __/ |_  | |  __/ | | | (_| | |_| | | |
 * |___/\___|\__/_/ \__, |\___|\__| |_|\___|_| |_|\__, |\__|_| |_|
 *                  |___/                         |___/
 */
static void gdas_sequencer2_get_length(GtkObject* object, GdamArg* arg)
{
  GdasSequencer2 *sequencer = GDAS_SEQUENCER2 (object);
  GDAM_ARG_INT (arg) = sequencer->sequence_length;
}
static void gdas_sequencer2_set_length(GtkObject* object, GdamArg* arg)
{
  int sequence_length = GDAM_ARG_INT (arg);
  GdasSequencer2 *sequencer = GDAS_SEQUENCER2 (object);
  if (sequence_length <= 0 || sequence_length > 1000 * 1000 * 1000)
    {
      gdam_debug ("sequence_length ridiculous... sorry");
      return;
    }
  sequencer->sequence_length = sequence_length;
}

static void 
gdas_sequencer2_set_scale_position (GtkObject *object, GdamArg *arg)
{
  GdasSequencer2 *sequencer = GDAS_SEQUENCER2 (object);
  sequencer->position *= GDAM_ARG_DOUBLE (arg);
}

static void gdas_sequencer2_class_init (GdasSourceClass* source_class)
{
  GtkObjectClass *object_class = (GtkObjectClass *) source_class;
  source_class->read = gdas_sequencer2_read;
  source_class->miss_data = gdas_sequencer2_miss_data;
  source_class->get_position = gdas_sequencer2_get_position;
  source_class->seek = gdas_sequencer2_seek;
  object_class->destroy = gdas_sequencer2_destroy;
  gdam_arg_register (object_class,
		     "elements",
		     GDAM_ARG_TYPE_INT_ARRAY,
		     gdas_sequencer2_set_elements,
		     gdas_sequencer2_get_elements, NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "scale_position",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_sequencer2_set_scale_position,
		     NULL, NULL, "0.0,", 0);
  gdam_arg_register (object_class,
		     "sequence_length",
		     GDAM_ARG_TYPE_INT,
		     gdas_sequencer2_set_length,
		     gdas_sequencer2_get_length, NULL, "0,", GDAM_ARG_VITAL);
}

GtkType gdas_sequencer2_get_type()
{
  static GtkType sequencer2_type = 0;
  if (!sequencer2_type) {
    static const GtkTypeInfo sequencer2_info =
    {
      "GdasSequencer2",
      sizeof(GdasSequencer2),
      sizeof(GdasSequencer2Class),
      (GtkClassInitFunc) gdas_sequencer2_class_init,
      (GtkObjectInitFunc) gdas_sequencer2_init,
      /* reserved_1 */ NULL,
      /* reserved_2 */ NULL,
      (GtkClassInitFunc) NULL
    };
    GtkType parent = GDAS_TYPE_SOURCE;
    sequencer2_type = gtk_type_unique (parent, &sequencer2_info);
    parent_class = gtk_type_class (parent);
  }
  return sequencer2_type;
}
