/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdasbender.h"
#include <gdam/common/gdamplugin.h>
#include <gdam/common/alloca.h>

#define MIN_BUFFERING	(44100 * 2)

/* XXX: we could probably bear much larger numbers,
 *      but they really are of questionable value... */
#define MIN_RATE	(-1000 << 16)
#define MAX_RATE	(+1000 << 16)

static void        gdas_bender_miss_data        (GdasSource*      input,
					         int              lost);
static gboolean    gdas_bender_is_ready_to_read (GdasSource*      source);
static int	   gdas_bender_read             (GdasSource*      input,
					         GdamSample*      samples,
					         int              max_read);
static GdamTime    gdas_bender_get_position     (GdasSource* source);
static void        gdas_bender_get_latency_info (GdasSource*      source,
                                                 GdasSource*      subsource,
                                                 GdamLatencyInfo* latency);

static void remove_change(GdasBender* bender);
static void remove_new_position_changes(GdasBender* bender);
static void gdas_bender_init(GdasBender* bender)
{
  int i;
  bender->rate = 1 << 16;
  bender->fade_window = 64;
  for (i=0; i < GDAS_BENDER_NUM_POSITION_MARKERS; i++)
    bender->position_markers[i] = -1;
}
static void append_changes_to_list(GdasBenderChange** p_last, 
                                   int                num_changes,
				   int*               time_rates)
{
  int i;
  for (i = 0; i < num_changes; i++)
    {
      GdasBenderChange *change;

      change = g_new (GdasBenderChange, 1);
      *p_last = change;
      p_last = &change->next;
      change->next = NULL;
      change->num_samples = time_rates[i * 2 + 0];
      if (change->num_samples <= 0)
	{
	  change->new_rate = time_rates[i * 2 + 1];
	  change->is_jump = TRUE;
	}
      else
	{
	  int rate;
	  rate = time_rates[i * 2 + 1] >> (GdamFixed_LOG_ONE - 16);
	  change->new_rate = CLAMP (rate, MIN_RATE, MAX_RATE);
	  change->is_jump = FALSE;
	}
    }
}
static void gdas_bender_arg_set_changes(GtkObject* object, GdamArg* arg)
{
  GdasBender *bender = GDAS_BENDER (object);
  int num_changes = GDAM_ARG_INT_ARRAY (arg).length / 2;
  int *time_rates = GDAM_ARG_INT_ARRAY (arg).array;
  while (bender->changes)
    remove_change (bender);
  append_changes_to_list (&bender->changes, num_changes, time_rates);

  /* If we are in one of the time-tracking modes, then setting the 
   * changes resets the mark to the current playback position. */
  if (bender->change_mode != GDAS_BENDER_CHANGE_KEEP_TIME
      && bender->change_mode != GDAS_BENDER_CHANGE_KEEP_IN_BEAT
      && bender->change_mode != GDAS_BENDER_CHANGE_SNAP_TIME)
    /* set mark to current position.  */
    bender->end_offset = bender->backbuffer_offset;
  remove_new_position_changes (bender);
}
static void gdas_bender_arg_append_changes(GtkObject* object, GdamArg* arg)
{
  int num_changes = GDAM_ARG_INT_ARRAY (arg).length / 2;
  int *time_rates = GDAM_ARG_INT_ARRAY (arg).array;
  GdasBender *bender = GDAS_BENDER (object);
  GdasBenderChange **p_last = &bender->changes;
  gboolean was_empty = (bender->changes == NULL);
  while (*p_last != NULL)
    p_last = &((*p_last)->next);
  append_changes_to_list (p_last, num_changes, time_rates);
  /* If we didn't have any previous changes, then we have to make sure
   * we don't leave position changes at the front of the list */
  if (was_empty)
    {
      /* If we are in one of the time-tracking modes, then setting the 
       * changes from NULL resets the mark to the current playback position. */
      if (bender->change_mode != GDAS_BENDER_CHANGE_KEEP_TIME
          && bender->change_mode != GDAS_BENDER_CHANGE_KEEP_IN_BEAT
	  && bender->change_mode != GDAS_BENDER_CHANGE_SNAP_TIME)
        /* set mark to current position.  */
	bender->end_offset = bender->backbuffer_offset;
      remove_new_position_changes (bender);
    }
}
static void gdas_bender_arg_get_changes(GtkObject* object, GdamArg* arg)
{
  int i, num_changes = 0;
  GdasBenderChange *change;
  GdasBender *bender = GDAS_BENDER (object);
  int *array;
  for (change = bender->changes; change != NULL; change = change->next)
    {
      num_changes++;
    }
  GDAM_ARG_INT_ARRAY (arg).length = num_changes * 2;
  GDAM_ARG_INT_ARRAY (arg).array = array = g_new (int, num_changes * 2);
  i = 0;
  for (change = bender->changes; change != NULL; change = change->next)
    {
      if (change->is_jump)
	{
	  array[i++] = 0;
	  array[i++] = change->new_rate;
	}
      else
	{
	  array[i++] = change->num_samples;
	  array[i++] = change->new_rate << (GdamFixed_LOG_ONE - 16);
	}
    }
}
static void gdas_bender_arg_set_change_mode(GtkObject* object, GdamArg* arg)
{
  const char *string = GDAM_ARG_STRING (arg);
  GdasBender *bender = GDAS_BENDER (object);
  if (strcmp (string, "continuous") == 0)
    {
      bender->change_mode = GDAS_BENDER_CHANGE_CONTINUOUS;
    }
  else if (strcmp (string, "keep_time") == 0)
    {
      bender->change_mode = GDAS_BENDER_CHANGE_KEEP_TIME;
      /* set mark to current position.  */
      bender->end_offset = bender->backbuffer_offset;
    }
  else if (strcmp (string, "keep_in_beat") == 0)
    {
      bender->change_mode = GDAS_BENDER_CHANGE_KEEP_IN_BEAT;
      /* set mark to current position.  */
      bender->end_offset = bender->backbuffer_offset;
    }
  else if (strcmp (string, "snap_time") == 0)
    {
      bender->change_mode = GDAS_BENDER_CHANGE_SNAP_TIME;
      /* set mark to current position.  */
      bender->end_offset = bender->backbuffer_offset;
    }
  else if (strcmp (string, "return") == 0)
    {
      bender->change_mode = GDAS_BENDER_CHANGE_RETURN;
    }
  else
    {
      gdam_debug ("gdas_bender no such change_mode: %s", string);
    }
}

static void gdas_bender_arg_get_change_mode(GtkObject* object, GdamArg* arg)
{
  const char *string = NULL;
  GdasBender *bender = GDAS_BENDER (object);
  switch (bender->change_mode)
    {
    case GDAS_BENDER_CHANGE_CONTINUOUS:
      string = "continuous";
      break;
    case GDAS_BENDER_CHANGE_KEEP_TIME:
      string = "keep_time";
      break;
    case GDAS_BENDER_CHANGE_KEEP_IN_BEAT:
      string = "keep_in_beat";
      break;
    case GDAS_BENDER_CHANGE_SNAP_TIME:
      string = "snap_time";
      break;
    case GDAS_BENDER_CHANGE_RETURN:
      string = "return";
      break;
    default:
      g_warning ("change_mode was invalid in bender");
      return;
    }
  GDAM_ARG_STRING (arg) = g_strdup (string);
}

/* Jumps to mark modulo block_size samples (block_size is the arg value) */
/* TODO: use bender->block_size?? */
static void 
gdas_bender_arg_jump_to_nearest_mark (GtkObject* object, GdamArg* arg)
{
  GdasBender *bender = GDAS_BENDER (object);
  gint64 block_size = (gint64) GDAM_ARG_INT (arg) << 16;
  gint64 distance_to_mark = bender->end_offset - bender->backbuffer_offset;
  while (distance_to_mark > block_size)
   distance_to_mark -= block_size;
  while (distance_to_mark < -1 * block_size)
   distance_to_mark += block_size;
  bender->backbuffer_offset += distance_to_mark;
  bender->fade_remaining = bender->fade_window;
}

/* beat_length is the number of samples in a block when keeping in beat
 * and calculating the nearest mark. */
static void 
gdas_bender_arg_beat_length (GtkObject* object, GdamArg* arg)
{
  GdasBender *bender = GDAS_BENDER (object);
  bender->beat_size = (gint64) GDAM_ARG_INT (arg) << 16;
}

/* Jump our current position to the mark point + offset
 * (offset is the arg value) */
static void gdas_bender_arg_jump_to_mark(GtkObject* object, GdamArg* arg)
{
  GdasBender *bender = GDAS_BENDER (object);
  gint64 offset = (gint64) GDAM_ARG_INT (arg) << 16;
  if (bender->change_mode == GDAS_BENDER_CHANGE_CONTINUOUS)
    return;
  bender->backbuffer_offset = bender->end_offset + offset;
  bender->fade_remaining = bender->fade_window;
}

/* set the mark to our current position + offset (offset is the arg value) */
static void gdas_bender_arg_set_mark(GtkObject* object, GdamArg* arg)
{
  GdasBender *bender = GDAS_BENDER (object);
  gint64 offset = (gint64) GDAM_ARG_INT (arg) << 16;
  bender->end_offset = bender->backbuffer_offset + offset;
}

static void 
gdas_bender_arg_set_position_marker (GtkObject *object, GdamArg *arg)
{
  GdasBender *bender = GDAS_BENDER (object);
  int idx =  GDAM_ARG_INT (arg);
  bender->position_markers[idx] = bender->backbuffer_offset;
}

static void
gdas_bender_arg_jump_to_position_marker (GtkObject *object, GdamArg *arg)
{
  GdasBender *bender = GDAS_BENDER (object);
  int idx =  GDAM_ARG_INT (arg);
  if (bender->change_mode == GDAS_BENDER_CHANGE_CONTINUOUS)
    return;
  if (bender->position_markers[idx] < 0)
    return;
  bender->backbuffer_offset = bender->position_markers[idx];
  bender->fade_remaining = bender->fade_window;
}

static void gdas_bender_arg_set_rate(GtkObject* object, GdamArg* arg)
{
  GDAS_BENDER (object)->rate = GDAM_ARG_INT (arg);
}
static void gdas_bender_arg_get_rate(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAS_BENDER (object)->rate;
}
static void gdas_bender_arg_set_readahead(GtkObject* object, GdamArg* arg)
{
  GDAS_BENDER (object)->readahead = GDAM_ARG_INT (arg);
}
static void gdas_bender_arg_get_readahead(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAS_BENDER (object)->readahead;
}
static void gdas_bender_arg_set_buffer_length(GtkObject* object, GdamArg* arg)
{
  gdas_buffering_filter_ensure_space (GDAS_BUFFERING_FILTER (object),
				      GDAM_ARG_INT (arg));
}
static void gdas_bender_arg_get_buffer_length(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAS_BUFFERING_FILTER (object)->buffer_length;
}

static void gdas_bender_arg_set_fade_window(GtkObject* object, GdamArg* arg)
{
  GdasBender *bender = GDAS_BENDER (object);
  int new_window = GDAM_ARG_INT (arg);
  if (bender->fade_window == 0)
    bender->fade_remaining = 0;
  else
    bender->fade_remaining = bender->fade_remaining
      * new_window / bender->fade_window;
  bender->fade_window = new_window;
}

static void gdas_bender_arg_get_fade_window(GtkObject* object, GdamArg* arg)
{
  GdasBender *bender = GDAS_BENDER (object);
  GDAM_ARG_INT (arg) = bender->fade_window;
}

#if 0
static void set_relative_smoothed_position(GdasBender* bender, int relpos)
{
  int projected;
  if (bender->last_smoothed_length <= 2)
    projected = 4096;
  else
    projected = bender->last_smoothed_length;
  bender->last_smoothed_length = 0;

  /*
   * We model the situation by the differential equation:
   *
   *       v'(t) = (-drag) * v(t) + k
   *
   * where `k' is the accelation we are going to apply.
   *
   * It has the following boundary conditions,
   *
   *      v(0)       = v0
   *      v(endtime) = endpos
   *
   */
  /* The solution for the first-order DE is taken from a book... */
  c2 = (gdouble) relpos + rate * projected
    / (1.0 + drag * relpos - exp (-drag * relpos));
  k = c2 * drag * drag - rate * drag;

  /* Then position and velocity are given:

   *      p(t) = -c2 * exp(-drag*t) + k*t/drag + c2
   *      v(t) = -c2 * drag * exp(-drag*t) + k*t/drag + c2
   */

...}

static void gdas_bender_arg_set_smoothed_position(GtkObject* object,
                                                  GdamArg* arg)
{
  GdasBender *bender = GDAS_BENDER (object);
  gint64 position = GDAM_ARG_INT64 (arg);
  GdasSource *filtered = GDAS_FILTER (object)->filtered;
  gint64 our_pos;
  int rel_pos;
  if (filtered == NULL)
    return;
  our_pos = gdas_source_get_position (filtered) - bender->readahead;
  rel_pos = (int) (our_pos - position);

  set_relative_smoothed_position (bender, rel_pos);
}
#endif

static void gdas_bender_class_init(GdasBenderClass* bender_class)
{
  GdasSourceClass *source_class = GDAS_SOURCE_CLASS (bender_class);
  /*if (!parent_class)
     parent_class = gtk_type_class(GDAS_TYPE_BUFFERING_FILTER); */
  source_class->read = gdas_bender_read;
  source_class->miss_data = gdas_bender_miss_data;
  source_class->is_ready_to_read = gdas_bender_is_ready_to_read;
  source_class->get_latency_info = gdas_bender_get_latency_info;
  source_class->get_position = gdas_bender_get_position;
  gdam_arg_register (GTK_OBJECT_CLASS (bender_class), "changes", GDAM_ARG_TYPE_INT_ARRAY, gdas_bender_arg_set_changes, gdas_bender_arg_get_changes, NULL, NULL,	/* XXX: needs validator! */
		     GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (bender_class), "append_changes", GDAM_ARG_TYPE_INT_ARRAY, gdas_bender_arg_append_changes, NULL, NULL, NULL,	/* XXX: needs validator! */
		     0);
  gdam_arg_register (GTK_OBJECT_CLASS (bender_class), "change_mode", GDAM_ARG_TYPE_STRING, gdas_bender_arg_set_change_mode, gdas_bender_arg_get_change_mode, NULL, NULL,	/* XXX: needs validator! */
		     GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (bender_class), "rate", GDAM_ARG_TYPE_INT, gdas_bender_arg_set_rate, gdas_bender_arg_get_rate, NULL, NULL,	/* XXX: needs validator! */
		     GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (bender_class), "readahead", GDAM_ARG_TYPE_INT, gdas_bender_arg_set_readahead, gdas_bender_arg_get_readahead, NULL, NULL,	/* XXX: needs validator! */
		     GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (bender_class),
		     "jump_to_mark",
		     GDAM_ARG_TYPE_INT,
		     gdas_bender_arg_jump_to_mark, NULL, NULL, NULL, 0);
  /* jump to nearest beat, relative to the mark.  arg is beat length. */
  gdam_arg_register (GTK_OBJECT_CLASS (bender_class),
		     "jump_to_nearest_mark",
		     GDAM_ARG_TYPE_INT,
		     gdas_bender_arg_jump_to_nearest_mark, NULL, NULL, "0,", 0);
  gdam_arg_register (GTK_OBJECT_CLASS (bender_class),
		     "set_mark",
		     GDAM_ARG_TYPE_INT,
		     gdas_bender_arg_set_mark, NULL, NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (bender_class),
		     "buffer_length",
		     GDAM_ARG_TYPE_INT,
		     gdas_bender_arg_set_buffer_length,
		     gdas_bender_arg_get_buffer_length,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (bender_class),
		     "beat_length",
		     GDAM_ARG_TYPE_INT,
		     gdas_bender_arg_beat_length, NULL, NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (bender_class),
		     "fade_window",
		     GDAM_ARG_TYPE_INT,
		     gdas_bender_arg_set_fade_window,
		     gdas_bender_arg_get_fade_window,
		     NULL, "0,", GDAM_ARG_VITAL);
  /* sets one of the position markers, arg value is index of marker */
  /* TODO: convert to int array and take offset value ala set_mark? */
  /* TODO: range derived from GDAS_BENDER_NUM_POSITION_MARKERS ?? */
  gdam_arg_register (GTK_OBJECT_CLASS (bender_class),
		     "set_position_marker",
		     GDAM_ARG_TYPE_INT,
		     gdas_bender_arg_set_position_marker, 
		     NULL,
		     NULL, "0,3", 0);
  gdam_arg_register (GTK_OBJECT_CLASS (bender_class),
		     "jump_to_position_marker",
		     GDAM_ARG_TYPE_INT,
		     gdas_bender_arg_jump_to_position_marker,
		     NULL,
		     NULL, "0,3", 0);
}

GtkType gdas_bender_get_type()
{
  static GtkType bender_type = 0;
  if (!bender_type)
    {
      static const GtkTypeInfo bender_info = {
	"GdasBender",
	sizeof (GdasBender),
	sizeof (GdasBenderClass),
	(GtkClassInitFunc) gdas_bender_class_init,
	(GtkObjectInitFunc) gdas_bender_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_BUFFERING_FILTER;
      bender_type = gtk_type_unique (parent, &bender_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return bender_type;
}

static void gdas_bender_miss_data(GdasSource*      input,
				  int              lost)
{
  GdasSource *filtered;
  filtered = GDAS_FILTER (input)->filtered;
  if (filtered != NULL)
    gdas_source_miss_data (filtered, lost);
  else
    GDAS_BENDER (input)->backbuffer_offset += ((gint64) lost << 16);
}

static void gdas_bender_get_latency_info(GdasSource* source,
                                         GdasSource* subsource,
                                         GdamLatencyInfo* latency)
{
  GdasBender *bender = GDAS_BENDER (source);
  int bb_offset = (int) (bender->backbuffer_offset >> 16);
  int bb_len = GDAS_BUFFERING_FILTER (source)->buffer_current_amount;
  latency->latency = bb_len - bb_offset;
  latency->resample = (GdamFixed) bender->rate << (GdamFixed_LOG_ONE - 16);
}

static gboolean gdas_bender_is_ready_to_read(GdasSource*      source)
{
  GdasBender *bender;
  GdamFixed rate;
  GdasSource *filtered;
  bender = GDAS_BENDER (source);
  if (bender->changes != NULL)
    {
      g_assert (!bender->changes->is_jump);
      rate = bender->changes->new_rate;
    }
  else
    rate = bender->rate;
  if (rate <= 0)
    return TRUE;
  filtered = GDAS_FILTER (source)->filtered;
  if (filtered == NULL)
    return TRUE;
  return gdas_source_is_ready_to_read (filtered);
}

/* Read from our source.
 * NOTE: filtered_read may change buffer offsets!! */
#define MAX_SANE_READ 50000
static int do_filtered_read(GdasBender* bender, int num)
{
  GdasSource *filtered = ((GdasFilter *) bender)->filtered;
  int num_read, discard;
  GdasBufferingFilter *buffering_filter = (GdasBufferingFilter *) bender;
  GdamSample *buf;
  g_return_val_if_fail (num <= MAX_SANE_READ, 0);
  /*  check vs. size allocated?? */
  buf = alloca (sizeof (GdamSample) * num);
  num_read = gdas_source_read (filtered, buf, num);
  if (num_read <= 0)
    return 0;
  /* discard is the amount of distant-past data lost to accomodate recent-past
   * data appended to the end of the buffer.  we need to adjust any offsets
   * into the buffer by the amount discarded so that the offsets point to
   * the same samples.
   */
  discard = gdas_buffering_filter_append (buffering_filter, buf, num_read);
  if (discard)
    {
      int i;
      bender->backbuffer_offset -= ((gint64) discard << 16);
      bender->end_offset -= ((gint64) discard << 16);
      /* for any valid position marker, adjust offset */
      for (i=0; i < GDAS_BENDER_NUM_POSITION_MARKERS; i++)
	if (bender->position_markers[i] >= 0)
	  bender->position_markers[i] -= ((gint64) discard << 16);
    }
  return num_read;
}

static void remove_change(GdasBender* bender)
{
  GdasBenderChange *next;
  g_return_if_fail (bender->changes != NULL);
  next = bender->changes->next;
  g_free (bender->changes);
  bender->changes = next;
  /* If we are out of changes and in snap_time mode, then jump to mark. */
  if (next == NULL && bender->change_mode == GDAS_BENDER_CHANGE_SNAP_TIME)
    {
      bender->backbuffer_offset = bender->end_offset;
      bender->fade_remaining = bender->fade_window;
    }
}

/* Remove any pending changes which are position seeks, and collect
 * their value into backbuffer_offset.  We also set fade_remaining
 * so that _read() will smooth over the seek. */
static void remove_new_position_changes(GdasBender* bender)
{
  while (bender->changes && bender->changes->is_jump)
    {
      gint64 offset;
      offset = (gint64) bender->changes->new_rate << 16;
      bender->backbuffer_offset += offset;
      remove_change (bender);
      bender->fade_remaining = bender->fade_window;
    }
}

/*=====================================================================*
 *                             do_fading                               *
 *                                                                     *
 * Obscures a position change by doing a crossfade from a dc signal    *
 * (which is the last sample from the old position, repeated) to the   *
 * sample data at the new position.                                    *
 *                                                                     *
 * The buffer 'sample' should be filled with at least 'num' audio      *
 * samples from the new position.  This buffer is modified in-place    *
 * to add the crossfade.                                               *
 *                                                                     *
 * This method is a little cheesy, and introduces a 2nd-order          *
 * discontinuity, but doesn't require us to read ahead from the old    *
 * position to get data for a real crossfade.                          *
 *=====================================================================*/
static void do_fading(GdasBender* bender,
                      GdamSample* sample,
		      int         num)
{
  int to_fade;
  int fade_remaining;
  int fade_window = bender->fade_window;
  GdamSample fade_sample = bender->fade_sample;
  fade_remaining = bender->fade_remaining;
  to_fade = num;
  if (to_fade > fade_remaining)
    to_fade = fade_remaining;
  while (to_fade-- > 0)
    {
      GdamSampleProduct tmp;

      tmp = (GdamSampleProduct) fade_remaining *fade_sample.s[0]
	+ (fade_window - fade_remaining) * sample->s[0];
      tmp /= fade_window;
      sample->s[0] = (GdamSampleValue) tmp;

      tmp = (GdamSampleProduct) fade_remaining *fade_sample.s[1]
	+ (fade_window - fade_remaining) * sample->s[1];
      tmp /= fade_window;
      sample->s[1] = (GdamSampleValue) tmp;

      sample++;
      fade_remaining--;
    }
  bender->fade_remaining = fade_remaining;
}

/*=====================================================================*
 *                             compute_clip                            *
 *                                                                     *
 * Determine how many samples from a range need to be zeroed out,      *
 * in the front and the back because they are outside the buffer.      *
 *                                                                     *
 * Parameters:                                                         *
 *     bender            the bender to clip for.                       *
 *     rate              the forward or backward velocity of the       *
 *                       position.                                     *
 *     offset            the position within the backbuffer.           *
 *     num               the number of samples of output at that rate. *
 *     prepad_out        returns the number of samples of silence to   *
 *                       write before sound starts.                    *
 *     postpad_out       returns the number of samples of silence to   *
 *                       write after sound ends.                       *
 *     return            whether any part was in the backbuffer.       *
 *=====================================================================*/
static gboolean
compute_clip(GdasBender*   bender,
             GdamFixed     rate,
	     gint64        offset,
	     int           num,
	     int*          prepad_out,
	     int*          postpad_out)
{
  gint64 end_offset;
  gint64 current_amount;
  gboolean starts_in_bounds, ends_in_bounds;
  gint rate_minus_one = rate - 1;
  current_amount = GDAS_BUFFERING_FILTER (bender)->buffer_current_amount;
  current_amount <<= 16;
  end_offset = offset + (gint64) rate *num;

  *prepad_out = 0;
  *postpad_out = 0;

  starts_in_bounds = (0 <= offset && offset < current_amount);
  ends_in_bounds = (0 <= end_offset && end_offset < current_amount);

  g_return_val_if_fail (rate != 0, FALSE);

  if (!starts_in_bounds)
    {
      gint64 in_point = rate > 0 ? 0 : current_amount - 1;
      *prepad_out = (in_point - offset + rate_minus_one) / rate;
      if (*prepad_out < 0)
	return FALSE;
    }
  if (!ends_in_bounds)
    {
      gint64 in_point = rate > 0 ? current_amount - 1 : 0;
      *postpad_out = (end_offset - in_point + rate_minus_one) / rate;
      if (*postpad_out < 0)
	return FALSE;
    }
  if (*prepad_out + *postpad_out >= num)
    return FALSE;
  return TRUE;
}

/*=====================================================================*
 *                      resample_from_back_buffer                      *
 *                                                                     *
 * Fills a buffer with sound (possibly padded with silence at          *
 * the ends) from the back buffer.  It resamples forward               *
 * or backward.                                                        *
 *                                                                     *
 * Parameters:                                                         *
 *     bender            the bender to read data from.                 *
 *     rate              the resampling rate (may be negative).        *
 *     offset            the backbuffer offset (with fraction).        *
 *     num_from_back     the number of samples to read from the        *
 *                       backbuffer.                                   *
 *     output            the output buffer.                            *
 *=====================================================================*/
static void resample_from_back_buffer(GdasBender* bender,
                                      GdamFixed   rate,
				      gint64      offset,
				      int         num_from_back,
				      GdamSample* output)
{
  gint64 max_offset;
  int num_left;
  int prepad, postpad;
  GdasBufferingFilter *buf_filter = (GdasBufferingFilter *) bender;
  gboolean had_new_sample = FALSE;

  num_left = num_from_back;

  if (rate == 0)
    {
      gdam_sample_zero (output, num_from_back);
      return;
    }

  /* Now clip.  Set prepad and postpad to the number of
   * zeroes where data is unavailable,
   * and adjust num_left/output to be the number of samples
   * to write from the real backbuffer. */
  if (!compute_clip (bender, rate, offset, num_left, &prepad, &postpad))
    {
      gdam_sample_zero (output, num_from_back);
      return;
    }

  gdam_sample_zero (output, prepad);
  output += prepad;
  num_left -= (postpad + prepad);
  offset += (gint64) rate *prepad;
  max_offset = (gint64) buf_filter->buffer_current_amount << 16;

  g_return_if_fail (offset >= 0);
  g_return_if_fail (offset <= max_offset);
  while (num_left-- > 0)
    {
      *output++ = gdas_buffering_filter_sample_from_start
	(buf_filter, (int) (offset >> 16));
      offset += rate;
      had_new_sample = TRUE;
    }
  if (offset > max_offset)
    {
      g_message ("offset=%lld, max_offset=%lld, rate=%d, num=%d",
		 offset, max_offset, rate, num_from_back);
    }
  g_return_if_fail (offset >= 0);
  g_return_if_fail (offset <= max_offset);
}

static GdamTime gdas_bender_get_position(GdasSource* source)
{
  GdasSource *filtered;
  GdasBufferingFilter *buf_filter;
  GdamTime rv;
  GdasBender *bender;
  buf_filter = GDAS_BUFFERING_FILTER (source);
  filtered = GDAS_FILTER (source)->filtered;
  bender = GDAS_BENDER (source);
  if (filtered == NULL)
    return -1;
  rv = gdas_source_get_position (filtered);
  {
    int latency;
    latency = (bender->backbuffer_offset >> 16)
      - (buf_filter->buffer_current_amount);
    if (bender->rate != 0)
      latency = ((gint64) latency << 16) / bender->rate;
    rv -= latency;
  }
  return rv;
}

static int  gdas_bender_read(GdasSource*      source,
			     GdamSample*      samples,
			     int              max_read)
{
  GdasBender *bender;
  GdasSource *filtered;
  GdasBufferingFilter *buffering;
  GdamSample *pending_fade_pointer = NULL;
  int rv = 0;
  int pending_fade_max_time = 0;

  bender = GDAS_BENDER (source);
  filtered = GDAS_FILTER (source)->filtered;
  buffering = GDAS_BUFFERING_FILTER (source);
  if (filtered == NULL)
    return 0;

  /* If we have a pending fade from the last call to _read() then prepare
   * to make that fade with the first audio data this _read()
   * Or, if we have a pending fade since the last read, prepare to do fading.
   */
  if (bender->pending_fade_pointer != NULL || bender->fade_remaining > 0)
    {
      bender->fade_sample = bender->last_sample;
      pending_fade_pointer = samples;
      pending_fade_max_time = max_read;
    }

  /* Make sure we have enough space in the buffer. */
  if (buffering->buffer_length < MIN_BUFFERING)
    gdas_buffering_filter_ensure_space (buffering, MIN_BUFFERING);

  /* If we are keeping in beat, then move the mark by even beats to minimize
   * distance to current play position. */
  if (bender->change_mode == GDAS_BENDER_CHANGE_KEEP_IN_BEAT
      && bender->beat_size > 0)
    while ((bender->end_offset - bender->backbuffer_offset) > bender->beat_size)
      bender->end_offset -= bender->beat_size;

  /* If we keeping in time, we may need to read ahead.  Also, we may
   * have bender->readahead set, meaning we need to read ahead to
   * accomodate recent changes (???)  This block gets any input samples
   * we need. */
  if (bender->change_mode == GDAS_BENDER_CHANGE_KEEP_TIME
      || bender->change_mode == GDAS_BENDER_CHANGE_SNAP_TIME
      || bender->change_mode == GDAS_BENDER_CHANGE_KEEP_IN_BEAT
      || bender->readahead > 0)
    {
      int advance;
      gint64 end_time;
      end_time = bender->backbuffer_offset + bender->readahead;
      /* If we're keeping time, then we need to read one sample from
       * our input for every sample we output.  This maintains a 1:1
       * ratio overall */
      if (bender->change_mode == GDAS_BENDER_CHANGE_KEEP_TIME
          || bender->change_mode == GDAS_BENDER_CHANGE_KEEP_IN_BEAT
	  || bender->change_mode == GDAS_BENDER_CHANGE_SNAP_TIME)
	{
	  gint64 est_end;
	  est_end = (gint64) max_read *bender->rate + bender->end_offset;
	  if (end_time < est_end)
	    {
	      end_time = est_end;
	    }
	}
      end_time -= ((gint64) buffering->buffer_current_amount << 16);
      /* advance is the number of samples we need to read from input */
      advance = (int) (end_time >> 16);
      if (advance > 0 && gdas_source_is_ready_to_read (filtered))
	(void) do_filtered_read (bender, advance);
    }

  /* Produce output by reading from buffer. */
  while (rv < max_read)
    {
      int max_time;
      GdamFixed rate;
      gboolean had_new_sample = FALSE;

      /* Compute `rate' and `max_time'. */
      max_time = max_read - rv;

      /* If we are through with a change, remove it and handle any immediately
       * pending seeks.   
       */
      while (bender->changes && bender->changes->num_samples <= 0)
	{
	  g_assert (!bender->changes->is_jump);
	  remove_change (bender);
	  remove_new_position_changes (bender);
	}
      if (bender->changes != NULL)
	{
	  g_assert (!bender->changes->is_jump);
	  rate = bender->changes->new_rate;
	  if (max_time > bender->changes->num_samples)
	    max_time = bender->changes->num_samples;
	}
      else
	{
	  rate = bender->rate;
	}

      /*
       * If we are resampling forward, sometimes we 
       * should try to read from the source to keep up.  
       * Make sure all samples we need are in back_buffer.
       */
      if (rate > 0)
	{
	  gint64 end_position;
	  int current_length = buffering->buffer_current_amount;
	  end_position = bender->backbuffer_offset
	    + (gint64) rate * (max_time + 1);
	  end_position >>= 16;
	  if (end_position > current_length)
	    {
	      int to_read;
	      to_read = end_position + 1 - current_length;
	      g_return_val_if_fail (to_read < 100 * max_read, 1);
	      if (!gdas_source_is_ready_to_read (filtered))
		{
		  /* XXX: this can really happen!!! */
		  g_return_val_if_fail (rv > 0, 1);
		  break;
		}
	      if (!do_filtered_read (bender, to_read))
		break;
	      had_new_sample = TRUE;
	    }
	}

      /* Resample data from back buffer and copy to output. */
      resample_from_back_buffer (bender,
				 rate,
				 bender->backbuffer_offset,
				 max_time, samples + rv);
      //g_message ("filled data at %p (max_time %d)", samples + rv, max_time);
      rv += max_time;

      /* If we were waiting for the above data in order to fade, then do so. */
      if (pending_fade_pointer != NULL)
	{
	  /* check that pending ptr is within the area we just filled */
	  g_assert (pending_fade_pointer >= samples + rv - max_time &&
	            pending_fade_pointer <  samples + rv);
	  do_fading (bender, pending_fade_pointer, pending_fade_max_time);
	  pending_fade_pointer = NULL;
	}

      /* If we are working on a change, then count down towards the end of the
       * change by decreasing num_samples.  If num_samples == 0, then we
       * are done with this change.  Discard this change, and handle any
       * immediately pending seeks. */
      if (bender->changes != NULL)
	{
	  g_assert (!bender->changes->is_jump);
	  bender->changes->num_samples -= max_time;
	  if (bender->changes->num_samples <= 0)
	    {
	      remove_change (bender);
	      //g_message ("maybe removing position at output %p", samples + rv);
	      remove_new_position_changes (bender);
	    }
	}
      /* Move our current position marker forward for the data we just output */
      bender->backbuffer_offset += max_time * rate;

      /* If we have a pending fade, and have not finished filling the samples
       * buffer, then prepare to do fading after the next read. */
      max_time = max_read - rv;
      if (max_time > 0 && bender->fade_remaining > 0)
	{
	  g_assert (rv >= 1);
	  bender->fade_sample = *(samples + rv - 1);
	  pending_fade_pointer = samples + rv;
	  pending_fade_max_time = max_time;
	}
      /* Else, if a fade is pending but we have filled the samples buffer,
       * then prepare to do fading next time bender->read() is called. */
      else if (bender->fade_remaining > 0)
	{
	  g_assert (rv >= 1);
	  bender->fade_sample = *(samples + rv - 1);
	  pending_fade_pointer = samples + rv;
	  pending_fade_max_time = max_time;
          //g_message ("GdasBender: missing fade at read boundary??");
	}
    }

  /* After all reads are done, move the mark by the amount output */
  bender->end_offset += (gint64) bender->rate * rv;
  
  /* Save value of the last output sample for fading during next read */
  if (rv > 0)
    bender->last_sample = *(samples + rv - 1);
  else
    g_warning ("bad rv!!");

  /* If we have a pending_fade_pointer, store it at exit.  Althought the
   * pointer may be invalid, we check for != NULL next read, and set it
   * the beginning of the next output buffer */
  bender->pending_fade_pointer = pending_fade_pointer;
  if (pending_fade_pointer)
    gdam_debug ("bender: pending_fade_pointer across reads...");
  return rv;
}
