/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include <math.h>
#include <stdlib.h>
#include "gdasfader.h"
#include <gdam/common/gdamplugin.h>

static int gdas_fader_read(GdasSource* source, GdamSample* buf, int n);
static void gdas_fader_init(GdasFader* fader)
{
}

void gdas_fader_set_mode(GdasFader* fader, GdasFaderMode mode)
{
  if (mode == GDAS_FADER_LOOPS && fader->num_time_vols > 1)
    fader->position %= fader->time_vols[fader->num_time_vols * 2 - 2];
  fader->mode = mode;
}
static void gdas_fader_arg_set_mode(GtkObject* object, GdamArg* arg)
{
  GdasFader *fader = GDAS_FADER (object);
  if (strcmp (GDAM_ARG_STRING (arg), "terminates") == 0)
    {
      gdas_fader_set_mode (fader, GDAS_FADER_TERMINATES);
    }
  else if (strcmp (GDAM_ARG_STRING (arg), "loops") == 0)
    {
      gdas_fader_set_mode (fader, GDAS_FADER_LOOPS);
    }
  else if (strcmp (GDAM_ARG_STRING (arg), "shelf") == 0)
    {
      gdas_fader_set_mode (fader, GDAS_FADER_SHELF);
    }
  else
    {
      gdam_debug ("fader: no mode %s", GDAM_ARG_STRING (arg));
    }
}

static void gdas_fader_arg_get_mode(GtkObject* object, GdamArg* arg)
{
  GdasFader *fader = GDAS_FADER (object);
  if (fader->mode == GDAS_FADER_TERMINATES)
    {
      GDAM_ARG_STRING (arg) = g_strdup ("terminates");
    }
  else if (fader->mode == GDAS_FADER_LOOPS)
    {
      GDAM_ARG_STRING (arg) = g_strdup ("loops");
    }
  else if (fader->mode == GDAS_FADER_SHELF)
    {
      GDAM_ARG_STRING (arg) = g_strdup ("shelf");
    }
  else
    {
      gdam_debug ("fader: get mode: no mode %d", fader->mode);
    }
}

static int gint64_sorter(const void* i1, const void* i2)
{
  gint64 i_1 = *(gint64 *) i1;
  gint64 i_2 = *(gint64 *) i2;
  if (i_1 < i_2)
    return -1;
  if (i_1 > i_2)
    return +1;
  return 0;
}

/* Offset is how much to trim off.  If negative,
 * silence is inserted. */
static void set_keypoints(GdasFader* fader,
                          int num_times,
			  gint64* array,
			  gint64 offset)
{
  int i;
  if (num_times <= 1 || array[(num_times - 1) * 2] <= 0)
    {
      gdam_debug ("set_keypoints too small");
      return;
    }
  fader->num_time_vols = num_times;
  g_free (fader->time_vols);
  fader->time_vols = g_new (gint64, fader->num_time_vols * 2);
  memcpy (fader->time_vols, array,
	  sizeof (gint64) * fader->num_time_vols * 2);
  for (i = 0; i < num_times; i++)
    fader->time_vols[2 * i] -= offset;
  qsort (fader->time_vols,
	 fader->num_time_vols, sizeof (gint64) * 2, gint64_sorter);
  i = 0;
  while (i < fader->num_time_vols - 1 && fader->time_vols[2 * (i + 1)] < 0)
    i++;
  if (i < fader->num_time_vols && fader->time_vols[2 * i] < 0)
    {
      gint64 vol = -fader->time_vols[2 * i];
      vol *= (fader->time_vols[2 * i + 3] - fader->time_vols[2 * i + 1]);
      vol /= (fader->time_vols[2 * i + 2] - fader->time_vols[2 * i + 0]);
      vol += fader->time_vols[2 * i + 1];
      fader->time_vols[2 * i] = 0;
      fader->time_vols[2 * i + 1] = vol;
    }
  if (i == fader->num_time_vols)
    {
      g_free (fader->time_vols);
      fader->time_vols = NULL;
      fader->num_time_vols = 0;
      fader->position = 0;
      return;
    }

  memmove (fader->time_vols,
	   fader->time_vols + 2 * i,
	   (fader->num_time_vols - i) * sizeof (gint64) * 2);
  fader->num_time_vols -= i;

  if (fader->num_time_vols >= 1)
    {
      gint64 length = fader->time_vols[fader->num_time_vols * 2 - 2];
      if (length > 0 && fader->mode == GDAS_FADER_LOOPS)
	fader->position %= length;
    }
}

static void gdas_fader_arg_set_keypoints(GtkObject* object, GdamArg* arg)
{
  GdasFader *fader = GDAS_FADER (object);
  int num_times = GDAM_ARG_INT64_ARRAY (arg).length / 2;
  gint64 *array = GDAM_ARG_INT64_ARRAY (arg).array;
  set_keypoints (fader, num_times, array, 0);
}

static void gdas_fader_arg_get_keypoints(GtkObject* object, GdamArg* arg)
{
  GdasFader *fader = GDAS_FADER (object);
  GDAM_ARG_INT64_ARRAY (arg).length = fader->num_time_vols * 2;
  GDAM_ARG_INT64_ARRAY (arg).array = g_new (gint64, fader->num_time_vols * 2);
  memcpy (GDAM_ARG_INT64_ARRAY (arg).array,
	  fader->time_vols, sizeof (gint64) * fader->num_time_vols * 2);
}

static void gdas_fader_arg_set_fader_position(GtkObject* object, GdamArg* arg)
{
  GDAS_FADER (object)->position = GDAM_ARG_INT64 (arg);
}
static void gdas_fader_arg_get_fader_position(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT64 (arg) = GDAS_FADER (object)->position;
}

static void gdas_fader_arg_set_phase_sync(GtkObject* object, GdamArg* arg)
{
  int num_double = GDAM_ARG_DOUBLE_ARRAY (arg).length;
  gdouble *doubles = GDAM_ARG_DOUBLE_ARRAY (arg).array;
  GdasSource *filtered;
  GdamTime offset;
  gdouble start_time, period, new_pos;

  filtered = GDAS_FILTER (object)->filtered;

  if (num_double != 2)
    {
      gdam_debug ("set_phase_sync: needs exactly two arguments");
      return;
    }

  if (filtered == NULL)
    return;

  start_time = doubles[0] * 44100.0;
  period = doubles[1] * 44100.0;

  if (period <= 0.01)
    {
      gdam_debug ("period was too small");
      return;
    }

  offset = gdas_source_get_position (filtered);
  new_pos = fmod (offset - start_time, period);
  if (new_pos < 0.0)
    new_pos += period;
  GDAS_FADER (object)->position = (gint64) new_pos;
}

static void gdas_fader_class_init(GdasFaderClass* fader_class)
{
  GDAS_SOURCE_CLASS (fader_class)->read = gdas_fader_read;
  gdam_arg_register (GTK_OBJECT_CLASS (fader_class),
		     "mode",
		     GDAM_ARG_TYPE_STRING,
		     gdas_fader_arg_set_mode,
		     gdas_fader_arg_get_mode, NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (fader_class), "keypoints", GDAM_ARG_TYPE_INT64_ARRAY, gdas_fader_arg_set_keypoints, gdas_fader_arg_get_keypoints, NULL, NULL,	/* XXX: needs validator! */
		     GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (fader_class), "fader_position", GDAM_ARG_TYPE_INT64, gdas_fader_arg_set_fader_position, gdas_fader_arg_get_fader_position, NULL, NULL,	/* XXX: needs validator! */
		     GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (fader_class), "phase_sync", GDAM_ARG_TYPE_DOUBLE_ARRAY, gdas_fader_arg_set_phase_sync, NULL, NULL, NULL,	/* XXX: needs validator! */
		     0);
}

static int find_index(GdasFader* fader)
{
  int last_possible = fader->num_time_vols - 1, first_possible = 0;
  while (first_possible < last_possible)
    {
      int mid = (first_possible + last_possible) / 2;
      if (mid == first_possible)
	mid++;
      if (fader->time_vols[mid * 2] <= fader->position)
	first_possible = mid;
      else
	last_possible = mid - 1;
    }
  return first_possible;
}

static void gdam_sample_fade(GdamSample*   at,
                             GdamFixed     volume,
			     guint32       volume_frac,
                             GdamFixed     volume_incr,
			     guint32       volume_incr_frac,
			     int           len)
{
  while (len-- > 0)
    {
      guint32 new_volume_frac;
      at->s[0] = ((GdamSampleProduct) at->s[0] * volume) / GdamFixed_ONE;
      at->s[1] = ((GdamSampleProduct) at->s[1] * volume) / GdamFixed_ONE;
      at++;
      volume += volume_incr;
      new_volume_frac = volume_frac + volume_incr_frac;
      if (new_volume_frac < volume_frac)
	volume++;
      volume_frac = new_volume_frac;
    }
}

static int gdas_fader_read(GdasSource* source, GdamSample* buf, int n)
{
  GdasFader *fader = GDAS_FADER (source);
  GdasSource *filtered = GDAS_FILTER (source)->filtered;
  int index = find_index (fader);
  int offset;
  int left_to_scale;
  GdamSample *scale_at;
  if (index == -1)
    {
      if (fader->terminates)
	return 0;
      gdam_sample_zero (buf, n);
      return n;
    }

  if (filtered == NULL)
    {
      return 0;
    }
  n = gdas_source_read (filtered, buf, n);
  if (n <= 0)
    {
      return n;
    }

  if (fader->num_time_vols == 0)
    {
      /* XXX: silence it? */
      gdam_sample_zero (buf, n);
      return n;
    }
  offset = (int) (fader->position - fader->time_vols[index * 2]);
  left_to_scale = n;
  scale_at = buf;
  while (left_to_scale > 0)
    {
      while (index < fader->num_time_vols - 1
	     && (fader->time_vols[index * 2 + 0]
		 == fader->time_vols[index * 2 + 2]))
	index++;
      if (index == fader->num_time_vols - 1)
	{
	  if (fader->mode == GDAS_FADER_TERMINATES)
	    break;
	  if (fader->mode == GDAS_FADER_LOOPS)
	    {
	      index = 0;
	      offset = 0;
	      fader->position = 0;
	    }
	  else
	    {
	      gdam_sample_scale (scale_at,
				 left_to_scale,
				 fader->time_vols[index * 2 + 1]);
	      fader->position += left_to_scale;
	      left_to_scale = 0;
	      break;
	    }
	}
      else
	{
	  GdamFixed volume;
	  guint32 volume_frac;
	  GdamFixed volume_incr;
	  guint32 volume_incr_frac;
	  gdouble vol_incr_d, vol_d;
	  int amt;
	  int delta_volume = (int) (fader->time_vols[index * 2 + 3]
				    - fader->time_vols[index * 2 + 1]);
	  int delta_time = (int) (fader->time_vols[index * 2 + 2]
				  - fader->time_vols[index * 2 + 0]);
	  int in_segment = (int) (fader->position
				  - fader->time_vols[index * 2 + 0]);
	  g_assert (0 <= in_segment);
	  g_assert (in_segment < delta_time);
	  vol_incr_d = (double) delta_volume / delta_time;
	  vol_d = (gdouble) fader->time_vols[index * 2 + 1]
	    + vol_incr_d * in_segment;
	  volume = (int) (floor (vol_d));
	  volume_frac = (guint32) ((vol_d - volume) * 65536.0);
	  volume_incr = (int) (floor (vol_incr_d));
	  volume_incr_frac = (guint32) ((vol_incr_d - volume_incr) * 65536.0);
	  /* unsigned value cannot be less than 0.. */
	  /*g_assert(0 <= volume_frac); */
	  /*g_assert(0 <= volume_incr_frac); */
	  g_assert (volume_frac < 65536);
	  g_assert (volume_incr_frac < 65536);
	  volume_frac <<= 16;
	  volume_incr_frac <<= 16;

	  /* Do the fade... */
	  amt = (delta_time - in_segment);
	  if (amt > left_to_scale)
	    amt = left_to_scale;
	  gdam_sample_fade (scale_at,
			    volume,
			    volume_frac, volume_incr, volume_incr_frac, amt);
	  scale_at += amt;
	  fader->position += amt;
	  left_to_scale -= amt;
	  index++;
	}
    }
  return n - left_to_scale;
}

GtkType gdas_fader_get_type()
{
  static GtkType fader_type = 0;
  if (!fader_type)
    {
      static const GtkTypeInfo fader_info = {
	"GdasFader",
	sizeof (GdasFader),
	sizeof (GdasFaderClass),
	(GtkClassInitFunc) gdas_fader_class_init,
	(GtkObjectInitFunc) gdas_fader_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      fader_type = gtk_type_unique (parent, &fader_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return fader_type;
}
