/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdasflange.h"
#include <gdam/common/gdamplugin.h>
#include <gdam/server/gdamconnection.h>
#include <gdam/common/alloca.h>

static int gdas_flange_read(GdasSource*    source,
                            GdamSample*    buffer,
			    int            max_read)
{
  GdasFlange *flange = GDAS_FLANGE (source);
  GdasBufferingFilter *buffering_filter = GDAS_BUFFERING_FILTER (source);
  gboolean chorus = flange->chorus;

  GdamSample *modulation = NULL;
  int num_read;
  int i;
  int effect = flange->effect;

  num_read = gdas_filter_sub_read (GDAS_FILTER (source), buffer, max_read);

  /* Fill modulation buffer. */
  if (num_read > 0)
    {
      int mod_read = 0;
      modulation = alloca (sizeof (GdamSample) * num_read);
      if (flange->modulator != NULL)
	{
	  mod_read = gdas_source_read (flange->modulator,
				       modulation, num_read);
	}
      if (mod_read < num_read)
	gdam_sample_zero (modulation + mod_read, num_read - mod_read);
    }

  for (i = 0; i < num_read; i++)
    {
      int channel;
      int init[2];
      GdamSample tmp_sample;
      init[0] = (GdamSampleProduct) buffer[i].s[0] * effect / GdamFixed_ONE;
      init[1] = (GdamSampleProduct) buffer[i].s[1] * effect / GdamFixed_ONE;
      if (chorus)
	{
	  GdamSample tmp;
	  tmp.s[0] = init[0];
	  tmp.s[1] = init[1];
	  gdas_buffering_filter_append_sample (buffering_filter, tmp);
	}
      for (channel = 0; channel < 2; channel++)
	{
	  int mod;
	  int value = buffer[i].s[channel];
	  int echo;
	  GdamFixed factor;
	  int sum;
	  int extra_value;

	  sum = 0;
	  extra_value = value - init[channel];

	  mod = modulation[i].s[channel] + GDAM_SAMPLE_ONE;
	  factor = (GdamFixed) (((gint64) mod
				 * flange->delta_echo_factor)
				>> (GDAM_SAMPLE_LOG + 1));
	  factor += flange->min_echo_factor;

	  for (echo = 0; echo < flange->num_echoes; echo++)
	    {
	      int delay;
	      int value;
	      int volume;
	      delay = flange->echoes[echo].delay;
	      delay = (gint64) factor *delay / GdamFixed_ONE;
	      value = gdas_buffering_filter_sample_at
		(buffering_filter, delay).s[channel];
	      volume = flange->echoes[echo].volume;
	      sum += GDAM_FIXED_MUL (volume, value);
	    }

	  tmp_sample.s[channel] = GDAM_SAMPLE_CLAMP (sum);
	  sum += extra_value;
	  buffer[i].s[channel] = GDAM_SAMPLE_CLAMP (sum);
	}
      if (!chorus)
	{
	  gdas_buffering_filter_append_sample (buffering_filter, tmp_sample);
	}
    }
  return num_read;
}

static void do_ensure_space(GdasFlange* flange)
{
  int max_delay = 0;
  int i;
  int required;
  GdamFixed max_echo_factor;
  for (i = 0; i < flange->num_echoes; i++)
    {
      int delay = flange->echoes[i].delay;
      if (delay > max_delay)
	max_delay = delay;
    }
  max_echo_factor = flange->min_echo_factor + flange->delta_echo_factor;
  required = (int) (((gint64) max_delay * max_echo_factor) / GdamFixed_ONE);
  gdas_buffering_filter_ensure_space ((GdasBufferingFilter *) flange,
				      required + 1);
}

static void gdas_flange_arg_set_min_max(GtkObject* object, GdamArg* arg)
{
  gdouble *array = GDAM_ARG_DOUBLE_ARRAY (arg).array;
  int length = GDAM_ARG_DOUBLE_ARRAY (arg).length;
  GdasFlange *flange = GDAS_FLANGE (object);
  if (length != 2 || array[0] < 0.0 || array[0] > array[1])
    {
      gdam_debug ("array to flange::min_max was malformed");
      return;
    }
  if (array[1] > 100.0)
    {
      gdam_debug ("flange: scale too large");
      return;
    }
  flange->min_echo_factor = GDAM_FIXED_FROM_DOUBLE (array[0]);
  flange->delta_echo_factor = GDAM_FIXED_FROM_DOUBLE (array[1] - array[0]);
  do_ensure_space (flange);
}
static void gdas_flange_arg_get_min_max(GtkObject* object, GdamArg* arg)
{
  gdouble *array;
  int max_echo_factor;
  GdasFlange *flange;
  array = g_new (gdouble, 2);
  GDAM_ARG_DOUBLE_ARRAY (arg).array = array;
  GDAM_ARG_DOUBLE_ARRAY (arg).length = 2;
  flange = GDAS_FLANGE (object);
  array[0] = GDAM_FIXED_TO_DOUBLE (flange->min_echo_factor);
  max_echo_factor = flange->min_echo_factor + flange->delta_echo_factor;
  array[1] = GDAM_FIXED_TO_DOUBLE (max_echo_factor);
}
static void gdas_flange_arg_set_echoes(GtkObject* object, GdamArg* arg)
{
  int num_echoes = GDAM_ARG_DOUBLE_ARRAY (arg).length / 2;
  GdasEchoElement *elements = g_new (GdasEchoElement, num_echoes * 2);
  gdouble *array = GDAM_ARG_DOUBLE_ARRAY (arg).array;
  int i;
  GdasFlange *flange = GDAS_FLANGE (object);
  for (i = 0; i < num_echoes; i++)
    {
      elements[i].volume = (GdamFixed) (array[2 * i + 0] * GdamFixed_ONE);
      elements[i].delay = array[2 * i + 1];
    }
  if (flange->echoes != NULL)
    g_free (flange->echoes);
  flange->echoes = elements;
  flange->num_echoes = num_echoes;
  do_ensure_space (flange);
}
static void gdas_flange_arg_get_echoes(GtkObject* object, GdamArg* arg)
{
  GdasFlange *flange = GDAS_FLANGE (object);
  int num_echoes = flange->num_echoes;
  gdouble *array = g_new (gdouble, num_echoes * 2);
  int i;
  for (i = 0; i < num_echoes; i++)
    {
      array[2 * i + 0] = GDAM_FIXED_TO_DOUBLE (flange->echoes[i].volume);
      array[2 * i + 1] = flange->echoes[i].delay;
    }
  GDAM_ARG_DOUBLE_ARRAY (arg).length = num_echoes * 2;
  GDAM_ARG_DOUBLE_ARRAY (arg).array = array;
}

static void gdas_flange_arg_set_chorus(GtkObject* object, GdamArg* arg)
{
  int chorus = GDAM_ARG_INT (arg);
  GDAS_FLANGE (object)->chorus = chorus ? TRUE : FALSE;
}
static void gdas_flange_arg_get_chorus(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAS_FLANGE (object)->chorus;
}

static void gdas_flange_arg_set_effect(GtkObject* object, GdamArg* arg)
{
  GdasFlange *flange = GDAS_FLANGE (object);
  flange->effect = GDAM_ARG_DOUBLE (arg) * GdamFixed_ONE;
  flange->effect = CLAMP (flange->effect, 0, GdamFixed_ONE);
}
static void gdas_flange_arg_get_effect(GtkObject* object, GdamArg* arg)
{
  GdasFlange *flange = GDAS_FLANGE (object);
  GDAM_ARG_DOUBLE (arg) = (gdouble) flange->effect / GdamFixed_ONE;
}

static void gdas_flange_arg_set_modulator(GtkObject* object, GdamArg* arg)
{
  int source_id = GDAM_ARG_INT (arg);
  GdasSource *sub = gdam_connection_get_source (NULL, source_id);
  gdas_flange_set_modulator (GDAS_FLANGE (object), sub);
  if (sub != NULL)
    gtk_object_unref (GTK_OBJECT (sub));
}
static void gdas_flange_arg_get_modulator(GtkObject* object, GdamArg* arg)
{
  GdasSource *modulator = GDAS_FLANGE (object)->modulator;
  GDAM_ARG_INT (arg) = modulator
    ? GDAM_SERVER_OBJECT (modulator)->global_id : 0;
}

static void gdas_flange_init(GdasFlange* flange)
{
  flange->min_echo_factor = GdamFixed_ONE;
  flange->effect = GdamFixed_ONE;
}
static void gdas_flange_class_init(GdasFlangeClass* flange_class)
{
  GtkObjectClass *object_class;

  GDAS_SOURCE_CLASS (flange_class)->read = gdas_flange_read;

  object_class = GTK_OBJECT_CLASS (flange_class);
  gdam_arg_register (object_class,
		     "min_max",
		     GDAM_ARG_TYPE_DOUBLE_ARRAY,
		     gdas_flange_arg_set_min_max,
		     gdas_flange_arg_get_min_max, NULL, ",", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "echoes",
		     GDAM_ARG_TYPE_DOUBLE_ARRAY,
		     gdas_flange_arg_set_echoes,
		     gdas_flange_arg_get_echoes, NULL, ",", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "modulator",
		     GDAM_ARG_TYPE_INT,
		     gdas_flange_arg_set_modulator,
		     gdas_flange_arg_get_modulator,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "chorus",
		     GDAM_ARG_TYPE_INT,
		     gdas_flange_arg_set_chorus,
		     gdas_flange_arg_get_chorus,
		     NULL, "boolean", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "effect",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_flange_arg_set_effect,
		     gdas_flange_arg_get_effect, NULL, "0,3", GDAM_ARG_VITAL);
}

GtkType gdas_flange_get_type()
{
  static GtkType flange_type = 0;
  if (!flange_type)
    {
      static const GtkTypeInfo flange_info = {
	"GdasFlange",
	sizeof (GdasFlange),
	sizeof (GdasFlangeClass),
	(GtkClassInitFunc) gdas_flange_class_init,
	(GtkObjectInitFunc) gdas_flange_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_BUFFERING_FILTER;
      flange_type = gtk_type_unique (parent, &flange_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return flange_type;
}
void gdas_flange_set_modulator(GdasFlange* flange, GdasSource* mod)
{
  if (flange->modulator != NULL)
    {
      GdasSource *old_mod = flange->modulator;
      flange->modulator = NULL;
      old_mod->parent = NULL;
      gtk_object_unref (GTK_OBJECT (old_mod));
    }
  if (mod != NULL)
    {
      GtkObject *object = GTK_OBJECT (mod);
      gtk_object_ref (object);
      gtk_object_sink (object);
    }
  flange->modulator = mod;
}
