/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdasscrub.h"

/*
 * Credits:
 *      This source was suggested by ix@ntw.net.
 */

/*
 *                  Overlapping Strategy
 *
 * In order to reduce the pop that occurs with an
 * abrupt change of audio stream causes,
 * we fade from the old buffer to the new one (the fade cuts
 * out on the most recent samples).
 */


/* the minimum back buffer size [we only increase the size of this buffer] */
#define DEFAULT_SCRUB_BUFFER_LEN	1024

/* the maximum number of samples to fade over when going
 * from the end back to the start of the backbuffer. */
#define FADE_LENGTH                     64


/*
 * Update `overlap_countdown' and `overlap_length' for
 * a change of scrub length.
 */
static void reset_overlap_data(GdasScrub* scrub)
{
  scrub->nonoverlap_size = scrub->scrub_len - FADE_LENGTH;
  if (scrub->nonoverlap_size < 0)
    scrub->nonoverlap_size = 0;
  scrub->overlap_countdown = scrub->nonoverlap_size;
  scrub->overlap_length = scrub->scrub_len = scrub->overlap_countdown;
}

static void gdas_scrub_init(GdasScrub* scrub)
{
  scrub->input_rate = GdamFixed_ONE;
  scrub->buffer_len = DEFAULT_SCRUB_BUFFER_LEN;
  scrub->buffer = NULL;
  scrub->scrub_len = 0;
  scrub->write_index = scrub->read_index = 0;
  reset_overlap_data (scrub);
}

static gboolean gdas_scrub_is_ready_to_read(GdasSource* source)
{
  GdasScrub *scrub;
  GdasSource *filtered;
  scrub = GDAS_SCRUB (source);
  filtered = GDAS_FILTER (source)->filtered;
  /*g_message("is_ready_to_read: read/write_index=%d/%d; filtered=%p",
     scrub->read_index , scrub->write_index, filtered); */

  if (scrub->read_index != scrub->write_index)
    return TRUE;
  if (filtered == NULL)
    return TRUE;
  return gdas_source_is_ready_to_read (filtered);
}

/*
 * Interpolate two samples, for fading.
 *
 * Parameters:
 *         sample:      array containing both of the samples to interpolate
 *         index_1:     index of the first sample to interpolate
 *         frac_1:      proportion of the first sample use
 *         index_2:     index of the second sample to interpolate
 *         frac_2:      proportion of the second sample use
 *
 * Returns:
 *         the interpolated samples.
 */
static inline GdamSample
interp_sample (GdamSample*    samples,
               int            index_1,
	       int            frac_1,
               int            index_2,
	       int            frac_2)
{
  GdamSample rv;
  GdamSampleProduct tmp;
  int total = frac_1 + frac_2;
  tmp = (GdamSampleProduct) samples[index_1].s[0] * frac_1
    + (GdamSampleProduct) samples[index_1].s[0] * frac_2;
  rv.s[0] = (GdamSampleValue) (tmp / total);

  tmp = (GdamSampleProduct) samples[index_1].s[1] * frac_1
    + (GdamSampleProduct) samples[index_1].s[1] * frac_2;
  rv.s[1] = (GdamSampleValue) (tmp / total);

  return rv;
}

static int gdas_scrub_read(GdasSource* source, 
                           GdamSample* buffer,
			   int         max_read)
{
  GdamSample *new_samples;
  GdasScrub *scrub;
  int sub_read_amt;
  int read_index;
  int write_index;
  int scrub_len;
  int remaining;
  int buffer_len;
  int overlap_countdown;
  int overlap_length;
  GdamFixed fraction;
  GdamFixed input_rate;
  GdamSample *scrub_buffer;

  scrub = GDAS_SCRUB (source);

  if (scrub->buffer == NULL)
    scrub->buffer = g_new0 (GdamSample, scrub->buffer_len);

  sub_read_amt = ((gint64) scrub->input_rate * max_read
		  + scrub->fraction_at) >> GdamFixed_LOG_ONE;
  /*g_message("max_read=%d, sub_read_amt=%d", max_read,sub_read_amt); */
  if (sub_read_amt > 0)
    {
      int num_sub_read;
      new_samples = alloca (sizeof (GdamSample) * sub_read_amt);
      num_sub_read = gdas_filter_sub_read (GDAS_FILTER (source),
					   new_samples, sub_read_amt);

      /* update max_read if we got too few samples. */
      if (num_sub_read < sub_read_amt)
	{
	  /* compute the number of samples we can output. */
	  gint64 tmp;
	  tmp = (GdamFixed_ONE - 1 - scrub->fraction_at);
	  tmp += (gint64) num_sub_read *GdamFixed_ONE;
	  tmp /= scrub->input_rate;
	  g_assert (tmp <= max_read);
	  max_read = tmp;
	}
    }
  else
    {
      new_samples = NULL;
    }

  /* input data from the buffer while cycling and outputting. */
  remaining = max_read;
  read_index = scrub->read_index;
  write_index = scrub->write_index;
  scrub_len = scrub->scrub_len;
  fraction = scrub->fraction_at;
  input_rate = scrub->input_rate;
  scrub_buffer = scrub->buffer;
  buffer_len = scrub->buffer_len;
  overlap_countdown = scrub->overlap_countdown;
  overlap_length = scrub->overlap_length;
  while (remaining-- > 0)
    {
      GdamFixed new_fraction = fraction + input_rate;
      if (new_fraction >= GdamFixed_ONE)
	{
	  new_fraction -= GdamFixed_ONE;
	  scrub_buffer[write_index] = *new_samples++;
	  if (++write_index == buffer_len)
	    write_index = 0;
	}
      fraction = new_fraction;

      overlap_countdown--;
      if (overlap_countdown < 0 && overlap_length > 0)
	{
	  GdamSample avg;
	  int fade_index = read_index - scrub_len;
	  if (fade_index < 0)
	    fade_index += buffer_len;
	  avg = interp_sample
	    (scrub_buffer,
	     read_index, overlap_length + overlap_countdown,
	     fade_index, -overlap_countdown);
	  *buffer++ = avg;
	}
      else
	{
	  *buffer++ = scrub_buffer[read_index];
	}
      if (++read_index == buffer_len)
	read_index = 0;
      if (read_index == write_index)
	{
	  read_index = write_index - scrub_len;
	  if (read_index < 0)
	    read_index += buffer_len;
	  overlap_countdown = scrub->nonoverlap_size;
	}
    }
  scrub->read_index = read_index;
  scrub->write_index = write_index;
  scrub->fraction_at = fraction;
  scrub->overlap_countdown = overlap_countdown;
  scrub->overlap_length = overlap_length;

  return max_read;
}

static void gdas_scrub_ensure_buffer_length(GdasScrub* scrub, int min_len)
{
  GdamSample *new_buffer;
  int old_buf_len = scrub->buffer_len;
  if (scrub->buffer_len >= min_len)
    return;

  scrub->buffer_len = min_len;
  if (scrub->buffer == NULL)
    return;

  new_buffer = g_new (GdamSample, min_len);
  gdam_sample_copy (new_buffer,
		    scrub->buffer + scrub->write_index,
		    old_buf_len - scrub->write_index);
  gdam_sample_zero (new_buffer + old_buf_len - scrub->write_index,
		    min_len - old_buf_len);
  gdam_sample_copy (new_buffer + min_len - scrub->write_index,
		    scrub->buffer, scrub->write_index);
  scrub->read_index -= scrub->write_index;
  while (scrub->read_index < 0)
    scrub->read_index += min_len;
  scrub->write_index = 0;

  g_free (scrub->buffer);
  scrub->buffer = new_buffer;
}

static void gdas_scrub_arg_set_scrub_length(GtkObject* object, GdamArg* arg)
{
  GdasScrub *scrub;
  int scrub_len = GDAM_ARG_DOUBLE (arg);
  int delta_scrub_len;
  scrub = GDAS_SCRUB (object);
  delta_scrub_len = scrub_len - scrub->scrub_len;

  gdas_scrub_ensure_buffer_length (scrub, scrub_len + FADE_LENGTH);
  scrub->scrub_len = scrub_len;
  scrub->overlap_length += delta_scrub_len;

  /* hack! */
  if (scrub->overlap_length < -FADE_LENGTH)
    scrub->overlap_length = 0;

  gdas_scrub_ensure_buffer_length (scrub, scrub_len + FADE_LENGTH);
}


static void gdas_scrub_arg_get_scrub_length(GtkObject* object, GdamArg* arg)
{
  GdasScrub *scrub;
  scrub = GDAS_SCRUB (object);
  GDAM_ARG_DOUBLE (arg) = scrub->scrub_len;
}

static void gdas_scrub_arg_set_buffer_length(GtkObject* object, GdamArg* arg)
{
  GdasScrub *scrub;
  scrub = GDAS_SCRUB (object);

  gdas_scrub_ensure_buffer_length (scrub, GDAM_ARG_INT (arg));
}
static void gdas_scrub_arg_get_buffer_length(GtkObject* object, GdamArg* arg)
{
  GdasScrub *scrub;
  scrub = GDAS_SCRUB (object);

  GDAM_ARG_INT (arg) = scrub->buffer_len;
}

static void gdas_scrub_arg_set_input_rate(GtkObject* object, GdamArg* arg)
{
  GdasScrub *scrub;
  gdouble rate;
  scrub = GDAS_SCRUB (object);
  rate = GDAM_ARG_DOUBLE (arg);
  rate = CLAMP (rate, 0.0, 1.0);
  scrub->input_rate = GDAM_FIXED_FROM_DOUBLE (rate);
}

static void gdas_scrub_arg_get_input_rate(GtkObject* object, GdamArg* arg)
{
  GdasScrub *scrub;
  scrub = GDAS_SCRUB (object);
  GDAM_ARG_DOUBLE (arg) = GDAM_FIXED_TO_DOUBLE (scrub->input_rate);
}

static void gdas_scrub_class_init(GdasScrubClass* scrub_class)
{
  GtkObjectClass *object_class;
  GdasSourceClass *source_class;
  object_class = GTK_OBJECT_CLASS (scrub_class);
  source_class = GDAS_SOURCE_CLASS (scrub_class);
  source_class->read = gdas_scrub_read;
  source_class->is_ready_to_read = gdas_scrub_is_ready_to_read;

  gdam_arg_register (object_class,
		     "scrub_length",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_scrub_arg_set_scrub_length,
		     gdas_scrub_arg_get_scrub_length,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "buffer_length",
		     GDAM_ARG_TYPE_INT,
		     gdas_scrub_arg_set_buffer_length,
		     gdas_scrub_arg_get_buffer_length,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "input_rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_scrub_arg_set_input_rate,
		     gdas_scrub_arg_get_input_rate,
		     NULL, "0,1", GDAM_ARG_VITAL);
}

GtkType gdas_scrub_get_type()
{
  static GtkType scrub_type = 0;
  if (!scrub_type)
    {
      static const GtkTypeInfo scrub_info = {
	"GdasScrub",
	sizeof (GdasScrub),
	sizeof (GdasScrubClass),
	(GtkClassInitFunc) gdas_scrub_class_init,
	(GtkObjectInitFunc) gdas_scrub_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      scrub_type = gtk_type_unique (parent, &scrub_info);
    }
  return scrub_type;
}
