/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdasxmmsvisfilter.h"
#include "../gdamfft.h"
#include <gdam/common/gdamconnect.h>
#include <unistd.h>
#include <math.h>

/* This number is set by the XMMS plugin standard. */
#define XMMS_BUF_LEN			512

/* This is the priority of a xmms-buffer-filling thread. */
#define G_XMMS_PRIORITY			G_PRIORITY_LOW

/* Maximum amount of pending outgoing data allowed before
 * we stop appending new data */
#define XMMS_MAX_ALLOWED_CONGESTION	32*1024

static GtkObjectClass* parent_class = NULL;

static void gdas_xmms_vis_filter_init(GdasXmmsVisFilter* xmms_vis_filter)
{
  xmms_vis_filter->cur_amount = 0;
  xmms_vis_filter->analysis_length = XMMS_BUF_LEN;
  xmms_vis_filter->fd = -1;
  xmms_vis_filter->is_connecting = 0;
  xmms_vis_filter->lr_analysis_buffer = g_new (GdamSample, XMMS_BUF_LEN);
  xmms_vis_filter->analysis_buffer = g_new (gint16, XMMS_BUF_LEN * 2);
  gdam_buffer_construct (&xmms_vis_filter->outgoing_buffer);
}

#define FFT_SCALE	4.0

static void do_fft(gint16* array, int length)
{
  gfloat *farray = alloca (sizeof (gfloat) * length);
  //const gfloat* window;
  int i;
  //window = gdam_fft_get_hamming_window_f(length);
  for (i = 0; i < length; i++)
    farray[i] = (gfloat) array[i] / 32768.0;
  gdam_real_fft_f (farray, length / 2, TRUE);
  {
    gfloat tmp = FFT_SCALE * farray[0];
    array[0] = CLAMP (tmp, -32767, 32767);
  }
  for (i = 1; i < length / 2; i++)
    {
      gfloat tmp = farray[0] * farray[0] + farray[1] * farray[1];
      tmp *= (32768.0 * FFT_SCALE);
      array[i] = CLAMP (tmp, -32767, 32767);
      farray += 2;
    }
}

static void flush_xmms_buffer(GdasXmmsVisFilter* filter)
{
  GdamBuffer *buffer = &filter->outgoing_buffer;
  if (buffer->size < XMMS_MAX_ALLOWED_CONGESTION)
    {
      GdamSample *lr_analysis_buffer = filter->lr_analysis_buffer;
      gint16 *analysis_buffer = filter->analysis_buffer;
      int analysis_length = filter->analysis_length;
      char header[4];

      /* Shimmy the samples into left then right */
      {
	gint16 *tmp_l = analysis_buffer + 0;
	gint16 *tmp_r = analysis_buffer + analysis_length;
	GdamSample *tmp_in = lr_analysis_buffer;
	int num = analysis_length;
	while (num-- > 0)
	  {
	    *tmp_l = (tmp_in->s[0] >> (GDAM_SAMPLE_LOG - 15));
	    *tmp_r = (tmp_in->s[1] >> (GDAM_SAMPLE_LOG - 15));
	    tmp_l++;
	    tmp_r++;
	    tmp_in++;
	  }
      }

      /* write out sample data */
      header[0] = 0;		/* ! freq_data */
      header[1] = 1;		/*   stereo */
      header[2] = (G_BYTE_ORDER == G_BIG_ENDIAN) ? 0 : 1;
      header[3] = 0;		/* unused */
      gdam_buffer_append (buffer, header, 4);
      gdam_buffer_append (buffer,
			  analysis_buffer, filter->analysis_length * 4);

      /* write out frequency data */
      header[0] = 1;		/*   freq_data */
      header[1] = 1;		/*   stereo */
      header[2] = (G_BYTE_ORDER == G_BIG_ENDIAN) ? 0 : 1;
      header[3] = 0;		/* unused */
      gdam_buffer_append (buffer, header, 4);
      do_fft (analysis_buffer, analysis_length);
      gdam_buffer_append (buffer, analysis_buffer, (analysis_length / 2) * 2);
      do_fft (analysis_buffer + analysis_length, analysis_length);
      gdam_buffer_append (buffer,
			  analysis_buffer + analysis_length,
			  (analysis_length / 2) * 2);
    }
  filter->cur_amount = 0;
}

static int  process_one(GdasXmmsVisFilter*       xmms_filter,
                        GdamSample*              buffer,
			int                      n)
{
  int max_to_fill = n;
  int analysis_length = xmms_filter->analysis_length;
  if (max_to_fill + xmms_filter->cur_amount > analysis_length)
    {
      max_to_fill = analysis_length - xmms_filter->cur_amount;
    }
  gdam_sample_copy (xmms_filter->lr_analysis_buffer
		    + xmms_filter->cur_amount, buffer, max_to_fill);
  xmms_filter->cur_amount += max_to_fill;
  if (xmms_filter->cur_amount == analysis_length)
    flush_xmms_buffer (xmms_filter);
  return max_to_fill;
}

static int  gdas_xmms_vis_read(GdasSource*       source,
                               GdamSample*       buffer,
			       int               max_read)
{
  GdasXmmsVisFilter *xmms_filter;
  int num_read;
  int remaining;
  GdamSample *at;
  xmms_filter = GDAS_XMMS_VIS_FILTER (source);
  num_read = gdas_filter_sub_read (GDAS_FILTER (source), buffer, max_read);

  if (num_read <= 0)
    return num_read;

  remaining = num_read;
  at = buffer;
  while (remaining > 0)
    {
      int processed;
      processed = process_one (xmms_filter, at, remaining);
      remaining -= processed;
    }
  return num_read;
}

/*
 * outgoing data source
 */
static gboolean gdas_xmms_vis_filter_prepare  (gpointer  source_data, 
			                       GTimeVal *current_time,
			                       gint     *timeout,
			                       gpointer  user_data)
{
  GdasXmmsVisFilter *filter = user_data;
  if (filter->is_connecting)
    filter->poll_fd.events = G_IO_IN | G_IO_OUT | G_IO_ERR;
  else if (filter->outgoing_buffer.size > 0)
    filter->poll_fd.events = G_IO_OUT;
  else
    filter->poll_fd.events = 0;
  return FALSE;
}
static gboolean gdas_xmms_vis_filter_check    (gpointer  source_data,
			                       GTimeVal *current_time,
			                       gpointer  user_data)
{
  GdasXmmsVisFilter *filter = user_data;
  return (filter->poll_fd.revents != 0);
}
static gboolean gdas_xmms_vis_filter_dispatch (gpointer  source_data, 
			                       GTimeVal *current_time,
			                       gpointer  user_data)
{
  GdasXmmsVisFilter *filter = user_data;
  int fd = filter->fd;
  if (filter->is_connecting)
    {
      if (!gdam_connect_finish (fd, TRUE))
	goto error;
      filter->is_connecting = FALSE;
    }
  else
    {
      if (!gdam_buffer_write_one (&filter->outgoing_buffer, fd))
	goto error;
    }
  return TRUE;

error:
  g_warning ("error processing xmms-vis-filter connection");
  close (filter->fd);
  filter->fd = -1;
  g_main_remove_poll (&filter->poll_fd);
  filter->source_id = 0;
  return FALSE;
}

static GSourceFuncs xmms_source_funcs = {
	gdas_xmms_vis_filter_prepare,
	gdas_xmms_vis_filter_check,
	gdas_xmms_vis_filter_dispatch,
	NULL
};

static GdamConnectResult finish_connect(GdasXmmsVisFilter* xmms_filter,
              				GdamConnectResult  result)
{
  switch (result)
    {
    case GDAM_CONNECT_CONNECTING:
      xmms_filter->is_connecting = TRUE;
      break;
    case GDAM_CONNECT_CONNECTED:
      xmms_filter->is_connecting = FALSE;
      break;
    case GDAM_CONNECT_FAILED:
      xmms_filter->fd = -1;
      return FALSE;
    }
  xmms_filter->poll_fd.fd = xmms_filter->fd;
  g_main_add_poll (&xmms_filter->poll_fd, G_XMMS_PRIORITY);
  xmms_filter->source_id = g_source_add (G_XMMS_PRIORITY,
					 FALSE,
					 &xmms_source_funcs,
					 NULL, xmms_filter, NULL);
}
gboolean gdas_xmms_vis_filter_connect_unix(GdasXmmsVisFilter* xmms_filter,
                                           const char*        path)
{
  GdamConnectResult result;
  if (xmms_filter->fd >= 0)
    {
      gdam_debug ("cannot connect twice in GdasXmmsVisFilter");
      return FALSE;
    }
  result = gdam_connect_unix (path, &xmms_filter->fd);
  return finish_connect (xmms_filter, result);

}
gboolean gdas_xmms_vis_filter_connect_tcp (GdasXmmsVisFilter* xmms_filter,
                                           const char*        host,
					   int                port)
{
  GdamConnectResult result;
  if (xmms_filter->fd >= 0)
    {
      gdam_debug ("cannot connect twice in GdasXmmsVisFilter");
      return FALSE;
    }
  result = gdam_connect_tcp (host, port, &xmms_filter->fd);
  return finish_connect (xmms_filter, result);
}

static void gdas_xmms_vis_filter_func_connect(GtkObject*        object,
                                              int               num_args_in,
                                              GdamArg*          args_in,
                                              int*              num_args_out,
                                              GdamArg**         args_out)
{
  int port = -1;
  const char *host = NULL;
  const char *path = NULL;
  int i;
  gboolean success;
  GdasXmmsVisFilter *xmms_filter;
  for (i = 0; i < num_args_in; i++)
    {
      GdamArg *arg = args_in + i;
      if (strcmp (arg->name, "path") == 0
	  || arg->type == GDAM_ARG_TYPE_STRING)
	{
	  path = GDAM_ARG_STRING (arg);
	  continue;
	}
      if (strcmp (arg->name, "host") == 0
	  || arg->type == GDAM_ARG_TYPE_STRING)
	{
	  host = GDAM_ARG_STRING (arg);
	  continue;
	}
      if (strcmp (arg->name, "port") == 0 || arg->type == GDAM_ARG_TYPE_INT)
	{
	  port = GDAM_ARG_INT (arg);
	  continue;
	}
    }
  success = FALSE;
  xmms_filter = GDAS_XMMS_VIS_FILTER (object);
  if (host != NULL && port >= 0)
    {
      /* tcp connect style */
      if (path != NULL)
	goto only_one_method_may_be_specified;
      success = gdas_xmms_vis_filter_connect_tcp (xmms_filter, host, port);
    }
  else if (path != NULL)
    {
      /* unix connect style */
      if (host != NULL || port >= 0)
	goto only_one_method_may_be_specified;
      success = gdas_xmms_vis_filter_connect_unix (xmms_filter, path);
    }
  else
    {
      gdam_debug ("neither tcp or unix connect method was specified");
    }

  *num_args_out = 1;
  *args_out = gdam_arg_new_from_valist (1, "success", GDAM_ARG_TYPE_INT,
					(int) (success ? 1 : 0));
  return;

only_one_method_may_be_specified:
  gdam_debug ("host/port and path are mutually exclusive");
  return;
}
static void gdas_xmms_vis_filter_destroy(GtkObject* object)
{
  GdasXmmsVisFilter *xmms;
  xmms = GDAS_XMMS_VIS_FILTER (object);

  /* this is probably wrong -- wrong if the source keeps a ref count */
  if (xmms->source_id != 0)
    {
      g_source_remove (xmms->source_id);
      g_main_remove_poll (&xmms->poll_fd);
      xmms->source_id = 0;
    }
  if (xmms->fd >= 0)
    {
      close (xmms->fd);
      xmms->fd = -1;
    }
  (*parent_class->destroy) (object);
}

static void gdas_xmms_vis_filter_class_init(GdasXmmsVisFilterClass* klass)
{
  GtkObjectClass *object_class;
  GdasSourceClass *source_class;
  object_class = GTK_OBJECT_CLASS (klass);
  source_class = GDAS_SOURCE_CLASS (klass);
  object_class->destroy = gdas_xmms_vis_filter_destroy;
  gdam_arg_function_register (object_class, "connect",
			      gdas_xmms_vis_filter_func_connect);
  source_class->read = gdas_xmms_vis_read;
}

GtkType gdas_xmms_vis_filter_get_type()
{
  static GtkType xmms_vis_filter_type = 0;
  if (!xmms_vis_filter_type)
    {
      static const GtkTypeInfo xmms_vis_filter_info = {
	"GdasXmmsVisFilter",
	sizeof (GdasXmmsVisFilter),
	sizeof (GdasXmmsVisFilterClass),
	(GtkClassInitFunc) gdas_xmms_vis_filter_class_init,
	(GtkObjectInitFunc) gdas_xmms_vis_filter_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      parent_class = gtk_type_class (parent);
      xmms_vis_filter_type = gtk_type_unique (parent, &xmms_vis_filter_info);
    }
  return xmms_vis_filter_type;
}
