/* GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdambufferviewport.h"
#include <gdam/common/gdam-protocol.h>
#include <gdam/common/gdamsample.h>
#include <gdam/component/gdambuffermanager.h>
#include <gdam/component/gdamutil.h>
#include <gtk/gtkmain.h>
#include "gdamskinutil.h"
#include <gtk/gtkeventbox.h>
#include <math.h>
#include <stdlib.h>
#include <gdk/gdktypes.h>
#include <gdk/gdkkeysyms.h>

enum {
	SELECTION_CHANGED,
	VIEW_CHANGED,
	BEAT_LINES_CHANGED,
	LENGTH_CHANGED,
	BUFFER_CHANGED,
	LAST_SIGNAL
};
#define MIN_BEAT_PERIOD			(0.01)

static void gdam_buffer_viewport_real_beat_lines_changed
                                              (GdamBufferViewport* viewport);

static void  gdam_buffer_viewport_view_changed(GdamBufferViewport* viewport);
gdouble gdam_buffer_viewport_get_sample_length(GdamBufferViewport* viewport);
void       gdam_buffer_viewport_set_length    (GdamBufferViewport* viewport,
                                               gdouble             length);
static void maybe_update_cursor 	      (GdamBufferViewport* viewport,
					       guint               state);
static guint viewport_signals[LAST_SIGNAL] = {0};
static void gdam_buffer_viewport_destroy(GtkObject* object);
static GtkObjectClass* parent_class = NULL;

static void gdam_buffer_viewport_init(GdamBufferViewport* buffer_viewport)
{
  buffer_viewport->draw_width = 256;
  buffer_viewport->draw_height = 256;
  buffer_viewport->time_start = 0.0;
  buffer_viewport->time_length = 1.0;
  buffer_viewport->has_selection = FALSE;
  buffer_viewport->selection_start = 0.0;
  buffer_viewport->selection_length = 1.0;
  buffer_viewport->is_selectable = TRUE;
  buffer_viewport->marker_format = g_strdup ("%5.5g");
  buffer_viewport->marker_unit = 1.00;
  buffer_viewport->use_buffer_manager = TRUE;
  buffer_viewport->sample_length = 0.0;
  buffer_viewport->samples_per_second = 44100;
  buffer_viewport->density = 5;
  buffer_viewport->beat_drag_index = -1;
  buffer_viewport->beat_radius = 3;
}

static void zoom_changed(GtkAdjustment*      adjustment,
                         GdamBufferViewport* viewport)
{
  gdouble zoom = exp (adjustment->value);
  if (zoom > 10000.0)
    {
      return;
    }
  gdam_buffer_viewport_set_zoom (viewport, zoom);
  return;
}
static void viewport_zoom_changed(GdamBufferViewport *viewport,
                                  GtkAdjustment      *adjustment)
{
  gdouble zoom = viewport->time_length;
  if (zoom < 0.0001)
    return;
  gtk_adjustment_set_value (adjustment, log (zoom));
}

static gboolean gdam_buffer_viewport_log_zoom_adjustment_handler
                                (GdamSkin          *skin,
                                 GtkObject         *object,
                                 const char        *signal_name,
                                 int                num_handler_names,
                                 char             **handler_names)
{
  GtkAdjustment *adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed", GTK_SIGNAL_FUNC (zoom_changed), skin);
  gtk_signal_connect_while_alive
    (GTK_OBJECT (skin), "view_changed", 
     GTK_SIGNAL_FUNC (viewport_zoom_changed), adj, GTK_OBJECT (adj));
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static void center_changed(GtkAdjustment      *adjustment,
                           GdamBufferViewport *viewport)
{
  gdam_buffer_viewport_set_center (viewport, adjustment->value);
}

static void viewport_center_changed(GdamBufferViewport *viewport,
                                    GtkAdjustment      *adjustment)
{
  /* TODO:  what's with use_buffer_manager?  Need this be a special case */
  //if (viewport->use_buffer_manager){
  //g_message("returning from center_changed, using bufman");
  //return;
  //}
  adjustment->page_size = viewport->time_length;
  if (viewport->time_length < viewport->sample_length)
    {
      adjustment->page_increment = viewport->time_length / 10.0;
      adjustment->step_increment = viewport->time_length / 100.0;
    }
  else
    {
      adjustment->page_increment = viewport->sample_length / 10.0;
      adjustment->step_increment = viewport->sample_length / 100.0;
    }
  adjustment->upper = viewport->sample_length + adjustment->page_size;
  adjustment->value = viewport->time_start + viewport->time_length / 2;
  gtk_adjustment_changed (adjustment);
}

static gboolean gdam_buffer_viewport_center_adjustment_handler
                                (GdamSkin          *skin,
                                 GtkObject         *object,
                                 const char        *signal_name,
                                 int                num_handler_names,
                                 char             **handler_names)
{
  GtkAdjustment *adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (center_changed), skin);
  gtk_signal_connect_while_alive
    (GTK_OBJECT (skin),
     "length_changed", GTK_SIGNAL_FUNC (viewport_center_changed), 
     adj, GTK_OBJECT (adj));
  gtk_signal_connect_while_alive
    (GTK_OBJECT (skin),
     "view_changed", GTK_SIGNAL_FUNC (viewport_center_changed), 
     adj, GTK_OBJECT (adj));
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static void on_buffer_id_received (const char       *text,
				   gpointer          data)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (data);
  int id = atoi (text);
  gdam_debug ("buffer id %d dropped", id);
  gdam_buffer_viewport_set_buffer_id (viewport, id);
}

static gboolean gdam_buffer_viewport_buffer_id_dropped_handler
                                (GdamSkin          *skin,
                                 GtkObject         *object,
                                 const char        *signal_name,
                                 int                num_handler_names,
                                 char             **handler_names)
{
  gdam_widget_string_dest_set (GTK_WIDGET (object),
			       on_buffer_id_received,
			       "gdam-buffer-id-text", skin);
  return TRUE;
}


static void gdam_buffer_viewport_zoom_to_selection(GdamBufferViewport *viewport)
{
  if (viewport->has_selection)
    {
      viewport->time_start = viewport->selection_start;
      viewport->time_length = viewport->selection_length;
      gdam_buffer_viewport_changed (viewport);
    }
}

static void gdam_buffer_viewport_select_all (GdamBufferViewport *viewport)
{
  viewport->selection_start = viewport->time_start;
  viewport->selection_length = viewport->time_length;
  gdam_buffer_viewport_queue_redraw (viewport);
}

static GdamSkinGenericConnect connect_infos[] = {
	{ "gdam_buffer_viewport_log_zoom_adjustment",
           gdam_buffer_viewport_log_zoom_adjustment_handler	},
	{ "gdam_buffer_viewport_center_adjustment",
           gdam_buffer_viewport_center_adjustment_handler	},
	{ "gdam_buffer_viewport_buffer_id_dropped",
           gdam_buffer_viewport_buffer_id_dropped_handler	},
};

static GdamSkinSimpleConnect simple_connect_infos[] = {
	{ "gdam_buffer_viewport_whole_sample",
           (GdamSkinFunc)gdam_buffer_viewport_whole_sample },
	{ "gdam_buffer_viewport_zoom_to_selection",
           (GdamSkinFunc)gdam_buffer_viewport_zoom_to_selection },
	{ "gdam_buffer_viewport_select_all",
           (GdamSkinFunc)gdam_buffer_viewport_select_all },
};

static void 
arg_get_zoom (GtkObject *object, GdamArg *arg)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (object);
  GDAM_ARG_DOUBLE (arg) = viewport->time_length;
}

static void 
arg_set_zoom (GtkObject *object, GdamArg *arg)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (object);
  gdouble zoom = GDAM_ARG_DOUBLE (arg);
  gdouble center;
  center = viewport->time_start + viewport->time_length / 2.0;
  viewport->time_start = center - zoom / 2.0;
  viewport->time_length = zoom;
  gdam_buffer_viewport_changed (viewport);
}

static void 
arg_get_time_start (GtkObject *object, GdamArg *arg)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (object);
  GDAM_ARG_DOUBLE (arg) = viewport->time_start;
}

static void 
arg_set_time_start (GtkObject *object, GdamArg *arg)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (object);
  viewport->time_start = GDAM_ARG_DOUBLE (arg);
  gdam_buffer_viewport_changed (viewport);
}

static void 
arg_set_nudge_time_start (GtkObject *object, GdamArg *arg)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (object);
  viewport->time_start += GDAM_ARG_DOUBLE (arg) * viewport->time_length;
  gdam_buffer_viewport_changed (viewport);
}

static void gdam_buffer_viewport_class_init(GdamBufferViewportClass* klass)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  GdamSkinClass *skin_class = GDAM_SKIN_CLASS (klass);
  GtkType type = object_class->type;
  object_class->destroy = gdam_buffer_viewport_destroy;
  parent_class = gtk_type_class (GDAM_TYPE_SKIN);
  klass->beat_lines_changed = gdam_buffer_viewport_real_beat_lines_changed;
  viewport_signals[SELECTION_CHANGED]
    = gtk_signal_new ("selection_changed",
		      GTK_RUN_FIRST,
		      type,
		      GTK_SIGNAL_OFFSET (GdamBufferViewportClass,
					 selection_changed),
		      gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  viewport_signals[VIEW_CHANGED]
    = gtk_signal_new ("view_changed",
		      GTK_RUN_FIRST,
		      type,
		      GTK_SIGNAL_OFFSET (GdamBufferViewportClass,
					 view_changed),
		      gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  viewport_signals[BEAT_LINES_CHANGED]
    = gtk_signal_new ("beat_lines_changed",
		      GTK_RUN_FIRST,
		      type,
		      GTK_SIGNAL_OFFSET (GdamBufferViewportClass,
					 beat_lines_changed),
		      gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  viewport_signals[LENGTH_CHANGED]
    = gtk_signal_new ("length_changed",
		      GTK_RUN_FIRST,
		      type,
		      GTK_SIGNAL_OFFSET (GdamBufferViewportClass,
					 length_changed),
		      gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  viewport_signals[BUFFER_CHANGED]
    = gtk_signal_new ("buffer_changed",
		      GTK_RUN_FIRST,
		      type,
		      GTK_SIGNAL_OFFSET (GdamBufferViewportClass,
					 buffer_changed),
		      gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class, viewport_signals, LAST_SIGNAL);
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
  gdam_skin_class_register_simple_connections 
    				      (skin_class,
				       G_N_ELEMENTS (simple_connect_infos),
				       simple_connect_infos);
  gdam_arg_register (object_class,
		     "zoom",
		     GDAM_ARG_TYPE_DOUBLE,
		     arg_set_zoom,
		     arg_get_zoom,
		     NULL, "0.01,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "time_start",
		     GDAM_ARG_TYPE_DOUBLE,
		     arg_set_time_start,
		     arg_get_time_start,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "nudge_time_start",
		     GDAM_ARG_TYPE_DOUBLE,
		     arg_set_nudge_time_start,
		     NULL,
		     NULL, NULL, 0);
  klass->view_changed = gdam_buffer_viewport_view_changed;
}

/*=====================================================================*
 *                           get_x_from_time                           *
 *                                                                     *
 * Parameters:                                                         *
 *     viewport          the viewport with the coordinate frame.       *
 *     time              the time in seconds to convert to screen      *
 *                       coordinates.                                  *
 *=====================================================================*/
static inline int     get_x_from_time(GdamBufferViewport* viewport, 
                                      gdouble             time)
{
  time -= viewport->time_start;
  if (viewport->time_length > 0.0001)
    time /= viewport->time_length;
  time *= viewport->draw_width;
  return (int) time;
}

/*=====================================================================*
 *                           get_time_from_x                           *
 *                                                                     *
 * Parameters:                                                         *
 *     viewport          the viewport.                                 *
 *     x                 the screen coordinate (in the window)         *
 *     return            the time into the sample in seconds.          *
 *=====================================================================*/
static inline gdouble get_time_from_x(GdamBufferViewport* viewport, 
                                      int                 screen_x)
{
  gdouble time = screen_x;
  time /= viewport->draw_width;
  time *= viewport->time_length;
  time += viewport->time_start;
  return time;
}

GtkType gdam_buffer_viewport_get_type()
{
  static GtkType buffer_viewport_type = 0;
  if (!buffer_viewport_type)
    {
      static const GtkTypeInfo buffer_viewport_info = {
	"GdamBufferViewport",
	sizeof (GdamBufferViewport),
	sizeof (GdamBufferViewportClass),
	(GtkClassInitFunc) gdam_buffer_viewport_class_init,
	(GtkObjectInitFunc) gdam_buffer_viewport_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      buffer_viewport_type = gtk_type_unique (parent, &buffer_viewport_info);
    }
  return buffer_viewport_type;
}
GdamBufferViewport* gdam_buffer_viewport_new_shown(GdamChannel* channel)
{
  GdamBufferViewport *view;
  GtkType type = GDAM_TYPE_BUFFER_VIEWPORT;
  view = GDAM_BUFFER_VIEWPORT (gtk_type_new (type));
  GDAM_SKIN (view)->channel = channel;
  gdam_buffer_viewport_new_event_box (view);
  gtk_widget_show (GDAM_SKIN (view)->main_widget);
  return view;
}

static void update_buffer_info(GdamBufferManagerInfo*        info,
                               gpointer                      viewport_ptr)
{
  GdamBufferViewport *viewport;
  gdouble length;
  viewport = GDAM_BUFFER_VIEWPORT (viewport_ptr);
  length = (gdouble) info->length / viewport->samples_per_second;
  gdam_buffer_viewport_set_length (viewport, length);
}

void       gdam_buffer_viewport_set_buffer_id (GdamBufferViewport* viewport,
                                               guint               buffer_id)
{
  gdouble length;
  viewport->buffer_id = buffer_id;
  length = gdam_buffer_viewport_get_sample_length (viewport);
  gdam_buffer_viewport_set_length (viewport, length);
  gdam_buffer_viewport_view_changed (viewport);
  if (viewport->use_buffer_manager)
    {
      if (viewport->connection_id == 0)
	viewport->connection_id
	  = gdam_buffer_manager_connect (buffer_id,
					 update_buffer_info, viewport);
      else
	gdam_buffer_manager_reconnect (viewport->connection_id, buffer_id);
    }
  gtk_signal_emit (GTK_OBJECT (viewport), viewport_signals[BUFFER_CHANGED]);
}

void       gdam_buffer_viewport_set_view      (GdamBufferViewport* viewport,
                                               gdouble             start,
					       gdouble             length)
{
  if (viewport->time_start == start && viewport->time_length == length)
    return;
  viewport->time_start = start;
  viewport->time_length = length;
  gdam_debug ("gdam_buffer_viewport_set_view: start=%4.4g, length=%4.4g",
	      start, length);

  /* Is this okay? */
  gdam_buffer_viewport_changed (viewport);
}


gdouble gdam_buffer_viewport_get_sample_length(GdamBufferViewport* viewport)
{
  if (viewport->use_buffer_manager)
    {
      GdamBufferManagerInfo *info;
      info = gdam_buffer_manager_info_by_id (viewport->buffer_id);
      if (info == NULL)
	{
	  g_warning ("buffer_viewport: manager has no info "
		     "on buffer %d", viewport->buffer_id);
	  return 0;
	}
      return (gdouble) info->length / viewport->samples_per_second;
    }
  else
    {
      return (gdouble) viewport->sample_length;
    }
}

void gdam_buffer_viewport_remove_beat_times(GdamBufferViewport* viewport)
{
  if (viewport->num_beat_times == 0)
    return;

  if (viewport->beat_times != NULL)
    {
      g_free (viewport->beat_times);
      viewport->beat_times = NULL;
    }
  gdam_buffer_viewport_queue_redraw (viewport);
}

void
delete_nth_beat_time (GdamBufferViewport *viewport, int n)
{
  gdouble *beat_times;
  int i;
  g_return_if_fail (n < viewport->num_beat_times);
  beat_times = g_new (gdouble, viewport->num_beat_times - 1);
  for (i = 0; i < n; i++)
    {
      beat_times[i] = viewport->beat_times[i];
    }
  for (i = n; i < viewport->num_beat_times - 1; i++)
    {
      beat_times[i] = viewport->beat_times[i+1];
    }
  g_free (viewport->beat_times);
  viewport->beat_times = beat_times;
  viewport->num_beat_times--;
  gdam_buffer_viewport_beat_lines_changed (viewport);
}

void       gdam_buffer_viewport_set_beat_times(GdamBufferViewport* viewport,
                                               int                 num_times,
					       const gdouble*      times)
{
  if (viewport->num_beat_times > 0)
    g_free (viewport->beat_times);

  viewport->num_beat_times = num_times;
  viewport->beat_times = g_memdup (times, sizeof (gdouble) * num_times);

  gdam_buffer_viewport_beat_lines_changed (viewport);
  gdam_buffer_viewport_queue_redraw (viewport);
}

void gdam_buffer_viewport_selection_changed(GdamBufferViewport* viewport)
{
  GtkObject *object = GTK_OBJECT (viewport);
  if (gtk_signal_n_emissions (object, viewport_signals[SELECTION_CHANGED]))
    return;
  gtk_signal_emit (object, viewport_signals[SELECTION_CHANGED]);
}

void       gdam_buffer_viewport_set_selection (GdamBufferViewport* viewport,
                                               gdouble             start,
					       gdouble             length)
{
  if (start < 0 && length < 0)
    {
      viewport->has_selection = FALSE;
      viewport->selection_start = 0;
      viewport->selection_length = 0;
    }
  else
    {
      viewport->has_selection = TRUE;
      viewport->selection_start = start;
      viewport->selection_length = length;
    }
  gdam_buffer_viewport_selection_changed (viewport);
  gdam_buffer_viewport_queue_redraw (viewport);
}

/*=====================================================================*
 *                  gdam_buffer_viewport_whole_sample                  *
 *                                                                     *
 * Snap the window to the exact length of the sample.                  *
 *=====================================================================*/
void gdam_buffer_viewport_whole_sample  (GdamBufferViewport* viewport)
{
  gdouble length;

  if (viewport->buffer_id == 0)
    return;

  length = gdam_buffer_viewport_get_sample_length (viewport);
  if (length < 0.0001)
    return;
  gdam_buffer_viewport_set_view (viewport, 0.0, length);
}

void       gdam_buffer_viewport_buffer_changed(GdamBufferViewport* viewport)
{
  gdam_buffer_viewport_view_changed (viewport);
}

/*=====================================================================*
 *                             got_samples                             *
 *                                                                     *
 * Callback for when the server responds to our request for samples.   *
 *=====================================================================*/
static void got_samples(int num_samples, int* samples, gpointer user_data)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (user_data);
  gboolean must_redo = viewport->redo_query;
  g_free (viewport->min_last_samples);
  g_free (viewport->max_last_samples);
  viewport->getting_profile = FALSE;
  viewport->redo_query = FALSE;
  if (viewport->density <= 1)
    {
      int i;
      viewport->min_last_samples = samples;
      viewport->max_last_samples = g_new0 (gint, num_samples);
      viewport->num_last_samples = num_samples;
      for (i = 0; i < num_samples; i++)
	{
	  int in = viewport->min_last_samples[i];
	  if (in > 0)
	    {
	      viewport->max_last_samples[i] = in;
	      viewport->min_last_samples[i] = 0;
	    }
	}
    }
  else
    {
      int density = viewport->density, i;
      int *at = samples;
      int num_out = num_samples / density;
      viewport->min_last_samples = g_new (gint, num_out);
      viewport->max_last_samples = g_new (gint, num_out);
      viewport->num_last_samples = num_out;
      for (i = 0; i < num_out; i++)
	{
	  int minimum, maximum, j;
	  minimum = maximum = *at++;
	  for (j = 1; j < density; j++)
	    {
	      if (*at > maximum)
		maximum = *at;
	      else if (*at < minimum)
		minimum = *at;
	      at++;
	    }
	  viewport->min_last_samples[i] = minimum;
	  viewport->max_last_samples[i] = maximum;
	}
    }
  if (must_redo)
    gdam_buffer_viewport_view_changed (viewport);
  else
    gdam_buffer_viewport_queue_redraw (viewport);
  gtk_object_unref (GTK_OBJECT (viewport));
}

static void failed_to_get_samples(gpointer user_data)
{
  g_warning ("failed to get buffer's samples");
  gtk_object_unref (GTK_OBJECT (user_data));
}

/*=====================================================================*
 *                       compute_marker_locations                      *
 *                                                                     *
 * Compute where the conveniently-spaced                               *
 * markers should be drawn.                                            *
 *=====================================================================*/
static void compute_marker_locations(GdamBufferViewport* viewport)
{
  int i;
  gdouble num_markers;
  gdouble marker_skip;
  gdouble start_in_marks, end_in_marks;
  if (viewport->markers != NULL)
    {
      for (i = 0; i < viewport->num_markers; i++)
	g_free (viewport->markers[i].text);
      g_free (viewport->markers);
      viewport->markers = NULL;
    }

  if (viewport->marker_font == NULL)
    return;

  num_markers = viewport->time_length / viewport->marker_unit;
  marker_skip = 0.001;
  while (num_markers / marker_skip > 20)
    {
      marker_skip *= 2;
      if (num_markers / marker_skip <= 20)
	break;
      marker_skip *= 5;
    }

  /* In even markers * marker_skip. */
  start_in_marks = viewport->time_start / viewport->marker_unit;
  end_in_marks = start_in_marks + num_markers;
  start_in_marks = ceil (start_in_marks / marker_skip);
  end_in_marks = floor (end_in_marks / marker_skip);

  if (start_in_marks > end_in_marks)
    return;

  viewport->num_markers = (int) (end_in_marks - start_in_marks + 1);
  viewport->markers = g_new (GdamBufferViewportMarker, viewport->num_markers);
  for (i = 0; i < viewport->num_markers; i++)
    {
      GdamBufferViewportMarker *mark = viewport->markers + i;
      gdouble mark_number = (start_in_marks + i) * marker_skip;
      gdouble t = viewport->marker_unit * mark_number;
      const char *fmt;
      int cent_x;
      int text_width;
      char *str;
      t -= viewport->time_start;
      cent_x = t * viewport->draw_width / viewport->time_length;
      if (marker_skip < 0.01)
	fmt = "%7.03f";
      else if (marker_skip < 0.1)
	fmt = "%6.02f";
      else
	fmt = "%5.01f";
      str = g_strdup_printf (fmt, (gdouble) mark_number);
      text_width = gdk_string_width (viewport->marker_font, str);
      mark->draw_x = cent_x - text_width / 2;
      mark->draw_y = viewport->draw_height - 5;
      mark->draw_width = text_width;
      mark->text = str;
    }
}

static GdamProtocolProfileMode map_profile_mode(GdamBufferViewportMode mode)
{
  switch (mode)
    {
    case GDAM_BUFFER_VIEWPORT_LEFT:
      return GDAM_PROTOCOL_PROFILE_LEFT;
    case GDAM_BUFFER_VIEWPORT_RIGHT:
      return GDAM_PROTOCOL_PROFILE_RIGHT;
    case GDAM_BUFFER_VIEWPORT_AVERAGE:
    default:
      return GDAM_PROTOCOL_PROFILE_AVERAGE;
    }
}

static void  gdam_buffer_viewport_view_changed(GdamBufferViewport* viewport)
{
  GdamRequest *request;
  int samps_per_sec, num_points;
  double start, length;
  compute_marker_locations (viewport);
  if (viewport->buffer_id == 0)
    {
      gdam_buffer_viewport_queue_redraw (viewport);
      return;
    }
  if (viewport->getting_profile)
    {
      viewport->redo_query = TRUE;
      return;
    }
  gtk_object_ref (GTK_OBJECT (viewport));
  viewport->getting_profile = TRUE;
  samps_per_sec = viewport->samples_per_second;

  start = viewport->time_start;
  num_points = viewport->draw_width * viewport->density;
  length = viewport->time_length;

  /* assumes that start + length > 0 */
  if (start < 0)
    {
      double percent_neg = start / length;
      double percent_pos = 1.0 - percent_neg;
      num_points *= percent_pos;
      length *= percent_pos;
      start = 0;
    }
  request = gdam_request_get_buffer_profile
    (viewport->buffer_id,
     map_profile_mode (viewport->profile_mode),
     (int) (start * samps_per_sec),
     (int) (length * samps_per_sec),
     num_points, got_samples, failed_to_get_samples, viewport);
  gdam_channel_append_request (GDAM_SKIN (viewport)->channel, request);
}


/*=====================================================================*
 *                             get_segments                            *
 *                                                                     *
 * Get the currently visible GdkSegments.                              *
 *                                                                     *
 * Parameters:                                                         *
 *     viewport          the viewport.                                 *
 *     x                 the leftmost pixel to render                  *
 *     width             width of the area to render in pixels.        *
 *     fg_segments       the return buffer of "foreground" segments    *
 *                       [at most 2 * width]                           *
 *     num_fg_segments   return address for number of foreground       *
 *                       segments.                                     *
 *     bg_segments       the return buffer of "background" segments    *
 *                       [at most 2 * width]                           *
 *     num_bg_segments   return address for number of background       *
 *                       segments.                                     *
 *=====================================================================*/
static void get_segments(GdamBufferViewport* viewport, int x, int width,
                         GdkSegment* fg_segments, int* num_fg_segments,
                         GdkSegment* bg_segments, int* num_bg_segments)
{
  gdouble sel_start, sel_end;
  gboolean has_sel;
  int draw_height, center_y, i;
  int *min_samples = viewport->min_last_samples;
  int *max_samples = viewport->max_last_samples;
  int num_samples = viewport->num_last_samples;
  int seg_count = 0;
  int neg_count = 0;
  int pos_count = 0;
  int post_count = 0;
  int origin_offset;
  gdouble time_per_pixel;
  gdouble time_start, time_width;
  time_per_pixel = viewport->time_length / viewport->draw_width;
  time_start = viewport->time_start;
  time_width = time_per_pixel * width;
  //g_message("get_segments: x=%d, w=%d  ts: %f, tw: %f", x, width,
  //time_start+time_per_pixel*x, 
  //time_start+time_per_pixel*width);
  has_sel = viewport->has_selection;
  sel_start = viewport->selection_start;
  sel_end = viewport->selection_start + viewport->selection_length;
  *num_fg_segments = *num_bg_segments = 0;
  draw_height = viewport->draw_height;
  center_y = draw_height / 2;
  origin_offset = 0;

  g_return_if_fail (x < viewport->draw_width);
  g_return_if_fail (width >= 0);
  g_return_if_fail (x >= 0);

  /*  If we're not drawing form left edge and part of display is negative...  */
  if (x > 0 && time_start < 0)
    {
      origin_offset =
	(-time_start / viewport->time_length) * viewport->draw_width;
      origin_offset = x - origin_offset;
      //g_message("origin offset by %d", origin_offset);
    }
  /*  If we're not drawing from left edge */
  else if (x > 0)
    {
      origin_offset = x;
    }


  /* Draw the visible portion of the song, one time unit at a time. */
  for (i = x; i < x + width; i++)
    {
      gdouble cur = time_start + time_per_pixel * i;
      GdkSegment *outer;
      GdkSegment *inner;
      int min_sample_y, max_sample_y;

      /* If the 'cur' time is within the selection, invert the colors. */
      if (has_sel && sel_start <= cur && cur < sel_end + time_per_pixel)
	{
	  outer = fg_segments + (*num_fg_segments);
	  inner = bg_segments + (*num_bg_segments);
	  (*num_bg_segments) += 1;
	  (*num_fg_segments) += 2;
	}
      else
	{
	  outer = bg_segments + (*num_bg_segments);
	  inner = fg_segments + (*num_fg_segments);
	  (*num_bg_segments) += 2;
	  (*num_fg_segments) += 1;
	}
      seg_count++;

      /* If this time is out of range, then don't display it. */
      if (cur < 0)
	{
	  neg_count++;
	  min_sample_y = 0;
	  max_sample_y = 0;
	}
      else if (cur > viewport->sample_length)
	{
	  post_count++;
	  min_sample_y = 0;
	  max_sample_y = 0;
	}
      else
	{
	  /*  hmm.. perhaps these should add up, but they dont */
	  //int j = seg_count - neg_count;
	  //g_return_if_fail(j < 0);
	  int j;
	  pos_count++;
	  j = pos_count + origin_offset;
	  //g_return_if_fail(j>=0);
	  //g_return_if_fail(j<num_samples);
	  if ((j < 0) || (j >= num_samples))
	    {
	      /*  XXX: this shouldn't happen (i think) */
	      min_sample_y = 0;
	      max_sample_y = 0;
	    }
	  else
	    {
	      min_sample_y = center_y + (GdamSampleProduct) center_y
		* min_samples[j] / GDAM_SAMPLE_ONE;
	      max_sample_y = center_y + (GdamSampleProduct) center_y
		* max_samples[j] / GDAM_SAMPLE_ONE;
	    }
	}
      inner[0].x1 = i;
      inner[0].x2 = i;
      inner[0].y1 = min_sample_y;
      inner[0].y2 = max_sample_y;

      outer[0].x1 = outer[0].x2 = outer[1].x1 = outer[1].x2 = i;

      outer[0].y1 = 0;
      outer[0].y2 = min_sample_y - 1;
      outer[1].y1 = max_sample_y + 1;
      outer[1].y2 = draw_height;
    }
  //g_message("neg: %d, pos: %d, post: %d", neg_count, pos_count, 
  //post_count);
}

  /* XXX hackery to deal with bpm == 0, beat_period == inf */
#define GDAM_REALLY_BIG_BEAT_PERIOD 1000000000000

/*=====================================================================*
 *                           draw_beat_lines                           *
 *                                                                     *
 * Render beat lines in some horizontal range.                         *
 *                                                                     *
 * Parameters:                                                         *
 *     viewport          the buffer viewport.                          *
 *     window            the window on which to render.                *
 *     x                 the left-coordinate to render.                *
 *     width             the width (in pixels) to render.              *
 *=====================================================================*/
static void draw_beat_lines(GdamBufferViewport* viewport,
                            GdkWindow*          window, 
                            int                 x,
			    int                 width)
{
  gdouble time_per_pixel;
  gdouble time_start, time_width;
  gdouble view_start, view_width;
  gdouble first_beat, last_beat;
  gdouble b_period;  /* beat period hackery. */
  int which_beat, last_beat_index, num_lines;
  int i, right_clip;
  gdouble end_in_beats;
  GdkSegment *segments;

  which_beat = 0;
  b_period = 0;
  last_beat_index = 0;

  /* XXX: cope with / prevent beat_period = inf?? */
  if (viewport->num_beats == 0)
    return;

  time_per_pixel = viewport->time_length / viewport->draw_width;
  time_start = viewport->time_start;
  time_width = time_per_pixel * width;
  view_start = time_start + time_per_pixel * x;
  right_clip = viewport->draw_width - x - width;
  if (right_clip < 0)
    right_clip = 0;
  view_width = time_width - time_per_pixel * (x + right_clip);

  /*  If we have a beat period (and therefore bpm) 
   */
  if (viewport->beat_period > 0 
      && viewport->beat_period < GDAM_REALLY_BIG_BEAT_PERIOD)
    {
      b_period = viewport->beat_period;
      /*  Calculate which beats are visible, and the number of beat lines
       *  to draw.
       */
      end_in_beats = gdam_buffer_viewport_get_sample_length (viewport);
      end_in_beats -= viewport->start_beat;
      end_in_beats /= viewport->beat_period;
      first_beat = (view_start - viewport->start_beat) / viewport->beat_period;
      last_beat = first_beat + view_width / viewport->beat_period;
      if (last_beat > end_in_beats)
	last_beat = end_in_beats;
      /* Calculate the first beat which is visible. */
      which_beat = floor (first_beat);
      if (which_beat < 0)
	which_beat = 0;
      /* Calculate the last beat which is visible. */
      last_beat_index = (int) ceil (last_beat);
      if (last_beat_index > viewport->num_beats)
	last_beat_index = viewport->num_beats;
      num_lines = last_beat_index - which_beat;
    }
  /* If the bpm is zero but the first beat is visible, draw the first
   * beat. 
   */ 
  else if (view_start < viewport->start_beat 
           && viewport->start_beat < view_start + view_width)
    {
      b_period = 0.0;
      num_lines = 1;
      which_beat = 0;
      last_beat_index = 1;
    }
  else
    {
      num_lines = 0;
    }

  if (num_lines <= 0)
    return;

  segments = alloca (sizeof (GdkSegment) * num_lines);
  i = 0;
  for (; which_beat < last_beat_index; which_beat++)
    {
      gdouble t;
      gdouble time_offset;
      int x;
      t = viewport->start_beat + b_period * which_beat;
      time_offset = t - viewport->time_start;
      x = (int) (time_offset / viewport->time_length * viewport->draw_width);
      segments[i].x1 = x;
      segments[i].y1 = 0;
      segments[i].x2 = x;
      segments[i].y2 = viewport->draw_height;
      i++;
    }
  gdk_draw_segments (window, viewport->beat_line_gc, segments, i);
}

/* XXX: this should definitely be rewritten to use a binary-search. */
static void get_visible_beats(gdouble start_time, gdouble duration,
		              int num_beats, gdouble* beats,
			      int* first_beat_out, int* num_beats_out)
{
  int first, num;
  gdouble end_time = start_time + duration;
  for (first = 0; first < num_beats; first++)
    /* the test for 0.0 == 0.0 commonly fails without the tiny delta value */
    if (beats[first] + 0.000001 >= start_time)
      break;
  for (num = 0; num + first < num_beats; num++)
    if (beats[first + num] > end_time)
      break;
  *first_beat_out = first;
  *num_beats_out = num;
}

static void redraw_subset(GdamBufferViewport* viewport,
                          int x,
			  int width)
{
  GdkSegment *fg_segments;
  GdkSegment *bg_segments;
  int num_fg_segments, num_bg_segments;
  GtkWidget *widget = GDAM_SKIN (viewport)->main_widget;
  GdkWindow *window;

  //g_message("redraw_subset: x=%d, w=%d", x, width);

  g_return_if_fail (viewport != NULL);
  g_return_if_fail (widget != NULL);
  window = widget->window;
  g_return_if_fail (window != NULL);

  if (x > viewport->draw_width)
    return;

  if (x < 0)
    {
      width += x;
      x = 0;
    }
  if (width <= 0)
    return;
  /* Attempt to clean up after drag outline... led to seg fault. */
  if (viewport->is_dragging)
    width += 1;
  if (x + width > viewport->draw_width)
    width -= ((x + width) - viewport->draw_width);
  g_return_if_fail (width > 0);
  fg_segments = (GdkSegment *) alloca (sizeof (GdkSegment) * width * 2);
  bg_segments = (GdkSegment *) alloca (sizeof (GdkSegment) * width * 2);
  get_segments (viewport, x, width,
		fg_segments, &num_fg_segments, bg_segments, &num_bg_segments);
  gdk_draw_segments (window, viewport->bg_gc, bg_segments, num_bg_segments);
  draw_beat_lines (viewport, window, x, width);
  gdk_draw_segments (window, viewport->fg_gc, fg_segments, num_fg_segments);

  if (viewport->markers != NULL
      && viewport->marker_font != NULL && viewport->marker_gc != NULL)
    {
      int i;
      int num_markers = viewport->num_markers;
      for (i = 0; i < num_markers; i++)
	{
	  GdamBufferViewportMarker *mark = viewport->markers + i;
	  if (mark->draw_x >= x + width
	      || mark->draw_x + mark->draw_width <= x)
	    continue;
	  gdk_draw_string (window,
			   viewport->marker_font,
			   viewport->marker_gc,
			   mark->draw_x, mark->draw_y, mark->text);
	}
    }
  /* Draw beat markers from the beat_times array */
  if (viewport->num_beat_times > 0 && viewport->beat_gc != NULL)
    {
      int first_beat, num_beat;
      gdouble start_time = get_time_from_x (viewport, x);
      gdouble duration;
      duration = (gdouble) width *viewport->time_length
	/ viewport->draw_width;
      g_return_if_fail (viewport->beat_times != NULL);
      get_visible_beats (start_time,
			 duration,
			 viewport->num_beat_times,
			 viewport->beat_times, &first_beat, &num_beat);
      if (num_beat > 0)
	{
	  int i;
	  int radius = viewport->beat_radius;
	  int y = viewport->beat_radius * 4;
	  int drag_index = viewport->beat_drag_index;
	  for (i = 0; i < num_beat; i++)
	    {
	      int pos;
	      gdouble time;
	      gboolean fill = (i + first_beat != drag_index);
	      g_return_if_fail (viewport->beat_times != NULL);
	      time = viewport->beat_times[i + first_beat];
	      pos = get_x_from_time (viewport, time);
	      gdk_draw_arc (window,
			    viewport->beat_gc,
			    fill,
			    pos - radius + 1,
			    y, radius * 2 - 1, radius * 2 - 1, 0, 360 * 64);
	    }
	}
    }
}

static gboolean redraw_from_buffer_thread(gpointer user_data)
{
  int draw_width;
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (user_data);
  viewport->redraw_thread = 0;
  /* XXX: does this make sense when num_last is display subset? */
  draw_width = MIN (viewport->draw_width, viewport->num_last_samples);
  redraw_subset (viewport, 0, draw_width);
  gtk_object_unref (GTK_OBJECT (viewport));
  return FALSE;
}

void       gdam_buffer_viewport_queue_redraw  (GdamBufferViewport* viewport)
{
  if (viewport->redraw_thread != 0)
    return;

  gtk_object_ref (GTK_OBJECT (viewport));
  viewport->redraw_thread = g_idle_add (redraw_from_buffer_thread, viewport);
}

static gint buffer_viewport_expose         (GtkEventBox*        event_box,
                                            GdkEventExpose*     expose_event,
				            GdamBufferViewport* viewport)
{
  gdam_buffer_viewport_queue_redraw (viewport);
  return TRUE;
}

static gint buffer_viewport_configure      (GtkEventBox*        event_box,
                                            GdkEventConfigure*  config_event,
				            GdamBufferViewport* viewport)
{
  viewport->draw_width = config_event->width;
  viewport->draw_height = config_event->height;
  gdam_buffer_viewport_queue_redraw (viewport);
  return FALSE;
}

/* See if (x,y) falls on a beat marker from the beat_times array */
/* XXX: should definitely use a binary search!!!!!!! 
        it is called a *LOT* (once per motion) */
static int scan_for_beat_index(GdamBufferViewport*    buffer_viewport,
                               int                    x,
			       int                    y)
{
  int radius = buffer_viewport->beat_radius;
  int height = buffer_viewport->draw_height;
  int circle_y;
  int i;
  if (y > 8 * radius)
    return -1;

  circle_y = radius * 4;
  for (i = 0; i < buffer_viewport->num_beat_times; i++)
    {
      int circle_x, radius_squared;
      g_return_val_if_fail (buffer_viewport->beat_times != NULL, 0);
      circle_x = get_x_from_time (buffer_viewport,
				  buffer_viewport->beat_times[i]);
      radius_squared = (x - circle_x) * (x - circle_x)
	+ (y - circle_y) * (y - circle_y);
      if (radius_squared < (3 * radius * radius))
	return i;
    }
  return 0;
}

/* Derive beat positions from bpm, fb and find the beat closest to 'time' */
static int get_beat_line (GdamBufferViewport* viewport, gdouble time)
{
  gdouble beat;
  beat = (time - viewport->start_beat) / viewport->beat_period;
  return rint (beat);
}

static int
key_to_state (gint key)
{
  switch (key)
    {
      case GDK_Alt_L:
      case GDK_Alt_R:
	return GDK_MOD1_MASK;
      case GDK_Shift_L:
      case GDK_Shift_R:
	return GDK_SHIFT_MASK;
      case GDK_Control_L:
      case GDK_Control_R:
	return GDK_CONTROL_MASK;
      default:
	return 0;
    }
}

static gint buffer_viewport_key_press      (GtkEventBox*        event_box,
                                            GdkEventKey*        event,
				            GdamBufferViewport* viewport)
{
  guint state = 0;
  state = event->state;

  if (event->keyval == GDK_Alt_L     || event->keyval == GDK_Alt_R ||
      event->keyval == GDK_Shift_L   || event->keyval == GDK_Shift_R ||
      event->keyval == GDK_Control_L || event->keyval == GDK_Control_R)
    {
      state |= key_to_state (event->keyval);
    }
  maybe_update_cursor (viewport, state);
  return 0;
}


static gint buffer_viewport_key_release    (GtkEventBox*        event_box,
                                            GdkEventKey*        event,
				            GdamBufferViewport* viewport)
{
  guint state = 0;
  state = event->state;

  if (event->keyval == GDK_Alt_L     || event->keyval == GDK_Alt_R ||
      event->keyval == GDK_Shift_L   || event->keyval == GDK_Shift_R ||
      event->keyval == GDK_Control_L || event->keyval == GDK_Control_R)
    {
      state &= ~key_to_state (event->keyval);
    }
  maybe_update_cursor (viewport, state);
  return TRUE;
}

static gint buffer_viewport_enter          (GtkEventBox*        event_box,
                                            GdkEventCrossing*   event,
				            GdamBufferViewport* viewport)
{
  gtk_widget_grab_focus (GTK_WIDGET (event_box));
  return 0;
}

/* TODO: should also be able to set. */
static void adjust_beat_to_x(GdamBufferViewport* viewport, 
			     int x, 
			     gboolean move,		/* shift */
			     gboolean set)		/* control */
{
  gdouble t;
  gdouble old_pos;
  gdouble delta;
  t = get_time_from_x (viewport, x);
  old_pos = viewport->start_beat
    + viewport->beat_period * viewport->dragging_beat_line;
  delta = t - old_pos;

  if (set)
    {
      viewport->start_beat = t;
    }
  else if (move)
    {
      viewport->start_beat += delta;
    }
  else
    {
      if (viewport->dragging_beat_line <= 0)
	return;
      viewport->beat_period += delta / viewport->dragging_beat_line;
      if (viewport->beat_period < MIN_BEAT_PERIOD)
	viewport->beat_period = MIN_BEAT_PERIOD;
    }
  viewport->last_drag = t;
  gdam_buffer_viewport_beat_lines_changed (viewport);
}


static gint buffer_viewport_press          (GtkEventBox*        event_box,
                                            GdkEventButton*     event,
				            GdamBufferViewport* viewport)
{
  int beat_index;

  gtk_grab_add (GDAM_SKIN (viewport)->main_widget);

  /* Events with certain buttons and masks are left for others to handle. */
  //if (event->button != 1 && event->button != 2)
    //return 0;
  if (event->button == 1 && 
      (GDK_MOD1_MASK & event->state) == GDK_MOD1_MASK)
    return 0;

  /* If there is nothing for us to do, exit immediately */
  if (viewport->is_dragging
      || !(viewport->is_selectable
	   || viewport->can_drag_beat_lines || viewport->num_beat_times > 0))
    return 1;

  if (viewport->has_selection)
    gdam_buffer_viewport_queue_redraw (viewport);

  viewport->last_drag = get_time_from_x (viewport, event->x);

  /* If we clicked on a beat from the beat_times arrary, begin dragging it. */
  beat_index = scan_for_beat_index (viewport, event->x, event->y);
  if (beat_index >= 0)
    {
      /* Right click deletes beat marker */
      if (event->button == 3)
	{
	  delete_nth_beat_time (viewport, beat_index);
	}
      /* Otherwise, drag the beat marker */
      else
	{
	  viewport->is_dragging = TRUE;
	  viewport->beat_drag_index = beat_index;
	}
      return 1;
    }

  if (event->button != 1 && event->button != 2)
    return 0;

  /* If we are using the select button, or select is the default action */
  if ((viewport->is_selectable)
      && ((viewport->can_drag_beat_lines && event->button == 2)
	  || (!viewport->can_drag_beat_lines)))
    {
      viewport->is_dragging = TRUE;
      viewport->drag_start = viewport->last_drag;
      viewport->has_selection = TRUE;
      viewport->selection_start = viewport->last_drag;
      viewport->selection_length = 0.0;
      gdam_buffer_viewport_selection_changed(viewport);
    }

  /* If we have beatlines derived from firstbeat and bpm values */
  else if (viewport->can_drag_beat_lines && viewport->num_beats > 0)
    {
      int which_beat_line;
      gboolean controlled, shifted;
      controlled = ((GDK_CONTROL_MASK & event->state) == GDK_CONTROL_MASK);
      shifted = ((GDK_SHIFT_MASK & event->state) == GDK_SHIFT_MASK);
      which_beat_line = get_beat_line (viewport, viewport->last_drag);

      /* If we clicked on a beatline, begin dragging it */
      if (which_beat_line >= 0 && which_beat_line < viewport->num_beats)
	{
	  viewport->is_dragging_beat_line = TRUE;
	  viewport->dragging_beat_line = which_beat_line;
	  adjust_beat_to_x (viewport, event->x, shifted, controlled);
	}
      /* A control click sets the first beat. */
      else if (controlled)
	{
	  adjust_beat_to_x (viewport, event->x, shifted, controlled);
	}
      /* If a firstbeat exists but no tempo has been defined, define the tempo
       * such that the second beats falls on the click location. */
      else 
	{
	  gdouble beat_length = viewport->last_drag - viewport->start_beat;
	  viewport->beat_period = beat_length;
	  gdam_buffer_viewport_beat_lines_changed (viewport);
	  return 0;
	}
    }
  return 1;
}

static void extend_drag_to_x(GdamBufferViewport* viewport,
                             int                 x)
{
  float start_redraw = viewport->last_drag;
  float end_redraw;
  int start_x, end_x;
  g_return_if_fail (viewport != NULL);
  viewport->last_drag = get_time_from_x (viewport, x);
  end_redraw = viewport->last_drag;
  viewport->selection_start = viewport->drag_start;
  viewport->selection_length = viewport->last_drag - viewport->drag_start;
  if (viewport->selection_length < 0.0)
    {
      gdouble length = (-viewport->selection_length);
      viewport->selection_length = length;
      viewport->selection_start -= length;
    }
  if (start_redraw > end_redraw)
    {
      gfloat tmp = end_redraw;
      end_redraw = start_redraw;
      start_redraw = tmp;
    }
  start_x = get_x_from_time (viewport, start_redraw) - 1;
  end_x = get_x_from_time (viewport, end_redraw) + 1;
  redraw_subset (viewport, start_x - 1, end_x - start_x + 2);
  gdam_buffer_viewport_selection_changed(viewport);
}

/* XXX: would be nice if we could just redraw the bottom area */
/* XXX: should check to see if things are off window before trying to draw
        them */
static void redraw_for_motion(GdamBufferViewport*  viewport,
                              gdouble              time1,
			      gdouble              time2)
{
  int radius = viewport->beat_radius;
  int x1, x2;
  if (time1 > time2)
    {
      gdouble tmp = time1;
      time1 = time2;
      time2 = tmp;
    }
  x1 = get_x_from_time (viewport, time1);
  x2 = get_x_from_time (viewport, time2);
#if 0				/* daveb: i wrote this, but don't understand it ... */
  if ((x2 - x1) > radius * 4)
    {
      /* Draw as two separate areas. */
      redraw_subset (viewport, x1 - radius, 2 * radius);
      redraw_subset (viewport, x2 - radius, 2 * radius);
    }
  else
    {
      /* Draw as one area. */
      redraw_subset (viewport, x1 - radius, x2 + radius - (x1 - radius) + 3);
    }
#endif
  redraw_subset (viewport, x1 - 1, x2 - x1 + 1);
}

static gint buffer_viewport_release        (GtkEventBox*        event_box,
                                            GdkEventButton*     event,
				            GdamBufferViewport* viewport)
{
  gboolean shifted, controlled;
  int x = event->x;
  if (x < 0)
    x = 0;
  else if (x >= viewport->draw_width)
    x = viewport->draw_width - 1;
  gtk_grab_remove (GDAM_SKIN (viewport)->main_widget);
  if (event->button != 1 && event->button != 2)
    return 0;
  shifted = ((GDK_SHIFT_MASK & event->state) == GDK_SHIFT_MASK);
  controlled = ((GDK_CONTROL_MASK & event->state) == GDK_CONTROL_MASK);

  /* If we are dragging a beat in the beat_times array */
  if (viewport->is_dragging && viewport->beat_drag_index >= 0)
    {
      int index = viewport->beat_drag_index;
      gdouble old_pos;
      gdouble new_pos;
      gdouble cur_time = get_time_from_x (viewport, x);
      g_return_val_if_fail (viewport->beat_times != NULL, 0);
      old_pos = viewport->beat_times[index];
      new_pos = old_pos + (cur_time - viewport->last_drag);
      viewport->beat_times[index] = new_pos;
      viewport->is_dragging = FALSE;
      viewport->beat_drag_index = -1;

      gdam_buffer_viewport_beat_lines_changed (viewport);
      viewport->last_drag = cur_time;

      /* Redraw the relevant areas. */
      redraw_for_motion (viewport, old_pos, new_pos);
      return 0;
    }

  /* If we are making a selection */
  if (viewport->is_dragging)
    {
      extend_drag_to_x (viewport, x);
      viewport->is_dragging = FALSE;
      return 1;
    }

  /* If we are dragging a beat line derived from the bpm */
  if (viewport->is_dragging_beat_line || controlled)
    {
      adjust_beat_to_x (viewport, x, shifted, controlled);
      viewport->is_dragging_beat_line = FALSE;
      return 1;
    }
  return 0;
}

static void maybe_update_cursor (GdamBufferViewport *viewport,
				 guint               state)
{
  GdamCursorType cursor;

  /* If viewport doesn't display beat markers */
  if (!viewport->can_drag_beat_lines)
    {
      if (viewport->is_selectable)
	cursor = GDAM_SELECT_CURSOR;
      else
	cursor = GDAM_MOUSE_CURSOR;
    }
  /* If we use beat markers from the beat_times arrary */
  else if (viewport->num_beat_times > 0)
    {
	cursor = GDAM_MOUSE_CURSOR;
    }
  /* Else beat locations are derived from bpm and firstbeat */
  else if ((GDK_SHIFT_MASK & state) == GDK_SHIFT_MASK)
    {
      cursor = GDAM_PHASE_CURSOR;
    }
  else if ((GDK_CONTROL_MASK & state) == GDK_CONTROL_MASK)
    {
      cursor = GDAM_FIRSTBEAT_CURSOR;
    }
  else if ((GDK_MOD1_MASK & state) == GDK_MOD1_MASK)
    {
      cursor = GDAM_INDEX_CURSOR;
    }
  else
    {
      cursor = GDAM_BEATLINE_CURSOR;
    }

  if (cursor == viewport->cursor)
    return;
  
  g_return_if_fail (GDAM_SKIN (viewport)->main_widget->window != NULL);
  gdam_set_window_cursor (GDAM_SKIN (viewport)->main_widget->window, cursor);
  viewport->cursor = cursor;
}

static gint buffer_viewport_motion         (GtkEventBox*        event_box,
                                            GdkEventMotion*     event,
				            GdamBufferViewport* viewport)
{
  int x, y;
  guint state;
  gboolean shifted, controlled;

  maybe_update_cursor (viewport, event->state);
  
  if (!viewport->is_dragging && !viewport->is_dragging_beat_line)
    return 0;

  if (event->is_hint)
    gdk_window_get_pointer (event->window, &x, &y, &state);
  else
    {
      x = event->x;
      y = event->y;
    }

  if (x < 0)
    x = 0;
  else if (x >= viewport->draw_width)
    x = viewport->draw_width - 1;

  /* If we are dragging a beat in the beat array */
  if (viewport->beat_drag_index >= 0)
    {
      int index = viewport->beat_drag_index;
      gdouble old_pos;
      gdouble cur_time = get_time_from_x (viewport, x);
      gdouble new_pos;
      g_return_val_if_fail (viewport->beat_times != NULL, 0);
      old_pos = viewport->beat_times[index];
      new_pos = old_pos + (cur_time - viewport->last_drag);
      viewport->beat_times[index] = new_pos;
      viewport->last_drag = cur_time;

      /* Redraw the relevant areas. */
      redraw_for_motion (viewport, old_pos, new_pos);
      gdam_buffer_viewport_beat_lines_changed (viewport);

      return 0;
    }
  shifted = ((GDK_SHIFT_MASK & event->state) == GDK_SHIFT_MASK);
  controlled = ((GDK_CONTROL_MASK & event->state) == GDK_CONTROL_MASK);

  /* If we are dragging a beat line derived from the bpm */
  if (viewport->is_dragging_beat_line)
    {
      adjust_beat_to_x (viewport, x, shifted, controlled);
      return 1;
    }

  /* If we are making a selection */
  extend_drag_to_x (viewport, x);
  return 1;
}

static void buffer_viewport_size_request   (GtkEventBox*        event_box,
                                            GtkRequisition*     requisition,
					    GdamBufferViewport* viewport)
{
  requisition->width = viewport->draw_width;
  requisition->height = viewport->draw_height;
}
static void buffer_viewport_size_allocate  (GtkEventBox*        event_box,
                                            GtkAllocation*      allocation,
					    GdamBufferViewport* viewport)
{
  gboolean changed = allocation->width != viewport->draw_width
    || allocation->height != viewport->draw_height;
  viewport->draw_width = allocation->width;
  viewport->draw_height = allocation->height;
  if (changed)
    gdam_buffer_viewport_view_changed (viewport);
}

static void buffer_viewport_realize        (GtkWidget*          widget,
                                            GdamBufferViewport* viewport)
{
  g_return_if_fail (widget->window != NULL);
  if (viewport->fg_gc == NULL)
    {
      /* We base most of our colors on the style of the viewport widget */
      viewport->fg_gc = widget->style->fg_gc[GTK_STATE_NORMAL];
      viewport->bg_gc = widget->style->bg_gc[GTK_STATE_NORMAL];
      viewport->marker_font = gdk_font_load ("fixed");
      viewport->marker_gc = widget->style->fg_gc[GTK_STATE_ACTIVE];
      viewport->beat_line_gc = widget->style->fg_gc[GTK_STATE_PRELIGHT];
      viewport->beat_gc = gdk_gc_new (widget->window);
      /* Make the beat bars yellow */
      gdk_gc_set_foreground (viewport->beat_gc,
			     util_alloc_color_cached (65535, 65535, 0));
    }
  gdam_set_window_cursor (widget->window, GDAM_SELECT_CURSOR);
}

void       gdam_buffer_viewport_new_event_box (GdamBufferViewport* viewport)
{
  GtkEventBox *event_box = GTK_EVENT_BOX (gtk_event_box_new ());
  GtkWidget *widget = GTK_WIDGET (event_box);
  GtkObject *object = GTK_OBJECT (event_box);
  gtk_widget_set_name (widget, "buffer_viewport_widget");
  gtk_widget_add_events (widget, GDK_EXPOSURE_MASK
			 | GDK_POINTER_MOTION_MASK
			 | GDK_POINTER_MOTION_HINT_MASK
			 | GDK_ENTER_NOTIFY_MASK
			 | GDK_KEY_PRESS_MASK    | GDK_KEY_RELEASE_MASK
			 | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
  GTK_WIDGET_SET_FLAGS (widget, GTK_CAN_FOCUS);

  gtk_signal_connect (object,
		      "expose_event",
		      GTK_SIGNAL_FUNC (buffer_viewport_expose), viewport);
  gtk_signal_connect (object,
		      "configure_event",
		      GTK_SIGNAL_FUNC (buffer_viewport_configure), viewport);
  gtk_signal_connect (object,
		      "button_press_event",
		      GTK_SIGNAL_FUNC (buffer_viewport_press), viewport);
  gtk_signal_connect (object,
		      "button_release_event",
		      GTK_SIGNAL_FUNC (buffer_viewport_release), viewport);
  gtk_signal_connect (object,
		      "key_press_event",
		      GTK_SIGNAL_FUNC (buffer_viewport_key_press), viewport);
  gtk_signal_connect (object,
		      "key_release_event",
		      GTK_SIGNAL_FUNC (buffer_viewport_key_release), viewport);
  gtk_signal_connect (object,
		      "enter_notify_event",
		      GTK_SIGNAL_FUNC (buffer_viewport_enter), viewport);
  gtk_signal_connect (object,
		      "motion_notify_event",
		      GTK_SIGNAL_FUNC (buffer_viewport_motion), viewport);
  gtk_signal_connect (object,
		      "size_request",
		      GTK_SIGNAL_FUNC (buffer_viewport_size_request),
		      viewport);
  gtk_signal_connect (object,
		      "size_allocate",
		      GTK_SIGNAL_FUNC (buffer_viewport_size_allocate),
		      viewport);
  gtk_signal_connect (object,
		      "realize",
		      GTK_SIGNAL_FUNC (buffer_viewport_realize), viewport);

  GDAM_SKIN (viewport)->main_widget = widget;
}

void gdam_buffer_viewport_changed(GdamBufferViewport* viewport)
{
  GtkObject *object = GTK_OBJECT (viewport);
  if (gtk_signal_n_emissions (object, viewport_signals[VIEW_CHANGED]))
    return;
  gtk_signal_emit (object, viewport_signals[VIEW_CHANGED]);
}

void       
gdam_buffer_viewport_set_zoom (GdamBufferViewport *viewport,
                               gdouble             zoom)
{
  gdam_arg_set (GTK_OBJECT (viewport), "zoom", zoom);
}

void       gdam_buffer_viewport_set_center    (GdamBufferViewport* viewport,
                                               gdouble             center)
{
  viewport->time_start = center - viewport->time_length / 2.0;
  gdam_buffer_viewport_changed (viewport);
  /* TODO: update position adjustment? */
}

void       gdam_buffer_viewport_set_mode      (GdamBufferViewport* viewport,
                                               GdamBufferViewportMode mode)
{
  viewport->profile_mode = mode;
  gdam_buffer_viewport_changed (viewport);
}
void       gdam_buffer_viewport_set_mode_string
                                              (GdamBufferViewport* viewport,
                                               const char*         string)
{
  GdamBufferViewportMode mode = 0;
  if (strcmp (string, "left") == 0)
    mode = GDAM_BUFFER_VIEWPORT_LEFT;
  else if (strcmp (string, "right") == 0)
    mode = GDAM_BUFFER_VIEWPORT_RIGHT;
  else if (strcmp (string, "average") == 0)
    mode = GDAM_BUFFER_VIEWPORT_AVERAGE;
  else
    {
      g_warning ("buffer_viewport: unknown profile mode: %s", string);
      return;
    }
  gdam_buffer_viewport_set_mode (viewport, mode);
}
const char* gdam_buffer_viewport_get_mode_string(GdamBufferViewport* viewport)
{
  switch (viewport->profile_mode)
    {
    case GDAM_BUFFER_VIEWPORT_LEFT:
      return "left";
    case GDAM_BUFFER_VIEWPORT_RIGHT:
      return "right";
    case GDAM_BUFFER_VIEWPORT_AVERAGE:
      return "average";
    }
  g_warning ("unknown buffer profile mode (getting string)");
  return "unknown";
}

static void gdam_buffer_viewport_destroy(GtkObject* object)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (object);
  gdam_buffer_viewport_maybe_uncache (viewport);
  if (viewport->connection_id != 0)
    {
      gdam_buffer_manager_disconnect (viewport->connection_id);
      viewport->connection_id = 0;
    }
  g_free (viewport->marker_format);
  (*parent_class->destroy) (object);
}
void       gdam_buffer_viewport_set_length    (GdamBufferViewport* viewport,
                                               gdouble             length)
{
  viewport->sample_length = length;
  gtk_signal_emit (GTK_OBJECT (viewport), viewport_signals[LENGTH_CHANGED]);
}
static void gdam_buffer_viewport_real_beat_lines_changed
                                              (GdamBufferViewport* viewport)
{
  g_return_if_fail (viewport != NULL);
  redraw_subset (viewport, 0, viewport->draw_width);
}

void gdam_buffer_viewport_beat_lines_changed(GdamBufferViewport* viewport)
{
  if (gtk_signal_n_emissions (GTK_OBJECT (viewport),
			      viewport_signals[BEAT_LINES_CHANGED]))
    return;

  gtk_signal_emit (GTK_OBJECT (viewport),
		   viewport_signals[BEAT_LINES_CHANGED]);
}

void gdam_buffer_viewport_maybe_uncache(GdamBufferViewport* viewport)
{
  g_return_if_fail (viewport != NULL);
  if (viewport->buffer_id == 0)
    return;
  gdam_channel_append_request (viewport->skin.channel,
			       gdam_request_unref (GDAM_PROTOCOL_BUFFER_ARG,
						   viewport->buffer_id,
						   NULL, NULL, NULL));
}
