/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamkeyboard.h"
#include <gdam/common/gdamdebug.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkeditable.h>
#include <gtk/gtkwindow.h>


static void gdam_keyboard_init(GdamKeyboard* keyboard)
{
  keyboard->trap_map = g_hash_table_new (NULL, NULL);
}

static GdamKeyboardTrap* lookup_trap(GdamKeyboard* keyboard,
                                     GtkWidget*    widget,
				     int           key)
{
  GHashTable *per_widget;
  gpointer rv;
  per_widget = g_hash_table_lookup (keyboard->trap_map, widget);
  if (per_widget == NULL)
    return NULL;
  rv = g_hash_table_lookup (per_widget, GINT_TO_POINTER (key));
  if (rv == NULL)
    return NULL;
  return GDAM_KEYBOARD_TRAP (rv);
}

static void handle_trap_destroy(GdamKeyboardTrap* trap,
                                GdamKeyboard* keyboard)
{
  GHashTable *per_widget;
  gpointer key_ptr = GINT_TO_POINTER (trap->key);
  per_widget = g_hash_table_lookup (keyboard->trap_map, trap->widget);
  g_return_if_fail (per_widget != NULL);
  g_return_if_fail (g_hash_table_lookup (per_widget, key_ptr) != NULL);
  g_hash_table_remove (per_widget, key_ptr);
  if (g_hash_table_size (per_widget) == 0)
    {
      g_hash_table_remove (keyboard->trap_map, trap->widget);
      g_hash_table_destroy (per_widget);
    }
}
static void itemize_keys(gpointer key, gpointer value, gpointer user_data)
{
  *(*(int **) user_data)++ = GPOINTER_TO_INT (key);
}
static void handle_widget_destroy(GtkWidget*    widget,
                                  GdamKeyboard* keyboard)
{
  GHashTable *per_widget;
  int *keys;
  int *keys_at;
  int num_keys, i;
  per_widget = g_hash_table_lookup (keyboard->trap_map, widget);
  if (per_widget == NULL)
    return;

  /* Create list of traps. */
  num_keys = g_hash_table_size (per_widget);
  keys = alloca (num_keys * sizeof (int));
  keys_at = keys;
  g_hash_table_foreach (per_widget, itemize_keys, &keys_at);
  g_assert (keys + num_keys == keys_at);

  gdam_debug ("keyboard:handle_widget_destroy: %d key bindings", num_keys);

  for (i = 0; i < num_keys; i++)
    gdam_keyboard_remove_trap (keyboard, widget, keys[i]);
}

static GdamKeyboardTrap* add_trap(GdamKeyboard* keyboard,
                                  GtkWidget*    widget,
				  int           key)
{
  GHashTable *per_widget;
  GdamKeyboardTrap *trap;
  gpointer old_val;
  GtkObject *object;
  per_widget = g_hash_table_lookup (keyboard->trap_map, widget);
  if (per_widget == NULL)
    {
      per_widget = g_hash_table_new (NULL, NULL);
      g_hash_table_insert (keyboard->trap_map, widget, per_widget);
      gtk_signal_connect (GTK_OBJECT (widget),
			  "destroy",
			  GTK_SIGNAL_FUNC (handle_widget_destroy), keyboard);
    }
  old_val = g_hash_table_lookup (per_widget, GINT_TO_POINTER (key));
  g_return_val_if_fail (old_val == NULL, NULL);
  object = gtk_type_new (GDAM_TYPE_KEYBOARD_TRAP);
  gtk_object_ref (object);
  gtk_object_sink (object);
  trap = GDAM_KEYBOARD_TRAP (object);
  trap->widget = widget;
  trap->key = key;
  trap->state = 0;
  g_hash_table_insert (per_widget, GINT_TO_POINTER (key), trap);
  gtk_signal_connect (GTK_OBJECT (trap),
		      "destroy",
		      GTK_SIGNAL_FUNC (handle_trap_destroy), keyboard);
  return trap;
}


GdamKeyboardTrap*   gdam_keyboard_get_trap   (GdamKeyboard*  keyboard,
                                              GtkWidget*     widget,
                                              int            key)
{
  GdamKeyboardTrap *rv;
  rv = lookup_trap (keyboard, widget, key);
  if (rv == NULL)
    rv = add_trap (keyboard, widget, key);
  g_return_val_if_fail (rv != NULL, NULL);
  return rv;
}

void                gdam_keyboard_remove_trap(GdamKeyboard*  keyboard,
                                              GtkWidget*     widget,
                                              int            key)
{
  GdamKeyboardTrap *trap;
  trap = lookup_trap (keyboard, widget, key);
  if (trap == NULL)
    return;
  gtk_object_unref (GTK_OBJECT (trap));
}

static void gdam_keyboard_class_init(GdamKeyboardClass* keyboard_class)
{
}

GtkType gdam_keyboard_get_type()
{
  static GtkType keyboard_type = 0;
  if (!keyboard_type)
    {
      static const GtkTypeInfo keyboard_info = {
	"GdamKeyboard",
	sizeof (GdamKeyboard),
	sizeof (GdamKeyboardClass),
	(GtkClassInitFunc) gdam_keyboard_class_init,
	(GtkObjectInitFunc) gdam_keyboard_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GTK_TYPE_OBJECT;
      keyboard_type = gtk_type_unique (parent, &keyboard_info);
    }
  return keyboard_type;
}

static guint map_state_to_gdam_shift(guint event_state)
{
  guint gdam_shift = 0;
  if ((event_state & GDK_CONTROL_MASK) == GDK_CONTROL_MASK)
    gdam_shift |= GDAM_KEYBOARD_CTRL_MASK;
  if ((event_state & GDK_MOD1_MASK) == GDK_MOD1_MASK)
    gdam_shift |= GDAM_KEYBOARD_ALT_MASK;
  return gdam_shift;
}

typedef struct _KeyHandlerInfo KeyHandlerInfo;
struct _KeyHandlerInfo {
	GtkWidget* widget;
	GtkWindow* toplevel;
};

static KeyHandlerInfo* key_handler_info_new(GtkWidget* widget,
		                            GtkWindow* toplevel)
{
  KeyHandlerInfo *info = g_new (KeyHandlerInfo, 1);
  info->widget = widget;
  info->toplevel = toplevel;
  return info;
}


static int keypress_event(KeyHandlerInfo* info, GdkEventKey* event)
{
  GtkWidget *widget = info->widget;
  GdamKeyboardTrap *trap;
  int key = event->keyval;
  if (info->toplevel != NULL || info->toplevel->focus_widget != NULL)
    {
      GtkWidget *focus = info->toplevel->focus_widget;
      if (GTK_IS_EDITABLE (focus))
	return 1;
    }
  key |= map_state_to_gdam_shift (event->state);
  gdam_debug ("gdam_keyboard_press_event: key=%08x", key);
  trap = lookup_trap (gdam_keyboard_get_instance (), widget, key);
  if (trap == NULL)
    return 0;
  gtk_signal_emit_by_name (GTK_OBJECT (trap), "pressed");
  return 1;
}
static int keyrelease_event(KeyHandlerInfo* info, GdkEventKey* event)
{
  GtkWidget *widget = info->widget;
  GdamKeyboardTrap *trap;
  int key = event->keyval;
  key |= map_state_to_gdam_shift (event->state);
  gdam_debug ("gdam_keyboard_release_event: key=%08x", key);
  trap = lookup_trap (gdam_keyboard_get_instance (), widget, key);
  if (trap == NULL)
    return 0;
  gtk_signal_emit_by_name (GTK_OBJECT (trap), "released");
  return 1;
}


void gdam_keyboard_connect_widget(GtkWidget* widget, int mode)
{
  GtkWidget *wwidget = widget;
  GtkWidget *top_widget = widget;
  GtkWindow *toplevel;

  while (wwidget != NULL && GTK_WIDGET_NO_WINDOW (wwidget))
    wwidget = wwidget->parent;
  while (top_widget != NULL && !GTK_IS_WINDOW (top_widget))
    top_widget = top_widget->parent;
  g_return_if_fail (wwidget != NULL);
  if (!top_widget)
    g_warning ("couldn't find a window to trap keyboard events!");
  wwidget  = top_widget == NULL ? wwidget : top_widget;
  toplevel = top_widget == NULL ? NULL : GTK_WINDOW (top_widget);
  gdam_debug ("settled on %s to bind keys to, top widget %s, toplevel %p",
	      gtk_type_name (GTK_OBJECT (wwidget)->klass->type),
	      gtk_type_name (GTK_OBJECT (top_widget)->klass->type),
	      toplevel);

  if (mode & GDAM_KEYBOARD_CONNECT_PRESS)
    {
      gtk_widget_add_events (wwidget, GDK_KEY_PRESS_MASK);
      gtk_signal_connect_full
	(GTK_OBJECT (wwidget),
	 "key_press_event",
	 GTK_SIGNAL_FUNC (keypress_event),
	 NULL, key_handler_info_new (widget, toplevel), g_free, TRUE, FALSE);
    }
  if (mode & GDAM_KEYBOARD_CONNECT_RELEASE)
    {
      gtk_widget_add_events (wwidget, GDK_KEY_RELEASE_MASK);
      gtk_signal_connect_full
	(GTK_OBJECT (wwidget),
	 "key_release_event",
	 GTK_SIGNAL_FUNC (keyrelease_event),
	 NULL, key_handler_info_new (widget, toplevel), g_free, TRUE, FALSE);
    }
}

GdamKeyboard* gdam_keyboard_get_instance()
{
  static GdamKeyboard *singleton = NULL;
  GtkObject *object;
  if (singleton != NULL)
    return singleton;
  object = gtk_type_new (GDAM_TYPE_KEYBOARD);
  gtk_object_ref (object);
  gtk_object_sink (object);
  return singleton = GDAM_KEYBOARD (object);
}
