/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdammixskin.h"
#include "gdamskinfactory.h"
#include <gdam/common/gdamplugin.h>
#include <gtk/gtkdrawingarea.h>
#include <gtk/gtksignal.h>
#include "gdamskinutil.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>

static GtkObjectClass* parent_class = NULL;

enum {
	GDAM_MIX_SELECTION_ADD,
	GDAM_MIX_SELECTION_TOGGLE,
	GDAM_MIX_SELECTION_REPLACE,
};

static GdamSkinFactory* gdam_mix_skin_factory_get();
static void gdam_mix_skin_stop_dragging(GdamMixSkin* skin);

static void gdam_mix_skin_init(GdamMixSkin* mix_skin)
{
  mix_skin->max_beat = 30.0;
  mix_skin->min_beat = 0.0;
}
static void alloc_colors(GdamMixSkin* skin);

int gdam_mix_skin_prototype_index (GdamMixSkinPrototype* prototype)
{
  GdamMixSkin *mix_skin;
  int i;
  mix_skin = prototype->mix_skin;
  for (i = 0; i < mix_skin->num_prototypes; i++)
    if (mix_skin->prototypes[i] == prototype)
      return i;
  return -1;
}

/* Traps for when the prototypes' models are removed or added. */
static void gdam_mix_skin_child_added   (GtkObject*     parent,
                                         GtkObject*     child,
					 const char*    child_name,
					 gpointer       user_data);
static void gdam_mix_skin_child_removed (GtkObject*     parent,
                                         GtkObject*     child,
					 const char*    child_name,
					 gpointer       user_data);

/*
 *                             __ __        _                   _
 *  ___ __ _ _ ___ ___ _ _    / /_\ \   ___| |___ _ __  ___ _ _| |_
 * (_-</ _| '_/ -_) -_) ' \  < <___> > / -_) / -_) '  \/ -_) ' \  _|
 * /__/\__|_| \___\___|_||_|  \_\ /_/  \___|_\___|_|_|_\___|_||_\__|
 * 
 * Various helpful coordinate lookup functions.
 */

static void
gdam_mix_skin_element_make_screen_data_valid(GdamMixSkinElement*   element,
                                             GdamMixSkin*          mix_skin)
{
  GdamMixElement *mix_element = element->element;
  GdamProducer *producer = mix_element->producer;
  GdamMixSkinPrototype *prototype = element->prototype;
  gdouble tmp;
  int selected_arg;
  gint64 length;
  const char *arg_name;
  int num_args;
  GdamArg *args;

  /* Samples per beat. */
  gdouble beat_length;

  /* Top of the element, in the range [0,1]. */
  gdouble fraction_y;

  GdamMix *mix;
  mix = GDAM_MIX (GDAM_SKIN (mix_skin)->model);
  beat_length = mix->beat_length;
  if (element->screen_data_valid)
    {
      return;
    }

  /* Compute the left edge of the element. */
  tmp = mix_element->start_time;
  tmp /= beat_length;
  tmp -= mix_skin->min_beat;
  tmp /= (mix_skin->max_beat - mix_skin->min_beat);
  tmp *= mix_skin->draw_width;
  element->screen_x = tmp;


  /* Compute the top edge of the element. */
  selected_arg = prototype->selected_arg;
  if (selected_arg > prototype->num_args || selected_arg < 0)
    arg_name = NULL;
  else
    arg_name = prototype->arg_names[selected_arg];

  if (arg_name != NULL)
    {
      gdam_arg_get_by_name (GTK_OBJECT (mix_element->producer),
			    1, &arg_name, &num_args, &args);
      if (num_args != 1)
	g_warning ("couldn't get argument %s", arg_name);
    }
  else
    {
      num_args = 0;
      args = NULL;
    }

  if (num_args == 1)
    {
      gdouble min_bound = prototype->min_bounds[selected_arg];
      gdouble max_bound = prototype->max_bounds[selected_arg];
      fraction_y = GDAM_ARG_DOUBLE (args);
      fraction_y -= min_bound;
      fraction_y /= (max_bound - min_bound);
    }
  else
    {
      fraction_y = element->natural_y_frac;
    }
  element->screen_y = (int) (fraction_y * mix_skin->draw_height);

  /* The element's height is constant here... */
  element->screen_height = 30;

  /* The element's width is determined by its length. */
  length = gdam_model_get_length (GDAM_MODEL (producer));
  if (length == -1)
    {
      element->screen_width = mix_skin->draw_width;
    }
  else
    {
      gdouble length_in_beats;
      gdouble visible_beats;
      length_in_beats = (gdouble) length / beat_length;
      visible_beats = (mix_skin->max_beat - mix_skin->min_beat);
      element->screen_width = length_in_beats
	* mix_skin->draw_width / visible_beats;
    }
}

static void
gdam_mix_skin_element_make_drag_data_valid(GdamMixSkinElement*   element,
                                           GdamMixSkin*          mix_skin)
{
  int delta_x, delta_y;
  gdam_mix_skin_element_make_screen_data_valid (element, mix_skin);
  delta_x = mix_skin->drag_current_x - mix_skin->drag_start_x;
  delta_y = mix_skin->drag_current_y - mix_skin->drag_start_y;
  element->drag_data_valid = TRUE;
  element->drag_x = element->screen_x + delta_x;
  element->drag_y = element->screen_y + delta_y;
  element->drag_width = element->screen_width;
  element->drag_height = element->screen_height;
}

static void 
get_string_offset(GdkFont* font, const char* string,
                  int* offset_left, int* offset_center, int* offset_y,
		  int* width)
{
  int lbearing, rbearing, ascent, descent;
  gdk_string_extents (font, string,
		      &lbearing, &rbearing, width, &ascent, &descent);
  *offset_center = (lbearing + rbearing) / 2;
  *offset_left = lbearing;
  *offset_y = (descent - ascent) / 2;
}

static inline void
gdam_mix_skin_prototype_make_string_offset_valid(GdamMixSkinPrototype* proto)
{
  GdamMixSkin *mix_skin = proto->mix_skin;

  if (!proto->has_string_offset)
    {
      get_string_offset (mix_skin->drawing_area->style->font,
			 proto->prototype_name,
			 &proto->string_offset_left,
			 &proto->string_offset_center,
			 &proto->string_offset_y, &proto->string_width);
      proto->has_string_offset = TRUE;
    }
}
GdamMixSkinElement* find_element(GdamMixSkin* skin, int x, int y)
{
  GSList *list;
  GdamMix *mix = GDAM_MIX (GDAM_SKIN (skin)->model);
  for (list = mix->elements; list; list = list->next)
    {
      GdamMixSkinElement *element;
      element = g_hash_table_lookup (skin->skin_elements, list->data);
      g_return_val_if_fail (element != NULL, NULL);
      gdam_mix_skin_element_make_screen_data_valid (element, skin);
      g_message ("considering element %d,%d x %d,%d",
		 element->screen_x, element->screen_y,
		 element->screen_width, element->screen_height);

      if (!element->prototype->is_hidden
	  && element->screen_x <= x
	  && element->screen_x + element->screen_width > x
	  && element->screen_y <= y
	  && element->screen_y + element->screen_height > y)
	return element;
    }
  return NULL;
}
static gdouble gdam_mix_skin_find_press_time(GdamMixSkin* mix_skin, int x)
{
  gdouble fraction;
  gdouble beat;
  GdamMix *mix = (GdamMix *) ((GdamSkin *) mix_skin)->model;
  fraction = (gdouble) x / mix_skin->draw_width;
  beat = fraction * (mix_skin->max_beat - mix_skin->min_beat)
    + mix_skin->min_beat;

  return (beat * mix->beat_length);
}


static void handle_submodel_arg_set(GdamMixSkinPrototype* prototype,
                                    GdamArg*              arg,
				    GdamModel*            submodel)
{
  int index;
  GSList *list;
  g_message ("handle_submodel_arg_set: argname=%s", arg->name);
  for (index = 0; index < prototype->num_submodels; index++)
    if (prototype->submodels[index] == submodel)
      break;
  g_return_if_fail (index < prototype->num_submodels);
  g_message ("handle_submodel_arg_set: index=%d", index);
  for (list = prototype->elements; list != NULL; list = list->next)
    {
      GdamMixSkinElement *skin_element = list->data;
      GdamModel *sub = skin_element->models[index];
      g_message ("handle_submodel_arg_set: handle one prototype");
      gdam_arg_set_object_arg (GTK_OBJECT (sub), arg);
    }
}

void gdam_mix_skin_start_prototype_tracking(GdamMixSkin*           mix_skin,
                                            GdamMixSkinPrototype*  prototype,
				            int                    model_index)
{
  GdamModel *submodel = prototype->submodels[model_index];
  gdam_arg_connect (GTK_OBJECT (submodel),
		    NULL, GDAM_ARG_FUNC (handle_submodel_arg_set), prototype);
}


/*
 *  ____                _           _
 * |  _ \ ___ _ __   __| | ___ _ __(_)_ __   __ _
 * | |_) / _ \ '_ \ / _` |/ _ \ '__| | '_ \ / _` |
 * |  _ <  __/ | | | (_| |  __/ |  | | | | | (_| |
 * |_| \_\___|_| |_|\__,_|\___|_|  |_|_| |_|\__, |
 *                                          |___/
 */

/* Update the display from the backbuffer, within the given rectangle.  */
static inline void
copy_backing_area(GdamMixSkin* skin, int x, int y, int w, int h)
{
  /* The widget to copy the data on to. */
  GtkWidget *widget = skin->drawing_area;

  /* The graphics context to use for the transfer. */
  GdkGC *gc = widget->style->fg_gc[GTK_WIDGET_STATE (widget)];

  /* The backing pixmap to copy from. */
  GdkPixmap *src_pixmap = skin->backing_pixmap;

  gdk_draw_pixmap (widget->window, gc, src_pixmap, x, y, x, y, w, h);
}

typedef struct _GdamMixSkinRenderInfo GdamMixSkinRenderInfo;
struct _GdamMixSkinRenderInfo {
	int 			x, y, width, height;
	GdamMixSkin*		mix_skin;
};


static void render_subsample(GdamMixElement*         mix_element,
                             GdamMixSkinRenderInfo*  render_info)
{
  GdamMixSkin *mix_skin = render_info->mix_skin;
  GdamMixSkinElement *element;
  GdkPixmap *pixmap = mix_skin->backing_pixmap;
  GdkGC *gc_inner;
  GdkGC *gc_outer;
  GdamMixSkinPrototype *prototype;
  GdamMix *mix;
  int screen_x, screen_y, screen_width, screen_height;
  int x;

  element = g_hash_table_lookup (mix_skin->skin_elements, mix_element);
  g_return_if_fail (element != NULL);
  prototype = element->prototype;
  mix = GDAM_MIX (GDAM_SKIN (mix_skin)->model);

  if (element->is_dragging)
    return;

  if (prototype->is_hidden)
    return;

  gdam_mix_skin_element_make_screen_data_valid (element, mix_skin);
  screen_x = element->screen_x;
  screen_y = element->screen_y;
  screen_width = element->screen_width;
  screen_height = element->screen_height;

  /*
   * Does the box around x/y touch the exposed area?
   */
  if (screen_x > render_info->x + render_info->width
      || screen_x + screen_width < render_info->x
      || screen_y > render_info->y + render_info->height
      || screen_y + screen_height < render_info->y)
    return;

  gc_inner = mix_skin->rect_gc;
  gc_outer = mix_skin->rect_boundary_gc;

  if (element->is_selected)
    {
      GdkGC *tmp = gc_inner;
      gc_inner = gc_outer;
      gc_outer = tmp;
    }

  if (mix->is_looping && screen_x + screen_width > mix_skin->draw_width)
    {
      int width_1 = mix_skin->draw_width - screen_x;
      int width_2 = screen_width - width_1;
      GdkSegment segments[6];
      /* Draw the wrapped rectangle:

       *     +-----------------------------------------+
       *     |                                         |
       *     |*******                         *********|
       *     |FFFFFF*                         *FFFFFFFF|
       *     |FFFFFF*                         *FFFFFFFF|
       *     |*******                         *********|
       *     |                                         |
       *     |                                         |
       *     |   Legend:                               |
       *     |        *   Lines to draw with gc_outer. |
       *     |        F   Area to draw with gc_inner.  |
       *     |      |+-   Window boundaries.           |
       */
      gdk_draw_rectangle (pixmap,	/* right rectangle */
			  gc_inner,
			  TRUE, screen_x, screen_y, width_1, screen_height);
      gdk_draw_rectangle (pixmap,	/* left rectangle */
			  gc_inner,
			  TRUE, 0, screen_y, width_2, screen_height);
      /* Using the above chart we fill in the six segments:
       *         0                                 3
       *     |*******                         *********|
       *     |FFFFFF* 1                     4 *FFFFFFFF|
       *     |FFFFFF*                         *FFFFFFFF|
       *     |*******                         *********|
       *         2                                 5
       */
      segments[0].x1 = 0;
      segments[0].x2 = width_2;
      segments[0].y1 = segments[0].y2 = screen_y;
      segments[1].x1 = segments[1].x2 = width_2;
      segments[1].y1 = screen_y;
      segments[1].y2 = screen_y + screen_height;
      segments[2].x1 = 0;
      segments[2].x2 = width_2;
      segments[2].y1 = segments[2].y2 = screen_y + screen_height;
      segments[3].x1 = screen_width - width_1;
      segments[3].x2 = mix_skin->draw_width;
      segments[3].y1 = segments[3].y2 = screen_y;
      segments[4].x1 = segments[4].x2 = screen_x;
      segments[4].y1 = screen_y;
      segments[4].y2 = screen_y + screen_height;
      segments[5].x1 = screen_width - width_1;
      segments[5].x2 = mix_skin->draw_width;
      segments[5].y1 = segments[5].y2 = screen_y + screen_height;
      gdk_draw_segments (pixmap, gc_outer, segments, 6);
    }
  else
    {
      gdk_draw_rectangle (pixmap,
			  gc_inner,
			  TRUE,
			  screen_x, screen_y, screen_width, screen_height);
      gdk_draw_rectangle (pixmap,
			  gc_outer,
			  FALSE,
			  screen_x, screen_y, screen_width, screen_height);
    }


  gdam_mix_skin_prototype_make_string_offset_valid (prototype);

  if (prototype->string_width > screen_width)
    {
      x = (screen_x + screen_width) % mix_skin->draw_width;
      x -= prototype->string_offset_left;
    }
  else
    {
      x = (screen_x + screen_width / 2) % mix_skin->draw_width;
      x -= prototype->string_offset_center;
    }

  gdk_draw_string (pixmap,
		   mix_skin->font,
		   mix_skin->text_gc,
		   x,
		   screen_y + screen_height + prototype->string_offset_y,
		   prototype->prototype_name);
}


static void redraw_area(GdamMixSkin* skin, int x, int y, int width, int height)
{
  GdamMixSkinRenderInfo render_info;
  GdamMix *mix;

  render_info.x = x;
  render_info.y = y;
  render_info.width = width;
  render_info.height = height;
  render_info.mix_skin = skin;

  if (skin->rect_gc == NULL)
    alloc_colors (skin);

  if (skin->backing_pixmap == NULL)
    return;

  gdk_draw_rectangle (skin->backing_pixmap,
		      skin->background, TRUE, x, y, width, height);

  /* XXX: rewrite grid drawing code here to do it algorithmically. */
  mix = GDAM_MIX (GDAM_SKIN (skin)->model);
  g_slist_foreach (mix->elements, (GFunc) render_subsample, &render_info);
  copy_backing_area (skin, x, y, width, height);
}

static void redraw_element_area(GdamMixSkin* skin, GdamMixSkinElement* element)
{
  int screen_width;
  gdam_mix_skin_element_make_screen_data_valid (element, skin);
  screen_width = element->screen_width + element->prototype->string_width;
  if (element->screen_x + screen_width > skin->draw_width)
    {
      int width_1, width_2;
      width_1 = element->screen_x + screen_width - skin->draw_width;
      width_2 = screen_width - width_1;
      redraw_area (skin,
		   0, element->screen_y, width_2, element->screen_height);
      redraw_area (skin,
		   skin->draw_width - width_1, element->screen_y,
		   width_1, element->screen_height);
    }
  else
    {
      redraw_area (skin,
		   element->screen_x, element->screen_y,
		   screen_width, element->screen_height);
    }
}

/* Queuing draws. */
static gboolean redraw_in_idle_thread(gpointer skin_ptr)
{
  GdamMixSkin *mix_skin;
  mix_skin = GDAM_MIX_SKIN (skin_ptr);
  redraw_area (mix_skin, 0, 0, mix_skin->draw_width, mix_skin->draw_height);
  mix_skin->redraw_thread = 0;
  gtk_object_unref (GTK_OBJECT (mix_skin));
  return FALSE;
}

static void delayed_redraw(GdamMixSkin* mix_skin)
{
  if (mix_skin->redraw_thread == 0 && !GTK_OBJECT_DESTROYED (mix_skin))
    {
      gtk_object_ref (GTK_OBJECT (mix_skin));
      mix_skin->redraw_thread = g_idle_add (redraw_in_idle_thread, mix_skin);
    }
}

void gdam_mix_skin_edit_prototype(GdamMixSkin*         mix_skin,
                                 int                  prototype_index)
{
  GdamMixSkinPrototype *old_prototype = mix_skin->selected_prototype;
  GdamMixSkinPrototype *new_prototype;
  GtkObject *old_skin;
  if (prototype_index >= 0)
    new_prototype = mix_skin->prototypes[prototype_index];
  else
    new_prototype = NULL;

  if (new_prototype == old_prototype)
    return;

  /* Destroy old_prototype's skin. */
  if (old_prototype != NULL && old_prototype->skin != NULL)
    {
      GtkWidget *old_main_widget = old_prototype->skin->main_widget;
      gtk_widget_hide (old_main_widget);
    }

  mix_skin->selected_prototype = NULL;
  if (mix_skin->prototype_name_entry != NULL && new_prototype != NULL)
    {
      gtk_entry_set_text (GTK_ENTRY (mix_skin->prototype_name_entry),
			  new_prototype->prototype_name);
    }
  mix_skin->selected_prototype = new_prototype;

  if (new_prototype == NULL)
    {
      gtk_widget_hide (mix_skin->prototype_name_entry);
      return;
    }

  /* Build new_prototype's skin. */
  gtk_widget_show (new_prototype->skin->main_widget);
  gtk_widget_show (mix_skin->prototype_name_entry);
}


/*
 *  ____                        _
 * |  _ \ _ __ __ _  __ _  __ _(_)_ __   __ _
 * | | | | '__/ _` |/ _` |/ _` | | '_ \ / _` |
 * | |_| | | | (_| | (_| | (_| | | | | | (_| |
 * |____/|_|  \__,_|\__, |\__, |_|_| |_|\__, |
 *                  |___/ |___/         |___/
 */
static void
draw_one_dragged_sample(GdamMixSkinElement* element, GdamMixSkin* skin)
{
  gdam_mix_skin_element_make_drag_data_valid (element, skin);

  /* For now, the rectangle is the entire screen area
   * of the element. */
  gdk_draw_rectangle (skin->drawing_area->window,
		      skin->drag_gc,
		      FALSE,
		      element->drag_x,
		      element->drag_y,
		      element->drag_width, element->drag_height);
}
static void draw_dragged_samples(GdamMixSkin* skin)
{
  g_slist_foreach (skin->dragging, (GFunc) draw_one_dragged_sample, skin);
}
static void
undraw_one_dragged_sample(GdamMixSkinElement* element,
                          GdamMixSkin*        skin)
{
  gdam_mix_skin_element_make_drag_data_valid (element, skin);
  copy_backing_area (skin,
		     element->drag_x - 1,
		     element->drag_y - 1,
		     element->drag_width + 2, element->drag_height + 2);
}

static void undraw_dragged_samples(GdamMixSkin* skin)
{
  g_slist_foreach (skin->dragging, (GFunc) undraw_one_dragged_sample, skin);
}

static void mark_dragging_state(GdamMixSkinElement* element, gpointer bool)
{
  element->is_dragging = GPOINTER_TO_INT (bool);
}

static void
begin_dragging(GdamMixSkin*          skin,
               GdamMixSkinElement*   element,
	       guint                 mask,
	       int                   mouse_x,
	       int                   mouse_y)
{
  gboolean copy = (mask & GDK_SHIFT_MASK) == GDK_SHIFT_MASK;

  gdam_mix_skin_stop_dragging (skin);

  skin->drag_start_x = mouse_x;
  skin->drag_start_y = mouse_y;

  g_message ("begin_dragging");
  delayed_redraw (skin);

  /* Set up:
   *      skin->in_transit
   *      skin->num_dragging
   *      skin->dragging. */
  if (element->is_selected)
    {
      if (copy)
	{
	  g_warning ("COPY not supported yet");
	}
      skin->dragging = g_slist_copy (skin->selected_elements);
      skin->was_dragging_selected = TRUE;
    }
  else
    {
      if (copy)
	{
	  g_warning ("COPY not supported yet");
	}
      skin->dragging = g_slist_prepend (skin->dragging, element);
      skin->was_dragging_selected = FALSE;
    }
  g_slist_foreach (skin->dragging,
		   (GFunc) mark_dragging_state, (gpointer) TRUE);

  /* TODO: should only redraw the in_transit area. */
  redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
  draw_dragged_samples (skin);
}

static void adjust_argument(GdamMixSkinElement*       element,
                            gdouble                   y_delta)
{
  const char *name;
  int index = element->prototype->selected_arg;
  int which_model;
  GdamModel *model;
  GdamArg *out;
  int num_out;
  double new_val;
  guint flags;
  gboolean is_log;
  GtkObject *object;
  GdamMixSkinPrototype *prototype = element->prototype;
  name = prototype->arg_names[index];
  which_model = prototype->arg_models[index];
  model = element->models[which_model];
  object = GTK_OBJECT (model);
  gdam_arg_get_by_name (object, 1, &name, &num_out, &out);
  if (num_out == 0)
    {
      g_warning ("invalid arg %s, to gdam_mix_select_arg", name);
      return;
    }

  if (out->type != GDAM_ARG_TYPE_DOUBLE)
    {
      g_warning ("invalid type for arg %s to gdam_mix_select_arg", name);
      return;
    }

  new_val = GDAM_ARG_DOUBLE (out);
  flags = prototype->arg_flags[index];
  is_log = (flags & GDAM_MIX_SKIN_FLAGS_LOGARITHMIC) != 0;

  if (is_log)
    new_val = (new_val <= 0.0) ? -10 : log (new_val);

  new_val += y_delta;

  if (is_log)
    new_val = exp (new_val);
  GDAM_ARG_DOUBLE (out) = new_val;
  gdam_arg_set_object_arg (object, out);
  gdam_arg_free_array (1, out);
}

static void gdam_mix_skin_stop_dragging(GdamMixSkin* skin)
{
  GdamMix *mix = GDAM_MIX (GDAM_SKIN (skin)->model);
  int move_x, move_y;
  gdouble delta_start;
  gdouble frac_y;
  gdouble screen_beats = (skin->max_beat - skin->min_beat);
  GSList *list;
  move_x = skin->drag_current_x - skin->drag_start_x;
  move_y = skin->drag_current_y - skin->drag_start_y;

  delta_start = screen_beats * move_x / skin->draw_width;
  frac_y = (gdouble) move_y / skin->draw_height;

  gdam_mix_freeze (mix);
  //unselect_all_elements(skin);

  for (list = skin->dragging; list != NULL; list = list->next)
    {
      GdamMixSkinElement *element = list->data;
      GdamMixElement *mix_element;
      if (element->prototype->selected_arg < 0)
	{
	  element->natural_y_frac += frac_y;
	}
      else
	{
	  adjust_argument (element, frac_y);
	}
      mix_element = element->element;
      gdam_mix_set_start (mix, mix_element->producer,
			  mix_element->start_time + delta_start);
      element->is_dragging = FALSE;
    }
  g_slist_free (skin->dragging);
  skin->dragging = NULL;

  gdam_mix_thaw (mix);
  delayed_redraw (skin);
}


/*
 *  ____       _           _   _
 * / ___|  ___| | ___  ___| |_(_) ___  _ __
 * \___ \ / _ \ |/ _ \/ __| __| |/ _ \| '_ \
 *  ___) |  __/ |  __/ (__| |_| | (_) | | | |
 * |____/ \___|_|\___|\___|\__|_|\___/|_| |_|
 * 
 */
static void begin_selection(GdamMixSkin*    skin,
                            int             press_x,
			    int             press_y)
{
  g_return_if_fail (skin->selected_elements == NULL);

  skin->has_selection = TRUE;
  skin->selection[0].x = skin->selection[1].x = press_x;
  skin->selection[0].y = skin->selection[1].y = press_y;
}

static void delete_one(GdamMixSkinElement* element, GdamMixSkin* mix_skin)
{
  gdam_mix_skin_remove (mix_skin, element);
}

static void delete_selected(GdamMixSkin* mix_skin)
{
  GdamMix *mix = GDAM_MIX (GDAM_SKIN (mix_skin)->model);
  GSList *selected = mix_skin->selected_elements;
  mix_skin->selected_elements = NULL;
  gdam_mix_freeze (mix);
  g_slist_foreach (selected, (GFunc) delete_one, mix_skin);
  gdam_mix_thaw (mix);
  delayed_redraw (mix_skin);
  g_slist_free (selected);
}

/*
 *     _                         _
 *    / \   ___ _   _ _ __   ___| |__  _ __ ___  _ __   ___  _   _ ___
 *   / _ \ / __| | | | '_ \ / __| '_ \| '__/ _ \| '_ \ / _ \| | | / __|
 *  / ___ \\__ \ |_| | | | | (__| | | | | | (_) | | | | (_) | |_| \__ \
 * /_/   \_\___/\__, |_| |_|\___|_| |_|_|  \___/|_| |_|\___/ \__,_|___/
 *              |___/
 *  ____       _   _                    _____                 _
 * |  _ \ __ _| |_| |_ ___ _ __ _ __   |_   _|_ _ _ __  _ __ (_)_ __   __ _
 * | |_) / _` | __| __/ _ \ '__| '_ \    | |/ _` | '_ \| '_ \| | '_ \ / _` |
 * |  __/ (_| | |_| ||  __/ |  | | | |   | | (_| | |_) | |_) | | | | | (_| |
 * |_|   \__,_|\__|\__\___|_|  |_| |_|   |_|\__,_| .__/| .__/|_|_| |_|\__, |
 *                                               |_|   |_|            |___/
 */
static void do_step_asynch(GdamMixSkin* skin)
{
  GdamMix *mix = GDAM_MIX (GDAM_SKIN (skin)->model);
  gdam_debug ("autostepping.");
  skin->asynch_position += skin->asynch_step_size * mix->beat_length;
  if (mix->is_looping && skin->asynch_position >= mix->loop_length)
    skin->asynch_position -= mix->loop_length;
  if (skin->asynch_position_label)
    {
      char txt[16];
      sprintf (txt, "%3.2f", skin->asynch_position);
      gtk_label_set_text (skin->asynch_position_label, txt);
    }
}
static GdamProducer*
gdam_mix_skin_prototype_make_producer(GdamMixSkinPrototype* prototype,
                                      GdamModel**           submodels_out)
{
  GdamProducer *producer;
  GdamModel *proto;
  GdamModel *proto_model;
  GdamModel *new_model;
  int i;

  proto = prototype->model;
  proto_model = GDAM_MODEL (proto);

  new_model = gdam_model_clone (proto_model);
  g_return_val_if_fail (new_model != NULL, NULL);
  producer = GDAM_PRODUCER (new_model);
  g_return_val_if_fail (producer != NULL, NULL);

  for (i = 0; i < prototype->num_submodels; i++)
    {
      GtkObject *sub_object;
      sub_object = gdam_arg_get_child (GTK_OBJECT (producer),
				       prototype->submodel_ident[i], -1);
      submodels_out[i] = GDAM_MODEL (sub_object);
    }
  return producer;
}

static gboolean find_which_submodel(GdamMixSkinElement* element,
				    GdamModel*          model,
                                    gboolean*           is_submodel_out,
				    int*                index_out)
{
  /* find out which prototype it was. */
  if (model == (GdamModel *) element->element->producer)
    {
      *is_submodel_out = FALSE;
    }
  else
    {
      int i;
      int num_sub = element->prototype->num_submodels;
      for (i = 0; i < num_sub; i++)
	{
	  if (element->models[i] == model)
	    break;
	}
      if (i == num_sub)
	{
	  g_warning ("child not found (in mixskin)");
	  return FALSE;
	}
      *is_submodel_out = TRUE;
      *index_out = i;
    }
  return TRUE;
}

static void gdam_mix_skin_child_added   (GtkObject*     parent,
                                         GtkObject*     child,
					 const char*    child_name,
					 gpointer       user_data)
{
  GdamMixSkinElement *skin_element;
  GdamMixElement *element;
  char *total_child_name;
  gboolean parent_is_submodel;
  gint parent_index;
  GdamModel *model;
  GdamMixSkinPrototype *prototype;
  int new_index, new_num_submodels;
  int i;
  GSList *list;
  GdamMixSkin *mix_skin;

  skin_element = (GdamMixSkinElement *) user_data;
  prototype = skin_element->prototype;
  mix_skin = prototype->mix_skin;
  element = skin_element->element;

  if (!find_which_submodel (skin_element, GDAM_MODEL (parent),
			    &parent_is_submodel, &parent_index))
    return;

  if (parent_is_submodel)
    {
      const char *parent_name;
      parent_name = prototype->submodel_ident[parent_index];
      total_child_name = g_strconcat (parent_name, "::", child_name, NULL);
    }
  else
    {
      total_child_name = g_strdup (child_name);
    }
  new_index = prototype->num_submodels;
  new_num_submodels = new_index + 1;
  prototype->submodel_ident
    = g_renew (gchar *, prototype->submodel_ident, new_num_submodels);
  prototype->submodel_ident[new_index] = total_child_name;

  /* track the new prototype. */
  skin_element->subchild_trap_ids
    = g_renew (guint, skin_element->subchild_trap_ids, new_num_submodels);
  skin_element->subchild_trap_ids[new_index] =
    gdam_arg_child_trap (GTK_OBJECT (element->producer),
			 gdam_mix_skin_child_added,
			 gdam_mix_skin_child_removed, skin_element);

  prototype->num_submodels = new_num_submodels;

  /* do the equivalent addition in each of the prototype's elements. */
  for (list = prototype->elements; list != NULL; list = list->next)
    {
      GdamMixElement *element;
      GdamMixSkinElement *skin_element;
      GdamModel *new_model;
      GdamModel *parent;
      element = list->data;
      skin_element = g_hash_table_lookup (mix_skin->skin_elements, element);
      g_return_if_fail (skin_element != NULL);

      new_model = gdam_model_clone (GDAM_MODEL (child));
      if (parent_is_submodel)
	parent = skin_element->models[parent_index];
      else
	parent = GDAM_MODEL (element->producer);
      g_return_if_fail (new_model != NULL);

      gdam_arg_set_child (GTK_OBJECT (parent),
			  child_name, -1, GTK_OBJECT (new_model));

      /* all the model to the array. */
      skin_element->models = g_renew (GdamModel *,
				      skin_element->models,
				      new_num_submodels);
      skin_element->models[new_index] = new_model;
    }
}

static void gdam_mix_skin_child_removed (GtkObject*     parent,
                                         GtkObject*     child,
					 const char*    child_name,
					 gpointer       user_data)
{
  GdamMixSkinElement *skin_element;
  gboolean parent_is_submodel;
  gint parent_index;
  gboolean child_is_submodel;
  gint child_index;
  guint trap_id;
  int num_subs;
  GdamMixSkinPrototype *prototype;
  GSList *list;
  GdamMixSkin *mix_skin;

  skin_element = (GdamMixSkinElement *) user_data;
  prototype = skin_element->prototype;
  mix_skin = prototype->mix_skin;

  /* find out which index was the parent. */
  if (!find_which_submodel (skin_element, GDAM_MODEL (parent),
			    &parent_is_submodel, &parent_index))
    return;
  if (!find_which_submodel (skin_element, GDAM_MODEL (parent),
			    &child_is_submodel, &child_index))
    return;

  g_return_if_fail (child_is_submodel);

  /* stop tracking the child. */
  trap_id = skin_element->subchild_trap_ids[child_index];
  gdam_arg_child_trap_remove (child, trap_id);
  num_subs = prototype->num_submodels;

  /* remove the equivalent from the prototype's elements. */
  for (list = prototype->elements; list != NULL; list = list->next)
    {
      GdamMixElement *equiv_element = list->data;
      GdamMixSkinElement *equiv;
      GdamModel **submodels;
      GdamProducer *prod;
      GtkObject *parent_obj;

      equiv = g_hash_table_lookup (mix_skin->skin_elements, equiv_element);
      g_return_if_fail (equiv != NULL);
      submodels = equiv->models;
      prod = equiv_element->producer;

      if (parent_is_submodel)
	{
	  parent_obj = GTK_OBJECT (submodels[parent_index]);
	}
      else
	{
	  parent_obj = GTK_OBJECT (prod);
	}
      g_error ("XXX: unimplemented");
      /*gdam_arg_remove_child(parent_obj,
         child_name,
         child_index); */
      g_memmove (submodels + child_index,
		 submodels + child_index + 1,
		 sizeof (GdamModel *) * (num_subs - child_index - 1));
    }
}


static GdamMixSkinElement* do_mix_add(GdamMixSkin*    skin,
                                      gdouble         position,
				      gdouble         y_fraction)
{
  GdamMix *mix = GDAM_MIX (GDAM_SKIN (skin)->model);
  GdamMixSkinPrototype *prototype = skin->selected_prototype;
  GdamProducer *new_prod;
  GdamModel **submodels;
  GdamMixElement *element;
  GdamMixSkinElement *skin_element;
  GdamModel *submodel_for_arg;
  int arg_index;
  int i;
  gdouble value;
  GdamArg arg;

  g_return_val_if_fail (prototype != NULL, NULL);

  /* Clone the prototype. */
  submodels = g_new (GdamModel *, prototype->num_submodels);
  new_prod = gdam_mix_skin_prototype_make_producer (prototype, submodels);

  arg_index = prototype->selected_arg;
  if (arg_index >= 0)
    {
      value = prototype->max_bounds[arg_index] * y_fraction
	+ prototype->min_bounds[arg_index] * (1.0 - y_fraction);
      arg.name = prototype->arg_names[arg_index];
      GDAM_ARG_DOUBLE (&arg) = value;
      submodel_for_arg = submodels[prototype->arg_models[arg_index]];
      gdam_arg_set_object_arg (GTK_OBJECT (submodel_for_arg), &arg);
    }

  element = gdam_mix_add (mix, new_prod, position);
  skin_element = g_new (GdamMixSkinElement, 1);
  skin_element->prototype = prototype;
  skin_element->element = element;
  skin_element->models = submodels;
  skin_element->is_dragging = FALSE;
  skin_element->is_selected = FALSE;
  skin_element->screen_data_valid = FALSE;
  skin_element->drag_data_valid = FALSE;
  skin_element->natural_y_frac = y_fraction;
  skin_element->child_trap_id =
    gdam_arg_child_trap (GTK_OBJECT (element->producer),
			 gdam_mix_skin_child_added,
			 gdam_mix_skin_child_removed, skin_element);
  skin_element->subchild_trap_ids = g_new (guint, prototype->num_submodels);
  for (i = 0; i < prototype->num_submodels; i++)
    {
      guint trap_id;
      trap_id = gdam_arg_child_trap (GTK_OBJECT (submodels[i]),
				     gdam_mix_skin_child_added,
				     gdam_mix_skin_child_removed,
				     skin_element);
      skin_element->subchild_trap_ids[i] = trap_id;
    }
  prototype->elements = g_slist_prepend (prototype->elements, skin_element);

  if (skin->skin_elements == NULL)
    skin->skin_elements = g_hash_table_new (NULL, NULL);
  g_hash_table_insert (skin->skin_elements, element, skin_element);
  return skin_element;
}

static void do_add_asynch(GdamMixSkin* skin)
{
  /* XXX: probably 0.5 isn't the best way to set the async y-value */
  do_mix_add (skin, skin->asynch_position, 0.5);
  if (skin->asynch_autostep)
    do_step_asynch (skin);
  delayed_redraw (skin);
}

static void do_autostep_asynch(GdamMixSkin* skin)
{
  gboolean rv;
  if (!skin->autostep_toggle_button)
    return;
  rv = gtk_toggle_button_get_active (skin->autostep_toggle_button);
  skin->asynch_autostep = rv;
}

#if  0
static void do_add_immediate(IntInfo* info)
{
  GdamMixSkin *skin = info->skin;
  GdamMix *mix = GDAM_MIX (GDAM_SKIN (skin)->model);
  int buffer_id;
  gboolean must_unref = FALSE;

  ...delayed_redraw (skin);
}
#endif

/*
 *  ____        _     _ _        ___       _             __
 * |  _ \ _   _| |__ | (_) ___  |_ _|_ __ | |_ ___ _ __ / _| __ _  ___ ___
 * | |_) | | | | '_ \| | |/ __|  | || '_ \| __/ _ \ '__| |_ / _` |/ __/ _ \
 * |  __/| |_| | |_) | | | (__   | || | | | ||  __/ |  |  _| (_| | (_|  __/
 * |_|    \__,_|_.__/|_|_|\___| |___|_| |_|\__\___|_|  |_|  \__,_|\___\___|
 * 
 */
void   gdam_mix_skin_remove        (GdamMixSkin*        mix_skin,
                                    GdamMixSkinElement* element)
{
  GdamMixElement *mix_element = element->element;
  GdamMix *mix;

  mix = GDAM_MIX (GDAM_SKIN (mix_skin)->model);

  g_hash_table_remove (mix_skin->skin_elements, mix_element);
  mix_skin->elements = g_slist_remove (mix_skin->elements, mix_element);
  mix_skin->dragging = g_slist_remove (mix_skin->dragging, element);
  gdam_mix_remove (mix, mix_element->producer);

  element->prototype->elements
    = g_slist_remove (element->prototype->elements, element);

  g_free (element->models);
  if (mix->freeze_count == 0)
    delayed_redraw (mix_skin);
}
/*
 *  ____                     _                _
 * |  _ \ _ __ __ ___      _(_)_ __   __ _   / \   _ __ ___  __ _
 * | | | | '__/ _` \ \ /\ / / | '_ \ / _` | / _ \ | '__/ _ \/ _` |
 * | |_| | | | (_| |\ V  V /| | | | | (_| |/ ___ \| | |  __/ (_| |
 * |____/|_|  \__,_| \_/\_/ |_|_| |_|\__, /_/   \_\_|  \___|\__,_|
 *                                   |___/
 *
 * Where we connect tons of signals to the drawingarea.
 */

static gint
gdam_mix_skin_button_press(GtkDrawingArea* area,
                           GdkEventButton* button_event,
		           GdamMixSkin*    skin)
{
  int press_x = button_event->x;
  int press_y = button_event->y;
  int width, height;
  GdamMixSkinElement *element;
  gdouble start_time;
  GdamMix *mix = GDAM_MIX (GDAM_SKIN (skin)->model);
  g_return_val_if_fail (skin->dragging == NULL, 0);
  if (button_event->button == 2)
    {
      int state = button_event->state;
      gboolean is_shifted, is_ctrl;
      begin_selection (skin, press_x, press_y);
      is_ctrl = ((state & GDK_CONTROL_MASK) == GDK_CONTROL_MASK);
      is_shifted = ((state & GDK_SHIFT_MASK) == GDK_SHIFT_MASK);
      if (is_shifted)
	skin->selection_mode = GDAM_MIX_SELECTION_ADD;
      else if (is_ctrl)
	skin->selection_mode = GDAM_MIX_SELECTION_TOGGLE;
      else
	skin->selection_mode = GDAM_MIX_SELECTION_REPLACE;
      return TRUE;
    }
  if ((element = find_element (skin, press_x, press_y)) != NULL)
    {
      g_message ("found element at %d,%d", press_x, press_y);
      if (button_event->button == 3)
	{
	  if (element->is_selected)
	    delete_selected (skin);
	  else
	    gdam_mix_skin_remove (skin, element);
	  return TRUE;
	}
      if (skin->dragging == NULL)
	begin_dragging (skin, element, button_event->state, press_x, press_y);
      return TRUE;
    }
  g_message ("didn't find element at %d,%d", press_x, press_y);

  if (button_event->button == 3)
    return 0;

  gdam_mix_freeze (mix);

  width = skin->draw_width;
  height = skin->draw_height;

  start_time = gdam_mix_skin_find_press_time (skin, press_x);
  element = do_mix_add (skin, start_time, (gdouble) press_y / height);
  g_return_val_if_fail (element != NULL, TRUE);

  gdam_mix_skin_prototype_make_string_offset_valid (element->prototype);
  redraw_element_area (skin, element);
  gdam_mix_thaw (mix);
  return TRUE;
}
static void update_selection(GdamMixSkin*        skin,
                             int                 x,
			     int                 y)
{
  g_error ("XXX: unimplemented");
}
static void finish_selection(GdamMixSkin*        skin,
                             int                 x,
			     int                 y)
{
  g_error ("XXX: unimplemented");
}

static gint
gdam_mix_skin_button_motion(GtkDrawingArea*     area,
                            GdkEventMotion*     button_event,
		            GdamMixSkin*        skin)
{
  int x, y;
  guint state;
  if (button_event->is_hint)
    gdk_window_get_pointer (button_event->window, &x, &y, &state);
  else
    {
      x = button_event->x;
      y = button_event->y;
      state = button_event->state;
    }

  if (skin->has_selection)
    {
      update_selection (skin, button_event->x, button_event->y);
      return TRUE;
    }
  if (skin->dragging)
    {
      undraw_dragged_samples (skin);
      skin->drag_current_x = button_event->x;
      skin->drag_current_y = button_event->y;
      draw_dragged_samples (skin);
    }
  return TRUE;
}

static gint
gdam_mix_skin_button_release(GtkDrawingArea*     area,
                             GdkEventButton*     button_event,
		             GdamMixSkin*        skin)
{
  if (skin->has_selection)
    {
      finish_selection (skin, button_event->x, button_event->y);
      return 0;
    }
  if (!skin->dragging)
    return 0;

  g_message ("undragging...");
  undraw_dragged_samples (skin);

  skin->drag_current_x = button_event->x;
  skin->drag_current_y = button_event->y;

  gdam_mix_skin_stop_dragging (skin);
  return TRUE;
}

static gint
gdam_mix_skin_configure(GtkWidget*         widget,
                        GdkEventConfigure* configure_event,
		        GdamMixSkin*       skin)
{
  if (skin->backing_pixmap)
    gdk_pixmap_unref (skin->backing_pixmap);
  skin->backing_pixmap = gdk_pixmap_new (widget->window,
					 widget->allocation.width,
					 widget->allocation.height, -1);
  skin->draw_width = widget->allocation.width;
  skin->draw_height = widget->allocation.height;
  return FALSE;
}

static gint
gdam_mix_skin_expose(GtkWidget*      widget,
                     GdkEventExpose* expose_event,
		     GdamMixSkin*    mix_skin)
{
  delayed_redraw (mix_skin);
  return FALSE;
}


static void gdam_mix_skin_connect_drawing_area(GdamMixSkin*    skin,
                                               GtkWidget*      drawing_area)
{
  skin->drawing_area = GTK_WIDGET (drawing_area);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "button_press_event",
		      GTK_SIGNAL_FUNC (gdam_mix_skin_button_press), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "motion_notify_event",
		      GTK_SIGNAL_FUNC (gdam_mix_skin_button_motion), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "button_release_event",
		      GTK_SIGNAL_FUNC (gdam_mix_skin_button_release), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "configure_event",
		      GTK_SIGNAL_FUNC (gdam_mix_skin_configure), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "expose_event",
		      GTK_SIGNAL_FUNC (gdam_mix_skin_expose), skin);
  gtk_widget_add_events (drawing_area,
			 GDK_EXPOSURE_MASK
			 | GDK_BUTTON_PRESS_MASK
			 | GDK_BUTTON_RELEASE_MASK
			 | GDK_POINTER_MOTION_MASK
			 | GDK_POINTER_MOTION_HINT_MASK);
}


/*
 *        _ _                         _
 *   __ _| | | ___   ___     ___ ___ | | ___  _ __ ___
 *  / _` | | |/ _ \ / __|   / __/ _ \| |/ _ \| '__/ __|
 * | (_| | | | (_) | (__   | (_| (_) | | (_) | |  \__ \
 *  \__,_|_|_|\___/ \___|___\___\___/|_|\___/|_|  |___/
 *                     |_____|
 */
static void alloc_colors(GdamMixSkin* skin)
{
  int state = GTK_WIDGET_STATE (skin->drawing_area);
  if (skin->backing_pixmap == NULL)
    return;
  skin->rect_gc = skin->drawing_area->style->fg_gc[state];
  skin->background = skin->drawing_area->style->bg_gc[state];
  skin->text_gc = gdk_gc_new (skin->backing_pixmap);
  gdk_gc_set_foreground (skin->text_gc,
			 util_alloc_color_cached (65535, 0, 32768));
  skin->font = skin->drawing_area->style->font;
  skin->grid_gc = gdk_gc_new (skin->backing_pixmap);
  gdk_gc_set_foreground (skin->grid_gc,
			 util_alloc_color_cached (65535, 65535, 0));
  skin->selection_gc = gdk_gc_new (skin->backing_pixmap);
  gdk_gc_set_foreground (skin->selection_gc,
			 util_alloc_color_cached (0, 65535, 0));
  skin->rect_boundary_gc = gdk_gc_new (skin->backing_pixmap);
  gdk_gc_set_foreground (skin->rect_boundary_gc,
			 util_alloc_color_cached (65535, 0, 0));

  skin->drag_gc = gdk_gc_new (skin->drawing_area->window);
  gdk_gc_set_foreground (skin->drag_gc,
			 util_alloc_color_cached (65535, 0, 65535));
}

/* Make the prototype into the default prototype for the skin. */
static void select_prototype (GtkButton*            button,
                              GdamMixSkinPrototype* prototype)
{
  int index;
  index = gdam_mix_skin_prototype_index (prototype);
  g_return_if_fail (index >= 0);
  gdam_mix_skin_edit_prototype (prototype->mix_skin, index);
}

/* Make the prototype into the default prototype for the skin. */
static void hide_prototype   (GtkToggleButton*      toggle_button,
                              GdamMixSkinPrototype* prototype)
{
  prototype->is_hidden = gtk_toggle_button_get_active (toggle_button);
  delayed_redraw (prototype->mix_skin);
}

/* Make the prototype into the default prototype for the skin. */
static void destroy_prototype(GtkButton*            button,
                              GdamMixSkinPrototype* prototype)
{
  int index;
  index = gdam_mix_skin_prototype_index (prototype);
  g_return_if_fail (index >= 0);
  gdam_mix_skin_delete_prototype (prototype->mix_skin, index);
}


/*
 * Make ``gdam_mix_arg_select''
 * a temporary handler of GdamSkin.
 *
 *    gdam_mix_arg_select:arg_name:min_value:max_value[:log]
 *
 * Make no arguments selected [free-y mode]:
 *    gdam_mix_arg_unselect
 */
static GSList* arg_selector_prototype_stack = NULL;

typedef struct _ConnectData ConnectData;
struct _ConnectData {
	GdamMixSkinPrototype* prototype;
	int			arg_index;
};
static void set_prototype_arg_index(ConnectData* connect_data)
{
  connect_data->prototype->selected_arg = connect_data->arg_index;
  if (connect_data->prototype->elements != NULL)
    delayed_redraw (connect_data->prototype->mix_skin);
}

static ConnectData* make_connect_data(GdamMixSkinPrototype* proto, int index)
{
  ConnectData *connect_data;
  connect_data = g_new (ConnectData, 1);
  connect_data->arg_index = index;
  connect_data->prototype = proto;
  return connect_data;
}

static gboolean gdam_mix_arg_select_handler
                                       (GdamSkin*          skin,
				        GtkObject*         object,
				        const char*        signal_name,
				        int                num_handler_names,
				        char**             handler_names)
{
  GdamMixSkinPrototype *proto;
  gdouble min_value, max_value;
  guint flags = 0;
  int arg_index;
  int new_num;
  const char *arg_name;
  g_return_val_if_fail (arg_selector_prototype_stack != NULL, TRUE);
  g_return_val_if_fail (num_handler_names >= 4, TRUE);

  proto = arg_selector_prototype_stack->data;

  arg_name = handler_names[1];
  min_value = strtod (handler_names[2], NULL);
  max_value = strtod (handler_names[3], NULL);
  if (num_handler_names >= 5)
    {
      char **subs;
      char **at;
      subs = g_strsplit (handler_names[4], ",", 0);
      for (at = subs; *at != NULL; at++)
	{
	  if (strcmp (*at, "log") == 0)
	    {
	      flags |= GDAM_MIX_SKIN_FLAGS_LOGARITHMIC;
	    }
	}
      g_strfreev (subs);
    }

  /* Resize the multitude of parallel arrays. */
  arg_index = proto->num_args;
  new_num = arg_index + 1;
  proto->arg_models = g_renew (int, proto->arg_models, new_num);
  proto->arg_names = g_renew (char *, proto->arg_models, new_num);
  proto->min_bounds = g_renew (gdouble, proto->min_bounds, new_num);
  proto->max_bounds = g_renew (gdouble, proto->max_bounds, new_num);
  proto->arg_flags = g_renew (guint, proto->arg_flags, new_num);
  proto->arg_models[arg_index] = proto->num_submodels - 1;
  proto->arg_names[arg_index] = g_strdup (arg_name);
  proto->min_bounds[arg_index] = min_value;
  proto->max_bounds[arg_index] = max_value;
  proto->arg_flags[arg_index] = flags;
  proto->num_args = new_num;

  /* Attach a signal to set proto->arg_index to arg_index. */
  gtk_signal_connect_full (object,
			   signal_name,
			   GTK_SIGNAL_FUNC (set_prototype_arg_index),
			   NULL,
			   make_connect_data (proto, arg_index),
			   g_free, TRUE, FALSE);

  return TRUE;
}
static gboolean gdam_mix_arg_unselect_handler
                                       (GdamSkin*          skin,
				        GtkObject*         object,
				        const char*        signal_name,
				        int                num_handler_names,
				        char**             handler_names)
{
  GdamMixSkinPrototype *prototype;
  g_return_val_if_fail (arg_selector_prototype_stack != NULL, TRUE);
  prototype = arg_selector_prototype_stack->data;
  gtk_signal_connect_full (object,
			   signal_name,
			   GTK_SIGNAL_FUNC (set_prototype_arg_index),
			   NULL,
			   make_connect_data (prototype, -1),
			   g_free, TRUE, FALSE);

  return TRUE;
}
static GdamSkinGenericConnect arg_selector_connect_infos[] = {
	{	"gdam_mix_arg_select",
		 gdam_mix_arg_select_handler		},
	{	"gdam_mix_arg_unselect",
		 gdam_mix_arg_unselect_handler		},
};

/* Use a global stack of prototypes;  that way they can be nested. */

/* Someday we might have connect_info_data or something that would
 * be better... but not now... */

/* redirect gdam_mix_arg_selector to refer to this prototype. */
static void push_arg_selector(GdamMixSkinPrototype* prototype)
{
  if (arg_selector_prototype_stack == NULL)
    {
      gdam_skin_class_register_generic_handlers
	((GdamSkinClass *) gtk_type_class (GDAM_TYPE_SKIN),
	 G_N_ELEMENTS (arg_selector_connect_infos),
	 arg_selector_connect_infos);
    }
  arg_selector_prototype_stack
    = g_slist_prepend (arg_selector_prototype_stack, prototype);
}
/* stop redirecting gdam_mix_arg_selector to the last prototype. */
static void pop_arg_selector()
{
  g_return_if_fail (arg_selector_prototype_stack != NULL);
  arg_selector_prototype_stack
    = g_slist_remove (arg_selector_prototype_stack,
		      arg_selector_prototype_stack->data);
  if (arg_selector_prototype_stack == NULL)
    {
      GdamSkinClass *klass;
      int i;
      klass = (GdamSkinClass *) gtk_type_class (GDAM_TYPE_SKIN);
      for (i = 0; i < G_N_ELEMENTS (arg_selector_connect_infos); i++)
	{
	  const char *name;
	  name = arg_selector_connect_infos[i].name;
	  gdam_skin_class_unregister_connection (klass, name);
	}
    }
}

int    gdam_mix_skin_new_prototype   (GdamMixSkin*        mix_skin,
                                      const char*         model_spec,
				      const char*         starting_name)
{
  GdamMixSkinPrototype *prototype;
  GdamModel *model;
  int new_num_prototypes;
  GdamSkin *skin;
  GtkBox *box;
  GtkWidget *widget;
  int i;

  model = gdam_model_from_spec (model_spec, GDAM_SKIN (mix_skin)->channel);
  if (model == NULL)
    {
      g_warning ("error creating prototype from %s", model_spec);
      return -1;
    }
  gtk_object_ref ((GtkObject *) model);
  gtk_object_sink ((GtkObject *) model);

  prototype = g_new0 (GdamMixSkinPrototype, 1);
  prototype->mix_skin = mix_skin;
  prototype->model = model;

  gdam_arg_get_children (GTK_OBJECT (model),
			 TRUE,
			 &prototype->num_submodels,
			 &prototype->submodel_ident,
			 (GtkObject ***) & prototype->submodels);
  prototype->is_hidden = FALSE;

  prototype->submodels = g_renew (GdamModel *,
				  prototype->submodels,
				  prototype->num_submodels + 1);
  prototype->submodel_ident = g_renew (char *,
				       prototype->submodel_ident,
				       prototype->num_submodels + 1);
  prototype->submodels[prototype->num_submodels] = model;
  prototype->submodel_ident[prototype->num_submodels] = g_strdup ("");
  prototype->num_submodels++;

  push_arg_selector (prototype);
  skin = gdam_skin_factory_skin (gdam_mix_skin_factory_get (), model);
  pop_arg_selector ();
  if (skin == NULL)
    {
      gtk_object_unref (GTK_OBJECT (model));
      g_free (prototype);
      return -1;
    }
  prototype->num_args = 0;
  prototype->arg_models = NULL;
  prototype->arg_names = NULL;
  prototype->min_bounds = NULL;
  prototype->max_bounds = NULL;
  prototype->skin = skin;
  /* prototype->{num_args,arg_models,arg_names,min_bounds,max_bounds}
   * are initialized now.. */

  prototype->has_string_offset = FALSE;
  prototype->selected_arg = -1;
  prototype->elements = NULL;
  prototype->prototype_name = g_strdup (starting_name);
  new_num_prototypes = mix_skin->num_prototypes + 1;
  mix_skin->prototypes = g_renew (GdamMixSkinPrototype *,
				  mix_skin->prototypes, new_num_prototypes);
  mix_skin->prototypes[mix_skin->num_prototypes] = prototype;
  mix_skin->num_prototypes = new_num_prototypes;

  gtk_widget_ref (skin->main_widget);
  if (skin->main_widget->parent != NULL)
    {
      gtk_container_remove (GTK_CONTAINER (skin->main_widget->parent),
			    skin->main_widget);
    }
  gtk_container_add (mix_skin->prototype_container, skin->main_widget);
  gtk_widget_unref (skin->main_widget);

  for (i = 0; i < prototype->num_submodels; i++)
    gdam_mix_skin_start_prototype_tracking (mix_skin, prototype, i);

  box = GTK_BOX (gtk_hbox_new (FALSE, 1));
  prototype->button_box = GTK_CONTAINER (box);

  widget = gtk_label_new (starting_name);
  prototype->name_label = GTK_LABEL (widget);
  gtk_widget_show (widget);
  gtk_box_pack_start_defaults (box, widget);

  widget = gtk_button_new_with_label ("Select");
  gtk_widget_show (widget);
  gtk_box_pack_start_defaults (box, widget);
  gtk_signal_connect (GTK_OBJECT (widget),
		      "clicked",
		      GTK_SIGNAL_FUNC (select_prototype), prototype);

  widget = gtk_toggle_button_new_with_label ("Hide");
  gtk_widget_show (widget);
  gtk_box_pack_start_defaults (box, widget);
  gtk_signal_connect (GTK_OBJECT (widget),
		      "toggled", GTK_SIGNAL_FUNC (hide_prototype), prototype);

  widget = gtk_button_new_with_label ("Destroy");
  gtk_widget_show (widget);
  gtk_box_pack_start_defaults (box, widget);
  gtk_signal_connect (GTK_OBJECT (widget),
		      "clicked",
		      GTK_SIGNAL_FUNC (destroy_prototype), prototype);

  gtk_widget_show (GTK_WIDGET (box));
  gtk_container_add (mix_skin->prototype_menu_box, GTK_WIDGET (box));

  return new_num_prototypes - 1;
}

void   gdam_mix_skin_delete_prototype(GdamMixSkin*        mix_skin,
                                      int                 prototype_id)
{
  GdamMix *mix;
  gboolean deleted_any = FALSE;
  GSList *copy;
  GdamMixSkinPrototype *prototype;
  GSList *at;
  GtkObject *object;

  mix = GDAM_MIX (GDAM_SKIN (mix_skin)->model);
  g_return_if_fail (0 <= prototype_id
		    && prototype_id < mix_skin->num_prototypes);
  prototype = mix_skin->prototypes[prototype_id];

  /* First, delete all the elements pertaining to this prototype. */
  gdam_mix_freeze (mix);
  mix = GDAM_MIX (GDAM_SKIN (mix_skin)->model);
  copy = g_slist_copy (prototype->elements);
  for (at = copy; at != NULL; at = at->next)
    {
      GdamMixSkinElement *skin_element = at->data;
      if (skin_element->prototype == prototype)
	{
	  gdam_mix_skin_remove (mix_skin, skin_element);
	  deleted_any = TRUE;
	}
    }
  g_slist_free (copy);
  gdam_mix_thaw (mix);
  if (deleted_any)
    delayed_redraw (mix_skin);

  object = GTK_OBJECT (prototype->model);
  gtk_object_destroy (object);

  /* Destroy the button box. */
  gtk_object_destroy (GTK_OBJECT (prototype->button_box));

  /* Remove the prototype from the array. */
  g_memmove (mix_skin->prototypes + prototype_id,
	     mix_skin->prototypes + prototype_id + 1,
	     (mix_skin->num_prototypes - prototype_id - 1)
	     * sizeof (GdamMixSkinPrototype *));

  if (mix_skin->selected_prototype == prototype)
    {
      mix_skin->selected_prototype = NULL;
      gtk_widget_hide (mix_skin->prototype_name_entry);
    }

  gtk_object_unref (object);
  if (prototype->skin != NULL)
    {
      gtk_object_destroy (GTK_OBJECT (prototype->skin));
      gtk_object_unref (GTK_OBJECT (prototype->skin));
    }

  g_free (prototype);
}

void gdam_mix_skin_set_prototype_name(GdamMixSkin*          mix_skin,
	                              GdamMixSkinPrototype* prototype,
				      const char*           new_name)
{
  char *name;
  if (strcmp (prototype->prototype_name, new_name) == 0)
    return;
  gtk_label_set_text (prototype->name_label, new_name);
  name = g_strdup (new_name);
  g_free (prototype->prototype_name);
  prototype->prototype_name = name;
  delayed_redraw (mix_skin);
}

/*
 *  _                        _
 * | | __ _ _   _ _ __   ___| |__
 * | |/ _` | | | | '_ \ / __| '_ \
 * | | (_| | |_| | | | | (__| | | |
 * |_|\__,_|\__,_|_| |_|\___|_| |_|
 * Creating a new prototype.
 */
typedef struct _LaunchInfo LaunchInfo;
struct _LaunchInfo {
	GdamMixSkin*		mix_skin;
	char*			model_spec;
	char*			name;
};
static void do_launch(LaunchInfo*    launch_info)
{
  GdamMixSkin *mix_skin = launch_info->mix_skin;
  int new_index;
  const char *model_spec = launch_info->model_spec;

  /* Create a new prototype. */
  new_index = gdam_mix_skin_new_prototype (mix_skin,
					   model_spec, launch_info->name);
  if (new_index < 0)
    {
      g_warning ("error creating new prototype");
      return;
    }
  g_return_if_fail (new_index < mix_skin->num_prototypes);


  /* Begin editting it. */
  gdam_mix_skin_edit_prototype (mix_skin, new_index);
}

static gboolean gdam_mix_launch_handler(GdamSkin*          skin,
				        GtkObject*         object,
				        const char*        signal_name,
				        int                num_handler_names,
				        char**             handler_names)
{
  LaunchInfo *launch_info;
  g_return_val_if_fail (num_handler_names == 3, TRUE);
  launch_info = g_new (LaunchInfo, 1);
  launch_info->mix_skin = GDAM_MIX_SKIN (skin);
  launch_info->model_spec = g_strdup (handler_names[1]);
  launch_info->name = g_strdup (handler_names[2]);
  gtk_signal_connect_full (object,
			   signal_name,
			   GTK_SIGNAL_FUNC (do_launch),
			   NULL, launch_info, g_free, TRUE, FALSE);
  return TRUE;
}

/*
 *        _                __   _
 *  _ __ | | __ _ _   _   / /__| |_ ___  _ __
 * | '_ \| |/ _` | | | | / / __| __/ _ \| '_ \
 * | |_) | | (_| | |_| |/ /\__ \ || (_) | |_) |
 * | .__/|_|\__,_|\__, /_/ |___/\__\___/| .__/
 * |_|            |___/                 |_|
 */
static gboolean gdam_mix_play_handler (GdamSkin*          skin,
				       GtkObject*         object,
				       const char*        signal_name,
				       int                num_handler_names,
				       char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (gdam_producer_play_now),
			     (GtkObject *) GDAM_PRODUCER (skin->model));
  return TRUE;
}

static gboolean gdam_mix_skin_build_new_model(GdamSkin*      skin,
				              GdamChannel*   channel,
					      int            num_options,
	                                      char**         options)
{
  GdamModel *model;
  model = GDAM_MODEL (gtk_type_new (GDAM_TYPE_MIX));
  model->channel = channel;
  skin->model = model;
  /*gtk_signal_connect(GTK_OBJECT(model),
     "set_beat_info",
     on_set_bi,
     skin);
     gtk_signal_connect(GTK_OBJECT(model),
     "sequence_changed",
     on_sequence_changed,
     skin); */
  return TRUE;
}

static void free_second_arg(gpointer a, gpointer b, gpointer c)
{
  g_free (b);
}

static void gdam_mix_skin_destroy(GtkObject* object)
{
  GdamMixSkin *mix_skin;
  mix_skin = GDAM_MIX_SKIN (object);
  g_slist_free (mix_skin->elements);
  mix_skin->elements = NULL;
  g_hash_table_foreach (mix_skin->skin_elements, free_second_arg, NULL);
  g_hash_table_destroy (mix_skin->skin_elements);
  mix_skin->skin_elements = NULL;
  (*parent_class->destroy) (object);
}

/*
 *   ____                            _     _____
 *  / ___|___  _ __  _ __   ___  ___| |_  |  ___|   _ _ __   ___
 * | |   / _ \| '_ \| '_ \ / _ \/ __| __| | |_ | | | | '_ \ / __|
 * | |__| (_) | | | | | | |  __/ (__| |_  |  _|| |_| | | | | (__
 *  \____\___/|_| |_|_| |_|\___|\___|\__| |_|   \__,_|_| |_|\___|
 * 
 */

static void on_beatnum_adjusted(GtkAdjustment*  adjustment,
                                GdamMixSkin*    mix_skin)
{
  GdamMix *mix;
  gdouble num_beats = adjustment->value;
  mix = GDAM_MIX (GDAM_SKIN (mix_skin)->model);
  //g_return_if_fail(mix->is_looping);
  mix->loop_length = num_beats * mix->beat_length;
}

static gboolean gdam_mix_add_asynch_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (do_add_asynch),
			     (GtkObject *) skin);
  return TRUE;
}
#if 0
static gboolean gdam_mix_add_immediate_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  int num;
  //IntInfo* info;
  char *endp;

  if (handler_names[1] == NULL)
    {
      num = 0;
    }
  else
    {
      num = strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
    }

  info = g_new (IntInfo, 1);
  info->skin = GDAM_SEQUENCER_SKIN (skin);
  info->num = num;

  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (do_add_immediate),
			     (GtkObject *) info);
  return TRUE;
}
#endif
static gboolean gdam_mix_asynch_position_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamMixSkin *mix_skin = (GdamMixSkin *) skin;
  mix_skin->asynch_position_label = GTK_LABEL (object);
  return TRUE;
}
static gboolean gdam_mix_autostep_asynch_tbutton_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamMixSkin *mix_skin;
  mix_skin = GDAM_MIX_SKIN (skin);
  mix_skin->autostep_toggle_button = GTK_TOGGLE_BUTTON (object);
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (do_autostep_asynch),
			     (GtkObject *) skin);
  return TRUE;
}
static gboolean gdam_mix_beatnum_adjuster_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (on_beatnum_adjusted), skin);
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static gboolean gdam_mix_prototype_menu_box_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  if (!GTK_IS_CONTAINER (object))
    {
      g_warning ("arg to %s must be a container", handler_names[0]);
      return TRUE;
    }
  if (GTK_IS_BIN (object))
    {
      g_warning ("arg to %s must not be a bin", handler_names[0]);
      return TRUE;
    }
  GDAM_MIX_SKIN (skin)->prototype_menu_box = GTK_CONTAINER (object);
  return TRUE;
}

static gboolean gdam_mix_prototype_container_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  if (!GTK_IS_CONTAINER (object))
    {
      g_warning ("arg to %s must be a container", handler_names[0]);
      return TRUE;
    }
  if (GTK_IS_BIN (object))
    {
      g_warning ("arg to %s must not be a bin", handler_names[0]);
      return TRUE;
    }
  GDAM_MIX_SKIN (skin)->prototype_container = GTK_CONTAINER (object);
  return TRUE;
}
static gboolean gdam_mix_drawing_area_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  if (!GTK_IS_DRAWING_AREA (object))
    {
      g_warning ("%s may only be connected to a drawing area",
		 handler_names[0]);
    }
  gdam_mix_skin_connect_drawing_area (GDAM_MIX_SKIN (skin),
				      GTK_WIDGET (object));
  return TRUE;
}
static void horizontal_pan(GtkAdjustment* adjustment, GdamMixSkin* skin)
{
  gdouble start_beat = adjustment->value;
  gdouble cur_width = skin->max_beat - skin->min_beat;
  skin->min_beat = start_beat;
  skin->max_beat = start_beat + cur_width;
  delayed_redraw (skin);
}
static gboolean gdam_mix_h_pan_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adjustment;
  GdamMixSkin *mix_skin = GDAM_MIX_SKIN (skin);
  adjustment = util_get_widget_adjustment (GTK_WIDGET (object));
  mix_skin->time_pan_adj = adjustment;
  gtk_signal_connect (GTK_OBJECT (adjustment),
		      "value_changed",
		      GTK_SIGNAL_FUNC (horizontal_pan), skin);
  gdam_delayed_adjustment_init (adjustment);
  return TRUE;
}
static void horizontal_zoom(GtkAdjustment* adjustment, GdamMixSkin* skin)
{
  gdouble cur_width = skin->max_beat - skin->min_beat;
  gdouble delta = adjustment->value - cur_width;
  //skin->min_beat -= delta / 2;
  //skin->max_beat += delta / 2;
  skin->max_beat += delta;
  delayed_redraw (skin);
}
static gboolean gdam_mix_h_zoom_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  GdamMixSkin *mix_skin = GDAM_MIX_SKIN (skin);
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  mix_skin->time_zoom_adj = adj;
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (horizontal_zoom), skin);
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static void on_asynch_step_size_adjusted(GtkAdjustment* adjustment, 
                                         GdamMixSkin*   mix_skin)
{
  GdamMix *mix = (GdamMix *) ((GdamSkin *) mix_skin)->model;
  mix_skin->asynch_step_size = adjustment->value * mix->beat_length;
}

static gboolean gdam_mix_set_asynch_step_size_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (on_asynch_step_size_adjusted), skin);
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}
static void on_lag_adjusted             (GtkAdjustment* adjustment, 
                                         GdamMixSkin*   mix_skin)
{
  float new_lag = adjustment->value;
  GdamMix *mix;
  GdamModel *model;
  model = GDAM_SKIN (mix_skin)->model;
  if (model == NULL)
    return;
  mix = GDAM_MIX (model);
  mix->immediate_lag = new_lag;
}


static gboolean gdam_mix_set_lag_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (on_lag_adjusted), skin);
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static void on_subbeats_adjusted(GtkAdjustment* adjustment,
                                 GdamMixSkin*   mix_skin)
{
  mix_skin->sub_beats = (int) adjustment->value;
}
	

static gboolean gdam_mix_set_sub_beats_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (on_subbeats_adjusted), skin);
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static gboolean gdam_mix_step_asynch_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (do_step_asynch),
			     (GtkObject *) skin);
  return TRUE;
}
static void on_tempo_adjusted(GtkAdjustment*     adjustment, 
                              GdamMixSkin*       mix_skin)
{
  GdamMix *mix = GDAM_MIX (GDAM_SKIN (mix_skin)->model);
  gdam_mix_set_beat_length (mix, adjustment->value);
}
static gboolean on_mix_set_beat_info(GdamModel*          model,
	                             const GdamBeatInfo* new_info,
				     GdamMixSkin*        mix_skin)
{
  gdouble bpm;
  GtkAdjustment *adjustment = mix_skin->tempo_adj;
  GtkObject *object = GTK_OBJECT (adjustment);
  gtk_signal_handler_block (object, mix_skin->tempo_adj_connect);
  bpm = gdam_beat_info_get_bpm (new_info, 0);	/* FIXME: fix position? */
  gtk_adjustment_set_value (adjustment, 60.0 / bpm);
  gtk_signal_handler_unblock (object, mix_skin->tempo_adj_connect);
  return TRUE;
}
static gboolean gdam_mix_tempo_adjuster_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamMixSkin *mix_skin = GDAM_MIX_SKIN (skin);
  GtkAdjustment *adj;
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  mix_skin->tempo_adj_connect
    = gtk_signal_connect (GTK_OBJECT (adj),
			  "value_changed",
			  GTK_SIGNAL_FUNC (on_tempo_adjusted), skin);
  gtk_signal_connect_object (GTK_OBJECT (skin->model),
			     "set_beat_info",
			     GTK_SIGNAL_FUNC (on_mix_set_beat_info),
			     GTK_OBJECT (adj));
  mix_skin->tempo_adj = adj;
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static void on_tempo_string_drop(const char* tempo, gpointer user_data)
{
  GdamMixSkin *skin = GDAM_MIX_SKIN (user_data);
  GdamModel *model = GDAM_SKIN (skin)->model;
  GdamBeatInfo *new_beat_info = gdam_beat_info_copy (&model->beat_info);
  gdam_beat_info_set_bpm (new_beat_info, strtod (tempo, NULL));
  gdam_model_set_beat_info (model, new_beat_info);
  gdam_beat_info_destroy (new_beat_info);
}

static gboolean gdam_mix_tempo_drop_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gdam_widget_string_dest_set (GTK_WIDGET (object),
			       on_tempo_string_drop, "gdam-tempo-text", skin);
  return TRUE;
}

static void prototype_name_entry(GtkEntry* entry, GdamMixSkin* mix_skin)
{
  if (mix_skin->selected_prototype == NULL)
    return;

  gdam_mix_skin_set_prototype_name (mix_skin,
				    mix_skin->selected_prototype,
				    gtk_entry_get_text (entry));
}

static gboolean gdam_mix_prototype_name_entry_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  g_return_val_if_fail (GTK_IS_ENTRY (object), TRUE);

  GDAM_MIX_SKIN (skin)->prototype_name_entry = GTK_WIDGET (object);
  gtk_signal_connect (object,
		      "changed",
		      GTK_SIGNAL_FUNC (prototype_name_entry), skin);
  gtk_widget_hide (GTK_WIDGET (object));
  return TRUE;
}

static GdamSkinGenericConnect connect_infos[] = {
  { "gdam_mix_drawing_area", gdam_mix_drawing_area_handler} ,
  { "gdam_mix_tempo_drop", gdam_mix_tempo_drop_handler} ,
  { "gdam_mix_launch", gdam_mix_launch_handler} ,
    /*{ "gdam_mix_add_immediate",
       gdam_mix_add_immediate_handler                       }, */
  { "gdam_mix_prototype_name_entry", gdam_mix_prototype_name_entry_handler} ,
  { "gdam_mix_add_asynch", gdam_mix_add_asynch_handler} ,
  { "gdam_mix_asynch_position_label", gdam_mix_asynch_position_label_handler} ,
  { "gdam_mix_step_asynch", gdam_mix_step_asynch_handler} ,
  { "gdam_mix_autostep_asynch_tbutton",
     gdam_mix_autostep_asynch_tbutton_handler} ,
  { "gdam_mix_set_sub_beats", gdam_mix_set_sub_beats_handler} ,
  { "gdam_mix_set_asynch_step_size", gdam_mix_set_asynch_step_size_handler} ,
  { "gdam_mix_set_lag", gdam_mix_set_lag_handler} ,
  { "gdam_mix_beatnum_adjuster", gdam_mix_beatnum_adjuster_handler} ,
  { "gdam_mix_tempo_adjuster", gdam_mix_tempo_adjuster_handler} ,
  { "gdam_mix_h_zoom", gdam_mix_h_zoom_handler} ,
  { "gdam_mix_h_pan", gdam_mix_h_pan_handler} ,
  { "gdam_mix_prototype_menu_box", gdam_mix_prototype_menu_box_handler} ,
  { "gdam_mix_prototype_container", gdam_mix_prototype_container_handler}
};

static void gdam_mix_skin_class_init(GdamMixSkinClass* mix_skin_class)
{
  GdamSkinClass *skin_class;
  skin_class = GDAM_SKIN_CLASS (mix_skin_class);
  skin_class->build_new_model = gdam_mix_skin_build_new_model;
  GTK_OBJECT_CLASS (skin_class)->destroy = gdam_mix_skin_destroy;
  gdam_skin_class_register_generic_handlers
    (skin_class, G_N_ELEMENTS (connect_infos), connect_infos);
}

GtkType gdam_mix_skin_get_type()
{
  static GtkType mix_skin_type = 0;
  if (!mix_skin_type)
    {
      static const GtkTypeInfo mix_skin_info = {
	"GdamMixSkin",
	sizeof (GdamMixSkin),
	sizeof (GdamMixSkinClass),
	(GtkClassInitFunc) gdam_mix_skin_class_init,
	(GtkObjectInitFunc) gdam_mix_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      parent_class = gtk_type_class (parent);
      mix_skin_type = gtk_type_unique (parent, &mix_skin_info);
    }
  return mix_skin_type;
}


/*
 *  ____  _    _         _____          _
 * / ___|| | _(_)_ __   |  ___|_ _  ___| |_ ___  _ __ _   _
 * \___ \| |/ / | '_ \  | |_ / _` |/ __| __/ _ \| '__| | | |
 *  ___) |   <| | | | | |  _| (_| | (__| || (_) | |  | |_| |
 * |____/|_|\_\_|_| |_| |_|  \__,_|\___|\__\___/|_|   \__, |
 *                                                    |___/
 */
typedef struct _SkinTypeInfo SkinTypeInfo;
struct _SkinTypeInfo {
	const char*		model_type_name;
	const char*		xml_name_spec;
	const char*		skin_type_name;

	/* These are computed from xml_name... */
	int 			num_options;
	char**			options;
	char*			xml_name;

	/* Computed from skin_type_name. */
	GtkType			type;
};
#define SKIN_TYPE_INFO_DEFAULT_ARGS	0,NULL,NULL,0

static GdamSkin* generic_builder_func(GdamModel* model, gpointer builder_data)
{
  SkinTypeInfo *type_info = builder_data;
  GdamSkin *rv;
  GtkType type;
  const char *skin_type_name = type_info->skin_type_name;
  type = type_info->type;
  if (type == 0)
    {
      type = gdam_plugin_get_type_by_name ("skin", skin_type_name);
      type_info->type = type;
    }
  if (type == 0)
    {
      g_warning ("couldn't find skin of type %s", skin_type_name);
      return NULL;
    }
  rv = gdam_skin_new_with_model (type,
				 model,
				 type_info->xml_name,
				 type_info->num_options, type_info->options);
  if (rv != NULL)
    {
      GtkObject *object;
      object = GTK_OBJECT (rv);
      gtk_object_ref (object);
      gtk_object_sink (object);
    }
  return rv;
}

static void skin_type_info_add_array(GdamSkinFactory*      factory,
                                     SkinTypeInfo*         type_info,
				     int                   num_type_info)
{
  int i;
  for (i = 0; i < num_type_info; i++)
    {
      const char *colon;
      const char *xml = type_info[i].xml_name_spec;
      colon = strchr (xml, ':');
      if (colon == NULL)
	{
	  type_info[i].xml_name = g_strdup (xml);
	  type_info[i].num_options = 0;
	  type_info[i].options = NULL;
	}
      else
	{
	  type_info[i].xml_name = g_new (char, colon - xml + 1);
	  memcpy (type_info[i].xml_name, xml, colon - xml);
	  type_info[i].xml_name[colon - xml] = 0;
	  type_info[i].num_options = 0;
	  type_info[i].options = g_strsplit (colon + 1, ":", 0);
	  while (type_info[i].options[type_info[i].num_options])
	    type_info[i].num_options++;
	}
      gdam_skin_factory_add_method (factory,
				    type_info[i].model_type_name,
				    generic_builder_func, type_info + i);
    }
}
static SkinTypeInfo skin_type_info[] = {
	{ "GdamGain",
	  "submixer-gain.xml",
	  "GdamGainSkin",
	  SKIN_TYPE_INFO_DEFAULT_ARGS },
	{ "GdamBuilder",
	  "submixer-builder.xml",
	  "GdamBuilderSkin",
	  SKIN_TYPE_INFO_DEFAULT_ARGS },
	{ "GdamProducerWavetable",
	  "submixer-wavetable.xml",
	  "GdamWavetableSkin",
	  SKIN_TYPE_INFO_DEFAULT_ARGS },
};

static GdamSkinFactory* gdam_mix_skin_factory_get()
{
  static GdamSkinFactory *the_factory = NULL;

  if (the_factory != NULL)
    return the_factory;

  the_factory = gdam_skin_factory_new ();
  skin_type_info_add_array (the_factory,
			    skin_type_info, G_N_ELEMENTS (skin_type_info));
  return the_factory;
}
