/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamturntableskin.h"
#include <gdam/component/gdamproducer.h>
#include <stdlib.h>
#include <string.h>
#include <gdk/gdk.h>
#include <glade/glade.h>
#include <gtk/gtkdnd.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkrange.h>
#include <gtk/gtklabel.h>
#include <gdam/component/gdamutil.h>
#include "gdamdnd.h"
#include "gdamskin.h"
#include "gdamskinutil.h"
#include "gdamkeyboardtrap.h"
#include "gtktrackdisplay.h"

enum {
	MATCH_TEMPO,
	LAST_SIGNAL
};
static guint turntable_signals[LAST_SIGNAL] = {0};


static gboolean gdam_turntable_build_new_model(GdamSkin*      skin,
					       GdamChannel*   channel,
					       int            num_options,
				               char**         options);
static void gdam_turntable_skin_synchronize   (GdamView*      view);

/*    ____ _                 ____  _          __  __ 
 *   / ___| | __ _ ___ ___  / ___|| |_ _   _ / _|/ _|
 *  | |   | |/ _` / __/ __| \___ \| __| | | | |_| |_ 
 *  | |___| | (_| \__ \__ \  ___) | |_| |_| |  _|  _|
 *   \____|_|\__,_|___/___/ |____/ \__|\__,_|_| |_|  
 */                                                 

static void gdam_turntable_skin_class_init(GdamTurntableSkinClass* klass);

static void gdam_turntable_skin_init(GdamTurntableSkin* skin)
{
  skin->fine_rate_control = 1;
}

GtkType gdam_turntable_skin_get_type()
{
  static GtkType turntable_skin_type = 0;
  if (!turntable_skin_type)
    {
      static const GtkTypeInfo turntable_skin_info = {
	"GdamTurntableSkin",
	sizeof (GdamTurntableSkin),
	sizeof (GdamTurntableSkinClass),
	(GtkClassInitFunc) gdam_turntable_skin_class_init,
	(GtkObjectInitFunc) gdam_turntable_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      turntable_skin_type = gtk_type_unique (GDAM_TYPE_SKIN,
					     &turntable_skin_info);
    }
  return turntable_skin_type;
}


/*  ____  _    _         _____                 _   _                 
 * / ___|| | _(_)_ __   |  ___|   _ _ __   ___| |_(_) ___  _ __  ___ 
 * \___ \| |/ / | '_ \  | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
 *  ___) |   <| | | | | |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
 * |____/|_|\_\_|_| |_| |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
 *
 */                                                                 

static void on_tempo_string_drop(const char* 		new_bpm_str,
				 gpointer		user_data)
{
  GdamTurntable *tt = GDAM_TURNTABLE (user_data);
  float new_bpm = atof (new_bpm_str);
  if (new_bpm > 0)
    turntable_set_tempo (tt, new_bpm);
}

/*  This function is called regularly to keep the sync preview up to date
 */ 
static int auto_sync_timeout(GdamTurntableSkin* skin)
{
  if (GDAM_SKIN (skin)->model->source_id == 0)
    {
      skin->timer_id = 0;
      return FALSE;
    }
  turntable_do_sync_preview (GDAM_TURNTABLE (GDAM_SKIN (skin)->model));
  return TRUE;
}

static void on_jump_adj_changed(GtkAdjustment* adj, GdamTurntable* turntable)
{
  gdam_debug ("tt:jump_adj changed - %f", adj->value);
  gdam_turntable_update_jump_preview (turntable,
				      (double) abs ((int) (adj->value)));
}

static void
on_jump_button_event(GtkWidget* widget, GdkEvent* ev, GdamTurntable* turntable)
{
  GtkRange *range = GTK_RANGE (widget);
  GtkAdjustment *adj = GTK_ADJUSTMENT (gtk_range_get_adjustment (range));
  if (ev->type == GDK_BUTTON_RELEASE && adj)
    {
      gfloat jump = adj->value;
      gdam_turntable_seek_by_beats (turntable, (int) jump);
      adj->value = 0;
      gtk_adjustment_changed (adj);
      gdam_turntable_update_jump_preview (turntable, 0);
    }
}

static void position_adj_changed(GtkAdjustment* adj, GdamTurntable* turntable)
{
  GdamModel *model = GDAM_MODEL (turntable);
  if (turntable->user_dragging_position)
    {
      gdam_producer_set_position (turntable->producer, adj->value);
      return;
    }
  if (model->source_id != 0 && turntable->producer != NULL)
    {
      GdamArg arg;
      arg.name = "playing";
      arg.type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (&arg) = 0;
      /* TODO: is it possible for this to change under us?? */
      GDAM_PRODUCER (turntable)->stopped_position = adj->value;
      gdam_arg_set_object_arg (GTK_OBJECT (turntable), &arg);
      //GdamPlayInfo play_info = {0};
      //play_info.position = adj->value;
      //gdam_producer_play(turntable->producer, &play_info);
    }
  else
    {
      gdam_producer_set_position (turntable->producer, adj->value);
    }
}


static void posscroll_button_event(GtkWidget*     wid, 
			    GdkEvent*      ev, 
			    GdamTurntable* turntable)
{
  if (ev->type == GDK_BUTTON_PRESS)
    {
      turntable->user_dragging_position = 1;
    }
  else if (ev->type == GDK_BUTTON_RELEASE)
    {
      if (turntable->producer
	  && GDAM_MODEL (turntable->producer)->source_id != 0)
	{
	  /* TODO:  this action must be deferred,
	   *        as it is not good here... */
	  gdam_producer_seek (GDAM_PRODUCER (turntable),
			      GDAM_PRODUCER (turntable)->stopped_position, 0);
	}
      turntable->user_dragging_position = 0;
    }
}

/*  These are from infoclient code...
 *
 */
static void
on_anchor_rbutton_toggled(GtkToggleButton*      button,
			  GdamTurntable*  	turntable)
{
  gdam_turntable_set_anchored (turntable,
			       gtk_toggle_button_get_active (button));
}

static void on_halve_check_button_clicked(GtkButton* button,
					  GdamTurntable* turntable)
{
  GdamBeatInfo *new_beat;
  const GdamBeatInfo *prod_beat;
  float bpm;
  float fb;
  if (GTK_TOGGLE_BUTTON (button)->active)
    turntable->bpm_factor = 2;
  else
    turntable->bpm_factor = 1;
  if (!turntable->producer)
    return;
  prod_beat = &(GDAM_MODEL (turntable->producer)->beat_info);
  bpm = gdam_turntable_get_current_bpm (turntable);
  fb = gdam_turntable_get_current_fb (turntable);
  new_beat = gdam_beat_info_new ();
  gdam_beat_info_set_bpm (new_beat, bpm);
  gdam_beat_info_set_first_beat (new_beat, fb);
  gdam_model_set_beat_info (GDAM_MODEL (turntable), new_beat);
  gdam_beat_info_destroy (new_beat);
}


/*
 * Functions to update display widgets
 *
 */
void turntable_update_pos_label(GdamTurntableSkin* skin, gdouble _new_time,
					GdamTurntable* turntable)
{
  char text[63];
  gdouble new_time;
  int secs;
  int centisecs;
  skin->last_time = _new_time;
  new_time = turntable->rate ? _new_time * turntable->rate : _new_time;
  new_time -= skin->timer_origin;
  secs = (int) (new_time);
  centisecs = (int) ((new_time - (float) secs) * 100.0);
  if (new_time < 0)
    sprintf (text, "-%02d:%02d.%02d",
	     abs (secs / 60), abs (secs % 60), abs (centisecs));
  else
    sprintf (text, "%03d:%02d.%02d", secs / 60, secs % 60, centisecs);
  gtk_label_set_text (skin->time_label, text);
}

void turntable_update_hour_label(GdamTurntableSkin* skin, gdouble _new_time,
					GdamTurntable* turntable)
{
  char text[63];
  int secs;
  gdouble new_time;
  skin->last_time = _new_time;
  new_time = turntable->rate ? _new_time * turntable->rate : _new_time;
  new_time -= skin->timer_origin;
  secs = (int) (new_time);
  if (new_time < 0)
    sprintf (text, "-%2d", abs (secs / 3600));
  else
    sprintf (text, "%2d", secs / 3600);
  gtk_label_set_text (skin->hour_label, text);
}

void turntable_update_minute_label(GdamTurntableSkin* skin, gdouble _new_time,
					GdamTurntable* turntable)
{
  char text[63];
  int secs;
  gdouble new_time;
  skin->last_time = _new_time;
  new_time = _new_time * turntable->rate;
  new_time -= skin->timer_origin;
  secs = (int) (new_time);
  sprintf (text, "%02d", abs (secs / 60));
  gtk_label_set_text (skin->minute_label, text);
}


void turntable_update_second_label(GdamTurntableSkin* skin, gdouble _new_time,
					GdamTurntable* turntable)
{
  char text[63];
  gdouble new_time;
  int secs;
  skin->last_time = _new_time;
  new_time = _new_time * turntable->rate;
  new_time -= skin->timer_origin;
  secs = (int) (new_time);
  sprintf (text, "%02d", abs (secs % 60));
  gtk_label_set_text (skin->second_label, text);
}


static void  turntable_update_beatnum(GtkLabel* label, gdouble new_beatnum,
					GtkObject* object)
{
  char text[63];
  sprintf (text, "%3.3f", new_beatnum);
  gtk_label_set_text (label, text);
}

static void update_sync_preview_adj(GtkAdjustment* adj, gdouble new_val,
					GtkObject* object)
{
  gtk_adjustment_set_value (adj, new_val);
}

static void update_sync_preview_label(GtkLabel* label, gdouble new_val,
					GtkObject* object)
{
  GdamTurntable *tt = GDAM_TURNTABLE (object);
  if (tt && !GDAM_MODEL (tt)->anchored)
    {
      char text[63];
      sprintf (text, "%1.3f", new_val);
      gtk_label_set_text (label, text);
    }
}

static void  turntable_update_jump_preview(GtkLabel* label, gdouble new_val,
						GtkObject* object)
{
  char text[63];
  gdam_debug ("tt:jump_prev - %f", new_val);
  sprintf (text, "%2.0f", new_val);
  gtk_label_set_text (label, text);
}

static void on_tempo_press_event(GtkWidget* widget, GdkEventButton* event,
				 gpointer udata)
{
  if (event->type == GDK_2BUTTON_PRESS)
    {
      GdamModel *model = GDAM_SKIN (udata)->model;
      GdamTurntable *turntable = GDAM_TURNTABLE (model);
      float bpm = gdam_infomanager_get_desired_bpm (model);
      if (bpm <= 0 || !turntable)
	{
	  /*failure */
	  return;
	}
      turntable_set_tempo (turntable, bpm);
    }
}

static void on_match_desired_tempo(gpointer udata)
{
  GdamModel *model = GDAM_SKIN (udata)->model;
  GdamTurntable *turntable = GDAM_TURNTABLE (model);
  float bpm = gdam_infomanager_get_desired_bpm (model);
  if (bpm <= 0 || !turntable)
    {
      /*failure */
      return;
    }
  turntable_set_tempo (turntable, bpm);
}


static void turntable_update_position_adj_value(GtkAdjustment* adj, 
					gdouble new_val, GtkObject* object)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (object);
  if (turntable && !turntable->user_dragging_position && adj)
    {
      adj->value = new_val * turntable->rate;
      if (adj->value > adj->upper)
	adj->value = adj->upper;
      else if (adj->value < adj->lower)
	adj->value = adj->lower;
      gtk_adjustment_changed (adj);
    }
}

static void track_display_length_changed(GdamTurntable *turntable,
					 gdouble length,
					 GtkTrackDisplay *display)
{
  g_return_if_fail  (GTK_IS_TRACK_DISPLAY (display));
  gtk_track_display_set_length (display, turntable->track_length);
}

static void track_display_tt_beatinfo_changed(GdamModel       *tt,
					      GdamBeatInfo    *beatinfo,
					      GtkTrackDisplay *display)
{
  GdamTurntable *turntable;
  turntable = GDAM_TURNTABLE (tt);
  g_return_if_fail  (GTK_IS_TRACK_DISPLAY (display));
  gtk_track_display_set_range_index (display, turntable->beatinfo_index);
}

static void track_display_beatinfo_changed(GdamModel       *producer,
					   GdamBeatInfo    *beatinfo,
					   GtkTrackDisplay *display)
{
  GdamBeatInfo *pbi;
  g_return_if_fail  (GTK_IS_TRACK_DISPLAY (display));
  pbi = beatinfo;
  gtk_track_display_set_beat_info (display, pbi);
}

static void do_cycle_beatinfo(GdamSkin* _skin){
	GdamTurntable* turntable = GDAM_TURNTABLE(_skin->model);
	gdam_turntable_cycle_beat_info(turntable);
}

static void nudge_timer_origin(GdamSkin* _skin){
	GdamTurntableSkin* skin = GDAM_TURNTABLE_SKIN(_skin);
	gdouble new_origin = 0;
	gdouble first_beat;
	gdouble length;
	GdamModel* model = _skin->model;
	GdamTurntable* turntable = GDAM_TURNTABLE(model);
	first_beat = gdam_turntable_get_current_producer_fb(turntable);
	length = turntable->track_length;
	if (first_beat > skin->timer_origin)
		new_origin = first_beat;
	else if (length > skin->timer_origin)
		new_origin = length;
	skin->timer_origin = new_origin;
	turntable_update_pos_label(skin, skin->last_time, turntable);
	if (skin->track_display)
	  gtk_track_display_set_origin (skin->track_display, new_origin);
}

static void timer_origin_zero(GdamSkin* _skin){
	GdamTurntableSkin* skin = GDAM_TURNTABLE_SKIN(_skin);
	GdamModel* model = _skin->model;
	GdamTurntable* turntable = GDAM_TURNTABLE(model);
	gdouble new_origin = 0;
	skin->timer_origin = new_origin;
	turntable_update_pos_label(skin, skin->last_time, turntable);
	if (skin->track_display)
	  gtk_track_display_set_origin (skin->track_display, new_origin);
}

static void timer_origin_length(GdamSkin* _skin){
	GdamTurntableSkin* skin = GDAM_TURNTABLE_SKIN(_skin);
	GdamModel* model = _skin->model;
	GdamTurntable* turntable = GDAM_TURNTABLE(model);
	gdouble new_origin = 0;
	new_origin = turntable->track_length;
	skin->timer_origin = new_origin;
	turntable_update_pos_label(skin, skin->last_time, turntable);
	if (skin->track_display)
	  gtk_track_display_set_origin (skin->track_display, new_origin);
}

static void turntable_update_position_adj_upper(GtkAdjustment* adj, 
					gdouble new_val, GtkObject* object)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (object);
  if (turntable && adj)
    {
      adj->upper = new_val * turntable->rate + adj->page_size;
      gtk_adjustment_changed (adj);
    }
}

/* A simple callback which traps the state of the most recent button_press callback */
static gboolean any_pressed_callback(GtkObject*         object,
                                     GdkEventButton*    event,
				     GdamTurntableSkin* skin)
{
  skin->event_state = event->state;
  return FALSE;
}

/* Connects the clicked signal of the button to the handler, and also connects 
 * to the button_press_event signal, in order to trap the state of the event 
 * which produced the click. */
static void connect_button_with_mask(GdamSkin*          skin, 
                                     GtkObject*         button,
				     const char*        signal_name,
				     GtkSignalFunc      clicked_callback,
				     gpointer           user_data)
{
  gtk_signal_connect (button, signal_name, clicked_callback, user_data);
  if (!GTK_IS_WIDGET (button))
    return;
  gtk_signal_connect (button,
		      "button_press_event",
		      GTK_SIGNAL_FUNC (any_pressed_callback), skin);
}

gboolean gdam_turntable_skin_set_producer(	GdamTurntableSkin* 	skin,
						const char* 	skin_type_name, 
						const char* 	skin_filename,
						int             num_options,
						char**          options)
{
  GdamModel *submodel = NULL;
  GdamSkin *submodel_skin = NULL;
  GdamTurntable *turntable = GDAM_TURNTABLE (GDAM_SKIN (skin)->model);
  if (!turntable || !skin_type_name || !skin_filename)
    {
      return FALSE;
    }
  /*gdam_debug("gdam_turntable_skin_set_producer: "
     "type=%s, filename=%s, options=%s",
     skin_type_name, skin_filename, options); */

  submodel_skin = gdam_skin_new_by_type_name
    (skin_type_name,
     skin_filename,
     num_options, options, GDAM_MODEL (turntable)->channel, &submodel);
  if (!GDAM_IS_PRODUCER (submodel))
    {
      g_warning ("sorry, submodel of turntable was not " "a producer");
      return FALSE;
    }
  gdam_turntable_set_producer (turntable, GDAM_PRODUCER (submodel));

#if 0
#endif

  if (skin->producer_container)
    {
      GtkWidget *widget = submodel_skin->main_widget;
      gdam_set_container (skin->producer_container, widget);
    }

  skin->producer_skin = submodel_skin;

  return TRUE;
}

/*  Connect callbacks to objects....  this function is used by gladelib
 *  to connect calback to objects at runtime.  We cannot let libglade 
 *  handle signal connection automatically, because we can't provide
 *  the user_data argument in that case.
 */
static gboolean gdam_turntable_frame_drag_data_received_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gdam_widget_string_dest_set (GTK_WIDGET (object),
			       on_tempo_string_drop,
			       "gdam-tempo-text",
			       GDAM_TURNTABLE (skin->model));
  return TRUE;
}

static void preview_adj_changed(GtkAdjustment* adj, GdamTurntable* turntable)
{
  //g_message("preview_adj val changed to %f", adj->value);
}


static void nudge_rate_adj_changed(GtkAdjustment* adj, GdamTurntableSkin* skin)
{
  GdamTurntable *turntable;
  float delval;
  delval = adj->value - skin->last_rate_nudge;
  turntable = GDAM_TURNTABLE (GDAM_SKIN (skin)->model);

  /*  the adjustment is set to zero on nudge dial hands off.  We ignore a
   *  change of zero.  */
  if (adj->value != 0)
    {
      float newrate = turntable->rate;
      newrate += (delval / 100);
      gdam_turntable_set_rate (turntable, newrate);

    }
  skin->last_rate_nudge = adj->value;
}

static void nudge_song_adj_changed(GtkAdjustment* adj, GdamTurntableSkin* skin)
{
  GdamTurntable *turntable;
  float delval;
  delval = adj->value - skin->last_song_nudge;
  turntable = GDAM_TURNTABLE (GDAM_SKIN (skin)->model);

  /*  the adjustment is set to zero on nudge dial hands off.  We ignore a
   *  change of zero.  */
  if (adj->value != 0)
    {
      gdouble bpm, del_time, percent, beat_length;
      gdam_turntable_seek_by_beats (turntable, delval);
      g_return_if_fail (turntable->producer != NULL);
      bpm = gdam_turntable_get_current_producer_bpm (turntable);
      beat_length = (bpm > 0) ? 60.0 / bpm : 1.0;
      skin->autotempo_nudge_amount += delval * beat_length;

      del_time = turntable->producer->stopped_position - 
		   skin->autotempo_time_start;
      percent = (del_time + skin->autotempo_nudge_amount) / del_time;
      gdam_debug ("nudged %f secs over %f of producer time: recommend %f percent of current speed", 
		 skin->autotempo_nudge_amount, del_time, percent);
    }
  skin->last_song_nudge = adj->value;
}

static void 
apply_auto_tempo (GdamTurntableSkin *skin)
{
  gdouble bpm, del_time, percent, beat_length, rate;
  GdamTurntable *turntable;
  turntable = GDAM_TURNTABLE (GDAM_SKIN (skin)->model);
  g_return_if_fail (turntable->producer != NULL);
  bpm = gdam_turntable_get_current_producer_bpm (turntable);
  beat_length = (bpm > 0) ? 60.0 / bpm : 1.0;
  del_time = turntable->producer->stopped_position -
	       skin->autotempo_time_start;
  percent = (del_time + skin->autotempo_nudge_amount) / del_time;
  gdam_debug ("nudged %f secs over %f of producer time: recommend %f percent of current speed", 
	     skin->autotempo_nudge_amount, del_time, percent);
  rate = turntable->rate;
  rate *= percent;
  gdam_turntable_set_rate (turntable, rate);
  skin->autotempo_nudge_amount = 0;
}

static void 
reset_auto_tempo (GdamTurntableSkin *skin)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (GDAM_SKIN (skin)->model);
  g_return_if_fail (turntable->producer != NULL);
  skin->autotempo_time_start = turntable->producer->stopped_position;
  skin->autotempo_nudge_amount = 0;
  gdam_debug ("resetting autotempo");
}


static void previewscroll_button_event(GtkWidget*     wid, 
				   GdkEvent*      ev, 
				   GdamTurntable* turntable)
{
  if (ev->type == GDK_BUTTON_PRESS)
    {
      //g_message("preview scroll button press");
    }
  else if (ev->type == GDK_BUTTON_RELEASE)
    {
      //g_message("preview scroll button release");
    }
}

static gboolean gdam_turntable_rate_adjustment_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  GdamTurntableSkin *ttskin;
  ttskin = GDAM_TURNTABLE_SKIN (skin);
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  ttskin->resample_adj = adj;
  return TRUE;
}

static gboolean gdam_turntable_nudge_rate_adjustment_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  GdamTurntable *turntable;
  turntable = GDAM_TURNTABLE (skin->model);
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (nudge_rate_adj_changed), skin);
  return TRUE;
}


static gboolean gdam_turntable_nudge_song_adjustment_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  GdamTurntable *turntable;
  turntable = GDAM_TURNTABLE (skin->model);
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (nudge_song_adj_changed), skin);
  return TRUE;
}




static gboolean gdam_turntable_sync_preview_adjustment_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  GdamTurntable *turntable;
  turntable = GDAM_TURNTABLE (skin->model);
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect_object (GTK_OBJECT (turntable),
			     "update_sync_preview",
			     GTK_SIGNAL_FUNC (update_sync_preview_adj),
			     GTK_OBJECT (adj));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (preview_adj_changed), turntable);
  gtk_signal_connect (object,
		      "button_press_event",
		      GTK_SIGNAL_FUNC (previewscroll_button_event),
		      turntable);
  gtk_signal_connect (object,
		      "button_release_event",
		      GTK_SIGNAL_FUNC (previewscroll_button_event),
		      turntable);
  return TRUE;
}

static void on_length_changed(GdamTurntableSkin* skin)
{
  if (skin->track_display)
    gtk_track_display_set_origin (skin->track_display, 0);
}


static gboolean gdam_turntable_position_adjustment_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  GdamTurntable *turntable;
  turntable = GDAM_TURNTABLE (skin->model);
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (position_adj_changed), turntable);
  gtk_signal_connect (object,
		      "button_press_event",
		      GTK_SIGNAL_FUNC (posscroll_button_event), turntable);
  gtk_signal_connect (object,
		      "button_release_event",
		      GTK_SIGNAL_FUNC (posscroll_button_event), turntable);
  gtk_signal_connect_object (GTK_OBJECT (turntable),
			     "set_position",
			     GTK_SIGNAL_FUNC
			     (turntable_update_position_adj_value),
			     GTK_OBJECT (adj));
  gtk_signal_connect_object (GTK_OBJECT (turntable), "update_length",
			     GTK_SIGNAL_FUNC
			     (turntable_update_position_adj_upper),
			     GTK_OBJECT (adj));
  gtk_signal_connect_object (GTK_OBJECT (turntable), "update_length",
			     GTK_SIGNAL_FUNC (on_length_changed),
			     GTK_OBJECT (skin));
  return TRUE;
}

static gboolean gdam_turntable_hour_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkLabel *label = GTK_LABEL (object);
  GdamTurntableSkin *ttskin = GDAM_TURNTABLE_SKIN (skin);
  if (ttskin->hour_label)
    g_warning ("multiple hour labels not allowed in turntable.");
  GDAM_TURNTABLE_SKIN (skin)->hour_label = label;

  gtk_signal_connect_object (GTK_OBJECT (skin->model),
			     "set_position",
			     GTK_SIGNAL_FUNC (turntable_update_hour_label),
			     GTK_OBJECT (skin));
  return TRUE;
}


static gboolean gdam_turntable_minute_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkLabel *label = GTK_LABEL (object);
  GdamTurntableSkin *ttskin = GDAM_TURNTABLE_SKIN (skin);
  if (ttskin->minute_label)
    g_warning ("multiple minute labels not allowed in turntable.");
  GDAM_TURNTABLE_SKIN (skin)->minute_label = label;

  gtk_signal_connect_object (GTK_OBJECT (skin->model),
			     "set_position",
			     GTK_SIGNAL_FUNC (turntable_update_minute_label),
			     GTK_OBJECT (skin));
  return TRUE;
}



static gboolean gdam_turntable_second_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkLabel *label = GTK_LABEL (object);
  GdamTurntableSkin *ttskin = GDAM_TURNTABLE_SKIN (skin);
  if (ttskin->second_label)
    g_warning ("multiple second labels not allowed in turntable.");
  GDAM_TURNTABLE_SKIN (skin)->second_label = label;

  gtk_signal_connect_object (GTK_OBJECT (skin->model),
			     "set_position",
			     GTK_SIGNAL_FUNC (turntable_update_second_label),
			     GTK_OBJECT (skin));
  return TRUE;
}



static gboolean gdam_turntable_time_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkLabel *label = GTK_LABEL (object);
  GdamTurntableSkin *ttskin = GDAM_TURNTABLE_SKIN (skin);
  if (ttskin->time_label)
    g_warning ("multiple time labels not allowed in turntable.");
  GDAM_TURNTABLE_SKIN (skin)->time_label = label;

  gtk_signal_connect_object (GTK_OBJECT (skin->model),
			     "set_position",
			     GTK_SIGNAL_FUNC (turntable_update_pos_label),
			     GTK_OBJECT (skin));
  return TRUE;
}

static gboolean gdam_turntable_timer_origin_zero_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (timer_origin_zero),
			     GTK_OBJECT (skin));
  return TRUE;
}

static gboolean gdam_turntable_timer_origin_length_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (timer_origin_length),
			     GTK_OBJECT (skin));
  return TRUE;
}


static gboolean gdam_turntable_nudge_timer_origin_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (nudge_timer_origin),
			     GTK_OBJECT (skin));
  return TRUE;
}

static gboolean gdam_turntable_cycle_beatinfo_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (do_cycle_beatinfo),
			     GTK_OBJECT (skin));
  return TRUE;
}


static gboolean gdam_turntable_beatnum_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkLabel *label = GTK_LABEL (object);
  gtk_signal_connect_object (GTK_OBJECT (skin->model),
			     "update_beatnum",
			     GTK_SIGNAL_FUNC (turntable_update_beatnum),
			     GTK_OBJECT (label));
  return TRUE;
}

static gboolean gdam_turntable_tempo_data_drag_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect (object,
		      "button_press_event",
		      GTK_SIGNAL_FUNC (on_tempo_press_event), skin);
  return TRUE;
}

static gboolean gdam_turntable_match_desired_tempo_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (on_match_desired_tempo), 
			     (gpointer)skin);
  return TRUE;
}


static gboolean gdam_turntable_jump_scrollbar_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkRange *range = GTK_RANGE (object);
  GtkAdjustment *adj = gtk_range_get_adjustment (range);
  GdamTurntable *turntable = GDAM_TURNTABLE (skin->model);
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (on_jump_adj_changed), turntable);
  gtk_signal_connect (object,
		      "button_release_event",
		      GTK_SIGNAL_FUNC (on_jump_button_event), turntable);
  return TRUE;
}

static gboolean gdam_turntable_halve_cbutton_pressed_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  connect_button_with_mask (skin,
			    object,
			    signal_name,
			    GTK_SIGNAL_FUNC (on_halve_check_button_clicked),
			    GDAM_TURNTABLE (skin->model));
  return TRUE;
}

static gboolean gdam_turntable_anchor_rbutton_pressed_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkRadioButton *radio = GTK_RADIO_BUTTON (object);
  GdamTurntable *turntable = GDAM_TURNTABLE (skin->model);
  gtk_signal_connect (object,
		      "clicked",
		      GTK_SIGNAL_FUNC (on_anchor_rbutton_toggled), turntable);
  gtk_signal_connect (object,
		      "toggled",
		      GTK_SIGNAL_FUNC (on_anchor_rbutton_toggled), turntable);
  return TRUE;
}

static gboolean gdam_turntable_preview_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkLabel *label = GTK_LABEL (object);
  GdamTurntable *turntable = GDAM_TURNTABLE (skin->model);
  GDAM_TURNTABLE_SKIN (skin)->anchored_label = label;
  gtk_signal_connect_object (GTK_OBJECT (turntable),
			     "update_sync_preview",
			     GTK_SIGNAL_FUNC (update_sync_preview_label),
			     GTK_OBJECT (label));
  return TRUE;
}


static gboolean gdam_turntable_jump_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkLabel *label = GTK_LABEL (object);
  gtk_signal_connect_object (GTK_OBJECT (skin->model),
			     "update_jump_preview",
			     GTK_SIGNAL_FUNC (turntable_update_jump_preview),
			     GTK_OBJECT (label));
  return TRUE;
}


static gint track_display_press          (GtkObject*          object,
                                          GdkEventButton*     event,
				          GdamTurntableSkin*  skin)
{
  gdouble percent;
  gdouble track_length;
  gdouble click_time;
  gdouble target_time;
  gboolean lock_to_index;
  GdamTurntable* turntable;
  GdamBeatInfo* prod_beat;
  GtkTrackDisplay *display;
  int draw_width;
  int beat_index;

  turntable = GDAM_TURNTABLE (GDAM_SKIN (skin)->model);
  display = GTK_TRACK_DISPLAY (object);
  draw_width = display->draw_width;

  if (event->button == 1)
    lock_to_index = TRUE;
  else
    lock_to_index = FALSE;

  track_length = display->length;
  percent = event->x / draw_width;
  click_time = percent * track_length;
  target_time = click_time;

  if (lock_to_index)
    {
      prod_beat = &(GDAM_MODEL (turntable->producer)->beat_info);
      beat_index = gdam_beat_info_guess_index_index (prod_beat, click_time);
      if (beat_index < 0)
	{
	  g_warning ("negative beat index in ttskin");
	  return 1;
	}
      target_time = gdam_beat_info_get_range_first_beat (prod_beat, beat_index);
    }
  gdam_turntable_seek_in_beat (turntable, target_time);
  return 1;
}



static void connect_track_display(GtkTrackDisplay*   display,
                                  GdamTurntableSkin* skin)
{
  GdamTurntable *turntable;
  turntable = GDAM_TURNTABLE (GDAM_SKIN (skin)->model);
  g_return_if_fail  (GTK_IS_TRACK_DISPLAY (display));
  gtk_signal_connect (GTK_OBJECT (display),
		      "button_press_event",
		      GTK_SIGNAL_FUNC (track_display_press), 
		      skin);
  gtk_signal_connect (GTK_OBJECT (turntable),
		      "update_length",
		      GTK_SIGNAL_FUNC (track_display_length_changed), 
		      GTK_OBJECT (display));
  gtk_signal_connect
    (GTK_OBJECT (turntable->producer),
     "set_beat_info",
     GTK_SIGNAL_FUNC (track_display_beatinfo_changed), 
     GTK_OBJECT (display));
  gtk_signal_connect
    (GTK_OBJECT (turntable),
     "set_beat_info",
     GTK_SIGNAL_FUNC (track_display_tt_beatinfo_changed), 
     GTK_OBJECT (display));
  if (skin->track_display) 
    {
      g_message ("auxilliary track displays semi-functional...");
      return;
    }
  skin->track_display = display;
}


static gboolean gdam_turntable_track_display_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  if (!GTK_IS_TRACK_DISPLAY (object))
    {
      g_warning ("%s may only be connected to a track display",
		 handler_names[0]);
      return TRUE;
    }
  if (GDAM_TURNTABLE (skin->model)->producer == NULL)
    return FALSE;
  connect_track_display (GTK_TRACK_DISPLAY (object),
			 GDAM_TURNTABLE_SKIN (skin));
  return TRUE;
}


static void do_seek_fb(gpointer data)
{
  GdamTurntable *turntable;
  GdamSkin *skin;
  gdouble fb;
  skin = GDAM_SKIN (data);
  g_return_if_fail (skin != NULL);
  turntable = GDAM_TURNTABLE (skin->model);
  fb = gdam_turntable_get_current_producer_fb (turntable);
  gdam_turntable_seek_in_beat (turntable, fb);
}

static gboolean gdam_turntable_seek_fb_in_time_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (do_seek_fb), (gpointer) skin);
  return TRUE;
}


static gboolean gdam_turntable_producer_container_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamTurntableSkin *turntable_skin = GDAM_TURNTABLE_SKIN (skin);
  gdam_debug ("howdy... matched prod_cont in connect_func");
  turntable_skin->producer_container = GTK_CONTAINER (object);
  if (turntable_skin->producer_skin)
    {
      GdamSkin *submodel_skin = turntable_skin->producer_skin;
      GtkWidget *widget = submodel_skin->main_widget;
      gdam_set_container (turntable_skin->producer_container, widget);
    }
  return TRUE;
}
static gboolean gdam_turntable_set_producer_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamTurntableSkin *turntable_skin = GDAM_TURNTABLE_SKIN (skin);
  if (!turntable_skin->producer_container)
    turntable_skin->producer_container = GTK_CONTAINER (object);
  if (handler_names[1] == NULL)
    {
      g_warning ("no name for %s", handler_names[0]);
      return TRUE;
    }
  if (num_handler_names < 3)
    {
      g_warning ("too few arguments to `%s' in turntable", handler_names[0]);
      return TRUE;
    }
  gdam_turntable_skin_set_producer (turntable_skin,
				    handler_names[1],
				    handler_names[2],
				    num_handler_names - 3, handler_names + 3);
  return TRUE;
}

static void on_turntable_anchored_set(gpointer		info_ptr,
				      GdamArg*		arg,
				      GtkObject*	object)
{
  GdamTurntableSkin *skin = GDAM_TURNTABLE_SKIN (info_ptr);
  GdamTurntable *tt = GDAM_TURNTABLE (GDAM_SKIN (skin)->model);
  if (tt && GDAM_MODEL (tt)->anchored && skin->anchored_label)
    {
      char text[63];
      sprintf (text, "-");
      gtk_label_set_text (skin->anchored_label, text);
    }
}

static void on_turntable_playing_set(gpointer	info_ptr,
				     GdamArg*	arg,
				     GtkObject*	object)
{
  GdamTurntableSkin *skin = GDAM_TURNTABLE_SKIN (info_ptr);
  GdamTurntable *turntable = GDAM_TURNTABLE (GDAM_SKIN (skin)->model);
  if ((GDAM_ARG_INT (arg) > 0) && (skin->timer_id == 0))
    {
      skin->timer_id = gtk_timeout_add (500,
					(GtkFunction) auto_sync_timeout,
					skin);
    }
  else if ((GDAM_ARG_INT (arg) <= 0) && (skin->timer_id != 0))
    {
      gtk_timeout_remove (skin->timer_id);
      skin->timer_id = 0;
    }
  if (GDAM_ARG_INT (arg) > 0 && GDAM_MODEL (turntable)->source_id == 0)
    reset_auto_tempo (skin);
}

static gboolean gdam_turntable_build_new_model     (GdamSkin*      skin,
					            GdamChannel*   channel,
						    int            num_options,
				                    char**         options)
{
  GdamTurntable *turntable = gdam_turntable_new (channel);
  skin->model = GDAM_MODEL (turntable);

  /* TODO: should we ever detach this?? */
  GDAM_TURNTABLE_SKIN (skin)->playing_arg_connect_id =
    gdam_arg_connect (GTK_OBJECT (turntable),
		      "playing", on_turntable_playing_set, skin);
  GDAM_TURNTABLE_SKIN (skin)->play_or_synch_arg_connect_id =
    gdam_arg_connect (GTK_OBJECT (turntable),
		      "play_or_synch", on_turntable_playing_set, skin);
  GDAM_TURNTABLE_SKIN (skin)->anchored_arg_connect_id =
    gdam_arg_connect (GTK_OBJECT (turntable),
		      "anchored", on_turntable_anchored_set, skin);

  if (num_options < 2)
    {
#if 0
      /* XXX TODO: this perhaps should be required... */
      g_warning ("gdam_turntable_build_new_model: "
		 "too few options, no producer");
      return FALSE;
      g_warning ("gdam_turntable_build_new_model: " "delaying set producer");
#endif
      return TRUE;
    }

  gdam_turntable_skin_set_producer (GDAM_TURNTABLE_SKIN (skin),
				    options[0],
				    options[1], num_options - 2, options + 2);

  return TRUE;
}

static void gdam_turntable_skin_synchronize(GdamView* comp_skin)
{
  GdamTurntableSkin *turntable_skin = GDAM_TURNTABLE_SKIN (comp_skin);
  GdamModel *model = GDAM_SKIN (comp_skin)->model;
  GdamTurntable *turntable = GDAM_TURNTABLE (model);
  GdamModel *producer = GDAM_MODEL (turntable->producer);
  GdamSkin *skin;
  if (producer == NULL)
    return;
  skin = gdam_skin_find_from_model (producer);
  turntable_skin->producer_skin = skin;
  if (turntable_skin->producer_container != NULL && skin->main_widget != NULL)
    {
      gdam_set_container (turntable_skin->producer_container,
			  skin->main_widget);
      gtk_widget_show (skin->main_widget);
    }

}

static GdamSkinGenericConnect connect_infos[] = {
	{ "gdam_turntable_frame_drag_data_received",
          gdam_turntable_frame_drag_data_received_handler	},
	{ "gdam_turntable_position_adjustment",
          gdam_turntable_position_adjustment_handler		},
	{ "gdam_turntable_sync_preview_adjustment",
          gdam_turntable_sync_preview_adjustment_handler	},
	{ "gdam_turntable_rate_adjustment",
          gdam_turntable_rate_adjustment_handler		},
	{ "gdam_turntable_nudge_rate_adjustment",
          gdam_turntable_nudge_rate_adjustment_handler		},
	{ "gdam_turntable_nudge_song_adjustment",
          gdam_turntable_nudge_song_adjustment_handler		},
	{ "gdam_turntable_time_label",
          gdam_turntable_time_label_handler			},
	{ "gdam_turntable_hour_label",
          gdam_turntable_hour_label_handler			},
	{ "gdam_turntable_minute_label",
          gdam_turntable_minute_label_handler			},
	{ "gdam_turntable_second_label",
          gdam_turntable_second_label_handler			},
	{ "gdam_turntable_beatnum_label",
          gdam_turntable_beatnum_label_handler			},
	{ "gdam_turntable_nudge_timer_origin",
          gdam_turntable_nudge_timer_origin_handler		},
	{ "gdam_turntable_timer_origin_zero",
          gdam_turntable_timer_origin_zero_handler		},
	{ "gdam_turntable_timer_origin_length",
          gdam_turntable_timer_origin_length_handler		},
	{ "gdam_turntable_cycle_beatinfo",
          gdam_turntable_cycle_beatinfo_handler			},
	{ "gdam_turntable_tempo_data_drag",
          gdam_turntable_tempo_data_drag_handler		},
	{ "gdam_turntable_match_desired_tempo",
          gdam_turntable_match_desired_tempo_handler		},
	{ "gdam_turntable_jump_scrollbar",
          gdam_turntable_jump_scrollbar_handler			},
	{ "gdam_turntable_halve_cbutton_pressed",
          gdam_turntable_halve_cbutton_pressed_handler		},
	{ "gdam_turntable_anchor_rbutton_pressed",
          gdam_turntable_anchor_rbutton_pressed_handler		},
	{ "gdam_turntable_preview_label",
	  gdam_turntable_preview_label_handler			},
	{ "gdam_turntable_jump_label",
	  gdam_turntable_jump_label_handler			},
	{ "gdam_turntable_producer_container",
	  gdam_turntable_producer_container_handler		},
	{ "gdam_turntable_seek_fb_in_time",
	  gdam_turntable_seek_fb_in_time_handler		},
	{ "gdam_turntable_track_display",
	  gdam_turntable_track_display_handler			},
	{ "gdam_turntable_set_producer",
	  gdam_turntable_set_producer_handler			}	};

static GdamSkinSimpleConnect simple_connect_infos[] = {
	{ "gdam_turntable_reset_autotempo",
           (GdamSkinFunc)reset_auto_tempo				},
	{ "gdam_turntable_apply_autotempo",
           (GdamSkinFunc)apply_auto_tempo				},
};

static void gdam_turntable_skin_destroy(GtkObject* object){
	GdamTurntableSkin* skin = GDAM_TURNTABLE_SKIN(object);
	if (skin->timer_id)
		gtk_timeout_remove(skin->timer_id);
}

static void gdam_turntable_arg_set_fine_rate_control(GtkObject* object, 
					             GdamArg*   arg)
{
  GdamTurntableSkin *turntable = GDAM_TURNTABLE_SKIN (object);
  turntable->fine_rate_control = GDAM_ARG_INT (arg);
  if (turntable->resample_adj)
    {
      float del;
      if (turntable->fine_rate_control)
	del = 0.12;
      else
	del = 1.00;
      turntable->resample_adj->upper = 1.0 + del + turntable->resample_adj->page_size;
      turntable->resample_adj->lower = 1.0 - del;
      gtk_adjustment_changed (turntable->resample_adj);
    }
}

static void gdam_turntable_arg_get_fine_rate_control(GtkObject* object, 
					             GdamArg*   arg)
{
  GdamTurntableSkin *turntable = GDAM_TURNTABLE_SKIN (object);
  GDAM_ARG_INT (arg) = turntable->fine_rate_control;
}


static void gdam_turntable_skin_class_init(GdamTurntableSkinClass* klass)
{
  GdamSkinClass *skin_class = GDAM_SKIN_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  GtkType type = object_class->type;
  skin_class->build_new_model = gdam_turntable_build_new_model;
  GDAM_VIEW_CLASS (klass)->synchronize = gdam_turntable_skin_synchronize;
  GTK_OBJECT_CLASS (klass)->destroy = gdam_turntable_skin_destroy;
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
  gdam_skin_class_register_simple_connections 
    				      (skin_class,
				       G_N_ELEMENTS (simple_connect_infos),
				       simple_connect_infos);
  gdam_arg_register (object_class,
		     "fine_rate_control",
		     GDAM_ARG_TYPE_INT,
		     gdam_turntable_arg_set_fine_rate_control,
		     gdam_turntable_arg_get_fine_rate_control,
		     NULL, "0,1", 0);
  turntable_signals[MATCH_TEMPO]
    = gtk_signal_new ("match_tempo",
		      GTK_RUN_FIRST,
		      type,
		      GTK_SIGNAL_OFFSET (GdamTurntableSkinClass,
					 match_tempo),
		      gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class, turntable_signals, LAST_SIGNAL);
}
