/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gtkpixsplitbar.h"
#include "pixmapcache.h"
#include "gdamskin.h"		/* needed for path stuff. */
#include <gtk/gtksignal.h>
#include <gtk/gtkmain.h>
#include <stdlib.h>
#include <string.h>

enum {
	MOTION,
	LAST_SIGNAL
};
static guint pix_splitbar_signals[LAST_SIGNAL] = { 0 };

static int  gtk_pix_splitbar_expose_event           (GtkWidget*       widget,
                                                     GdkEventExpose*  event);
static int  gtk_pix_splitbar_configure_event        (GtkWidget*       widget,
                                                     GdkEventConfigure*event);
static void gtk_pix_splitbar_draw                   (GtkWidget*       widget,
                                                     GdkRectangle*    area);
static void gtk_pix_splitbar_draw_default           (GtkWidget *widget);
static void gtk_pix_splitbar_draw_focus             (GtkWidget *widget);
static void gtk_pix_splitbar_size_request           (GtkWidget*       widget,
                                                     GtkRequisition*  event);
static int  gtk_pix_splitbar_press_event            (GtkWidget*       widget,
                                                     GdkEventButton*  event);
static int  gtk_pix_splitbar_release_event          (GtkWidget*       widget,
                                                     GdkEventButton*  event);
static int  gtk_pix_splitbar_enter_notify_event     (GtkWidget*       widget,
                                                     GdkEventCrossing*  event);
static int  gtk_pix_splitbar_leave_notify_event     (GtkWidget*       widget,
                                                     GdkEventCrossing*event);
static int  gtk_pix_splitbar_motion_event           (GtkWidget*       widget,
                                                     GdkEventMotion*  event);
static void update_adjustment                   (GtkPixSplitbar* pix_splitbar,
                                                 int              cur_x,
			                         int              cur_y);
static GtkObjectClass* parent_class;
static void load_images                     (GtkPixSplitbar* pix_splitbar);
static void gtk_pix_splitbar_init(GtkPixSplitbar* pix_splitbar)
{

  GTK_WIDGET_SET_FLAGS (pix_splitbar, GTK_CAN_FOCUS | GTK_RECEIVES_DEFAULT);
  gtk_widget_add_events (GTK_WIDGET (pix_splitbar),
			 GDK_EXPOSURE_MASK
			 | GDK_STRUCTURE_MASK
			 | GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK);
}
static void gtk_pix_splitbar_class_init(GtkPixSplitbarClass* klass)
{
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  widget_class->expose_event = gtk_pix_splitbar_expose_event;
  widget_class->configure_event = gtk_pix_splitbar_configure_event;
  widget_class->button_press_event = gtk_pix_splitbar_press_event;
  widget_class->button_release_event = gtk_pix_splitbar_release_event;
  widget_class->motion_notify_event = gtk_pix_splitbar_motion_event;
  widget_class->enter_notify_event = gtk_pix_splitbar_enter_notify_event;
  widget_class->leave_notify_event = gtk_pix_splitbar_leave_notify_event;
  widget_class->size_request = gtk_pix_splitbar_size_request;
  widget_class->draw = gtk_pix_splitbar_draw;
  widget_class->draw_focus = gtk_pix_splitbar_draw_focus;
  widget_class->draw_default = gtk_pix_splitbar_draw_default;
  pix_splitbar_signals[MOTION] =
    gtk_signal_new ("motion",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkPixSplitbarClass, motion),
		    gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class,
				pix_splitbar_signals, LAST_SIGNAL);
}

GtkType gtk_pix_splitbar_get_type()
{
  static GtkType pix_splitbar_type = 0;
  if (!pix_splitbar_type)
    {
      static const GtkTypeInfo pix_splitbar_info = {
	"GtkPixSplitbar",
	sizeof (GtkPixSplitbar),
	sizeof (GtkPixSplitbarClass),
	(GtkClassInitFunc) gtk_pix_splitbar_class_init,
	(GtkObjectInitFunc) gtk_pix_splitbar_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GTK_TYPE_EVENT_BOX;
      parent_class = gtk_type_class (parent);
      pix_splitbar_type = gtk_type_unique (parent, &pix_splitbar_info);
    }
  return pix_splitbar_type;
}

static void
gtk_pix_splitbar_draw_focus (GtkWidget *widget)
{
  gtk_widget_draw (widget, NULL);
}

static void
gtk_pix_splitbar_draw_default (GtkWidget *widget)
{
  gtk_widget_draw (widget, NULL);
}

static inline gfloat get_fraction(GtkPixSplitbar* pix_splitbar)
{
  gfloat fraction;
  if (pix_splitbar->adjustment != NULL)
    {
      GtkAdjustment *adjustment = pix_splitbar->adjustment;
      fraction = (adjustment->value - adjustment->lower)
	/ (adjustment->upper - adjustment->lower);
      fraction = CLAMP (fraction, 0.0, 1.0);
    }
  else
    {
      fraction = 0.5;
    }
  return fraction;
}


static void gtk_pix_splitbar_draw(GtkWidget* widget, GdkRectangle* area)
{
  GtkPixSplitbar *pix_splitbar = GTK_PIX_SPLITBAR (widget);
  GdkWindow *window = widget->window;
  GdkGCValues values;
  int window_width = widget->allocation.width;
  int window_height = widget->allocation.height;
  int x, y;
  int pixels;
  gfloat fraction;
  int area_width, area_height;

  GdkRectangle above_area, below_area;

  if (!GTK_WIDGET_DRAWABLE (widget))
    {
      return;
    }
  if (!pix_splitbar->loaded_images)
    load_images (pix_splitbar);

  gdk_window_set_back_pixmap (widget->window, NULL, TRUE);

  area_width = pix_splitbar->above_size.x;
  area_height = pix_splitbar->above_size.y;
  if (area_height > window_height)
    area_height = window_height;
  if (area_width > window_width)
    area_width = window_width;

  fraction = get_fraction (pix_splitbar);

  gdk_gc_get_values (widget->style->fg_gc[0], &values);

  switch (pix_splitbar->orientation)
    {
    case GTK_PIX_SPLITBAR_LEFT:
    case GTK_PIX_SPLITBAR_RIGHT:
      pixels = area_width * fraction;
      above_area.y = 0;
      below_area.y = 0;
      above_area.height = area_height;
      below_area.height = area_height;
      break;
    case GTK_PIX_SPLITBAR_TOP:
    case GTK_PIX_SPLITBAR_BOTTOM:
      pixels = area_height * fraction;
      above_area.x = 0;
      below_area.x = 0;
      above_area.width = area_width;
      below_area.width = area_width;
      break;
    default:
      {
	g_warning ("bad mode in PixSplitbar: %d", pix_splitbar->orientation);
	return;
      }
    }

  switch (pix_splitbar->orientation)
    {
    case GTK_PIX_SPLITBAR_LEFT:
      above_area.x = 0;
      above_area.width = pixels;
      below_area.x = pixels;
      below_area.width = area_width - pixels;
      break;

    case GTK_PIX_SPLITBAR_RIGHT:
      below_area.x = 0;
      below_area.width = area_width - pixels;
      above_area.x = area_width - pixels;
      above_area.width = pixels;
      break;

    case GTK_PIX_SPLITBAR_TOP:
      above_area.y = 0;
      above_area.height = pixels;
      below_area.y = pixels;
      below_area.height = area_height - pixels;
      break;

    case GTK_PIX_SPLITBAR_BOTTOM:
      below_area.y = 0;
      below_area.height = area_height - pixels;
      above_area.y = area_height - pixels;
      above_area.height = pixels;
      break;
    }
  //g_message("pixbar draw, fraction: %f", fraction);
  //g_message("pixsplitbar draw above: x,y = %d,%d  w,h = %d,%d", 
  //above_area.x, above_area.y, above_area.width, above_area.height);
  //g_message("pixsplitbar draw below: x,y = %d,%d  w,h = %d,%d", 
  //below_area.x, below_area.y, below_area.width, below_area.height);

  gdk_gc_set_clip_origin (widget->style->fg_gc[0], 0, 0);
  if (pix_splitbar->below_pixmap)
    {
      if (pix_splitbar->below_mask != NULL)
	gdk_gc_set_clip_mask (widget->style->fg_gc[0],
			      pix_splitbar->below_mask);
      gdk_window_copy_area (window,
			    widget->style->fg_gc[0],
			    below_area.x,
			    below_area.y,
			    pix_splitbar->below_pixmap,
			    below_area.x,
			    below_area.y,
			    below_area.width, below_area.height);
      gdk_gc_set_clip_mask (widget->style->fg_gc[0], NULL);
    }
  else
    {
      gtk_paint_box (widget->style, widget->window,
		     GTK_STATE_NORMAL, GTK_SHADOW_IN,
		     area, widget, "buttondefault",
		     area->x, area->y, area->width, area->height);
    }

  gdk_gc_set_clip_origin (widget->style->bg_gc[0], 0, 0);
  if (pix_splitbar->above_pixmap)
    {
      if (pix_splitbar->above_mask != NULL)
	gdk_gc_set_clip_mask (widget->style->bg_gc[0],
			      pix_splitbar->above_mask);
      gdk_window_copy_area (window,
			    widget->style->bg_gc[0],
			    above_area.x,
			    above_area.y,
			    pix_splitbar->above_pixmap,
			    above_area.x,
			    above_area.y,
			    above_area.width, above_area.height);
      gdk_gc_set_clip_mask (widget->style->bg_gc[0], NULL);
    }
  else
    {
      gtk_paint_box (widget->style, widget->window,
		     GTK_STATE_NORMAL, GTK_SHADOW_IN,
		     area, widget, "background",
		     area->x, area->y, area->width, area->height);
    }
}

static int gtk_pix_splitbar_press_event(GtkWidget*       widget,
                                         GdkEventButton*  event)
{
  GtkPixSplitbar *pix_splitbar = GTK_PIX_SPLITBAR (widget);
  gtk_grab_add (widget);
  update_adjustment (pix_splitbar, event->x, event->y);
  gtk_widget_draw_default (GTK_WIDGET (pix_splitbar));
  return FALSE;
}

static void update_adjustment(GtkPixSplitbar* pix_splitbar,
                              int              cur_x,
			      int              cur_y)
{
  int x, y;
  gfloat value;
  GtkAdjustment *adj;
  gfloat fraction;
  int width, height;
  g_return_if_fail (pix_splitbar->adjustment != NULL);
  adj = pix_splitbar->adjustment;

  width = pix_splitbar->above_size.x;
  height = pix_splitbar->above_size.y;

  switch (pix_splitbar->orientation)
    {
    case GTK_PIX_SPLITBAR_LEFT:
      cur_x = width - 1 - cur_x;
      /* fall through */
    case GTK_PIX_SPLITBAR_RIGHT:
      fraction = (gfloat) cur_x / width;
      break;
    case GTK_PIX_SPLITBAR_TOP:
      cur_y = height - 1 - cur_y;
      /* fall through */
    case GTK_PIX_SPLITBAR_BOTTOM:
      fraction = (gfloat) cur_y / height;
      break;
    default:
      {
	g_warning ("bad mode in PixSplitbar: %d", pix_splitbar->orientation);
	return;
      }
    }

  value = fraction * (adj->upper - adj->lower) + adj->lower;

  gtk_signal_handler_block (GTK_OBJECT (adj), pix_splitbar->connect_id);
  gtk_adjustment_set_value (adj, value);
  gtk_signal_handler_unblock (GTK_OBJECT (adj), pix_splitbar->connect_id);
  gtk_widget_draw_default (GTK_WIDGET (pix_splitbar));
}

static int gtk_pix_splitbar_release_event(GtkWidget*       widget,
                                           GdkEventButton*  event)
{
  gboolean clicked = FALSE;
  int x, y;
  guint state;
  GtkPixSplitbar *pix_splitbar = GTK_PIX_SPLITBAR (widget);
  gtk_grab_remove (widget);

  if (event->window != widget->window)
    {
      gdk_window_get_pointer (widget->window, &x, &y, &state);
    }
  else
    {
      x = event->x;
      y = event->y;
      state = event->state;
    }

  update_adjustment (pix_splitbar, x, y);
  gtk_pix_splitbar_motion (pix_splitbar);

  return TRUE;
}
static void load_images(GtkPixSplitbar* pix_splitbar)
{
  GdkWindow *window = GTK_WIDGET (pix_splitbar)->window;
  PixmapInfo *info;
  const char *fname;
  char *xpm;
  if (pix_splitbar->loaded_images)
    return;
  pix_splitbar->loaded_images = TRUE;
  xpm = pix_splitbar->above_file;
  if (xpm == NULL)
    {
      g_warning ("error: no filename for above in pixsplitbar");
    }
  else
    {
      info = pixmap_cache_get_pixmap (window, xpm);
      if (info == NULL)
	{
	  g_warning ("error loading pixmap %s", xpm);
	  return;
	}
      else
	{
	  pix_splitbar->above_pixmap = info->pixmap;
	  pix_splitbar->above_mask = info->mask;
	  pix_splitbar->above_size.x = info->width;
	  pix_splitbar->above_size.y = info->height;
	}
    }

  xpm = pix_splitbar->below_file;
  if (xpm == NULL)
    {
      g_warning ("error: no filename for below in pixsplitbar");
    }
  else
    {
      info = pixmap_cache_get_pixmap (window, xpm);
      if (info == NULL)
	{
	  g_warning ("error loading pixmap %s", xpm);
	  return;
	}
      else
	{
	  pix_splitbar->below_pixmap = info->pixmap;
	  pix_splitbar->below_mask = info->mask;
	  pix_splitbar->below_size.x = info->width;
	  pix_splitbar->below_size.y = info->height;
	}
    }
}

static int gtk_pix_splitbar_expose_event(GtkWidget*       widget,
                                          GdkEventExpose*  event)
{
  GtkPixSplitbar *pix_splitbar = GTK_PIX_SPLITBAR (widget);
  gtk_widget_draw_default (GTK_WIDGET (pix_splitbar));
  return FALSE;
}
static int gtk_pix_splitbar_configure_event(GtkWidget*       widget,
                                             GdkEventConfigure*event)
{
  GtkPixSplitbar *pix_splitbar = GTK_PIX_SPLITBAR (widget);
  if (!pix_splitbar->loaded_images)
    load_images (pix_splitbar);
  gtk_widget_draw_default (GTK_WIDGET (pix_splitbar));
  return FALSE;
}
static int gtk_pix_splitbar_enter_notify_event(GtkWidget*       widget,
                                                GdkEventCrossing*event)
{
  GtkPixSplitbar *pix_splitbar = GTK_PIX_SPLITBAR (widget);
  GtkWidget *event_widget;
  event_widget = gtk_get_event_widget ((GdkEvent *) event);
  if (event_widget != widget || event->detail == GDK_NOTIFY_INFERIOR)
    return FALSE;
  gtk_widget_draw_default (widget);
  return FALSE;
}
static int gtk_pix_splitbar_leave_notify_event(GtkWidget*        widget,
                                                GdkEventCrossing* event)
{
  GtkPixSplitbar *pix_splitbar = GTK_PIX_SPLITBAR (widget);
  GtkWidget *event_widget;
  event_widget = gtk_get_event_widget ((GdkEvent *) event);
  if (event_widget != widget || event->detail == GDK_NOTIFY_INFERIOR)
    return FALSE;
  gtk_widget_draw_default (widget);
  return FALSE;
}
static void gtk_pix_splitbar_size_request    (GtkWidget*       widget,
                                               GtkRequisition*  requisition)
{
  GtkPixSplitbar *pix_splitbar = GTK_PIX_SPLITBAR (widget);
  requisition->width = pix_splitbar->above_size.x;
  requisition->height = pix_splitbar->above_size.y;
}

static int  gtk_pix_splitbar_motion_event          (GtkWidget*       widget,
                                                     GdkEventMotion*  event)
{
  int x, y;
  guint state;
  GtkPixSplitbar *pix_splitbar;
  pix_splitbar = GTK_PIX_SPLITBAR (widget);
  if (event->is_hint)
    gdk_window_get_pointer (event->window, &x, &y, &state);
  else
    {
      if (event->window != widget->window)
	{
	  gdk_window_get_pointer (widget->window, &x, &y, &state);
	}
      else
	{
	  x = event->x;
	  y = event->y;
	  state = event->state;
	}
    }

  update_adjustment (pix_splitbar, x, y);
  gtk_pix_splitbar_motion (pix_splitbar);
  return 0;
}

void gtk_pix_splitbar_motion(GtkPixSplitbar* pix_splitbar)
{
  gtk_signal_emit (GTK_OBJECT (pix_splitbar), pix_splitbar_signals[MOTION]);
}


void gtk_pix_splitbar_set_pixmaps_by_filename(GtkPixSplitbar* pix_splitbar,
                                              const char*     above,
                                              const char*     below)
{
  pix_splitbar->above_file = g_strdup (above);
  pix_splitbar->below_file = g_strdup (below);
  if (GTK_WIDGET (pix_splitbar)->window)
    load_images (pix_splitbar);
}

static int string_to_orientation(const char* string)
{
  if (strcasecmp (string, "left") == 0)
    return GTK_PIX_SPLITBAR_LEFT;
  if (strcasecmp (string, "right") == 0)
    return GTK_PIX_SPLITBAR_RIGHT;
  if (strcasecmp (string, "top") == 0)
    return GTK_PIX_SPLITBAR_TOP;
  if (strcasecmp (string, "bottom") == 0)
    return GTK_PIX_SPLITBAR_BOTTOM;
  return -1;
}

/*
 * Create_func for gtk_pix_splitbar.
 *
 *   string1:    above:below
 *   string2:    orientation:cur_value:min_value:max_value
 *
 * Here is a description of the fields.
 *
 *     above:            the pixmap to display for the area above the
 *                       current value.
 *     below:            the pixmap to display for the area below the
 *                       current value.
 *                       the splitbar.
 *     orientation:      one of the following values:
 *          left              grow from right-to-left 
 *          right             grow from left-to-right 
 *          top               grow from bottom-to-top 
 *          bottom            grow from top-to-bottom 
 *     cur_value:        initial value for the adjustment.
 *     min_value:        minimum value for the adjustment.
 *     max_value:        maximum value for the adjustment.
 *
 *
 */
GtkWidget* gtk_pix_splitbar_create_func    (const char*   name,
                                            const char*   string1,
                                            const char*   string2,
                                            int           int1,
                                            int           int2)
{
  char **pixmaps = g_strsplit (string1, ":", 100);
  char **coordinates;
  gfloat cur_value, min_value, max_value, step_incr;
  GtkAdjustment *adjustment;
  GtkPixSplitbar *pix_splitbar;
  int id;
  pix_splitbar = GTK_PIX_SPLITBAR (gtk_type_new (GTK_TYPE_PIX_SPLITBAR));
  gtk_pix_splitbar_set_pixmaps_by_filename (pix_splitbar,
					    pixmaps[0], pixmaps[1]);
  g_strfreev (pixmaps);

  coordinates = g_strsplit (string2, ":", 100);
  id = string_to_orientation (coordinates[0]);
  if (id < 0)
    {
      g_warning ("orientation %s not allowed to GtkPixSplitbar",
		 coordinates[0]);
      return NULL;
    }

  pix_splitbar->orientation = id;

  cur_value = atof (coordinates[1]);
  min_value = atof (coordinates[2]);
  max_value = atof (coordinates[3]);
  step_incr = coordinates[4] ? atof (coordinates[4]) : 0.01;
  g_strfreev (coordinates);

  adjustment = (GtkAdjustment *) gtk_adjustment_new (cur_value,
						     min_value,
						     max_value,
						     step_incr,
						     step_incr * 5,
						     step_incr);
  gtk_pix_splitbar_set_adjustment (pix_splitbar, adjustment);

  return GTK_WIDGET (pix_splitbar);
}

void gtk_pix_splitbar_set_adjustment(GtkPixSplitbar*  pix_splitbar,
                                      GtkAdjustment*    adjustment)
{
  GtkAdjustment *old_adj = pix_splitbar->adjustment;
  if (old_adj == adjustment)
    return;
  if (old_adj != NULL)
    {
      gtk_signal_disconnect (GTK_OBJECT (old_adj), pix_splitbar->connect_id);
      gtk_object_unref (GTK_OBJECT (old_adj));
    }
  pix_splitbar->adjustment = adjustment;
  if (adjustment == NULL)
    pix_splitbar->connect_id = 0;
  else
    {
      GtkObject *object = GTK_OBJECT (adjustment);
      gtk_object_ref (object);
      gtk_object_sink (object);
      pix_splitbar->connect_id
	= gtk_signal_connect_object
	(object,
	 "value_changed",
	 GTK_SIGNAL_FUNC (gtk_widget_queue_draw), GTK_OBJECT (pix_splitbar));
    }
}
