/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gtktrackdisplay.h"
#include "gdamskin.h"		/* needed for path stuff. */
#include "gdamskinutil.h"
#include <gtk/gtksignal.h>
#include <gtk/gtkdrawingarea.h>
#include <gtk/gtkmain.h>
#include <string.h>


static GtkObjectClass* parent_class = NULL;

static int  gtk_track_display_expose_event        (GtkWidget*        widget,
                                                GdkEventExpose*   event);
static int  gtk_track_display_configure_event     (GtkWidget*        widget,
                                                GdkEventConfigure*event);
static void gtk_track_display_draw                (GtkWidget*        widget,
                                                GdkRectangle*     area);
static void gtk_track_display_draw_default        (GtkWidget*        widget);
static void gtk_track_display_draw_focus          (GtkWidget*        widget);
static void gtk_track_display_size_request        (GtkWidget*        widget,
                                                GtkRequisition*   event);
static int  gtk_track_display_button_press_event  (GtkWidget*        widget,
                                                GdkEventButton*   event);
static int  gtk_track_display_button_release_event(GtkWidget*        widget,
                                                GdkEventButton*   event);
static int  gtk_track_display_enter_notify_event  (GtkWidget*        widget,
                                                GdkEventCrossing* event);
static int  gtk_track_display_leave_notify_event  (GtkWidget*        widget,
                                                GdkEventCrossing* event);

static void gtk_track_display_init(GtkTrackDisplay* track_display)
{

  GTK_WIDGET_SET_FLAGS (track_display, GTK_CAN_FOCUS | GTK_RECEIVES_DEFAULT);
  gtk_widget_add_events (GTK_WIDGET (track_display),
			 GDK_EXPOSURE_MASK
			 | GDK_STRUCTURE_MASK | GDK_VISIBILITY_NOTIFY_MASK
			 | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
			 | GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK);
}

static void gtk_track_display_class_init(GtkTrackDisplayClass* klass)
{
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  widget_class->expose_event = gtk_track_display_expose_event;
  widget_class->configure_event = gtk_track_display_configure_event;
  widget_class->button_press_event = gtk_track_display_button_press_event;
  widget_class->button_release_event = gtk_track_display_button_release_event;
  widget_class->enter_notify_event = gtk_track_display_enter_notify_event;
  widget_class->leave_notify_event = gtk_track_display_leave_notify_event;
  widget_class->size_request = gtk_track_display_size_request;
  widget_class->draw = gtk_track_display_draw;
  widget_class->draw_default = gtk_track_display_draw_default;
  widget_class->draw_focus = gtk_track_display_draw_focus;
}

GtkType gtk_track_display_get_type()
{
  static GtkType track_display_type = 0;
  if (!track_display_type)
    {
      static const GtkTypeInfo track_display_info = {
	"GtkTrackDisplay",
	sizeof (GtkTrackDisplay),
	sizeof (GtkTrackDisplayClass),
	(GtkClassInitFunc) gtk_track_display_class_init,
	(GtkObjectInitFunc) gtk_track_display_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      //GtkType parent = GTK_TYPE_EVENT_BOX;
      GtkType parent = GTK_TYPE_DRAWING_AREA;
      parent_class = gtk_type_class (parent);
      track_display_type = gtk_type_unique (parent, &track_display_info);
    }
  return track_display_type;
}


static void alloc_colors(GtkTrackDisplay* display)
{
  int state;
  GtkWidget *widget;
  GdkGC *gc;
  if (!display->backing_pixmap)
    return;
  widget = GTK_WIDGET (display);
  state = GTK_WIDGET_STATE (widget);
  gc = gdk_gc_new (display->backing_pixmap);
  gdk_gc_set_foreground (gc, util_alloc_color_cached (65535, 0, 0));
  display->upper_gc = widget->style->fg_gc[GTK_STATE_NORMAL];
  display->lower_gc = widget->style->bg_gc[GTK_STATE_NORMAL];
  display->origin_gc = widget->style->fg_gc[GTK_STATE_ACTIVE];
  display->desc_gc = widget->style->bg_gc[GTK_STATE_ACTIVE];
  display->text_gc = widget->style->fg_gc[GTK_STATE_INSENSITIVE];
  display->font = gdk_font_load ("fixed");
}

static void render_range_description (GtkTrackDisplay *display,
				      int beat_index)
{
  gdouble len;
  gdouble firb;
  GdamBeatInfo *pbi;
  int draw_width = display->draw_width;
  int draw_height = display->draw_height;
  int cur_x;
  const char *name;
  int allowed_width;
  int next_index;

  len = display->length;
  pbi = display->beat_info;
  firb = gdam_beat_info_get_range_first_beat (pbi, beat_index);
  /* This is a common occurance, because the beat info and length 
   * are set seperately, and if they don't match when this is
   * called we may have values out of bound.  Once we can
   * atomicize the setting of those values, or all trackdisplay
   * users can freeze/thaw as needed, this should be an assertion.
   */
  //g_return_if_fail (firb <= len);
  if (firb > len)
    return;

  name = gdam_beat_info_get_range_name (pbi, beat_index);
  cur_x = (draw_width * firb / len) + 2;
  g_return_if_fail (cur_x <= draw_width);
  allowed_width = draw_width - cur_x;
  next_index = gdam_beat_info_guess_next_index (pbi, firb);
  if (next_index >= 0)
    {
      gdouble nextfirb;
      int next_x;
      nextfirb = gdam_beat_info_get_range_first_beat (pbi, next_index);
      next_x = (draw_width * nextfirb / len);
      allowed_width = next_x - cur_x;
    }
  g_return_if_fail (allowed_width >= 0);

  /* Render the "up"/"down" as triangles. */
  if (name != NULL && 
      (g_strcasecmp (name, "up") == 0 || 
       g_strcasecmp (name, "down") == 0))
    {
      GdkPoint points[3];
      int half_height = draw_height / 2;
      if (draw_height > allowed_width)
	half_height = allowed_width / 2;
      points[0].x = cur_x;
      points[1].x = cur_x + (half_height * 2);
      points[2].x = cur_x + half_height;;
      if (g_strcasecmp (name, "up") == 0)
	{
	  points[0].y = draw_height - 1;
	  points[1].y = draw_height - 1;
	  points[2].y = 1;
	}
      else
	{
	  points[0].y = 1;
	  points[1].y = 1;
	  points[2].y = draw_height - 1;
	}
      gdk_draw_polygon (display->backing_pixmap, display->text_gc,
			TRUE, points, 3);
    }
  else if (name != NULL && (g_strcasecmp (name, "break") == 0)) 
    {
      GdkPoint points[2];
      int quarter_height = draw_height / 4;
      if (draw_height > allowed_width)
	quarter_height = allowed_width / 4;
      points[0].x = cur_x ;
      points[1].x = cur_x + (quarter_height * 2);
      points[0].y = draw_height / 2 - 1;
      points[1].y = draw_height / 2 - 1;
      gdk_draw_arc (display->backing_pixmap, display->text_gc,
	            TRUE,
		    points[0].x, points[0].y,
		    quarter_height * 2 - 1, quarter_height * 2 - 1,
		    0, 360 * 64);
      gdk_draw_arc (display->backing_pixmap, display->text_gc,
	            TRUE,
		    points[1].x, points[1].y,
		    quarter_height * 2 - 1, quarter_height * 2 - 1,
		    0, 360 * 64);
    }
  else if (name != NULL && 
           (g_strcasecmp (name, "verse") == 0 || 
	    g_strcasecmp (name, "chorus") == 0)) 
    {
      GdkPoint points[3];
      int half_height = draw_height / 2;
      if (draw_height > allowed_width)
	half_height = allowed_width / 2;
      points[0].x = cur_x ;
      points[0].y = 1;
      //points[0].y = draw_height / 2 - 1;
      gdk_draw_arc (display->backing_pixmap, display->text_gc,
	            TRUE,
		    points[0].x, points[0].y,
		    half_height * 2 - 1, half_height * 2 - 1,
		    0, 360 * 64);
      if (g_strcasecmp (name, "verse") == 0)
	{
	  points[0].x = cur_x + 0;
	  points[1].x = cur_x + (half_height * 2) - 0;
	  points[2].x = cur_x + half_height - 1;
	  points[0].y = 1;
	  points[1].y = 1;
	  points[2].y = half_height + 1;
	}
      else
	{
	  points[0].x = cur_x + (half_height * 2) - 1;
	  points[1].x = cur_x + half_height - 1;
	  points[2].x = cur_x + (half_height * 2) - 1;
	  points[0].y = 2;
	  points[1].y = half_height + 1;
	  points[2].y = half_height * 2 - 1;
	}
      gdk_draw_polygon (display->backing_pixmap, display->lower_gc,
			TRUE, points, 3);
    }
  else if (name != NULL && (g_strcasecmp (name, "ambient") == 0)) 
    {
      GdkPoint points[1];
      int half_height = draw_height / 2;
      if (draw_height > allowed_width)
	half_height = allowed_width / 2;
      points[0].x = cur_x ;
      points[0].y = draw_height / 2 - 1;
      gdk_draw_arc (display->backing_pixmap, display->text_gc,
	            TRUE,
		    points[0].x, points[0].y,
		    half_height * 2 - 1, half_height * 2 - 1,
		    0, 180 * 64);
    }
  /* Render the name as text. */
  else if (name != NULL && strcmp(name, "") != 0)
    {
      gpointer nextdata;
      char *str;
      int text_width;
      str = g_strdup (name);
      text_width = gdk_string_width (display->font, str);
      if (text_width > allowed_width)
	{
	  int char_width = text_width / strlen (str);
	  int num_chars = allowed_width / char_width;
	  g_return_if_fail (num_chars >= 0);
	  g_return_if_fail (num_chars < 1000);
	  g_free (str);
	  str = g_new (char, (num_chars + 1));
	  if (num_chars)
	    strncpy (str, name, num_chars);
	  str[num_chars] = '\0';
	}
      gdk_draw_string (display->backing_pixmap,
		       display->font,
		       display->text_gc,
		       cur_x, 
		       draw_height - 1, 
		       str);
      g_free (str);
    }
}

static void redraw_track_area(GtkTrackDisplay *display)
{
  GdkSegment *segments;
  GdamBeatInfo *pbi;
  int num_segments;
  int num_ranges;
  int draw_width = display->draw_width;
  int draw_height = display->draw_height;
  gdouble fb;
  gdouble len;
  gdouble percent;
  int i;

  if (!display->backing_pixmap)
    return;

  if (display->freeze_count > 0)
    {
      display->changed_while_frozen = TRUE;
      return;
    }

  g_return_if_fail (draw_width > 0);

  if (display->upper_gc == NULL)
    alloc_colors (display);

  /*  Fill with background color */
  gdk_draw_rectangle (display->backing_pixmap,
		      display->lower_gc,
		      TRUE, 0, 0, draw_width, draw_height);
  len = display->length;
  pbi = display->beat_info;
  if (!pbi) 
    {
      gtk_widget_draw_default (GTK_WIDGET (display));
      return;
    }
  if (len <= 0) 
    {
      gtk_widget_draw_default (GTK_WIDGET (display));
      return;
    }
  num_ranges = 0;
  if (pbi->range_list)
    num_ranges = g_list_length (pbi->range_list);
  num_segments = num_ranges + 2;
  segments = alloca (sizeof (GdkSegment) * num_segments);

  /*  Draw bar for current first beat  */
  fb = gdam_beat_info_get_range_first_beat (pbi, display->range_index);
  percent = fb / len;
  segments[0].x1 = segments[0].x2 = percent * draw_width;
  segments[0].y1 = 0;
  segments[0].y2 = draw_height;

  /*  Draw tick for default first beat...  */
  fb = gdam_beat_info_get_range_first_beat (pbi, 0);
  percent = fb / len;
  segments[1].x1 = segments[1].x2 = percent * draw_width;
  segments[1].y1 = (draw_height / 2) - 1;
  segments[1].y2 = draw_height - 1;

  render_range_description (display, 0);

  /*  Draw ticks for auxilliary first beats...  */
  for (i = 0; i < num_ranges; i++)
    {
      gdouble firb;
      firb = gdam_beat_info_get_range_first_beat (pbi, i+1);
      segments[i + 2].x1 = segments[i + 2].x2 = draw_width * firb / len;
      /* If this BIR defines a tempo range, draw the lower half of the line. */
      if (gdam_beat_info_get_range_bpm (pbi, i+1) > 0)
	{
	  segments[i + 2].y1 = (draw_height / 2) - 1;
	  segments[i + 2].y2 = draw_height - 1;
	}
      /* If this BIR defines an index point, draw the top half. */
      else
	{
	  segments[i + 2].y1 = 0;
	  segments[i + 2].y2 = (draw_height / 2) - 1;
	}
    }
  gdk_draw_segments (display->backing_pixmap,
		     display->upper_gc, segments, num_segments);

  /* Render the description for each range, if possible. */
  for (i = 0; i < num_ranges; i++)
    render_range_description (display, i+1);

  /*  Draw dot for timer origin  */
  percent = display->timer_origin / len;
  segments[0].x1 = (percent * draw_width) - 1;
  if (segments[0].x1 < 0)
    segments[0].x1 = 0;
  segments[0].x2 = (percent * draw_width) + 1;
  if (segments[0].x2 >= draw_width)
    segments[0].x2 = draw_width - 1;
  segments[0].y1 = draw_height - 2;
  segments[0].y2 = draw_height - 2;
  gdk_draw_segments (display->backing_pixmap, display->origin_gc, segments, 1);

  gtk_widget_draw_default (GTK_WIDGET (display));
}

void gtk_track_display_freeze (GtkTrackDisplay *display)
{
  display->freeze_count++;
}

void gtk_track_display_thaw (GtkTrackDisplay *display)
{
  display->freeze_count--;
  if (display->freeze_count <= 0 && display->changed_while_frozen)
    {
      redraw_track_area (display);
      display->changed_while_frozen = FALSE;
    }
}

void gtk_track_display_set_range_index (GtkTrackDisplay *display,
				        int             idx)
{
  display->range_index = idx;
  redraw_track_area (display);
}

void gtk_track_display_set_beat_info (GtkTrackDisplay *display,
				      const GdamBeatInfo *bi)
{
  if (display->beat_info)
    gdam_beat_info_destroy (display->beat_info);
  display->beat_info = gdam_beat_info_copy (bi);
  redraw_track_area (display);
}

void gtk_track_display_set_length (GtkTrackDisplay *display,
				   gdouble 	   length)
{
  display->length = length;
  redraw_track_area (display);
}

void gtk_track_display_set_origin (GtkTrackDisplay *display,
				   gdouble 	   origin)
{
  display->timer_origin = origin;
  redraw_track_area (display);
}



static void gtk_track_display_draw(GtkWidget* widget, GdkRectangle* area)
{
  GdkRectangle tmp;
  GtkTrackDisplay *display;
  display = GTK_TRACK_DISPLAY (widget);

  if (!GTK_WIDGET_DRAWABLE (widget))
    {
      return;
    }

  if (!display->backing_pixmap)
    return;

  if (display->upper_gc == NULL)
    alloc_colors (display);

  if (area == NULL)
    {
      area = &tmp;
      tmp.x = 0;
      tmp.y = 0;
      tmp.width = display->draw_width;
      tmp.height = display->draw_height;
    }

  gdk_draw_pixmap (widget->window,
		   widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
		   display->backing_pixmap, area->x, area->y, area->x, area->y,
		   area->width, area->height);
}

static void
gtk_track_display_draw_focus (GtkWidget *widget)
{
  gtk_widget_draw (widget, NULL);
}

static void
gtk_track_display_draw_default (GtkWidget *widget)
{
  gtk_widget_draw (widget, NULL);
}

static int gtk_track_display_button_press_event(GtkWidget*       widget,
                                                GdkEventButton*  event)
{
  return TRUE;
  //return FALSE;
}

static int gtk_track_display_button_release_event(GtkWidget*       widget,
                                               GdkEventButton*  event)
{
  return TRUE;
}

static int gtk_track_display_expose_event(GtkWidget*       widget,
                                       GdkEventExpose*  event)
{
  gtk_widget_draw_default (widget);
  return FALSE;
}
static int gtk_track_display_configure_event       (GtkWidget*       widget,
                                                 GdkEventConfigure*event)
{
  GtkTrackDisplay *display;
  display = GTK_TRACK_DISPLAY (widget);
  if (display->draw_width == widget->allocation.width &&
      display->draw_height == widget->allocation.height)
    {
      return TRUE;
    }
  display->draw_width = widget->allocation.width;
  display->draw_height = widget->allocation.height;
  if (display->backing_pixmap)
    {
      gdk_pixmap_unref (display->backing_pixmap);
    }
  display->backing_pixmap = gdk_pixmap_new (widget->window,
					 widget->allocation.width,
					 widget->allocation.height, -1);
  if (display->upper_gc == NULL)
    alloc_colors (display);
  redraw_track_area (display);
  return FALSE;
}

static int gtk_track_display_enter_notify_event(GtkWidget*       widget,
                                             GdkEventCrossing*event)
{
  gtk_widget_draw_default (widget);
  return FALSE;
}
static int gtk_track_display_leave_notify_event(GtkWidget*       widget,
                                             GdkEventCrossing*event)
{
  gtk_widget_draw_default (widget);
  return FALSE;
}

static void gtk_track_display_size_request    (GtkWidget*       widget,
                                            GtkRequisition*  requisition)
{
  requisition->width = 100;
  requisition->height = 100;
}

GtkWidget* gtk_track_display_create_func   (const char*   name,
                                            const char*   string1,
                                            const char*   string2,
                                            int           int1,
                                            int           int2)
{
  //char *pixmaps_colon_sep = g_strdup (string1);
  //char **pixmaps = g_strsplit (pixmaps_colon_sep, ":", 100);
  GtkTrackDisplay *track_display;
  track_display = GTK_TRACK_DISPLAY (gtk_type_new (GTK_TYPE_TRACK_DISPLAY));
  //track_display->is_toggle = int1;
  //g_strfreev (pixmaps);
  //g_free (pixmaps_colon_sep);
  return GTK_WIDGET (track_display);
}

