/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdambenderskin.h"
#include "../gdamskinutil.h"
#include <gdam/component/plugins/gdammididevice.h>
#include <gdam/common/gdamsample.h>
#include <gdam/component/gdampitch.h>
#include <gdam/common/gdamsample.h>
#include <glade/glade.h>
#include <gtk/gtksignal.h>
#include <gtk/gtktogglebutton.h>
//#include "../gdamfactory.h"
#include <stdlib.h>
#include <gdam/common/gdamplugin.h>

static void gdam_bender_skin_class_init(GdamBenderSkinClass* klass);

static gboolean gdam_bender_build_new_model(GdamSkin*      skin,
                                              GdamChannel*   channel,
					      int            num_options,
					      char**         options);

/*    ____ _                 ____  _          __  __ 
 *   / ___| | __ _ ___ ___  / ___|| |_ _   _ / _|/ _|
 *  | |   | |/ _` / __/ __| \___ \| __| | | | |_| |_ 
 *  | |___| | (_| \__ \__ \  ___) | |_| |_| |  _|  _|
 *   \____|_|\__,_|___/___/ |____/ \__|\__,_|_| |_|  
 */                                                 

static void gdam_bender_skin_init(GdamBenderSkin* skin)
{
  skin->bend_mode = GDAM_BENDER_KEEP_TIME;
}

GtkType gdam_bender_skin_get_type()
{
  static GtkType bender_skin_type = 0;
  if (!bender_skin_type)
    {
      static const GtkTypeInfo bender_skin_info = {
	"GdamBenderSkin",
	sizeof (GdamBenderSkin),
	sizeof (GdamBenderSkinClass),
	(GtkClassInitFunc) gdam_bender_skin_class_init,
	(GtkObjectInitFunc) gdam_bender_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      bender_skin_type = gtk_type_unique (GDAM_TYPE_SKIN, &bender_skin_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return bender_skin_type;
}

/*  ____  _    _         _____                 _   _                 
 * / ___|| | _(_)_ __   |  ___|   _ _ __   ___| |_(_) ___  _ __  ___ 
 * \___ \| |/ / | '_ \  | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
 *  ___) |   <| | | | | |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
 * |____/|_|\_\_|_| |_| |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
 */                                                                 


typedef struct _StallInfo {
	GdamBenderSkin* skin;
	float amount;
} StallInfo;

static void stall_clicked(GtkButton* button, StallInfo* stall_info)
{
  GdamBender *bender;
  bender = GDAM_BENDER (GDAM_SKIN (stall_info->skin)->model);
  gdam_bender_slow_to_stall (bender, stall_info->amount);
}

static int delta_millis(GTimeVal* older, GTimeVal* younger)
{
  int millis = (older->tv_sec - younger->tv_sec) * 1000
    + (older->tv_usec - younger->tv_usec) / 1000;
  return millis;
}

static void bend_adj_changed(GtkAdjustment* adj, GdamBenderSkin* skin)
{
  GdamBender *bender = GDAM_BENDER (GDAM_SKIN (skin)->model);
  if (!skin->user_bending)
    {
      gdam_debug ("bend adj changed, but aborting.");
      return;
    }
  else
    {
      GTimeVal real_time;
      int num_points = 100;
      float total_time = 0.2;
      int *change_array = g_new (int, num_points * 2);
      int n, delmil;
      float last_rate = skin->last_rate;
      float new_rate;
      float cur_rate = last_rate;
      float step_rate;
      float step_time;
      new_rate = adj->value;	/* TODO: asymptotic mapping */

      if (bender->autotune)
	new_rate = gdam_pitch_get_magic_rate (new_rate);

      skin->last_rate = new_rate;
      step_rate = (new_rate - last_rate) / num_points;
      step_time = total_time / num_points;

      g_get_current_time (&real_time);
      delmil = delta_millis (&real_time, &skin->bend_adj_time);
      g_get_current_time (&skin->bend_adj_time);

      for (n = 0; n < num_points * 2; n += 2)
	{
	  cur_rate += step_rate;
	  change_array[n] = step_time * 44100;
	  if (n == num_points * 2 - 2)
	    change_array[n] = 10.0 * 44100;
	  change_array[n + 1] = cur_rate * GdamFixed_ONE;
	}

      gdam_bender_set_changes (bender, num_points, change_array);
    }
}

static void bend_adj_button_event(GtkWidget*      wid,
			    	  GdkEventButton* ev,
			    	  GdamBenderSkin* skin)
{
  GdamBender *bender = GDAM_BENDER (GDAM_SKIN (skin)->model);
  gdam_debug ("bend adj button_event");
  if (ev->type == GDK_BUTTON_PRESS)
    {
      skin->user_bending = 1;
      skin->last_rate = 1.0;
      gdam_bender_set_mode (bender, skin->bend_mode);
    }
  else if (ev->type == GDK_BUTTON_RELEASE)
    {
      /* TODO: is this necessary w/o SNAP_TIME? */
      if (skin->bend_mode == GDAM_BENDER_CONTINUOUS)
	gdam_bender_set_changes (bender, 0, NULL);
      else if ((ev->state & GDK_BUTTON2_MASK) == GDK_BUTTON2_MASK ||
	       (ev->state & GDK_BUTTON3_MASK) == GDK_BUTTON3_MASK)
	gdam_arg_set (GTK_OBJECT (bender), "jump_to_nearest_mark", 1);
      else
	gdam_bender_jump_to_mark (bender);
      skin->bend_adj->value = 1.0;
      skin->user_bending = 0;	/* set this before adj_v_changed */
      gtk_adjustment_value_changed (skin->bend_adj);
    }
}

static gboolean gdam_bender_bend_adjustment_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *bend_adj;
  GdamBenderSkin *bender_skin = GDAM_BENDER_SKIN (skin);
  bend_adj = util_get_widget_adjustment (GTK_WIDGET (object));
  bender_skin->bend_adj = bend_adj;
  gtk_signal_connect (GTK_OBJECT (bend_adj), "value_changed",
		      GTK_SIGNAL_FUNC (bend_adj_changed), bender_skin);
  gtk_signal_connect (object,
		      "button_press_event",
		      GTK_SIGNAL_FUNC (bend_adj_button_event), bender_skin);
  gtk_signal_connect (object,
		      "button_release_event",
		      GTK_SIGNAL_FUNC (bend_adj_button_event), bender_skin);
  gdam_delayed_adjustment_init (bend_adj);
  return TRUE;
}

static gboolean gdam_bender_stall_button_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkButton *button = GTK_BUTTON (object);
  float amount;
  char *endp;
  StallInfo *stall_info;
  if (handler_names[1] == NULL)
    {
      g_warning ("%s needs a real-valued argument", handler_names[0]);
      return TRUE;
    }
  amount = strtod (handler_names[1], &endp);
  if (endp == handler_names[1] || *endp != '\0')
    {
      return TRUE;
    }
  stall_info = g_new (StallInfo, 1);
  stall_info->skin = GDAM_BENDER_SKIN (skin);
  stall_info->amount = amount;
  gtk_signal_connect (GTK_OBJECT (button),
		      "clicked", GTK_SIGNAL_FUNC (stall_clicked), stall_info);
  return TRUE;
}
 
gboolean gdam_bender_skin_set_child(GdamBenderSkin* 	skin,
				     const char* 	skin_type_name, 
				     const char* 	skin_filename,
				     int                num_options,
				     char**             options)
{
  GdamModel *submodel;
  GdamSkin *subview;
  GdamBender *bender = GDAM_BENDER (GDAM_SKIN (skin)->model);
  gdam_debug ("gdam_builder_skin_set_child: "
	      "type=%s, filename=%s, options=%s",
	      skin_type_name, skin_filename, options);
  subview = gdam_skin_new_by_type_name
    (skin_type_name,
     skin_filename,
     num_options, options, GDAM_MODEL (bender)->channel, &submodel);
  if (subview == NULL)
    {
      g_warning ("error creating bender's child");
      return FALSE;
    }
  gdam_filter_set_child (GDAM_FILTER (bender), submodel);
  skin->child_skin = subview;
  if (skin->child_container)
    {
      gdam_debug ("gdam_builder_skin: adding child to container");
      gtk_container_add (skin->child_container, subview->main_widget);
    }
  return TRUE;
}

static void rewind_pressed(GtkObject*	object,
			       GdamBenderSkin*	skin)
{
  GdamBender *bender;
  bender = GDAM_BENDER (GDAM_SKIN (skin)->model);
  //gdam_bender_set_mode(bender, GDAM_BENDER_KEEP_TIME);
  gdam_bender_rewind (bender, -1);
}

static void rewind_released(GtkObject*	object,
			        GdamBenderSkin*	skin)
{
  GdamBender *bender;
  bender = GDAM_BENDER (GDAM_SKIN (skin)->model);
  gdam_bender_jump_to_mark (bender);
}

static gboolean gdam_bender_rewind_button_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkButton *button = GTK_BUTTON (object);
  gtk_signal_connect (GTK_OBJECT (button),
		      "pressed", GTK_SIGNAL_FUNC (rewind_pressed), skin);
  gtk_signal_connect (GTK_OBJECT (button),
		      "released", GTK_SIGNAL_FUNC (rewind_released), skin);
  return TRUE;
}


static void	   rub_pressed(GtkObject*	object,
			       GdamBenderSkin*	skin)
{
  GdamBender *bender;
  bender = GDAM_BENDER (GDAM_SKIN (skin)->model);
  gdam_bender_set_mode (bender, GDAM_BENDER_KEEP_TIME);
  gdam_bender_rub (bender, 0, 0);
}

static void	   rub_released(GtkObject*	object,
			        GdamBenderSkin*	skin)
{
  GdamBender *bender;
  bender = GDAM_BENDER (GDAM_SKIN (skin)->model);
  gdam_bender_jump_to_mark (bender);
}


static void	   free_clicked(GtkObject*	object,
			        GdamBenderSkin*	skin)
{
  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (object)))
    skin->bend_mode = GDAM_BENDER_CONTINUOUS;
  else
    skin->bend_mode = GDAM_BENDER_KEEP_TIME;
  gdam_bender_set_mode (GDAM_BENDER (GDAM_SKIN (skin)->model),
			skin->bend_mode);
}

static gboolean gdam_bender_rub_button_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkButton *button = GTK_BUTTON (object);
  gtk_signal_connect (GTK_OBJECT (button),
		      "pressed", GTK_SIGNAL_FUNC (rub_pressed), skin);
  gtk_signal_connect (GTK_OBJECT (button),
		      "released", GTK_SIGNAL_FUNC (rub_released), skin);
  return TRUE;
}

static gboolean gdam_bender_free_tbutton_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkToggleButton *tbutton = GTK_TOGGLE_BUTTON (object);
  gtk_signal_connect (GTK_OBJECT (tbutton),
		      "clicked", GTK_SIGNAL_FUNC (free_clicked), skin);
  return TRUE;
}
 
static gboolean gdam_bender_build_new_model(GdamSkin*      skin,
                                            GdamChannel*   channel,
					    int            num_options,
					    char**         options)
{
  GdamModel *model;
  model = GDAM_MODEL (gtk_type_new (GDAM_TYPE_BENDER));
  model->channel = channel;
  skin->model = model;
  if (num_options <= 0)
    {
      return TRUE;
    }
  if (num_options < 2)
    {
      g_warning ("too few options to bender filter");
      return FALSE;
    }
  gdam_bender_skin_set_child (GDAM_BENDER_SKIN (skin),
			      options[0], options[1],
			      num_options - 2, options + 2);
  return TRUE;
}

static GdamSkinGenericConnect connect_infos[] = {
  { "gdam_bender_stall_button", gdam_bender_stall_button_handler },
  { "gdam_bender_bend_adjustment", gdam_bender_bend_adjustment_handler },
  { "gdam_bender_rewind_button", gdam_bender_rewind_button_handler },
  { "gdam_bender_rub_button", gdam_bender_rub_button_handler },
  { "gdam_bender_free_tbutton", gdam_bender_free_tbutton_handler}
};


static void gdam_bender_skin_class_init(GdamBenderSkinClass* klass)
{
  GdamSkinClass *skin_class = GDAM_SKIN_CLASS (klass);
  skin_class->build_new_model = gdam_bender_build_new_model;
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
}

