/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamoscilloscopeskin.h"
#include <gdam/skin/gdamskinutil.h>
#include <gtk/gtkdrawingarea.h>
#include <string.h>
#include <gdam/component/plugins/gdamoscilloscope.h>
#include <gdam/common/gdamsample.h>
#include <gtk/gtksignal.h>
#include <gdam/skin/gdamskinutil.h>
#include <gdam/common/gdamplugin.h>

static GtkObjectClass* parent_class = NULL;
static void schedule_query_timeout(GdamOscilloscopeSkin* skin);

static void
gdam_oscilloscope_skin_init(GdamOscilloscopeSkin* oscilloscope_skin)
{
  schedule_query_timeout (oscilloscope_skin);
}
static void gdam_oscilloscope_skin_class_init(GdamSkinClass* skin_class);

GtkType gdam_oscilloscope_skin_get_type()
{
  static GtkType oscilloscope_skin_type = 0;
  if (!oscilloscope_skin_type)
    {
      static const GtkTypeInfo oscilloscope_skin_info = {
	"GdamOscilloscopeSkin",
	sizeof (GdamOscilloscopeSkin),
	sizeof (GdamOscilloscopeSkinClass),
	(GtkClassInitFunc) gdam_oscilloscope_skin_class_init,
	(GtkObjectInitFunc) gdam_oscilloscope_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      parent_class = gtk_type_class (parent);
      oscilloscope_skin_type
	= gtk_type_unique (parent, &oscilloscope_skin_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return oscilloscope_skin_type;
}

static void alloc_colors(GdamOscilloscopeSkin* skin)
{
  int state = GTK_WIDGET_STATE (skin->drawing_area);
  GdkGC *gc = gdk_gc_new (skin->backing_pixmap);
  gdk_gc_set_foreground (gc, util_alloc_color_cached (65535, 0, 0));
  skin->upper_gc = skin->drawing_area->style->bg_gc[state];
  skin->lower_gc = skin->drawing_area->style->fg_gc[state];
}


static void copy_backing_area(GdamOscilloscopeSkin* skin,
                              GdkRectangle*          area)
{
  GdkRectangle tmp;
  g_return_if_fail (!GTK_OBJECT_DESTROYED (GTK_OBJECT (skin)));
  g_return_if_fail (skin->drawing_area != NULL);
  g_return_if_fail (GTK_IS_WIDGET (skin->drawing_area));
  if (area == NULL)
    {
      area = &tmp;
      tmp.x = 0;
      tmp.y = 0;
      tmp.width = skin->draw_width;
      tmp.height = skin->draw_height;
    }
  gdk_draw_pixmap (skin->drawing_area->window,
		   skin->drawing_area->style->
		   fg_gc[GTK_WIDGET_STATE (skin->drawing_area)],
		   skin->backing_pixmap, area->x, area->y, area->x, area->y,
		   area->width, area->height);
}


static void gdam_oscilloscope_set_ints   (GdamOscilloscopeSkin* skin,
                                          int                   num_ints,
					  int*                  ints)
{
  GdkSegment *lower;
  int draw_width = skin->draw_width;
  int draw_height = skin->draw_height;
  int i;

  g_return_if_fail (draw_width > 0);
  g_return_if_fail (GTK_OBJECT (skin)->klass != NULL);
  g_return_if_fail (!GTK_OBJECT_DESTROYED (GTK_OBJECT (skin)));

  lower = alloca (sizeof (GdkSegment) * draw_width);
  for (i = 0; i < skin->draw_width; i++)
    {
      lower[i].x1 = lower[i].x2 = i;
      lower[i].y1 = (GDAM_SAMPLE_ONE
		     - ints[i * num_ints / draw_width])
	* draw_height / (2 * GDAM_SAMPLE_ONE);
      lower[i].y2 = draw_height / 2;
    }
  gdk_draw_rectangle (skin->backing_pixmap,
		      skin->upper_gc,
		      TRUE, 0, 0, skin->draw_width, skin->draw_height);
  gdk_draw_segments (skin->backing_pixmap,
		     skin->lower_gc, lower, skin->draw_width);
  copy_backing_area (skin, NULL);
}

static gint
oscilloscope_expose(GtkDrawingArea* area,
             GdkEventExpose* expose_event,
             GdamOscilloscopeSkin*  skin)
{
  if (skin->upper_gc == NULL)
    alloc_colors (skin);
  copy_backing_area (skin, &expose_event->area);
  return 0;
}

static gint
oscilloscope_configure(GtkWidget* widget,
                        GdkEventConfigure* configure_event,
		        GdamOscilloscopeSkin* skin)
{
  if (skin->backing_pixmap)
    gdk_pixmap_unref (skin->backing_pixmap);
  skin->backing_pixmap = gdk_pixmap_new (widget->window,
					 widget->allocation.width,
					 widget->allocation.height, -1);
  skin->draw_width = widget->allocation.width;
  skin->draw_height = widget->allocation.height;
  if (skin->upper_gc == NULL)
    alloc_colors (skin);
  return TRUE;
}

static void connect_drawing_area(GtkDrawingArea*    drawing_area,
                                 GdamOscilloscopeSkin* skin)
{
  skin->drawing_area = GTK_WIDGET (drawing_area);
  gtk_drawing_area_size (GTK_DRAWING_AREA (drawing_area), 800, 300);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "expose_event",
		      GTK_SIGNAL_FUNC (oscilloscope_expose), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "configure_event",
		      GTK_SIGNAL_FUNC (oscilloscope_configure), skin);
  gtk_widget_show (GTK_WIDGET (drawing_area));
  gtk_widget_add_events (GTK_WIDGET (drawing_area), GDK_EXPOSURE_MASK);
}


/* 
 *  ____  _    _         _____                 _   _                 
 * / ___|| | _(_)_ __   |  ___|   _ _ __   ___| |_(_) ___  _ __  ___ 
 * \___ \| |/ / | '_ \  | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
 *  ___) |   <| | | | | |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
 * |____/|_|\_\_|_| |_| |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
 */
static void got_sample_data(int num_args, GdamArg* args, gpointer skin_ptr)
{
  GdamOscilloscopeSkin *skin;
  int length;
  int *array;
  g_return_if_fail (num_args == 1);
  g_return_if_fail (args->type == GDAM_ARG_TYPE_INT_ARRAY);
  skin = GDAM_OSCILLOSCOPE_SKIN (skin_ptr);

  skin->request_pending = FALSE;

  length = GDAM_ARG_INT_ARRAY (args).length;
  array = GDAM_ARG_INT_ARRAY (args).array;

  if (!GTK_OBJECT_DESTROYED (skin) && length > 0)
    {
      gdam_oscilloscope_set_ints (skin, length, array);
    }

  gtk_object_unref (GTK_OBJECT (skin));
}

static void didnt_get_sample_data(gpointer skin_ptr)
{
  GdamOscilloscopeSkin *skin = skin_ptr;
  g_warning ("didn't get oscilloscope sample data");
  skin->request_pending = FALSE;
  gtk_object_unref (GTK_OBJECT (skin));
}

static gboolean do_query(GdamOscilloscopeSkin* skin)
{
  GdamModel *model;

  model = GDAM_SKIN (skin)->model;

  if (model == NULL || model->source_id == 0)
    return TRUE;

  if (!skin->request_pending)
    {
      const char *name = "if_modified_sample_data";
      gtk_object_ref (GTK_OBJECT (skin));
      gdam_channel_append_request
	(model->channel,
	 gdam_request_get_args
	 (model->source_id,
	  1, &name, got_sample_data, didnt_get_sample_data, skin));
      skin->request_pending = TRUE;
    }
  return TRUE;
}

	


static void schedule_query_timeout(GdamOscilloscopeSkin* skin)
{
  if (skin->query_timeout_id != 0)
    return;
  skin->query_timeout_id = g_timeout_add (100, (GSourceFunc) do_query, skin);
}



static gboolean gdam_oscilloscope_drawing_area_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  if (!GTK_IS_DRAWING_AREA (object))
    {
      g_warning ("%s may only be connected to a drawing area",
		 handler_names[0]);
      return TRUE;
    }
  connect_drawing_area (GTK_DRAWING_AREA (object),
			GDAM_OSCILLOSCOPE_SKIN (skin));
  return TRUE;
}

static GdamSkinGenericConnect connect_infos[] = { 	
	{ "gdam_oscilloscope_drawing_area",
	   gdam_oscilloscope_drawing_area_handler },
};

static void gdam_oscilloscope_skin_destroy(GtkObject* object)
{
  GdamOscilloscopeSkin *skin = GDAM_OSCILLOSCOPE_SKIN (object);
  g_message ("osc destroy");
  skin->drawing_area = NULL;
  if (skin->request_pending)
    g_warning ("Oscilloscope request pending, fear seg fault.");
  if (skin->query_timeout_id != 0)
    {
      g_source_remove (skin->query_timeout_id);
      skin->query_timeout_id = 0;
    }
  (*parent_class->destroy) (object);
}

static void gdam_oscilloscope_skin_class_init(GdamSkinClass* skin_class)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (skin_class);
  skin_class->model_type = GDAM_TYPE_OSCILLOSCOPE;
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
  object_class->destroy = gdam_oscilloscope_skin_destroy;
}
