/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamsamplingsequencerskin.h"
#include <gdam/skin/gdamskinutil.h>
#include <gdam/component/gdambuffermanager.h>
#include <gdam/common/gdam-protocol.h>
#include <gtk/gtk.h>
#include <stdlib.h>
#include <math.h>

typedef struct _GdamSubSampleInfo {
  int buffer_id;
  gdouble start_time;
  gdouble length;
  gdouble beatnumber_in_source;
} GdamSubSampleInfo;

static GdamBufferViewport* get_viewport (GdamSamplingSequencerSkin* ssskin)
{
  if (!ssskin->viewport)
    {
      GdamBufferViewport *viewport;
      GdamChannel* channel;
      g_return_val_if_fail (GDAM_SKIN (ssskin)->model != NULL, NULL);
      channel = GDAM_SKIN (ssskin)->model->channel;
      viewport = gdam_buffer_viewport_new_shown (channel);
      gtk_object_ref (GTK_OBJECT (viewport));
      gtk_object_sink (GTK_OBJECT (viewport));
      viewport->can_drag_beat_lines = TRUE;
      viewport->is_selectable = FALSE;
      ssskin->viewport = viewport;
    }
  return ssskin->viewport;
}

static inline GdamSequencer* get_sequencer(GdamSamplingSequencerSkin* ssskin)
{
    return GDAM_SEQUENCER (GDAM_SKIN (ssskin)->model);
}

int sort_ssinfo (gconstpointer a, gconstpointer b)
{
  GdamSubSampleInfo* ssinfoa = (GdamSubSampleInfo*)a;
  GdamSubSampleInfo* ssinfob = (GdamSubSampleInfo*)b;
  if (ssinfoa->buffer_id < ssinfob->buffer_id)
    return -1;
  else if (ssinfoa->buffer_id > ssinfob->buffer_id)
    return 1;
  else if (ssinfoa->start_time < ssinfob->start_time)
    return -1;
  else if (ssinfoa->start_time > ssinfob->start_time)
    return 1;
  return 0;
}

int  gdam_sampling_sequencer_skin_add_sample (GdamSamplingSequencerSkin* ssskin,
    					      gdouble		start_time,
					      gdouble		length)
{
  GdamSubSampleInfo* ssinfo;
  ssinfo = g_new (GdamSubSampleInfo, 1);
  ssinfo->start_time = start_time;
  ssinfo->length = length;
  ssinfo->buffer_id = ssskin->sampled_buffer_id;
  if (!ssinfo->buffer_id)
    g_warning ("missing buffer_id in skin_add_sample");
  ssskin->subsample_infos = g_list_append (ssskin->subsample_infos, 
      					   ssinfo);
  return g_list_index (ssskin->subsample_infos, ssinfo);
}

/* Make an array of beat times from the subsample list */
void update_buffer_beatlist (GdamSamplingSequencerSkin* ssskin)
{
  GList *tmp;
  GArray *array;
  GdamBufferViewport *viewport = get_viewport (ssskin);
  array = g_array_new (FALSE, FALSE, sizeof (gdouble));
  tmp = ssskin->subsample_infos;
  while (tmp)
    {
      GdamSubSampleInfo* ssinfo;
      ssinfo = (GdamSubSampleInfo*) tmp->data;
      g_array_append_val (array, ssinfo->start_time);
      tmp = tmp->next;
    }
  gdam_buffer_viewport_set_beat_times(viewport, array->len, (gdouble*)array->data);
  /* TODO: shouldn't be needed?? */
  gdam_buffer_viewport_queue_redraw (viewport);
}

/* Make an array of local info structs from the subsample list */
void update_sequence_local_infos (GdamSamplingSequencerSkin* ssskin)
{
  GList *tmp;
  GdamSequencer *seq = get_sequencer (ssskin);
  int index = 0;
  int num_infos = g_list_length  (ssskin->subsample_infos);
  gdam_sequencer_set_num_local_infos (seq, num_infos);
  tmp = ssskin->subsample_infos;
  while (tmp)
    {
      GdamSubSampleInfo* ssinfo;
      GdamLocalSampleInfo *linfo;
      ssinfo = (GdamSubSampleInfo*) tmp->data;
      linfo = &g_array_index (seq->local_infos, GdamLocalSampleInfo, index);
      linfo->buffer_id = ssinfo->buffer_id;
      linfo->sample_offset = (int)(ssinfo->start_time * 44100);
      linfo->output_length = (int)(ssinfo->length * 44100);
      tmp = tmp->next;
      index++;
    }
}

static void free_subsamples (GdamSamplingSequencerSkin *ssskin)
{
  while (ssskin->subsample_infos)
    {
      GdamSubSampleInfo* ssinfo;
      ssinfo = (GdamSubSampleInfo*) ssskin->subsample_infos->data;
      ssskin->subsample_infos = g_list_remove (ssskin->subsample_infos, ssinfo);
      g_free (ssinfo);
    }
}

static void calculate_subsample_lengths (GdamSamplingSequencerSkin *ssskin)
{
  GdamBufferViewport *viewport = get_viewport (ssskin);
  GList *sslist = ssskin->subsample_infos;
  GList *cur = ssskin->subsample_infos;
  while (sslist)
    {
      GdamSubSampleInfo *ssinfo = (GdamSubSampleInfo*) sslist->data;
      if (sslist->next)
	{
	  GdamSubSampleInfo *nextinfo = (GdamSubSampleInfo*) sslist->next->data;
	  ssinfo->length = nextinfo->start_time - ssinfo->start_time;
	}
      else
	ssinfo->length = gdam_buffer_viewport_get_sample_length (viewport)
			  - ssinfo->start_time;
      sslist = sslist->next;
    }
}

static void subsamples_from_beat_times (GdamSamplingSequencerSkin *ssskin)
{
  GdamBufferViewport *viewport = get_viewport (ssskin);
  GList *sslist = NULL;
  int c;
  for (c = 0; c < viewport->num_beat_times; c++)
    {
      GdamSubSampleInfo *ssinfo;
      ssinfo = g_new (GdamSubSampleInfo, 1);
      ssinfo->buffer_id = ssskin->sampled_buffer_id;
      if  (!ssinfo->buffer_id)
	g_warning ("missing buffer_id in samples_from_beat_times");
      ssinfo->start_time = viewport->beat_times[c];
      sslist = g_list_prepend (sslist, ssinfo);
    }
  sslist = g_list_reverse (sslist);
  free_subsamples (ssskin);
  ssskin->subsample_infos = sslist;
  calculate_subsample_lengths (ssskin);
}

void select_subsample (GdamSamplingSequencerSkin* ssskin, int idx)
{
  GdamArg arg;
  arg.name = "current_subsample_index";
  arg.type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (&arg) = idx;
  gdam_arg_set_object_arg (GTK_OBJECT (ssskin), &arg);
}

static void cycle_subsample (GdamSamplingSequencerSkin* ssskin)
{
  int num_ss = g_list_length (ssskin->subsample_infos);
  int new_idx;
  new_idx = ssskin->current_subsample_index + 1;
  if (new_idx >= num_ss)
    new_idx = 0;
  select_subsample (ssskin, new_idx);
}

static void extend_subsample (GdamSamplingSequencerSkin* ssskin)
{
  int rv;
  GdamBufferViewport* viewport = get_viewport (ssskin);
  rv = gdam_sampling_sequencer_skin_add_sample (ssskin, 0, 0);
  select_subsample (ssskin, rv);
  gdam_buffer_viewport_queue_redraw (viewport);
}

static void set_buffer_id (GdamSamplingSequencerSkin* ssskin, int id)
{
  GdamBufferViewport* viewport = get_viewport (ssskin);
  gdam_buffer_viewport_set_buffer_id (viewport, id);
  gdam_buffer_viewport_whole_sample (viewport);
  ssskin->sampled_buffer_id = id;
}

static void set_default_buffer (GdamSamplingSequencerSkin* ssskin)
{
  int id = gdam_buffer_manager_obtain_id (NULL);
  gdam_buffer_manager_unref (id, NULL);
  set_buffer_id (ssskin, id);
}

static gdouble get_bpm (GdamSamplingSequencerSkin *ssskin)
{
  GdamBufferManagerInfo *bufinfo;
  bufinfo = gdam_buffer_manager_info_by_id(ssskin->sampled_buffer_id);
  g_return_val_if_fail (bufinfo != NULL, 0);
  if (!bufinfo->beat_info)
    return 0;
  return gdam_beat_info_get_bpm (bufinfo->beat_info, 0.0);
}

/* Set the sequencer's length in beats and tempo to match our sampled buffer,
 * and clear the old sequence. */
static gboolean prepare_sequencer (GdamSamplingSequencerSkin* ssskin)
{
  GdamBufferViewport *viewport = get_viewport (ssskin);
  gdouble bpm, beat_length;
  gdouble sample_length = gdam_buffer_viewport_get_sample_length (viewport);
  GdamModel *model = GDAM_SKIN (ssskin)->model;
  GdamBeatInfo new_beat_info = model->beat_info;
  int num_beats;

  bpm = get_bpm (ssskin);
  g_return_val_if_fail (bpm > 0.0, FALSE);
  gdam_beat_info_set_bpm (&new_beat_info, bpm);
  gdam_model_set_beat_info (model, &new_beat_info);
  beat_length = 60.0 / bpm;
  num_beats = (int) (sample_length / beat_length + 0.5);
  gdam_sequencer_set_num_beats (get_sequencer (ssskin), num_beats);
  gdam_sequencer_set_elements (get_sequencer (ssskin), NULL);
  gdam_sequencer_skin_display_full (GDAM_SEQUENCER_SKIN (ssskin));
  update_sequence_local_infos (ssskin);
  return TRUE;
}

static GdamSequencerElement *add_element (GdamSequencer *seq,
    					  int		 buffer_id,
					  gdouble	 start_beat,
					  gdouble	 gain,
					  gdouble	 rate,
					  int		 start,
					  int		 length,
					  int		 local_id)
{
  GdamSequencerElement *element;
  element = gdam_sequencer_add_element (seq,
				        buffer_id,
				        start_beat, /* position */
				        gain, rate, 
				        start, /* offest */
				        length);
  element->local_id = local_id;
  return element;
}

/* Recreate original loop as a collection of subsamples, each at their
 * original position within the loop.
 */
static void export_sequence (GdamSamplingSequencerSkin *ssskin)
{
  GdamBufferViewport *viewport = get_viewport (ssskin);
  gdouble bpm, beat_length;
  gdouble sample_length = gdam_buffer_viewport_get_sample_length (viewport);
  GList *sslist = ssskin->subsample_infos;
  int samps_per_sec = viewport->samples_per_second;
  int index = 0;

  bpm = get_bpm (ssskin);
  g_return_if_fail (bpm > 0.0);
  beat_length = 60.0 / bpm;

  gdam_sequencer_freeze (get_sequencer (ssskin));

  if (!prepare_sequencer (ssskin))
    { 
      gdam_sequencer_thaw (get_sequencer (ssskin));
      return;
    }

  while (sslist)
    {
      GdamSubSampleInfo *ssinfo = (GdamSubSampleInfo*)sslist->data;
      gdouble start_beat = ssinfo->start_time / beat_length;
      /* XXX: these should not depend on samps per sec?? */
      int start = (int) (ssinfo->start_time * samps_per_sec);
      int length = (int) (ssinfo->length * samps_per_sec);
      GdamSequencerElement *element;
      element = add_element (get_sequencer (ssskin),
			     ssinfo->buffer_id,
			     start_beat, /* position */
			     1.0, 1.0, 
			     0,-1,
			     index);
      g_return_if_fail (element != NULL);
      sslist = sslist->next;
      index++;
    }
  gdam_sequencer_thaw (get_sequencer (ssskin));
  gdam_producer_play_now (GDAM_PRODUCER (get_sequencer(ssskin))); 
}

typedef struct _RollInfo {
	  /* absolute values, scale of 1.0 */
  	gdouble gain1;
  	gdouble gain2;
  	gdouble rate1;
  	gdouble rate2;
	  /* relative to subsample's values, units in seconds */
  	gdouble offset1;
  	gdouble offset2;
  	gdouble length1;
  	gdouble length2;
	  /* number of triggers per beat */
  	gdouble steps1;
  	gdouble steps2;
	  /* length of roll */
  	gdouble beats;
} RollInfo;

static RollInfo *new_rollinfo ()
{
  RollInfo *rv;
  rv = g_new (RollInfo, 1);
  rv->gain1   = rv->gain2   = 1.0;
  rv->rate1   = rv->rate2   = 1.0;
  rv->offset1 = rv->offset2 = 0.0;
  rv->length1 = rv->length2 = 0.0;
  rv->steps1  = rv->steps2  = 4;
  rv->beats   = 4;
  return rv;
}

static void init_rerhythminfo (RerhythmInfo *rv)
{
  rv->grain_size = 1.0;
  rv->leeway = 0.05;
  rv->whole_offset = 0;
  rv->whole_count = 1;
  rv->whole_delay = 0.125;
  rv->half_offset = 0;
  rv->half_count = 1;
  rv->half_delay = 0.125;
  rv->offbeat_count = 1;
}

static RerhythmInfo *new_rerhythminfo ()
{
  RerhythmInfo *rv;
  rv = g_new (RerhythmInfo, 1);
  init_rerhythminfo (rv);
  return rv;
}


static inline gdouble percval (gdouble perc, gdouble sval, gdouble eval)
{
  return sval + perc * (eval - sval);
}

/* Export a loop which is a single subsample tweaked over time */
static void export_roll_by_info (GdamSamplingSequencerSkin *ssskin, int idx,
    				 RollInfo *ri)
{
  GdamBufferViewport *viewport = get_viewport (ssskin);
  GdamSubSampleInfo *ssinfo;
  int samps_per_sec = viewport->samples_per_second;
  gdouble cur_beat = 0;

  ssinfo = (GdamSubSampleInfo*) g_list_nth_data (ssskin->subsample_infos, idx);
  g_return_if_fail (ssinfo != NULL);

  gdam_sequencer_freeze (get_sequencer (ssskin));

  if (!prepare_sequencer (ssskin))
    { 
      gdam_sequencer_thaw (get_sequencer (ssskin));
      return;
    }

  while (cur_beat < ri->beats)
    {
      GdamSequencerElement *element;
      gdouble gain, rate, del_start, del_length;
      gdouble perc = cur_beat / ri->beats;
      int start, len;
      gain       = percval (perc, ri->gain1,   ri->gain2);
      rate       = percval (perc, ri->rate1,   ri->rate2);
      del_start  = percval (perc, ri->offset1, ri->offset2);
      del_length = percval (perc, ri->length1, ri->length2);
      start = (int) ((ssinfo->start_time + del_start) * samps_per_sec);
      len = (int) ((ssinfo->length + del_length) * samps_per_sec);
      element = add_element (get_sequencer (ssskin),
			     ssinfo->buffer_id,
			     cur_beat, 
			     rate, gain, 
			     //start, len,
			     0, -1,
			     idx);
      cur_beat +=  1.0 / ri->steps1;
    }
  gdam_sequencer_thaw (get_sequencer (ssskin));
  gdam_producer_play_now (GDAM_PRODUCER (get_sequencer(ssskin))); 
}

/* TODO: don't ignore buffer's first_beat when calculating phase */
/* TODO: wrap beats which are shifted out of range (here or always) */
static void export_rerhythm_by_info (GdamSamplingSequencerSkin *ssskin,
    				     RerhythmInfo *ri)
{
  GdamBufferViewport *viewport = get_viewport (ssskin);
  GList *sslist;
  gdouble bpm, beat_length;
  int samps_per_sec = viewport->samples_per_second;
  int index = 0;

  gdam_sequencer_freeze (get_sequencer (ssskin));

  if (!prepare_sequencer (ssskin))
    { 
      gdam_sequencer_thaw (get_sequencer (ssskin));
      return;
    }

  sslist = ssskin->subsample_infos;
  bpm = get_bpm (ssskin);
  g_return_if_fail (bpm > 0.0);
  beat_length = 60.0 / bpm;

  while (sslist)
    {
      GdamSequencerElement *element;
      GdamSubSampleInfo *ssinfo;
      gdouble orig_beat, cur_beat;
      gdouble nearest_beat, nearest_half_beat, dist, half_dist;
      int start, len;
      ssinfo = (GdamSubSampleInfo*) sslist->data;
      g_return_if_fail (ssinfo != NULL);
      start = (int) (ssinfo->start_time * samps_per_sec);
      len = (int) (ssinfo->length * samps_per_sec);
      orig_beat = ssinfo->start_time / beat_length;

	/* find nearest even beat, scaling by grain_size so that rounding makes
	 * sense. */
      nearest_beat = (int)(orig_beat / ri->grain_size + 0.5);
      nearest_beat *= ri->grain_size;
      nearest_half_beat = (int)(orig_beat / (0.5 * ri->grain_size) + 0.5);
      nearest_half_beat *= (ri->grain_size * 0.5);

	/* get distance to nearest even beat */
      dist = fabs (orig_beat - nearest_beat);
      half_dist = fabs (orig_beat - nearest_half_beat);

      cur_beat = orig_beat;

      /* Subsamples which play on the downbeat */
      if (dist < ri->leeway)
	{
	  int n;
	  cur_beat += ri->whole_offset;
	  for (n = 0; n < ri->whole_count; n++)
	    {
	      element = add_element (get_sequencer (ssskin),
				     ssinfo->buffer_id,
				     cur_beat, 
				     1.0, 1.0, 
				     0, -1,
				     //start, len,
				     index);
	      cur_beat += ri->whole_delay;
	    }
	}
      /* Subsamples which play on the offbeat */
      else if (half_dist < ri->leeway)
	{
	  int n;
	  cur_beat += ri->half_offset;
	  for (n = 0; n < ri->half_count; n++)
	    {
	      element = add_element (get_sequencer (ssskin),
				     ssinfo->buffer_id,
				     cur_beat, 
				     1.0, 1.0, 
				     0, -1, 
				     index);
	      cur_beat += ri->half_delay;
	    }
	}
      else
	{
	  int n;
	  for (n = 0; n < ri->offbeat_count; n++)
	    {
	      element = add_element (get_sequencer (ssskin),
				     ssinfo->buffer_id,
				     cur_beat, 
				     1.0, 1.0, 
				     0, -1,
				     index);
	      dist /= 2;
	      cur_beat += dist;
	    }
	}
      sslist = sslist->next;
      index++;
    }
  gdam_sequencer_thaw (get_sequencer (ssskin));
}

static void set_internal_rerhythm_by_info (GdamSamplingSequencerSkin *ssskin,
    				           RerhythmInfo *ri)
{
  GtkObject *obj = GTK_OBJECT (ssskin);
  gdam_sequencer_freeze (get_sequencer (ssskin));
  gdam_arg_set (obj, "rerhythm_whole_offset", ri->whole_offset);
  gdam_arg_set (obj, "rerhythm_whole_delay", ri->whole_delay);
  gdam_arg_set (obj, "rerhythm_whole_count", ri->whole_count);
  gdam_arg_set (obj, "rerhythm_half_offset", ri->half_offset);
  gdam_arg_set (obj, "rerhythm_half_delay", ri->half_delay);
  gdam_arg_set (obj, "rerhythm_half_count", ri->half_count);
  gdam_arg_set (obj, "rerhythm_offbeat_count", ri->offbeat_count);
  gdam_arg_set (obj, "rerhythm_grain_size", ri->grain_size);
  gdam_sequencer_thaw (get_sequencer (ssskin));
}

static void export_rerhythm (GdamSamplingSequencerSkin *ssskin)
{
  RerhythmInfo *ri;
  ri = new_rerhythminfo ();
  ri->grain_size = 0.5;
  ri->leeway = 0.05;
  export_rerhythm_by_info (ssskin, ri);
  g_free (ri);
}

static void export_gain_roll (GdamSamplingSequencerSkin *ssskin, int idx)
{
  RollInfo *rinfo = new_rollinfo ();
  rinfo->gain1 = 0.0;
  export_roll_by_info (ssskin, idx, rinfo);
  g_free (rinfo);
}

static void export_gain_roll_down (GdamSamplingSequencerSkin *ssskin, int idx)
{
  RollInfo *rinfo = new_rollinfo ();
  rinfo->gain2 = 0.0;
  export_roll_by_info (ssskin, idx, rinfo);
  g_free (rinfo);
}

static void export_size_roll (GdamSamplingSequencerSkin *ssskin, int idx)
{
  RollInfo *rinfo = new_rollinfo ();
  GdamSubSampleInfo *ssinfo;
  ssinfo = (GdamSubSampleInfo*) g_list_nth_data (ssskin->subsample_infos, idx);
  g_return_if_fail (ssinfo != NULL);
  rinfo->length1 = -ssinfo->length;
  rinfo->length2 = 0;
  export_roll_by_info (ssskin, idx, rinfo);
  g_free (rinfo);
}

static void export_size_roll_down (GdamSamplingSequencerSkin *ssskin, int idx)
{
  RollInfo *rinfo = new_rollinfo ();
  GdamSubSampleInfo *ssinfo;
  ssinfo = (GdamSubSampleInfo*) g_list_nth_data (ssskin->subsample_infos, idx);
  g_return_if_fail (ssinfo != NULL);
  rinfo->length1 = 0;
  rinfo->length2 = -ssinfo->length;
  export_roll_by_info (ssskin, idx, rinfo);
  g_free (rinfo);
}

static void export_rate_roll_down (GdamSamplingSequencerSkin *ssskin, int idx)
{
  RollInfo *rinfo = new_rollinfo ();
  GdamSubSampleInfo *ssinfo;
  ssinfo = (GdamSubSampleInfo*) g_list_nth_data (ssskin->subsample_infos, idx);
  g_return_if_fail (ssinfo != NULL);
  rinfo->rate1 = 1.333;
  rinfo->rate2 = 1.0;
  rinfo->gain1 = 0.333;
  rinfo->gain2 = 1.0;
  export_roll_by_info (ssskin, idx, rinfo);
  g_free (rinfo);
}


void gdam_sampling_sequencer_skin_get_sample (GdamSamplingSequencerSkin* ssskin,
    					 int			index,
					 gdouble*		start_time,
					 gdouble*		length)
{
  GdamSubSampleInfo* ssinfo;
  ssinfo = g_list_nth_data (ssskin->subsample_infos, index);
  *start_time = ssinfo->start_time;
  *length = ssinfo->length;
}

/* TODO: remove sample */

void gdam_sampling_sequencer_skin_set_sampled_buffer 
					(GdamSamplingSequencerSkin* ssskin,
    					 int                        buffer_id)
{
  GList *sslist = ssskin->subsample_infos;
  GList *allsslist = ssskin->all_subsample_infos;
  GList *tmplist = NULL;
  /* Move old current subsamples into the full list */
  while (sslist)
    {
      allsslist = g_list_insert_sorted (allsslist, sslist->data, sort_ssinfo);
      sslist = g_list_remove (sslist, sslist->data); 
    }
  /* Move new current subsamples into the current list */
  while (allsslist)
    {
      GdamSubSampleInfo* ssinfo = (GdamSubSampleInfo*)allsslist->data;
      if (ssinfo->buffer_id == buffer_id)
	sslist = g_list_insert_sorted (sslist, allsslist->data, sort_ssinfo);
      else
	tmplist = g_list_insert_sorted (tmplist, allsslist->data, sort_ssinfo);
      allsslist = g_list_remove (allsslist, allsslist->data);
    }
  ssskin->subsample_infos = sslist;
  ssskin->all_subsample_infos = tmplist ;
  /* XXX: this function should be tied to set_buffer or removed */
}

/* Allow access to viewport's handlers */
static gboolean gdam_sampling_sequencer_skin_defer_to_viewport_handler
                                (GdamSkin          *skin,
                                 GtkObject         *object,
                                 const char        *signal_name,
                                 int                num_handler_names,
                                 char             **handler_names)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (skin);
  GdamSkin *subskin = GDAM_SKIN (get_viewport (ssskin));
  if (!subskin)
    {
      g_warning ("missing subskin in %s", handler_names[0]);
      return FALSE;
    }
  if (!gdam_skin_do_connect (subskin,
			     object,
			     signal_name,
			     num_handler_names - 1, 
			     handler_names + 1))
    {
      g_warning ("gdam_skin_do_connect failed on deferred handler: "
		 "num_args=%d, handler_name=%s, skintype=%s",
		 num_handler_names,
		 handler_names[1],
		 gtk_type_name (GTK_OBJECT (subskin)->klass->type));
      return FALSE;
    }
  return TRUE;
}

static gboolean gdam_sampling_sequencer_skin_viewport_handler
                                (GdamSkin          *skin,
                                 GtkObject         *object,
                                 const char        *signal_name,
                                 int                num_handler_names,
                                 char             **handler_names)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (skin);
  GdamBufferViewport* viewport = get_viewport (ssskin);
  if (!GTK_IS_CONTAINER (object))
    {
      g_warning ("cannot add buffer_viewport to non-container");
      return TRUE;
    }
  if (!viewport)
    {
      g_warning ("cannot add non-existant buffer_viewport");
      return FALSE;
    }
  gtk_container_add (GTK_CONTAINER (object),
		     GDAM_SKIN (viewport)->main_widget);
  return TRUE;
}

/* We trap the toplevel's delete event, and hide it instead.  return TRUE
 * to indicate we've handled the signal, and prevent the window from being 
 * destroyed. */
static int 
rerhythm_toplevel_delete_event (GtkWidget *widget, 
    				GdkEvent  *event,
			        GdamSkin  *skin)
{
  gdam_arg_set (GTK_OBJECT (skin), "rerhythm_visibility", 0);
  return TRUE;
}

static gboolean 
gdam_sampling_sequencer_skin_rerhythm_window_handler
                                (GdamSkin          *skin,
                                 GtkObject         *object,
                                 const char        *signal_name,
                                 int                num_handler_names,
                                 char             **handler_names)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (skin);
  ssskin->rerhythm_toplevel = GTK_WIDGET (object);
  gtk_object_ref (object);
  gtk_signal_connect (object,
		      "delete_event",
		      GTK_SIGNAL_FUNC (rerhythm_toplevel_delete_event), 
		      skin);
  /* TODO: trap toplevel delete_event */
  gtk_signal_connect_object_while_alive (GTK_OBJECT (skin),
					 "destroy", 
					 GTK_SIGNAL_FUNC (gtk_object_destroy), 
					 object);
  return TRUE;
}

static void on_buffer_id_received(const char*       text,
				  gpointer          data)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (data);
  set_buffer_id (ssskin, atoi (text));
}

static gboolean gdam_sampling_sequencer_skin_buffer_id_dropped_handler
                                (GdamSkin          *skin,
                                 GtkObject         *object,
                                 const char        *signal_name,
                                 int                num_handler_names,
                                 char             **handler_names)
{
  gdam_widget_string_dest_set (GTK_WIDGET (object),
			       on_buffer_id_received,
			       "gdam-buffer-id-text", skin);
  return TRUE;
}

static void play_by_idx(GdamSamplingSequencerSkin* ssskin, int idx)
{
  GdamBufferViewport* viewport = get_viewport (ssskin);
  GdamSequencerElement *sample_info;
  GdamSubSampleInfo* ssinfo;
  int num_ss = g_list_length (ssskin->subsample_infos);
  int samps_per_sec = viewport->samples_per_second;
  int start, length;
  if (idx >= num_ss)
    {
      g_warning ("index out of range.");
      return;
    }
  ssinfo = (GdamSubSampleInfo*)g_list_nth_data (ssskin->subsample_infos, idx);
  /* XXX: these should not depend on samps per sec?? */
  start = (int) (ssinfo->start_time * samps_per_sec);
  length = (int) (ssinfo->length * samps_per_sec);

  gdam_sequencer_freeze (get_sequencer (ssskin));
  gdam_sequencer_set_elements (get_sequencer (ssskin), NULL);
  sample_info = add_element (get_sequencer (ssskin),
			     ssinfo->buffer_id, 
			     0, 1.0, 1.0, 0, -1, idx);
  //g_return_if_fail (sample_info != NULL);

  gdam_sequencer_thaw (get_sequencer (ssskin));
  gdam_producer_play_now (GDAM_PRODUCER (get_sequencer(ssskin))); 
}

static void set_play_by_idx(GtkObject* object, GdamArg* arg)
{
  GdamSamplingSequencerSkin* ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  int id = GDAM_ARG_INT (arg);
  play_by_idx (ssskin, id);
}

static void set_play_roll_by_idx (GtkObject* object, GdamArg* arg)
{
  GdamSamplingSequencerSkin* ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  int id = GDAM_ARG_INT (arg);
  export_gain_roll (ssskin, id);
}

static void set_play_size_roll_by_idx (GtkObject* object, GdamArg* arg)
{
  GdamSamplingSequencerSkin* ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  int id = GDAM_ARG_INT (arg);
  export_size_roll (ssskin, id);
}

static void set_play_size_roll_down_by_idx (GtkObject* object, GdamArg* arg)
{
  GdamSamplingSequencerSkin* ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  int id = GDAM_ARG_INT (arg);
  export_size_roll_down (ssskin, id);
}

static void set_play_rate_roll_down_by_idx (GtkObject* object, GdamArg* arg)
{
  GdamSamplingSequencerSkin* ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  int id = GDAM_ARG_INT (arg);
  export_rate_roll_down (ssskin, id);
}

/* For now, hardcoded presets for testing */
static void set_play_rerhythm_by_info (GtkObject* object, GdamArg* arg)
{
  GdamSamplingSequencerSkin* ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  int id = GDAM_ARG_INT (arg);
  RerhythmInfo *ri;
  ri = g_new (RerhythmInfo, 1);
  init_rerhythminfo (ri);
  ri->grain_size = ssskin->rerhythm_info.grain_size;
  ri->leeway = ssskin->rerhythm_info.leeway;
  if (id == 1)
    ri->half_offset = 0.125;
  else if (id == 2)
    {
      ri->half_count = 2;
      ri->half_offset = -0.125;
    }
  else if (id == 3)
    {
      ri->half_count = 3;
      ri->half_delay = 0.333;
    }
  else if (id == 4)
    ri->whole_offset = 0.125;
  else if (id == 5)
    {
      ri->half_count = 0;
      ri->offbeat_count = 4;
    }
  set_internal_rerhythm_by_info (ssskin, ri);
}

static void play_internal_loop (GdamSamplingSequencerSkin *ssskin)
{
  if (!ssskin->sampled_buffer_id)
    return;
  export_rerhythm_by_info (ssskin, &ssskin->rerhythm_info);
}

/* TODO: use a switch statement for these? */
static void set_whole_offset (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  ssskin->rerhythm_info.whole_offset = GDAM_ARG_DOUBLE (arg);
  play_internal_loop (ssskin);
}
static void get_whole_offset (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  GDAM_ARG_DOUBLE (arg) = ssskin->rerhythm_info.whole_offset;
}
static void set_whole_count (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  ssskin->rerhythm_info.whole_count = GDAM_ARG_INT (arg);
  play_internal_loop (ssskin);
}
static void get_whole_count (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  GDAM_ARG_INT (arg) = ssskin->rerhythm_info.whole_count;
}
static void set_whole_delay (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  ssskin->rerhythm_info.whole_delay = GDAM_ARG_DOUBLE (arg);
  play_internal_loop (ssskin);
}
static void get_whole_delay (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  GDAM_ARG_DOUBLE (arg) = ssskin->rerhythm_info.whole_delay;
}

static void set_half_offset (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  ssskin->rerhythm_info.half_offset = GDAM_ARG_DOUBLE (arg);
  play_internal_loop (ssskin);
}
static void get_half_offset (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  GDAM_ARG_DOUBLE (arg) = ssskin->rerhythm_info.half_offset;
}
static void set_half_count (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  ssskin->rerhythm_info.half_count = GDAM_ARG_INT (arg);
  play_internal_loop (ssskin);
}
static void get_half_count (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  GDAM_ARG_INT (arg) = ssskin->rerhythm_info.half_count;
}
static void set_half_delay (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  ssskin->rerhythm_info.half_delay = GDAM_ARG_DOUBLE (arg);
  play_internal_loop (ssskin);
}
static void get_half_delay (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  GDAM_ARG_DOUBLE (arg) = ssskin->rerhythm_info.half_delay;
}

static void set_offbeat_count (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  ssskin->rerhythm_info.offbeat_count = GDAM_ARG_INT (arg);
  play_internal_loop (ssskin);
}
static void get_offbeat_count (GtkObject *object, GdamArg*arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  GDAM_ARG_INT (arg) = ssskin->rerhythm_info.offbeat_count;
}
static void set_grain_size (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  ssskin->rerhythm_info.grain_size = GDAM_ARG_DOUBLE (arg);
  play_internal_loop (ssskin);
}
static void get_grain_size (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  GDAM_ARG_DOUBLE (arg) = ssskin->rerhythm_info.grain_size;
}

static void set_amplitude_threshold (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  ssskin->amplitude_threshold = GDAM_ARG_DOUBLE (arg);
}
static void get_amplitude_threshold (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  GDAM_ARG_DOUBLE (arg) = ssskin->amplitude_threshold;
}

static void set_amplitude_ratio (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  ssskin->amplitude_ratio = GDAM_ARG_DOUBLE (arg);
}
static void get_amplitude_ratio (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  GDAM_ARG_DOUBLE (arg) = ssskin->amplitude_ratio;
}

static void set_window_size (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  ssskin->window_size = GDAM_ARG_DOUBLE (arg);
}
static void get_window_size (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  GDAM_ARG_DOUBLE (arg) = ssskin->window_size;
}

static void 
set_rerhythm_visibility (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  gboolean is_visible;
  gboolean visible = GDAM_ARG_INT (arg);

  if (ssskin->rerhythm_toplevel == NULL)
    return;

  is_visible = GTK_WIDGET_VISIBLE (ssskin->rerhythm_toplevel);
  if (is_visible == visible)
    return;
  if (visible)
    gtk_widget_show (ssskin->rerhythm_toplevel);
  else
    gtk_widget_hide (ssskin->rerhythm_toplevel);
}

static void 
get_rerhythm_visibility (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  if (ssskin->rerhythm_toplevel && 
      GTK_WIDGET_VISIBLE (ssskin->rerhythm_toplevel))
    GDAM_ARG_INT (arg) = 1;
  else
    GDAM_ARG_INT (arg) = 0;
}

static void set_min_sample_length (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  ssskin->min_sample_length = GDAM_ARG_DOUBLE (arg);
}
static void get_min_sample_length (GtkObject *object, GdamArg *arg)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  GDAM_ARG_DOUBLE (arg) = ssskin->min_sample_length;
}


static void set_current_subsample_index (GtkObject* object, GdamArg* arg)
{
  GdamSamplingSequencerSkin* ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  GdamSubSampleInfo* ssinfo;
  GdamBufferViewport* viewport = get_viewport (ssskin);
  int num_ss = g_list_length (ssskin->subsample_infos);
  int idx = GDAM_ARG_INT (arg);
  if (idx >= num_ss)
    {
      g_warning ("index out of range.");
      return;
    }
  ssinfo = (GdamSubSampleInfo*)g_list_nth_data (ssskin->subsample_infos, idx);
  ssskin->current_subsample_index = idx;
  gdam_buffer_viewport_queue_redraw (viewport);
}

static void get_current_subsample_index (GtkObject* object, GdamArg* arg)
{
  GdamSamplingSequencerSkin* ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (object);
  GDAM_ARG_INT (arg) = ssskin->current_subsample_index;
}

static void play_current (GdamSamplingSequencerSkin* ssskin)
{
  play_by_idx (ssskin, ssskin->current_subsample_index);
}

static int zero_threshold = (int)(0.005 * GDAM_SAMPLE_ONE);
static inline gboolean is_zero (int sample)
{
  if (sample < zero_threshold && sample > -1 * zero_threshold)
    return TRUE;
  return FALSE;
}

/* TODO: move this functionality to a server-side function. */
static void 
got_beat_times (int num_beat_times, gdouble *samples, gpointer user_data)
{
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (user_data);
  GList *sslist = NULL;
  int i;

  /* Walk through the buffer from the beginning */
  for (i=0; i < num_beat_times; i++)
    {
      GdamSubSampleInfo *ssinfo;
      /* Save the description of the beat we just found. */
      ssinfo = g_new (GdamSubSampleInfo, 1);
      ssinfo->buffer_id = ssskin->sampled_buffer_id;
      if  (!ssinfo->buffer_id)
	g_warning ("missing buffer_id in samples_from_beat_times");
      ssinfo->start_time = samples[i];
      sslist = g_list_prepend (sslist, ssinfo);
    }

  sslist = g_list_reverse (sslist);
  free_subsamples (ssskin);
  ssskin->subsample_infos = sslist;
  calculate_subsample_lengths (ssskin);
  update_buffer_beatlist (ssskin);
  export_sequence (ssskin);
}

static void 
got_beat_times_arg (int num_args, GdamArg *args, gpointer data)
{
  gtk_object_unref (GTK_OBJECT (data));
  if (num_args != 1
      || strcmp (args->name, "beat_times") != 0
      || args->type != GDAM_ARG_TYPE_DOUBLE_ARRAY)
    {
      g_warning ("missing or incorrect args from guess_beat_times");
      return;
    }
  got_beat_times (GDAM_ARG_DOUBLE_ARRAY (args + 0).length,
                  GDAM_ARG_DOUBLE_ARRAY (args + 0).array,
		  data);
}

static void 
guess_beat_times_from_arg_func (GdamSamplingSequencerSkin *ssskin)
{
  GdamBufferViewport *viewport = get_viewport (ssskin);
  GdamRequest *request;
  GdamArg args[4];

  gtk_object_ref (GTK_OBJECT (ssskin));
  args[0].name = "window_size";
  args[0].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (args + 0) = ssskin->window_size;
  args[1].name = "amplitude_threshold";
  args[1].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + 1) = 
     (int)(ssskin->amplitude_threshold * GDAM_SAMPLE_ONE);
  args[2].name = "amplitude_ratio";
  args[2].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (args + 2) = ssskin->amplitude_ratio;
  args[3].name = "min_sample_length";
  args[3].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (args + 3) = ssskin->min_sample_length;
  request = gdam_request_call_function (GDAM_PROTOCOL_BUFFER_ARG,
					viewport->buffer_id,
					"guess_beat_times",
					4,
					args,
					got_beat_times_arg,
					NULL, ssskin);

  gdam_channel_append_request (GDAM_SKIN (ssskin)->channel, request);
}

static GdamSkinGenericConnect connect_infos[] = {
	{ "gdam_sampling_sequencer_viewport",
           gdam_sampling_sequencer_skin_viewport_handler },
	{ "gdam_sampling_sequencer_rerhythm_window",
           gdam_sampling_sequencer_skin_rerhythm_window_handler },
	{ "defer_to_viewport",
           gdam_sampling_sequencer_skin_defer_to_viewport_handler },
	{ "gdam_sampling_sequencer_buffer_id_dropped",
           gdam_sampling_sequencer_skin_buffer_id_dropped_handler },
};

static GdamSkinSimpleConnect simple_connect_infos[] = {
	{ "gdam_sampling_sequencer_cycle",
           (GdamSkinFunc)cycle_subsample },
	{ "gdam_sampling_sequencer_extend",
           (GdamSkinFunc)extend_subsample },
	{ "gdam_sampling_sequencer_set_default",
           (GdamSkinFunc)set_default_buffer },
	{ "gdam_sampling_sequencer_export_sequence",
           (GdamSkinFunc)export_sequence },
	{ "gdam_sampling_sequencer_export_rerhythm",
           (GdamSkinFunc)export_rerhythm},
	{ "gdam_sampling_sequencer_play",
           (GdamSkinFunc)play_current },
	{ "gdam_sampling_sequencer_guess_beats",
           (GdamSkinFunc)guess_beat_times_from_arg_func },
};


/* When the viewport's beat lines change, rebuild our list to match it.
 */
static gint on_beat_lines_changed (GtkObject* object, 
                                   GdamSamplingSequencerSkin* ssskin)
{
  GdamBufferViewport *viewport = get_viewport (ssskin);
  int idx, len;
  subsamples_from_beat_times (ssskin);
  
  /* Safety check for invalid index... */
  idx = ssskin->current_subsample_index;
  len = g_list_length (ssskin->subsample_infos);
  if (idx >= len)
    {
      g_warning ("oversized current_subsample_index");
      ssskin->current_subsample_index = 0;
    }

  return TRUE;
}

static gint on_buffer_changed (GtkObject* object, 
                               GdamSamplingSequencerSkin* ssskin)
{
  GdamBufferViewport *viewport = get_viewport (ssskin);
  ssskin->num_last_samples = 0;
  guess_beat_times_from_arg_func (ssskin);
  return TRUE;
}

static GdamSkinClass* parent_class = NULL;
static gboolean       sampling_sequencer_build_new_model
                                                 (GdamSkin*      skin,
                                                  GdamChannel*   channel,
						  int		 num_options,
						  char**         options)
{
  GdamModel *model;
  GdamSamplingSequencerSkin *ssskin = GDAM_SAMPLING_SEQUENCER_SKIN (skin);
  gboolean rv;
  rv = (*parent_class->build_new_model) (skin, channel, num_options, options);
  gtk_signal_connect (GTK_OBJECT (get_viewport (ssskin)),
		      "beat_lines_changed",
		      GTK_SIGNAL_FUNC (on_beat_lines_changed),
		      ssskin);
  gtk_signal_connect (GTK_OBJECT (get_viewport (ssskin)),
		      "buffer_changed",
		      GTK_SIGNAL_FUNC (on_buffer_changed),
		      ssskin);
  return TRUE;
}

static void 
gdam_sampling_sequencer_skin_init (GdamSamplingSequencerSkin* ssskin)
{
  //gdam_sampling_sequencer_skin_add_sample (ssskin, 0, 0);
  ssskin->amplitude_threshold = 0.15;
  ssskin->amplitude_ratio = 2.0;
  ssskin->window_size = 0.03;
  ssskin->min_sample_length = 0.08;
  ssskin->samples_per_second = 44100;
  init_rerhythminfo (&ssskin->rerhythm_info);
}

static void 
gdam_sampling_sequencer_skin_class_init(GdamSamplingSequencerSkinClass* sampling_sequencer_skin_class)
{
  GdamSkinClass* skin_class = GDAM_SKIN_CLASS (sampling_sequencer_skin_class);
  parent_class = gtk_type_class (GDAM_TYPE_SEQUENCER_SKIN);
  skin_class->build_new_model = sampling_sequencer_build_new_model;
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
  gdam_skin_class_register_simple_connections 
    				      (skin_class,
				       G_N_ELEMENTS (simple_connect_infos),
				       simple_connect_infos);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "play_by_subbuffer_index",
		     GDAM_ARG_TYPE_INT, 
		     set_play_by_idx,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "play_roll_by_subbuffer_index",
		     GDAM_ARG_TYPE_INT, 
		     set_play_roll_by_idx,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "play_size_roll_by_subbuffer_index",
		     GDAM_ARG_TYPE_INT, 
		     set_play_size_roll_by_idx,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "play_size_roll_down_by_subbuffer_index",
		     GDAM_ARG_TYPE_INT, 
		     set_play_size_roll_down_by_idx,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "play_rate_roll_down_by_subbuffer_index",
		     GDAM_ARG_TYPE_INT, 
		     set_play_rate_roll_down_by_idx,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "play_rerhythm_by_info",
		     GDAM_ARG_TYPE_INT, 
		     set_play_rerhythm_by_info,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "current_subsample_index",
		     GDAM_ARG_TYPE_INT, 
		     set_current_subsample_index,
		     get_current_subsample_index,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "rerhythm_whole_offset",
		     GDAM_ARG_TYPE_DOUBLE, 
		     set_whole_offset,
		     get_whole_offset,
		     NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "rerhythm_whole_delay",
		     GDAM_ARG_TYPE_DOUBLE, 
		     set_whole_delay,
		     get_whole_delay,
		     NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "rerhythm_whole_count",
		     GDAM_ARG_TYPE_INT, 
		     set_whole_count,
		     get_whole_count,
		     NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "rerhythm_half_offset",
		     GDAM_ARG_TYPE_DOUBLE, 
		     set_half_offset,
		     get_half_offset,
		     NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "rerhythm_half_delay",
		     GDAM_ARG_TYPE_DOUBLE, 
		     set_half_delay,
		     get_half_delay,
		     NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "rerhythm_half_count",
		     GDAM_ARG_TYPE_INT, 
		     set_half_count,
		     get_half_count,
		     NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "rerhythm_offbeat_count",
		     GDAM_ARG_TYPE_INT, 
		     set_offbeat_count,
		     get_offbeat_count,
		     NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "rerhythm_grain_size",
		     GDAM_ARG_TYPE_DOUBLE, 
		     set_grain_size,
		     get_grain_size,
		     NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "amplitude_threshold",
		     GDAM_ARG_TYPE_DOUBLE, 
		     set_amplitude_threshold,
		     get_amplitude_threshold,
		     NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "amplitude_ratio",
		     GDAM_ARG_TYPE_DOUBLE, 
		     set_amplitude_ratio,
		     get_amplitude_ratio,
		     NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "window_size",
		     GDAM_ARG_TYPE_DOUBLE, 
		     set_window_size,
		     get_window_size,
		     NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "min_sample_length",
		     GDAM_ARG_TYPE_DOUBLE, 
		     set_min_sample_length,
		     get_min_sample_length,
		     NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
                     "rerhythm_visibility",
		     GDAM_ARG_TYPE_INT, 
		     set_rerhythm_visibility,
		     get_rerhythm_visibility,
		     NULL, "boolean", 0);
}

GtkType gdam_sampling_sequencer_skin_get_type()
{
	static GtkType sampling_sequencer_skin_type = 0;
	if (!sampling_sequencer_skin_type) {
		static const GtkTypeInfo sampling_sequencer_skin_info = {
			"GdamSamplingSequencerSkin",
			sizeof(GdamSamplingSequencerSkin),
			sizeof(GdamSamplingSequencerSkinClass),
			(GtkClassInitFunc) gdam_sampling_sequencer_skin_class_init,
			(GtkObjectInitFunc) gdam_sampling_sequencer_skin_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_SEQUENCER_SKIN;
		sampling_sequencer_skin_type = gtk_type_unique(parent,
		                                 &sampling_sequencer_skin_info);
	}
	return sampling_sequencer_skin_type;
}
